/*
 * Copyright (C) 2010-2012  Internet Systems Consortium, Inc. ("ISC")
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND ISC DISCLAIMS ALL WARRANTIES WITH
 * REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS.  IN NO EVENT SHALL ISC BE LIABLE FOR ANY SPECIAL, DIRECT,
 * INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
 * OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */

/* $Id: 131316dcd5f71a6794ed5113f0b92cd6914aaccf $ */

/*
 * PCP (port-control-protocol) client library include
 *
 * Francis_Dupont@isc.org, December 2010
 *
 * a la miniUPnP libnatpmp
 *
 * types: pcp_t, pcp_request_t, pcp_response_t
 */

#ifndef __PCP_H__
#define __PCP_H__

/* API version */

#define __PCP_API_VERSION__		0

#ifdef WIN32
#include <stdint.h>
#include <time.h>
#else
#include <sys/time.h>
#endif
#include <string.h>

#ifdef __cplusplus
extern "C" {
#endif


/* Types */

#ifndef PCP_MAX_OPTIONS
#define PCP_MAX_OPTIONS	4
#endif
#define PCP_VECOPT_SIZE	(PCP_MAX_OPTIONS + 1)

typedef struct {			/* PCP handler (opaque) */
#ifdef WIN32
	SOCKET s;			/* server socket */
	SOCKET ms;			/* rapid recovery socket */
#else
	int s;				/* server socket */
	int ms;				/* rapid recovery socket */
#endif
	int tries;			/* retry counter */
	uint8_t usable;			/* better than s < 0 */
	uint8_t rcusable;		/* better than ms < 0 */
	uint8_t reset;			/* reset flag */
	uint8_t autofill;		/* address and port automatic fill */
	struct timeval retry;		/* next retry date */
	uint16_t reqlen;		/* current request length */
	uint16_t resplen;		/* current response length */
	uint8_t *request;		/* current request */
#define PCP_RESPLENGTH	1040
	uint8_t *response;		/* current request */
} pcp_t;

typedef struct {			/* PCP option */
	uint8_t code;			/* option code */
	uint16_t length;		/* option data length */
	uint8_t *data;			/* option data */
	union {
		uint32_t _oa4;		/* option IPv4 address */
		uint8_t _oa6[16];	/* option IPv6 address */
	} _oa;
#define optaddr4	_oa._oa4
#define optaddr6	_oa._oa6
} pcp_option_t;

typedef struct {			/* PCP element of request */
	uint8_t opcode;			/* opcode */
	uint8_t protocol;		/* protocol */
	uint32_t lifetime;		/* lifetime */
	union {
		uint32_t _ia4;		/* internal IPv4 address */
		uint8_t _ia6[16];	/* internal IPv6 address */
	} _ia;
#define intaddr4	_ia._ia4
#define intaddr6	_ia._ia6
	union {
		uint32_t _ea4;		/* external IPv4 address */
		uint8_t _ea6[16];	/* external IPv6 address */
	} _ea;
#define extaddr4	_ea._ea4
#define extaddr6	_ea._ea6
	union {
		uint32_t _rp4;		/* remote IPv4 peer */
		uint8_t _rp6[16];	/* remote IPv6 perr */
	} _rp;
#define remotepeer4	_rp._rp4
#define remotepeer6	_rp._rp6
	uint16_t intport;		/* internal port */
	uint16_t extport;		/* external port */
	uint16_t peerport;		/* remote peer port */
} pcp_request_t;

typedef struct {			/* PCP response */
	uint8_t result;			/* result-code */
	uint8_t optcnt;			/* option count */
	uint32_t epoch;			/* second since start of epoch */
	pcp_request_t assigned;		/* assigned request */
	pcp_option_t options[PCP_VECOPT_SIZE];	/* returned options */
	unsigned int optoffs[PCP_VECOPT_SIZE];	/* options offsets */
} pcp_response_t;

/* Routines */

#ifdef _MSC_VER
#define pcp_inline	__inline
#else
#define pcp_inline	inline
#endif

extern int pcp_init(struct sockaddr *server,
		    struct sockaddr *source,
		    pcp_t *pcp);
extern int pcp_close(pcp_t *pcp);
extern int pcp_getsocket(pcp_t *pcp, int *sock);
extern int pcp_setrcsocket(pcp_t *pcp, int sock);
extern int pcp_gettries(pcp_t *pcp, int *tries);
extern int pcp_gettimeout(pcp_t *pcp, struct timeval *timeout, int relative);
extern int pcp_setreset(pcp_t *pcp, int reset, int *old);
extern int pcp_setautofill(pcp_t *pcp, int autofill, int *old);
extern int pcp_third_party4(pcp_option_t ***options, uint32_t addr);
extern int pcp_third_party6(pcp_option_t ***options, uint8_t *addr);
extern int pcp_prefer_failure(pcp_option_t ***options);
extern int pcp_filter4(pcp_option_t ***options,
		       int prefixlen, uint32_t addr, uint16_t port);
extern int pcp_filter6(pcp_option_t ***options,
		       int prefixlen, uint8_t *addr, uint16_t port);
extern int pcp_makerequest(pcp_t *pcp,
			   pcp_request_t *request,
			   pcp_option_t **options);
extern int pcp_sendrequest(pcp_t *pcp);
extern int pcp_recvresponse(pcp_t *pcp, pcp_response_t *response);
extern const char *pcp_strerror(int error);
extern int pcp_getmapping(pcp_t *pcp,
			  pcp_request_t *request,
			  pcp_option_t **options,
			  pcp_response_t *response);
extern int pcp_renewmapping(pcp_t *pcp, pcp_response_t *response);
extern int pcp_delmapping(pcp_t *pcp, pcp_response_t *response);
extern int _pcp_fdinit(pcp_t *pcp, int fd);
extern int _pcp_fillsrcaddr(pcp_t *pcp, pcp_request_t *request);
extern int _pcp_setrequest(pcp_t *pcp, uint8_t *request, uint16_t reqlen);
extern int _pcp_getrequest(pcp_t *pcp, uint8_t **request, uint16_t *reqlen);
extern int _pcp_setresponse(pcp_t *pcp, uint8_t *response, uint16_t resplen);
extern int _pcp_getresponse(pcp_t *pcp, uint8_t **response, uint16_t *resplen);
extern int _pcp_submitresponse(pcp_t *pcp, pcp_response_t *response);
extern int _pcp_setoption(pcp_option_t ***options, pcp_option_t *opt);
static pcp_inline void pcp_freeoptions(pcp_option_t **options);

static const uint8_t pcp_mapped[16] =
	{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	  0x00, 0x00, 0xff, 0xff, 0x00, 0x00, 0x00, 0x00 };

static pcp_inline void pcp_mapintaddr(pcp_request_t *request)
{
	memcpy(request->intaddr6 + 12, &request->intaddr4, 4);
	memcpy(request->intaddr6, pcp_mapped, 12);
}

static pcp_inline void pcp_mapextaddr(pcp_request_t *request)
{
	memcpy(request->extaddr6 + 12, &request->extaddr4, 4);
	memcpy(request->extaddr6, pcp_mapped, 12);
}

static pcp_inline void pcp_mapremotepeer(pcp_request_t *request)
{
	if (request->remotepeer4 != 0) {
		memcpy(request->remotepeer6 + 12, &request->remotepeer4, 4);
		memcpy(request->remotepeer6, pcp_mapped, 12);
	}
}

static pcp_inline void pcp_mapoptaddr(pcp_option_t *option)
{
	if (option->optaddr4 != 0) {
		memcpy(option->optaddr6 + 12, &option->optaddr4, 4);
		memcpy(option->optaddr6, pcp_mapped, 12);
	}
}

static pcp_inline int pcp_unmapintaddr(pcp_request_t *request)
{
	if (memcmp(request->intaddr6, pcp_mapped, 12) == 0) {
		uint32_t addr;

		memcpy(&addr, request->intaddr6 + 12, 4);
		memset(request->intaddr6, 0, 16);
		request->intaddr4 = addr;
		return 1;
	}
	return 0;
}

static pcp_inline int pcp_unmapextaddr(pcp_request_t *request)
{
	if (memcmp(request->extaddr6, pcp_mapped, 12) == 0) {
		uint32_t addr;

		memcpy(&addr, request->extaddr6 + 12, 4);
		memset(request->extaddr6, 0, 16);
		request->extaddr4 = addr;
		return 1;
	}
	return 0;
}

static pcp_inline int pcp_unmapremotepeer(pcp_request_t *request)
{
	if (memcmp(request->remotepeer6, pcp_mapped, 12) == 0) {
		uint32_t addr;

		memcpy(&addr, request->remotepeer6 + 12, 4);
		memset(request->remotepeer6, 0, 16);
		request->remotepeer4 = addr;
		return 1;
	}
	return 0;
}

static pcp_inline int pcp_unmapoptaddr(pcp_option_t *option)
{
	if (memcmp(option->optaddr6, pcp_mapped, 12) == 0) {
		uint32_t addr;

		memcpy(&addr, option->optaddr6 + 12, 4);
		memset(option->optaddr6, 0, 16);
		option->optaddr4 = addr;
		return 1;
	}
	return 0;
}

/* opcode values */
#define PCP_OPCODE_RESPONSE	128	/* response bit */
#define PCP_OPCODE_ANNOUNCE	0
#define PCP_OPCODE_MAP		1
#define PCP_OPCODE_PEER		2

/* option code values */
#define PCP_OPTION_THIRD_PARTY	1	/* third-party request */
#define PCP_OPTION_PREF_FAIL	2	/* prefer failure */
#define PCP_OPTION_FILTER	3	/* remote peer filter */
#define PCP_OPTION_OPTIONAL	128	/* flag bit */

/* offsets in the common header */
#define PCP_VERSION		0
#define PCP_OPCODE		1
#define PCP_RESULT_CODE		3
#define PCP_LIFETIME		4
#define PCP_EPOCH		8
#define PCP_CLIENT_ADDR		8
#define PCP_OPCODE_HDR		24

/* offsets in the MAP header */
#define PCP_MAP_PROTOCOL	(PCP_OPCODE_HDR + 0)
#define PCP_MAP_INTERNAL_PORT	(PCP_OPCODE_HDR + 4)
#define PCP_MAP_EXTERNAL_PORT	(PCP_OPCODE_HDR + 6)
#define PCP_MAP_EXTERNAL_ADDR	(PCP_OPCODE_HDR + 8)

/* offsets in the PEER header */
#define PCP_PEER_PROTOCOL	(PCP_OPCODE_HDR + 0)
#define PCP_PEER_INTERNAL_PORT	(PCP_OPCODE_HDR + 4)
#define PCP_PEER_EXTERNAL_PORT	(PCP_OPCODE_HDR + 6)
#define PCP_PEER_EXTERNAL_ADDR	(PCP_OPCODE_HDR + 8)
#define PCP_PEER_REMOTE_PORT	(PCP_OPCODE_HDR + 24)
#define PCP_PEER_REMOTE_ADDR	(PCP_OPCODE_HDR + 28)


/* offsets in option */
#define PCP_OPT_CODE		0
#define PCP_OPT_LEN		2
#define PCP_OPT_DATA		4
#define PCP_OPTDFLT_RSVD	0
#define PCP_OPTDFLT_PLEN	1
#define PCP_OPTDFLT_PORT	2
#define PCP_OPTDFLT_PREF	4

/* lengths */
#define PCP_LEN_ANNOUNCE	24
#define PCP_LEN_MAP		48
#define PCP_LEN_PEER		68
#define PCP_LEN_OPTHDR		4
#define PCP_LEN_OPT0		0
#define PCP_LEN_OPT3PTY		16
#define PCP_LEN_OPTFLT		20

#define PCP_ALIGN(x)		(((x) + 3) & ~3)

/* Errors */

/* OK */
#define PCP_OK			(0)	/* OK */
#define PCP_TRY_AGAIN		(1)	/* try again */
/* bad API use */
#define PCP_ERR_INVAL		(-1)	/* invalid arguments */
/* OS errors */
#define PCP_ERR_NOMEM		(-10)	/* malloc() failed */
#define PCP_ERR_SOCKET		(-11)	/* socket() syscall failed */
#define PCP_ERR_BIND		(-12)	/* bind() syscall failed */
#define PCP_ERR_CONNECT		(-13)	/* connect() syscall failed */
#define PCP_ERR_SEND		(-14)	/* send() syscall failed */
#define PCP_ERR_RECV		(-15)	/* recv() syscall failed */
#define PCP_ERR_SYSCALL		(-16)	/* miscellaneous syscall failed */
/* routine errors */
#define PCP_ERR_NOREQUEST	(-20)	/* no current request */
#define PCP_ERR_RECVBAD		(-21)	/* received a bad response */
#define PCP_ERR_TOOMANYOPTS	(-22)	/* too many options */
/* *mapping failure */
#define PCP_ERR_FAILURE		(-30)	/* *mapping() internal failure */
/* user application */
#define PCP_ERR_APP0		(-50)	/* user application error 0 */
#define PCP_ERR_APP1		(-51)	/* user application error 1 */
#define PCP_ERR_APP2		(-52)	/* user application error 2 */
/* protocol errors */
#define PCP_ERR_PROTOBASE	(-100)	/* base for protocol errors */
#define PCP_ERR_UNSUPVERSION	(-101)	/* unsupported version */
#define PCP_ERR_NOTAUTH		(-102)	/* not authorized */
#define PCP_ERR_BADREQUEST	(-103)	/* malformed request */
#define PCP_ERR_UNSUPOPCODE	(-104)	/* unsupported opcode */
#define PCP_ERR_UNSUPOPTION	(-105)	/* unsupported option */
#define PCP_ERR_BADOPTION	(-106)	/* malformed option */
#define PCP_ERR_NETFAILURE	(-107)	/* network failure */
#define PCP_ERR_NORESOURCES	(-108)	/* out of resources */
#define PCP_ERR_UNSUPPROTO	(-109)	/* unsupported protocol */
#define PCP_ERR_EXQUOTA		(-110)	/* user exceeded quota */
#define PCP_ERR_CANTPROVIDE	(-111)	/* cannot provide external */
#define PCP_ERR_ADDRMISMATCH	(-112)	/* address mismatch */
#define PCP_ERR_TOOMANYPEER	(-113)	/* excessive number of remote peers */
#define PCP_ERR_PROCERROR	(-292)	/* processing error */

/* Others */

/* Default port of PCP service */

#define PCP_PORT		5351

/* Default retry timer (in seconds) */

#define PCP_RETRY_TIMER		2

/* Maximum number of tries (can be set to 0 to disable it) */

#ifndef PCP_MAX_TRIES
#define PCP_MAX_TRIES		3
#endif

/* Maximum timeout between 2 retransmits: 15 mn */

#define PCP_MAX_TIMEOUT		900

/* Default reset threshold (in seconds) */

#ifndef PCP_RESET_THRESHOLD
#define PCP_RESET_THRESHOLD	60
#endif

/* Rapid Recovery */

#define PCP_RC_PORT		5350
#define PCP_RC_ADDR4		"224.0.0.1"
#define PCP_RC_ADDR6		"FF02::1"
#define PCP_RC_INITTRMT		250000
#define PCP_RC_MAXTRMT		10

/* Tunables & co */

/* weak definition so you can overwrite them */

#ifdef WIN32
#define pcp_malloc(s)	malloc(s)
#define pcp_free(p)	free(p)
#else
extern void *pcp_malloc(size_t size);
extern void pcp_free(void *ptr);
#endif

/* inline utility to free allocated options */

static pcp_inline void pcp_freeoptions(pcp_option_t **options)
{
	if (options != NULL) {
		unsigned int i;

		for (i = 0; options[i] != NULL; i++)
			pcp_free(options[i]);
		pcp_free(options);
	}
}

extern const char *err_app0_msg;
extern const char *err_app1_msg;
extern const char *err_app2_msg;

/* for bad response debugging */

#if defined(WIN32) && !defined(STATICLIB)
#ifdef PCP_EXPORTS
#define LIBSPEC __declspec(dllexport)
#else
#define LIBSPEC __declspec(dllimport)
#endif
#else
#define LIBSPEC
#endif

LIBSPEC int pcp_debug_offset;
LIBSPEC int pcp_debug_line;

/* support for a minimal pcp.conf file */

#ifdef WIN32
#define PCP_CONFFILE		"\\Drivers\\etc\\pcp.conf"
#else
#define PCP_CONFFILE		"/etc/pcp.conf"
#endif

#define PCP_NOCONFFILE		PCP_TRY_AGAIN

extern int _pcp_config(struct sockaddr_storage *server, socklen_t *len);

#ifdef __cplusplus
}
#endif

#endif
