/*
 * Copyright (C) 2010-2012  Internet Systems Consortium, Inc. ("ISC")
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND ISC DISCLAIMS ALL WARRANTIES WITH
 * REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS.  IN NO EVENT SHALL ISC BE LIABLE FOR ANY SPECIAL, DIRECT,
 * INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
 * OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */

/* $Id: 75400bf16512ca6632bc0d1ea00065a0245b3ba1 $ */

/* pseudo-firewall stuff for PCP */

#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <unistd.h>
#include <sys/errno.h>
#include <sys/socket.h>
#include <net/if.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "upnpglobalvars.h"
#include "upnpredirect.h"
#include "pcpiwf.h"

static uint32_t cpe;
static uint32_t hold;
static uint32_t lasttm;
static struct timeval last;

const char *err_app0_msg = "inet_pton failure";

/* OpenPCPServerSocket() :
 * setup the socket with the PCP server. */
int
OpenPCPServerSocket(struct transaction *t)
{
	struct sockaddr_in6 srv6;
	struct sockaddr_in6 loc6;

	memset(&srv6, 0, sizeof(srv6));
	srv6.sin6_family = AF_INET6;
	srv6.sin6_port = htons(PCP_PORT);
	if (inet_pton(AF_INET6, pcp_server, &srv6.sin6_addr) <= 0) {
		syslog(LOG_ERR,
		       "Failed to convert '%s' to local IPv6 address",
		       pcp_local);
		return PCP_ERR_APP0;
	}
	memset(&loc6, 0, sizeof(loc6));
	loc6.sin6_family = AF_INET6;
	if (inet_pton(AF_INET6, pcp_local, &loc6.sin6_addr) <= 0) {
		syslog(LOG_ERR,
		       "Failed to convert '%s' to server IPv6 address",
		       pcp_server);
		return PCP_ERR_APP0;		
	}
	return pcp_init((struct sockaddr *) &srv6,
			(struct sockaddr *) &loc6,
			&t->pcp);
}

static int
OpenPCPClientSocket(in_addr_t addr)
{
	int fd;
	struct sockaddr_in sin4;

	fd = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);
	if (fd < 0) {
		syslog(LOG_ERR, "socket(pcp): %m");
		goto bad;
	}
	memset(&sin4, 0, sizeof(sin4));
	sin4.sin_family = AF_INET;
	sin4.sin_addr.s_addr = addr;
	sin4.sin_port = htons(PCP_PORT);
	if (bind(fd, (struct sockaddr *) &sin4, sizeof(sin4)) < 0) {
		syslog(LOG_ERR, "bind(pcp): %m");
		goto bad;
	}
	if (setsockopt(fd, IPPROTO_IP, IP_MULTICAST_IF,
		       &sin4.sin_addr, sizeof(struct in_addr)) < 0) {
		syslog(LOG_ERR, "MULTICAST_IF(pcp): %m");
		goto bad;
	}

	return fd;
    bad:
	if (fd >= 0)
		(void)close(fd);
	return -1;
}

int
OpenPCPSockets(int *sockets)
{
	int i, j;
	struct lan_addr_s *lan_addr;

	hold = htonl(1800);

	for (i = 0, lan_addr = lan_addrs.lh_first;
	     lan_addr != NULL;
	     lan_addr = lan_addr->list.le_next, i++) {
		sockets[i] = OpenPCPClientSocket(lan_addr->addr.s_addr);
		if (sockets[i] < 0) {
			for (j = 0; j < i; j++) {
				(void)close(sockets[j]);
				sockets[j] = -1;
			}
			sockets[i] = -1;
			return -1;
		}
	}
	return 0;
}

int
OpenPCPRapidRecovery(void)
{
	struct sockaddr_in6 srv;
	struct ipv6_mreq mr;
	int fd, flags, on;

	memset(&srv, 0, sizeof(srv));
	srv.sin6_family = AF_INET6;
	srv.sin6_port = htons(PCP_RC_PORT);
	memset(&mr, 0, sizeof(mr));
	if (inet_pton(AF_INET6, PCP_RC_ADDR6, &mr.ipv6mr_multiaddr) <= 0) {
		syslog(LOG_ERR,
		       "Failed to convert '%s' to IPv6 address",
		       PCP_RC_ADDR6);
		return PCP_ERR_APP0;
	}
	mr.ipv6mr_interface = if_nametoindex(ext_if_name);
	if (mr.ipv6mr_interface == 0) {
		syslog(LOG_ERR,
		       "Failed to convert '%s' to interface index",
		       ext_if_name);
		return PCP_ERR_APP0;
	}
	fd = socket(PF_INET6, SOCK_DGRAM, IPPROTO_UDP);
	if (fd < 0)
		return PCP_ERR_SOCKET;
	if (((flags = fcntl(fd, F_GETFL, 0)) < 0) ||
	    (fcntl(fd, F_SETFL, flags | O_NONBLOCK) < 0)) {
		close(fd);
		return PCP_ERR_SYSCALL;
	}
	on = 1;
	/* hey, Linux guys: this MUST be REUSEPORT! */
#ifndef SO_REUSEPORT
#define SO_REUSEPORT	SO_REUSEADDR
#endif
	if (setsockopt(fd, SOL_SOCKET, SO_REUSEPORT, &on, sizeof(on)) < 0) {
		syslog(LOG_ERR, "Failed to set REUSEPORT");
		close(fd);
		return PCP_ERR_BIND;
	}
	if (bind(fd, (struct sockaddr *) &srv, sizeof(srv)) < 0) {
		syslog(LOG_ERR, "Failed to bind to %d", PCP_RC_PORT);
		close(fd);
		return PCP_ERR_BIND;
	}
	if (setsockopt(fd, IPPROTO_IPV6, IPV6_JOIN_GROUP,
		       &mr, sizeof(mr)) < 0) {
		syslog(LOG_ERR, "Failed to join '%s'", PCP_RC_ADDR6);
		close(fd);
		return PCP_ERR_SYSCALL;
	}
	return fd;
}

void
SendPCPNotification(int *spcp)
{
	struct lan_addr_s *lan_addr;
	struct sockaddr_in sockname;
	uint8_t notif[24];
	uint32_t epoch;
	int i, n;

	memset(notif, 0, sizeof(notif));
	/* version (1) */
	notif[PCP_VERSION] = 1;
	/* opcode (128) */
	notif[PCP_OPCODE] = PCP_OPCODE_ANNOUNCE | PCP_OPCODE_RESPONSE;
	/* seconds since "start of epoch" :
	 * time elapsed since the port mapping table was initialized on
	 * startup or reset for any other reason */
	epoch = htonl(time(NULL) - startup_time);
	memcpy(notif + PCP_EPOCH, &epoch, 4);

	memset(&sockname, 0, sizeof(sockname));
	sockname.sin_family = AF_INET;
	if (inet_pton(AF_INET, PCP_RC_ADDR4, &sockname.sin_addr) <= 0) {
		syslog(LOG_WARNING, "%s: inet_pton(%s)",
		       "SendPCPNotification", PCP_RC_ADDR4);
		return;
	}
	for (i = 0, lan_addr = lan_addrs.lh_first;
	     lan_addr != NULL;
	     lan_addr = lan_addr->list.le_next, i++) {
		if (spcp[i] < 0)
			continue;
		sockname.sin_port = htons(PCP_RC_PORT);
		n = sendto(spcp[i], notif, sizeof(notif), 0,
			   (struct sockaddr *) &sockname, sizeof(sockname));
		if (n < 0) {
			syslog(LOG_ERR, "%s: sendto(%d): %m",
			       "SendPCPNotification", spcp[i]);
			continue;
		}
	}
}

static void
pcp_boot_resend(void *arg)
{
	struct transaction *t = (struct transaction *) arg;
	int ret;

	t->status = PCP_TRY_AGAIN;
	ret = pcp_sendrequest(&t->pcp);
	if (ret != PCP_OK)
		syslog(LOG_WARNING, "pcp_boot_resend: %s", pcp_strerror(ret));
	t->retrans.when.tv_sec = now.tv_sec + 60;
	t->retrans.when.tv_usec = now.tv_usec;
	LIST_INSERT_HEAD(&timeouts, &t->retrans, chain);
}

static void
pcp_boot_renew(void *arg)
{
	struct transaction *t = (struct transaction *) arg;
	pcp_request_t req;
	pcp_option_t **options = NULL;

	syslog(LOG_NOTICE, "boot renew");
	memset(&req, 0, sizeof(req));
	/* opcode (map) */
	req.opcode = PCP_OPCODE_MAP;
	/* protocol (udp) */
	req.protocol = IPPROTO_UDP;
	/* lifetime (1 day) */
	req.lifetime = 86400U;
	/* internal port (9) */
	req.intport = 9;
	/* external port (0) */
	/* external address (0.0.0.0) */
	pcp_mapextaddr(&req);
	/* third party (192.0.0.2) */
	if (pcp_third_party4(&options, cpe) != PCP_OK)
		goto bad;
	if (pcp_makerequest(&t->pcp, &req, options) != PCP_OK)
		goto bad;
	pcp_freeoptions(options);
	pcp_send_request(t);
	return;

    bad:
	syslog(LOG_ERR, "boot renew failed");
	pcp_freeoptions(options);
	t->status = PCP_ERR_FAILURE;
	(t->handler)(t);
	return;
}

static void
pcp_boot_handler(struct transaction *t)
{
	uint32_t lt;

	if (t->status > PCP_OK)
		return;
	else if (t->status < PCP_OK) {
		syslog(LOG_WARNING,
		       "pcp_boot_handler: %s",
		       pcp_strerror(t->status));
		/* reset retrans timeout */
		t->retrans.action = pcp_boot_resend;
		t->retrans.arg = t;
		t->retrans.when.tv_sec = now.tv_sec + 60;
		t->retrans.when.tv_usec = now.tv_usec;
		LIST_INSERT_HEAD(&timeouts, &t->retrans, chain);
		return;
	}
	syslog(LOG_NOTICE, "boot succeed");
	/* reuse retrans timeout for renewal */
	t->retrans.action = pcp_boot_renew;
	t->retrans.arg = t;
	lt = t->response.assigned.lifetime;
	if (lt > 3600)
		lt = 3600;
	t->retrans.when.tv_sec = now.tv_sec + lt;
	t->retrans.when.tv_usec = now.tv_usec;
	LIST_INSERT_HEAD(&timeouts, &t->retrans, chain);
	return;
}

void
pcp_boot(int spcprc)
{
	struct transaction *t;
	pcp_request_t req;
	pcp_option_t **options = NULL;

	/* must not fail */
	if (cpe == 0)
		(void) inet_pton(AF_INET, "192.0.0.2", &cpe);
	if (cpe == 0)
		return;
	t = (struct transaction *) malloc(sizeof(*t));
	memset(t, 0, sizeof(*t));
	t->type = TTYPE_SYSTEM;
	if (OpenPCPServerSocket(t) != PCP_OK) {
		free(t);
		return;
	}
	if (spcprc >= 0)
		(void) pcp_setrcsocket(&t->pcp, spcprc);
	(void) pcp_setreset(&t->pcp, 1, NULL);
	t->status = PCP_NOT_READY;

	memset(&req, 0, sizeof(req));
	/* opcode (map) */
	req.opcode = PCP_OPCODE_MAP;
	/* protocol (udp) */
	req.protocol = IPPROTO_UDP;
	/* lifetime (1 day) */
	req.lifetime = 86400U;
	/* internal port (9) */
	req.intport = 9;
	/* external port (0) */
	/* external address (0.0.0.0) */
	pcp_mapextaddr(&req);
	/* third party (192.0.0.2) */
	if (pcp_third_party4(&options, cpe) != PCP_OK)
		goto bad;
	if (pcp_makerequest(&t->pcp, &req, options) != PCP_OK)
		goto bad;
	pcp_freeoptions(options);

	LIST_INSERT_HEAD(&transactions, t, chain);
	t->handler = pcp_boot_handler;
	pcp_send_request(t);
	return;

    bad:
	(void) pcp_close(&t->pcp);
	free(t);
	pcp_freeoptions(options);
	return;
}

struct client_request {
	struct transaction t;
	int s;
	struct in_addr client_addr;
	unsigned short client_port;
	unsigned short source_port;
	struct in6_addr source_addr;
	struct timeout expire;
};

static void
pcp_request_expire(void *arg)
{
	struct client_request *c = (struct client_request *) arg;

	(void) pcp_close(&c->t.pcp);
	LIST_REMOVE(&c->t, chain);
	if (c->t.retrans.action != NULL)
		LIST_REMOVE(&c->t.retrans, chain);
	free(c);
}

static uint16_t
in_cksum(uint8_t *p, u_int l)
{
	u_int sum = 0;

	while (l > 1) {
		sum += *p++ << 8;
		sum += *p++;
		l -= 2;
	}
	if (l == 1)
		sum += *p << 8;
	sum = ((sum >> 16) & 0xffff) + (sum & 0xffff);
	sum += sum >> 16;
	return htons((uint16_t) ~sum);
}

static void
pcp_icmp(struct client_request *c)
{
	struct sockaddr_in addr;
	uint8_t buf[36];
	socklen_t l;
	uint16_t cksum;
	int r;

	r = socket(PF_INET, SOCK_RAW, IPPROTO_ICMP);
	if (r < 0) {
		syslog(LOG_ERR, "pcp_icmp: socket");
		return;
	}
	memset(&addr, 0, sizeof(addr));
	l = sizeof(addr);
	if (getsockname(c->s, (struct sockaddr *) &addr, &l) < 0) {
		syslog(LOG_ERR, "pcp_icmp: getsockname: %m");
		return;
	}

	memset(buf, 0, sizeof(buf));
	/* type = UNREACH */
	buf[0] = 3;
	/* code = UNREACH_PORT */
	buf[1] = 3;
	/* checksum (use UDP one), pptr/gwaddr/void */
	/* ip[0] = version | hlen */
	buf[8] = 0x45;
	/* tos, len = 20+8+28 */
	buf[11] = 56;
	/* id, off, ttl */
	buf[16] = 64;
	/* protocol */
	buf[17] = IPPROTO_UDP;
	/* checksum, source */
	memcpy(buf + 20, &c->client_addr, 4);
	/* destination */
	memcpy(buf + 24, &addr.sin_addr, 4);
	/* UDP source port */
	memcpy(buf + 28, &c->client_port, 2);
	/* UDP destination port */
	memcpy(buf + 30, &addr.sin_port, 2);
	/* UDP length = 8+28 */
	buf[33] = 36;
	/* IP checksum */
	cksum = in_cksum(buf + 8, 20);
	memcpy(buf + 18, &cksum, 2);
	/* UDP checksum (to adjust ICMP one) */
	cksum = in_cksum(buf, sizeof(buf));
	memcpy(buf + 34, &cksum, 2);

	addr.sin_port = 0;
	if (bind(r, (struct sockaddr *) &addr, l) < 0) {
		syslog(LOG_ERR, "pcp_icmp: bind");
		return;
	}
	addr.sin_addr = c->client_addr;
	if (sendto(r, buf, sizeof(buf), 0, (struct sockaddr *) &addr, l) < 0)
		syslog(LOG_ERR, "pcp_icmp: sendto");
	(void) close(r);
}

static void
pcp_request_handler(struct transaction *t)
{
	struct client_request *c = t->parent;
	uint8_t *resp = NULL;
	uint16_t resplen;
	unsigned int optidx, off;
	struct sockaddr_in client;

	if (t->status > PCP_OK)
		return;
	/* special case for recv() syscall error */
	if (t->status == PCP_ERR_RECV) {
		pcp_icmp(c);
		return;
	}
	/* not protocol errors */
	if ((t->status < PCP_OK) && (t->status > PCP_ERR_PROTOBASE))
		return;
	if (_pcp_getresponse(&t->pcp, &resp, &resplen) != PCP_OK) {
		syslog(LOG_ERR, "pcp_request_handler(getresponse)");
		return;
	}
	/* get back client address */
	memset(&client, 0, sizeof(client));
	client.sin_family = AF_INET;
	client.sin_addr.s_addr = c->client_addr.s_addr;
	client.sin_port = c->client_port;
	/* pop third party */
	for (optidx = 0; optidx < t->response.optcnt; optidx++) {
		if (t->response.options[optidx].code != PCP_OPTION_THIRD_PARTY)
			continue;
		if (t->response.options[optidx].length != PCP_LEN_OPT3PTY)
			continue;
		if (memcmp(t->response.options[optidx].data,
			   pcp_mapped, 12) != 0)
			continue;
		if (memcmp(t->response.options[optidx].data + 12,
			   &client.sin_addr, 4) != 0)
			continue;
		off = t->response.optoffs[optidx];
#define OPTLEN	(PCP_OPT_DATA + PCP_LEN_OPT3PTY)
		if (off + OPTLEN > resplen)
			memmove(resp + off,
				resp + off + OPTLEN,
				resplen - (off + OPTLEN));
		resplen -= OPTLEN;
#undef OPTLEN
		break;
	}
	/* send response back to the client */
	if (sendto(c->s, resp, (size_t) resplen, 0,
		   (struct sockaddr *) &client, sizeof(client)) < 0)
		syslog(LOG_ERR, "sendto(pcp): %m");
}

void
ProcessPCPRequest(int s)
{
	struct transaction *t;
	struct client_request *c;
	unsigned char req[1048];
	struct sockaddr_in client;
	struct sockaddr_in6 source;
	socklen_t clen = sizeof(client);
	char clientstr[INET_ADDRSTRLEN];
	int cc, ret, fd;

	memset(req, 0, sizeof(req));
	cc = recvfrom(s, req, sizeof(req), 0,
		      (struct sockaddr *) &client, &clen);
	if (cc < 0) {
		syslog(LOG_ERR, "recvfrom(pcp): %m");
		return;
	}
	if (cc < 2) {
		syslog(LOG_WARNING, "ProcessPCPRequest(underrun): %d", cc);
		return;
	}
	/* discarding PCP responses silently */
	if ((req[PCP_OPCODE] & PCP_OPCODE_RESPONSE) != 0)
		return;
	if (!inet_ntop(AF_INET, &client.sin_addr,
		       clientstr, INET_ADDRSTRLEN)) {
		syslog(LOG_ERR, "ProcessPCPRequest(inet_ntop)");
		return;
	}
	syslog(LOG_INFO, "PCP request received from %s:%hu %d bytes",
	       clientstr, ntohs(client.sin_port), cc);
	if (req[PCP_VERSION] != 1) {
		/* UNSUPP_VERSION */
		req[PCP_RESULT_CODE] = 1;
		cc = 4;
		goto err;
	}
	if (cc < 4) {
		syslog(LOG_WARNING, "ProcessPCPRequest(underrun2): %d", cc);
		return;
	}
	if (cc > 1024) {
		syslog(LOG_WARNING, "ProcessPCPRequest(overrun): %d", cc);
		/* just let 4 extra octets go through */
		if (cc > 1028)
			cc = 1028;
	}
	/* let (cc & 3) != 0 go through */
	if (cc < PCP_OPCODE_HDR) {
		/* MALFORMED_REQUEST */
		req[PCP_RESULT_CODE] = 2;
		cc = PCP_OPCODE_HDR;
		goto errhdr;
	}
	if ((memcmp(req + PCP_CLIENT_ADDR, pcp_mapped, 12) != 0) ||
	    (memcmp(req + PCP_CLIENT_ADDR + 12, &client.sin_addr, 4) != 0)) {
		/* ADDRESS_MISMATCH */
		req[PCP_RESULT_CODE] = 7;
		goto errhdr;
	}

	/* nothing for secure mode as the THIRD_PARTY will always fail */

	/* check authorization for MAP create requests */
	if ((req[PCP_OPCODE] == PCP_OPCODE_MAP) && (cc >= PCP_LEN_MAP)) {
		uint32_t lifetime;
		uint16_t iport, eport;
		
		memcpy(&lifetime, req + PCP_LIFETIME, 4);
		lifetime = ntohl(lifetime);
		memcpy(&iport, req + PCP_MAP_INTERNAL_PORT, 2);
		iport = ntohs(iport);
		memcpy(&eport, req + PCP_MAP_EXTERNAL_PORT, 2);
		if (eport != 0)
			eport = ntohs(eport);
		else
			eport = iport;
		if ((lifetime != 0) &&
		    !check_permissions(upnppermlist, num_upnpperm,
				       eport, client.sin_addr, iport)) {
			/* NOT_AUTHORIZED */
			req[PCP_RESULT_CODE] = 2;
			goto errhdr;
		}
	}

	/* known? */
	LIST_FOREACH(t, &transactions, chain) {
		uint8_t *p = NULL;
		uint16_t l;

		if (t->type != TTYPE_PCP)
			continue;
		c = t->parent;
		if ((c->s == s) &&
		    (c->client_addr.s_addr == client.sin_addr.s_addr) &&
		    (c->client_port == client.sin_port) &&
		    (_pcp_getrequest(&t->pcp, &p, &l) == PCP_OK) &&
		    ((int) l == cc) &&
		    (memcmp(p, req, (size_t) l) == 0)) {
			if (t->status <= 0)
				pcp_request_handler(t);
			return;
		}
	}

	/* new one */
	c = (struct client_request *)malloc(sizeof(*c));
	if (c == NULL)
		return;
	memset(c, 0, sizeof(*c));
	/* fill client request structure */
	ret = OpenPCPServerSocket(&c->t);
	if (ret != PCP_OK) {
		syslog(LOG_WARNING,
		       "ProcessPCPRequest(init): %s",
		       pcp_strerror(ret));
		free(c);
		return;
	}
	/* copied from _pcp_fillsrcaddr */
	ret = pcp_getsocket(&c->t.pcp, &fd);
	if (ret != PCP_OK) {
		syslog(LOG_WARNING,
		       "ProcessPCPRequest(getsocket): %s",
		       pcp_strerror(ret));
		free(c);
		return;
	}
	memset(&source, 0, sizeof(source));
	clen = sizeof(source);
	if (getsockname(fd, (struct sockaddr *) &source, &clen) != 0) {
		syslog(LOG_WARNING, "ProcessPCPRequest(getsockname): %m");
		free(c);
		return;
	}
	memcpy(&c->source_addr, &source.sin6_addr, 16);
	memcpy(&c->source_port, &source.sin6_port, 2);
	c->t.parent = c;
	c->t.type = TTYPE_PCP;
	c->t.status = PCP_NOT_READY;
	c->s = s;
	c->client_addr.s_addr = client.sin_addr.s_addr;
	c->client_port = client.sin_port;
	memcpy(req + PCP_CLIENT_ADDR, &c->source_addr, 16);
	/* add a third party option if not announce request */
	if (req[PCP_OPCODE] != PCP_OPCODE_ANNOUNCE) {
		req[cc + PCP_OPT_CODE] = PCP_OPTION_THIRD_PARTY;
		req[cc + PCP_OPT_CODE + 1] = req[cc + PCP_OPT_LEN] = 0;
		req[cc + PCP_OPT_LEN + 1] = PCP_LEN_OPT3PTY;
		memcpy(req + cc + PCP_OPT_DATA, pcp_mapped, 12);
		memcpy(req + cc + PCP_OPT_DATA + 12, &client.sin_addr, 4);
		cc += PCP_OPT_DATA + PCP_LEN_OPT3PTY;
	}
	if (_pcp_setrequest(&c->t.pcp, req, (uint16_t) cc) != PCP_OK) {
		(void) pcp_close(&c->t.pcp);
		free(c);
		return;
	}
	c->expire.when.tv_sec = now.tv_sec + 120;
	c->expire.when.tv_usec = now.tv_usec;
	c->expire.action = pcp_request_expire;
	c->expire.arg = c;
	LIST_INSERT_HEAD(&timeouts, &c->expire, chain);

	/* booting */
	if (external_addr == 0) {
		c->t.status = PCP_ERR_RECV;
		pcp_icmp(c);
	}

	/* forward */
	LIST_INSERT_HEAD(&transactions, &c->t, chain);
	c->t.handler = pcp_request_handler;
	pcp_send_request(&c->t);
	return;

    errhdr:
	/* hold lifetime (1800s) */
	memcpy(req + PCP_LIFETIME, &hold, 4);
	/* epoch */
	if (lasttm != 0) {
		uint32_t epoch;

		epoch = lasttm + (now.tv_sec - last.tv_sec);
		if ((now.tv_usec - last.tv_usec) >= 500000)
			epoch++;
		else if ((now.tv_usec - last.tv_usec) < -500000)
			epoch--;
		epoch = htonl(epoch);
		memcpy(req + PCP_EPOCH, &epoch, 4);
	} else
		memset(req + PCP_EPOCH, 0, 4);
	/* rest of client address */
	memset(req + PCP_EPOCH + 4, 0, 12);

    err:
	/* send error response back to the client */
	req[PCP_OPCODE] |= PCP_OPCODE_RESPONSE;
	req[PCP_OPCODE + 1] = 0;
	if (sendto(s, req, (size_t) cc, 0,
		   (struct sockaddr *) &client, sizeof(client)) < 0)
		syslog(LOG_ERR, "sendto(err): %m");
}

void
check_epoch(uint32_t tm)
{
	uint32_t cdelta, sdelta;

	/* protect against not initialized and clock going backward */
	if ((lasttm != 0) &&
	    ((now.tv_sec > last.tv_sec) ||
	     ((now.tv_sec == last.tv_sec) && (now.tv_usec >= last.tv_usec)))) {
		if (tm < lasttm) {
			syslog(LOG_ERR, "epoch: going backward");
			exit(0);
		}
		sdelta = tm - lasttm;
		cdelta = now.tv_sec - last.tv_sec;
		if ((now.tv_usec - last.tv_usec) >= 500000)
			cdelta++;
		else if ((now.tv_usec - last.tv_usec) < -500000)
			cdelta--;
		if ((cdelta + 2 < (((sdelta << 4) - sdelta) >> 4)) ||
		    (sdelta + 2 < (((cdelta << 4) - cdelta) >> 4)) ||
		    (sdelta + 3600 < cdelta) ||
		    (cdelta + 3600 < sdelta)) {
			syslog(LOG_ERR, "epoch: server has rebootted");
			exit(0);
		}
	}
	lasttm = tm;
	last = now;
}

void
get_external_address(uint8_t *buf, uint16_t len)
{
	in_addr_t addr;
	struct in_addr in;

	if ((len < PCP_LEN_MAP) ||
	    (buf[PCP_VERSION] != 1) ||
	    (buf[PCP_OPCODE] != (PCP_OPCODE_MAP | PCP_OPCODE_RESPONSE))) {
		syslog(LOG_WARNING, "get_external_address: skip");
		return;
	}
	if (memcmp(buf + PCP_MAP_EXTERNAL_ADDR, pcp_mapped, 12) != 0) {
		syslog(LOG_WARNING, "not mapped external address: skip");
		return;
	}
	memcpy(&addr, buf + PCP_MAP_EXTERNAL_ADDR + 12, 4);
	if (addr == 0U) {
		syslog(LOG_WARNING, "no external address: skip");
		return;
	}
	if (external_addr != 0) {
		if (external_addr == addr)
			return;
		syslog(LOG_ERR, "external address: server has rebootted");
		exit(0);
	}
	external_addr = addr;
	in.s_addr = addr;
	syslog(LOG_NOTICE, "new external address '%s'", inet_ntoa(in));
	should_send_public_address_change_notif = 1;
}

void
ProcessPCPResponse(struct transaction *t)
{
	uint8_t *buf;
	uint16_t len;

	/* no transaction: rapid recovery, using the system one */
	if (t == NULL) {
		LIST_FOREACH(t, &transactions, chain)
			if (t->type == TTYPE_SYSTEM)
				break;
		/* sanity: should not happen */
		if (t == NULL)
			return;
		t->status = pcp_recvresponse(&t->pcp, &t->response);
		if (t->response.assigned.opcode ==
		    (PCP_OPCODE_ANNOUNCE | PCP_OPCODE_RESPONSE)) {
			check_epoch(t->response.epoch);
			return;
		}
	} else
		t->status = pcp_recvresponse(&t->pcp, &t->response);
	/* common processing */
	if (t->status == PCP_TRY_AGAIN)
		return;
	else if (t->status < PCP_OK)
		syslog(LOG_WARNING,
		       "ProcessPCPResponse: %s",
		       pcp_strerror(t->status));

	/* check server reboot */
	if ((t->status == PCP_OK) || (t->status <= PCP_ERR_PROTOBASE))
		check_epoch(t->response.epoch);
	if (((t->status == PCP_OK) || (t->status <= PCP_ERR_PROTOBASE)) &&
	    (_pcp_getresponse(&t->pcp, &buf, &len) == PCP_OK))
		get_external_address(buf, len);

	/* cancel retrans */
	if (t->retrans.action)
		LIST_REMOVE(&t->retrans, chain);
	memset(&t->retrans, 0, sizeof(t->retrans));

	if (t->handler != NULL)
		(t->handler)(t);
}

static void
pcp_resend(void *arg)
{
	struct transaction *t = (struct transaction *) arg;
	int ret, trycnt;

	ret = pcp_gettries(&t->pcp, &trycnt);
	if (ret < PCP_OK) {
		syslog(LOG_WARNING, "pcp_resend(pcp_gettries)");
		t->status = ret;
		return;
	} else if (trycnt > PCP_MAX_TRIES) {
		memset(&t->retrans, 0, sizeof(t->retrans));
		t->status = PCP_ERR_FAILURE;
		if (t->handler != NULL)
			(t->handler)(t);
		return;
	}
	t->status = PCP_TRY_AGAIN;
	ret = pcp_sendrequest(&t->pcp);
	if (ret != PCP_OK)
		syslog(LOG_WARNING, "pcp_resend: %s", pcp_strerror(ret));
	/* can't fail */
	(void) pcp_gettimeout(&t->pcp, &t->retrans.when, 0);
	LIST_INSERT_HEAD(&timeouts, &t->retrans, chain);
}

void
pcp_send_request(struct transaction *t)
{
	int ret;

	/* set status */
	t->status = PCP_TRY_AGAIN;
	/* send */
	ret = pcp_sendrequest(&t->pcp);
	if (ret != PCP_OK)
		syslog(LOG_WARNING, "pcp_send_request: %s", pcp_strerror(ret));
	/* set retrans timeout */
	t->retrans.action = pcp_resend;
	t->retrans.arg = t;
	/* can't fail */
	(void) pcp_gettimeout(&t->pcp, &t->retrans.when, 0);
	LIST_INSERT_HEAD(&timeouts, &t->retrans, chain);
}
