/*
 * Copyright (C) 2012  Internet Systems Consortium, Inc. ("ISC")
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND ISC DISCLAIMS ALL WARRANTIES WITH
 * REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS.  IN NO EVENT SHALL ISC BE LIABLE FOR ANY SPECIAL, DIRECT,
 * INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
 * OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */

/* $Id: deb35c790f6638f61736735945635e422987651d $ */

/*
 * PCP (port-control-protocol) client library simple test client (ANNOUNCE)
 *
 * Francis_Dupont@isc.org, January 2012
 *
 * ./announce <source> <server>
 */

#ifdef WIN32
#include <ws2tcpip.h>
#else
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#endif
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "pcp.h"

int
main(int argc, char *argv[])
{
	pcp_t pcp;
	pcp_request_t request;
	pcp_option_t *opt;
	pcp_response_t response;
	int ret, fd, tries;
	fd_set set;
	struct sockaddr_in source, server;
	struct timeval tv;
	struct in_addr in;
	char host6[INET6_ADDRSTRLEN];
	unsigned int optidx;
	uint16_t n16;
#ifdef WIN32
	WSADATA wsaData;
#endif

	if (argc != 3) {
		fprintf(stderr, "usage: %s <source> <server>\n", argv[0]);
		exit(1);
	}
#ifdef WIN32
	ret = WSAStartup(MAKEWORD(2,2), &wsaData);
	if (ret != 0) {
		fprintf(stderr, "WSAStartup() failed.\n");
		exit(1);
	}
#endif
	memset(&server, 0, sizeof(server));
	server.sin_family = AF_INET;
	server.sin_addr.s_addr = inet_addr(argv[2]);
	memset(&source, 0, sizeof(source));
	source.sin_family = AF_INET;
	source.sin_addr.s_addr = inet_addr(argv[1]);
	ret = pcp_init((struct sockaddr *) &server,
		       (struct sockaddr *) &source,
		       &pcp);
	if (ret < PCP_OK) {
		fprintf(stderr, "pcp_init: %s\n", pcp_strerror(ret));
		exit(1);
	}
	ret = pcp_getsocket(&pcp, &fd);
	if (ret < PCP_OK) {
		fprintf(stderr, "pcp_getsocket: %s\n", pcp_strerror(ret));
		exit(1);
	}
	memset(&request, 0, sizeof(request));
	request.opcode = PCP_OPCODE_ANNOUNCE;
	ret = pcp_makerequest(&pcp, &request, NULL);
	if (ret < PCP_OK) {
		fprintf(stderr, "pcp_makerequest: %s\n", pcp_strerror(ret));
		exit(1);
	}
	ret = pcp_sendrequest(&pcp);
	if (ret < PCP_OK) {
		fprintf(stderr, "pcp_sendrequest: %s\n", pcp_strerror(ret));
		exit(1);
	}
	for (;;) {
		ret = pcp_gettries(&pcp, &tries);
		if (ret < PCP_OK) {
			fprintf(stderr,
				"pcp_gettries: %s\n",
				pcp_strerror(ret));
			exit(1);
		}
		if ((PCP_MAX_TRIES > 0) && (tries > PCP_MAX_TRIES)) {
			fprintf(stderr, "too many retries: giving up\n");
			exit(1);
		}
		FD_ZERO(&set);
		FD_SET(fd, &set);
		ret = pcp_gettimeout(&pcp, &tv, 1);
		if (ret < PCP_OK) {
			fprintf(stderr,
				"pcp_gettimeout: %s\n",
				pcp_strerror(ret));
			exit(1);
		}
		ret = select(fd + 1, &set, NULL, NULL, &tv);
		if (ret < 0) {
#ifdef WIN32
			fprintf(stderr, "select: %d\n", WSAGetLastError());
#else
			fprintf(stderr, "select: %s\n", strerror(errno));
#endif
			continue;
		}
		if (FD_ISSET(fd, &set)) {
			memset(&response, 0, sizeof(response));
			ret = pcp_recvresponse(&pcp, &response);
			if (ret < PCP_OK) {
				fprintf(stderr,
					"pcp_recvresponse: %s\n",
					pcp_strerror(ret));
				if (ret == PCP_ERR_RECVBAD) {
					fprintf(stderr,
						"at line %d\n",
						pcp_debug_line);
					abort();
				}
				exit(1);
			}
			if (ret == PCP_TRY_AGAIN)
				fprintf(stderr, "try again\n");
			else
				break;
		}
		ret = pcp_sendrequest(&pcp);
		if (ret < PCP_OK) {
			fprintf(stderr,
				"pcp_sendrequest: %s\n",
				pcp_strerror(ret));
			exit(1);
		}
	}
	printf("epoch: %d\n", (int) response.epoch);
	if (response.assigned.lifetime != 0)
		printf("lifetime: %d\n", (int) response.assigned.lifetime);
	for (optidx = 0; optidx < response.optcnt; optidx++) {
		opt = &response.options[optidx];
		switch (opt->code) {
		case PCP_OPTION_THIRD_PARTY:
			if (opt->length != PCP_LEN_OPT3PTY) {
				printf("option: third party?\n");
				break;
			}
			if (pcp_unmapoptaddr(opt)) {
				in.s_addr = opt->optaddr4;
				printf("option: third party %s\n",
				       inet_ntoa(in));
			} else {
				(void) inet_ntop(AF_INET6,
						 opt->optaddr6,
						 host6,
						 INET6_ADDRSTRLEN);
				printf("option: third party %s\n", host6);
			}
			break;
		case PCP_OPTION_PREF_FAIL:
			printf("option: prefer failure\n");
			break;
		case PCP_OPTION_FILTER:
			if (opt->length != PCP_LEN_OPTFLT) {
				printf("option: filter\n");
				break;
			}
			memcpy(&n16, opt->data + PCP_OPTDFLT_PORT, 2);
			n16 = ntohs(n16);
			if (pcp_unmapoptaddr(opt)) {
				in.s_addr = opt->optaddr4;
				printf("option: filter %s/%d %hu\n",
				       inet_ntoa(in),
				       (int) opt->data[PCP_OPTDFLT_PLEN],
				       n16);
			} else {
				(void) inet_ntop(AF_INET6,
						 opt->optaddr6,
						 host6,
						 INET6_ADDRSTRLEN);
				printf("option: filter %s/%d %hu\n",
				       host6,
				       (int) opt->data[PCP_OPTDFLT_PLEN],
				       n16);
			}
			break;
		default:
			printf("option: %u\n", opt->code);
			break;
		}
	}
	ret = pcp_close(&pcp);
	if (ret < PCP_OK) {
		fprintf(stderr, "pcp_close: %s\n", pcp_strerror(ret));
		exit(1);
	}
#ifdef WIN32
	(void) WSACleanup();
#endif
	exit(0);
}
