#!/usr/bin/env python

# Copyright (C) 2011-2012  Internet Systems Consortium, Inc. ("ISC")
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND ISC DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
# AND FITNESS.  IN NO EVENT SHALL ISC BE LIABLE FOR ANY SPECIAL, DIRECT,
# INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
# LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
# OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
# PERFORMANCE OF THIS SOFTWARE.
#

"""
Interactive full PCP client

Francis_Dupont@isc.org, July 2011

Usage: python interactive
"""   

__version__ = "$Id: 8bdff594af31f9a13391b907e8c2691e744c41a3 $"


import libpcp
import select
import socket
import sys

try:
    if sys.version_info.major > 2:
        raw_input = input
except AttributeError:
    pass

def print_(*args):
    """print as a function"""
    prntstr = str(args[0])
    for arg in args[1:]:
        prntstr += ' ' + str(arg)
    print(prntstr)

def printoptions(pcp, options, debug):
    """print options"""
    for (code, data) in options:
        if code == libpcp.OPTION_FILTER:
            if len(data) != 20:
                if debug:
                    print_('option filter: [', repr(data), ']')
                else:
                    print_('option filter')
            else:
                if debug:
                    print_('option filter: [', repr(data), ']',
                           pcp.getoptinfo(code, data))
                else:
                    print_('option filter:', pcp.getoptinfo(code, data))
            continue
        elif code == libpcp.OPTION_PREF_FAIL:
            if len(data) != 0:
                if debug:
                    print_('option prefer_failure: [', repr(data), ']')
                else:
                    print_('option prefer_failure')
            else:
                print_('option prefer_failure')
            continue
        elif code == libpcp.OPTION_THIRD_PARTY:
            if (len(data) != 16):
                if debug:
                    print_('option third party: [', repr(data), ']')
                else:
                    print_('option third party')
            else:
                if debug:
                    print_('option third party: [', repr(data), ']',
                           pcp.getoptinfo(code, data))
                else:
                    print_('option third party:', pcp.getoptinfo(code, data))
            continue
        else:
            if debug:
                print_('unknown option:', code, '[', repr(data), ']')
            else:
                print_('unknown option:', code)
            continue

def loop(pcp, announce, peer, debug, verbose):
    """send/recv loop"""
    if debug:
        print_('calling sendrequest method')
    ret = pcp.sendrequest()
    if debug:
        print_('sendrequest method returns', ret)
    while True:
        pcpfd = pcp.getsocket()
        tries = pcp.gettries()
        if debug:
            print_('tries:', tries)
        if tries > libpcp.MAX_TRIES:
            print_('too many retries: giving up')
            return -1
        timeout = pcp.gettimeout(1)
        if debug:
            print_('timeout:', timeout)
        fds = select.select([pcpfd], [], [], timeout)
        if debug:
            print_('select() returns', fds)
        if len(fds[0]) > 0:
            (ret, resp) = pcp.recvresponse()
            if debug:
                print_('received response:')
                print_(' return code:', ret)
                print_(' response:', resp)
            if ret == libpcp.TRY_AGAIN:
                if verbose or debug:
                    print_('trying again...')
                continue
            if debug:
                print_('result:', resp['result'])
                print_('opcode:', resp['opcode'])
            print_('epoch:', resp['epoch'])
            if not announce:
                print_('internal port:', resp['intport'])
                print_('external address:', resp['extaddr'])
                print_('external port:', resp['extport'])
            if peer:
                print_('remote address:', resp['remotepeer'])
                print_('remote port:', resp['peerport'])
            if not announce:
                print_('lifetime:', resp['lifetime'])
            printoptions(pcp, resp['options'], debug)
            pcp.close()
            return resp['result']
        else:
            if verbose or debug:
                print_('timeout: retrying...')
            ret = pcp.sendrequest()
            if debug:
                print_('sendrequest method returns', ret)

def main():
    """main"""
    pcp = libpcp.PCP()
    if pcp is None:
        print_('PCP object creation failure?')
        return -1

    param = raw_input('verbose? [y|n] (default no) ')
    verbose = param == 'y'
    param = raw_input('debug? [y|n] (default no) ')
    debug = param == 'y'

    openkw = {}
    param = raw_input('transport address family? [4|6] (default 4) ')
    if param == '4':
        openkw['family'] = socket.AF_INET
    elif param == '6':
        openkw['family'] = socket.AF_INET6
    elif param != '':
        print_('bad transport address family', param)
        return -1
    param = raw_input('source address? (default is none) ')
    if param != '':
        openkw['source'] = param
    param = raw_input('server address? (no default) ')
    if param == '':
        print_('default server address is not yet supported')
        return -1
    openkw['server'] = param
    param = raw_input('server port? (or use default) ')
    if param != '':
        openkw['port'] = int(param)
    if debug:
        print_('calling open method with', openkw)
    pcp.open(**openkw)
    if debug:
        print_('file descriptor is', pcp.getsocket())

    requestkw = {}
    options = []
    announce = False
    peer = False
    param = raw_input('opcode? [announce|map|peer] (default map) ')
    if (param == '') or (param == 'map'):
        requestkw['opcode'] = libpcp.OPCODE_MAP
    elif param == 'announce':
        requestkw['opcode'] = libpcp.OPCODE_ANNOUNCE
        announce = True
    elif param == 'peer':
        requestkw['opcode'] = libpcp.OPCODE_PEER
        peer = True
    else:
        print_('unknown opcode', param)
        return -1
    param = raw_input('internal address? (default auto) ')
    if param != '':
        requestkw['intaddr'] = param
        param = raw_input('autofill? [y|n] (default no) ')
        if param != 'y':
            print_('previous autofill value was:', pcp.setautofill(0))
    if not announce:
        param = raw_input('internal port? (no default) ')
        if param == '':
            print_('an internal port number is required')
            return -1
        requestkw['intport'] = int(param)
        param = raw_input('protocol? (no default) ')
        if param == 'all':
            requestkw['protocol'] = 0
        elif param == 'udp':
            requestkw['protocol'] = socket.IPPROTO_UDP
        elif param == 'tcp':
            requestkw['protocol'] = socket.IPPROTO_TCP
        else:
            requestkw['protocol'] = int(param)
        param = raw_input('lifetime? (default is 0) ')
        if param != '':
            requestkw['lifetime'] = int(param)
        if peer:
            param = raw_input('remote peer address? ')
            if param != '':
                requestkw['remotepeer'] = param
                param = raw_input('remote peer port? ')
                if param != '':
                    requestkw['peerport'] = int(param)
        param = raw_input('external address? (default 0.0.0.0) ')
        if param == '':
            requestkw['extaddr'] = '0.0.0.0'
        else:
            requestkw['extaddr'] = param
        param = raw_input('external port? (default 0) ')
        if param != '':
            requestkw['extport'] = int(param)
    while True:
        param = raw_input('append an option? [y|n] (default no) ')
        if param != 'y':
            break
        param = raw_input('option code? ' + \
                              '[third_party|prefer_failure|filter] ')
        if param == 'third_party':
            family = socket.AF_INET
            param = raw_input('third party address family? '
                              '[4|6] (default 4) ')
            if param == '6':
                family = socket.AF_INET6
            elif (param != '4') and (param != ''):
                print_('bad third party address family', param)
            address = raw_input('third party address? (no default) ')
            options.append(pcp.third_party(family, address))
            continue
        elif param == 'prefer_failure':
            options.append(pcp.prefer_failure())
            continue
        elif param == 'filter':
            family = socket.AF_INET
            param = raw_input('filter family? [4|6] (default 4) ')
            if param == '6':
                family = socket.AF_INET6
            elif (param != '4') and (param != ''):
                print_('bad filter address family', param)
            address = raw_input('filter address? (no default) ')
            if family == socket.AF_INET:
                param = raw_input('filter prefix length? (default 32) ')
                if param == '':
                    param = '32'
            else:
                param = raw_input('filter prefix length? (default 128) ')
                if param == '':
                    param = '128'
            prefixlen = int(param)
            param = raw_input('filter port? (default is 0) ')
            if param == '':
                param = '0'
            options.append(pcp.filter(family, prefixlen, address, int(param)))
            continue
        else:
            print_('unknown option', param)
            continue
    if options:
        requestkw['options'] = options

    if debug:
        print_('calling makerequest method with', requestkw)
    pcp.makerequest(**requestkw)
    return loop(pcp, announce, peer, debug, verbose)

main()
