/*
 * Copyright (C) 2010-2012  Internet Systems Consortium, Inc. ("ISC")
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND ISC DISCLAIMS ALL WARRANTIES WITH
 * REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS.  IN NO EVENT SHALL ISC BE LIABLE FOR ANY SPECIAL, DIRECT,
 * INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
 * OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */

/* $Id: 821137c57c0a1c76dacf53109af062a99c0e20c4 $ */

/*
 * PCP (port-control-protocol) client library body
 *
 * Francis_Dupont@isc.org, December 2010
 *
 * a la miniUPnP libnatpmp
 */

#ifdef WIN32
#include <winsock2.h>
#include <ws2tcpip.h>
#else
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <fcntl.h>
#include <unistd.h>
#endif

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "pcp.h"

#ifdef WIN32
#define close(s)	closesocket(s)
#endif

/* Initialize a PCP handler */

int
pcp_init(struct sockaddr *server, struct sockaddr *source, pcp_t *pcp)
{
	struct sockaddr_in *server4 = (struct sockaddr_in *) server;
	struct sockaddr_in6 *server6 = (struct sockaddr_in6 *) server;
#ifdef WIN32
	u_long ioctlArg = 1;
#else
	int flags;
#endif

	if ((pcp == NULL) || (server == NULL) ||
	    ((source != NULL) && (server->sa_family != source->sa_family)))
		return PCP_ERR_INVAL;
	memset(pcp, 0, sizeof(pcp_t));
	pcp->autofill = 1;
	pcp->s = socket(server->sa_family == AF_INET6 ? PF_INET6 : PF_INET,
			SOCK_DGRAM, IPPROTO_UDP);
#ifdef WIN32
	if (pcp->s == INVALID_SOCKET)
#else
	if (pcp->s < 0)
#endif
		return PCP_ERR_SOCKET;
	pcp->usable = 1;
#ifdef WIN32
	if (ioctlsocket(pcp->s, FIONBIO, &ioctlArg) != 0)
#else
	if (((flags = fcntl(pcp->s, F_GETFL, 0)) < 0) ||
	    (fcntl(pcp->s, F_SETFL, flags | O_NONBLOCK) < 0))
#endif
	{
		(void) close(pcp->s);
		pcp->usable = 0;
		return PCP_ERR_SYSCALL;
	}
	if (server->sa_family == AF_INET6) {
		if (server6->sin6_port == 0)
			server6->sin6_port = htons(PCP_PORT);
		if ((source != NULL) &&
		    (bind(pcp->s, source, sizeof(*server6)) < 0)) {
			(void) close(pcp->s);
			pcp->usable = 0;
			return PCP_ERR_BIND;
		}
		if (connect(pcp->s, server, sizeof(*server6)) < 0) {
			(void) close(pcp->s);
			pcp->usable = 0;
			return PCP_ERR_CONNECT;
		}
	} else {
		if (server4->sin_port == 0)
			server4->sin_port = htons(PCP_PORT);
		if ((source != NULL) &&
		    (bind(pcp->s, source, sizeof(*server4)) < 0)) {
			(void) close(pcp->s);
			pcp->usable = 0;
			return PCP_ERR_BIND;
		}
		if (connect(pcp->s, server, sizeof(*server4)) < 0) {
			(void) close(pcp->s);
			pcp->usable = 0;
			return PCP_ERR_CONNECT;
		}
	}
	return PCP_OK;
}

/* close a PCP handler */

int
pcp_close(pcp_t *pcp)
{
	if (pcp == NULL)
		return PCP_ERR_INVAL;
	if (pcp->request != NULL)
		pcp_free(pcp->request);
	if (pcp->response != NULL)
		pcp_free(pcp->response);
	pcp->reqlen = 0;
	pcp->request = NULL;
	pcp->resplen = 0;
	pcp->response = NULL;
	pcp->rcusable = 0;
	if (pcp->usable == 0)
		return PCP_OK;
	pcp->usable = 0;
	if (close(pcp->s) < 0)
		return PCP_ERR_SYSCALL;
	return PCP_OK;
}

/* Get the socket file descriptor */

int
pcp_getsocket(pcp_t *pcp, int *sock)
{
	if ((pcp == NULL) || (pcp->usable == 0) || (sock == NULL))
		return PCP_ERR_INVAL;
	*sock = pcp->s;
	return PCP_OK;
}

/* Set the rapid recovery multicast socket file descriptor */

int
pcp_setrcsocket(pcp_t *pcp, int sock)
{
	if ((pcp == NULL) || (pcp->usable == 0))
		return PCP_ERR_INVAL;
	/* give a way to reset it */
#ifdef WIN32
	if (sock == INVALID_SOCKET) {
		pcp->rcusable = 0;
		pcp->ms = 0;
		return PCP_OK;
	}
#else
	if (sock < 0) {
		pcp->rcusable = 0;
		pcp->ms = 0;
		return PCP_OK;
	}
#endif
	pcp->ms = sock;
	pcp->rcusable = 1;
	return PCP_OK;
}

/* Get the number of tries */

int
pcp_gettries(pcp_t *pcp, int *tries)
{
	if ((pcp == NULL) || (tries == NULL))
		return PCP_ERR_INVAL;
	*tries = pcp->tries;
	return PCP_OK;
}

/* Get the timeout for the next retry */

#ifdef WIN32
int
gettimeofday(struct timeval* tp, void* tzp)
{
	DWORD t;

	t = timeGetTime();
	tp->tv_sec = t / 1000;
	tp->tv_usec = t % 1000;
	return 0;
}
#endif

int
pcp_gettimeout(pcp_t *pcp, struct timeval *timeout, int relative)
{
	struct timeval now;

	if ((pcp == NULL) || (timeout == NULL))
		return PCP_ERR_INVAL;
	if (pcp->request == NULL)
		return PCP_ERR_NOREQUEST;
	timeout->tv_sec = pcp->retry.tv_sec;
	timeout->tv_usec = pcp->retry.tv_usec;
	if (relative) {
		if (gettimeofday(&now, NULL) < 0)
			return PCP_ERR_SYSCALL;
		if ((pcp->retry.tv_sec == 0) && (pcp->retry.tv_usec == 0)) {
			timeout->tv_sec = now.tv_sec;
			timeout->tv_usec = now.tv_usec;
		} else {
			timeout->tv_sec -= now.tv_sec;
			timeout->tv_usec -= now.tv_usec;
		}
		if (timeout->tv_usec < 0) {
			timeout->tv_sec -= 1;
			timeout->tv_usec += 1000000;
		}
	}
	return PCP_OK;
}

/* Set the reset flag */

int
pcp_setreset(pcp_t *pcp, int reset, int *old)
{
	if (pcp == NULL)
		return PCP_ERR_INVAL;
	if (old != NULL)
		*old = (int) pcp->reset;
	if (reset)
		pcp->reset = 1;
	else
		pcp->reset = 0;
	return PCP_OK;
}

/* Set the fill flag */

int
pcp_setautofill(pcp_t *pcp, int autofill, int *old)
{
	if (pcp == NULL)
		return PCP_ERR_INVAL;
	if (old != NULL)
		*old = (int) pcp->autofill;
	if (autofill)
		pcp->autofill = 1;
	else
		pcp->autofill = 0;
	return PCP_OK;
}

/* Stack THIRD_PARTY IPv4 option */

int
pcp_third_party4(pcp_option_t ***options, uint32_t addr)
{
	pcp_option_t *o, **v = NULL;
	size_t sz;
	unsigned int i;

	if (options == NULL)
		return PCP_ERR_INVAL;
	if (*options == NULL) {
		sz = sizeof(pcp_option_t *) * PCP_VECOPT_SIZE;
		v = (pcp_option_t **) pcp_malloc(sz);
		if (v == NULL)
			return PCP_ERR_NOMEM;
		memset(v, 0, sz);
		*options = v;
	}
	for (i = 0; i < PCP_MAX_OPTIONS; i++)
		if ((*options)[i] == NULL)
			break;
	if (i >= PCP_MAX_OPTIONS)
		return PCP_ERR_TOOMANYOPTS;
	sz = sizeof(pcp_option_t) + PCP_LEN_OPT3PTY;
	o = (pcp_option_t *) pcp_malloc(sz);
	if (o == NULL) {
		if (*options == v) {
			*options = NULL;
			free(v);
		}
		return PCP_ERR_NOMEM;
	}
	memset(o, 0, sz);
	o->code = PCP_OPTION_THIRD_PARTY;
	o->length = PCP_LEN_OPT3PTY;
	o->data = (uint8_t *) (o + 1);
	o->optaddr4 = addr;
	pcp_mapoptaddr(o);
	memcpy(o->data, o->optaddr6, 16);
	(*options)[i] = o;
	return PCP_OK;
}

/* Stack THIRD_PARTY IPv6 option */

int
pcp_third_party6(pcp_option_t ***options, uint8_t *addr)
{
	pcp_option_t *o, **v = NULL;
	size_t sz;
	unsigned int i;

	if ((options == NULL) || (addr == NULL))
		return PCP_ERR_INVAL;
	if (*options == NULL) {
		sz = sizeof(pcp_option_t *) * PCP_VECOPT_SIZE;
		v = (pcp_option_t **) pcp_malloc(sz);
		if (v == NULL)
			return PCP_ERR_NOMEM;
		memset(v, 0, sz);
		*options = v;
	}
	for (i = 0; i < PCP_MAX_OPTIONS; i++)
		if ((*options)[i] == NULL)
			break;
	if (i >= PCP_MAX_OPTIONS)
		return PCP_ERR_TOOMANYOPTS;
	sz = sizeof(pcp_option_t) + PCP_LEN_OPT3PTY;
	o = (pcp_option_t *) pcp_malloc(sz);
	if (o == NULL) {
		if (*options == v) {
			*options = NULL;
			free(v);
		}
		return PCP_ERR_NOMEM;
	}
	memset(o, 0, sz);
	o->code = PCP_OPTION_THIRD_PARTY;
	o->length = PCP_LEN_OPT3PTY;
	o->data = (uint8_t *) (o + 1);
	memcpy(o->optaddr6, addr, 16);
	memcpy(o->data, o->optaddr6, 16);
	(*options)[i] = o;
	return PCP_OK;
}

/* Stack PREFER_FAILURE option */

int
pcp_prefer_failure(pcp_option_t ***options)
{
	pcp_option_t *o, **v = NULL;
	size_t sz;
	unsigned int i;

	if (options == NULL)
		return PCP_ERR_INVAL;
	if (*options == NULL) {
		sz = sizeof(pcp_option_t *) * PCP_VECOPT_SIZE;
		v = (pcp_option_t **) pcp_malloc(sz);
		if (v == NULL)
			return PCP_ERR_NOMEM;
		memset(v, 0, sz);
		*options = v;
	}
	for (i = 0; i < PCP_MAX_OPTIONS; i++)
		if ((*options)[i] == NULL)
			break;
	if (i >= PCP_MAX_OPTIONS)
		return PCP_ERR_TOOMANYOPTS;
	sz = sizeof(pcp_option_t);
	o = (pcp_option_t *) pcp_malloc(sz);
	if (o == NULL) {
		if (*options == v) {
			*options = NULL;
			free(v);
		}
		return PCP_ERR_NOMEM;
	}
	memset(o, 0, sz);
	o->code = PCP_OPTION_PREF_FAIL;
	o->length = PCP_LEN_OPT0;
	(*options)[i] = o;
	return PCP_OK;
}

/* Stack FILTER IPv4 option */

int
pcp_filter4(pcp_option_t ***options,
	    int prefixlen, uint32_t addr, uint16_t port)
{
	pcp_option_t *o, **v = NULL;
	size_t sz;
	unsigned int i;
	uint16_t n16;

	if ((options == NULL) || (prefixlen < 0) || (prefixlen > 32))
		return PCP_ERR_INVAL;
	if (*options == NULL) {
		sz = sizeof(pcp_option_t *) * PCP_VECOPT_SIZE;
		v = (pcp_option_t **) pcp_malloc(sz);
		if (v == NULL)
			return PCP_ERR_NOMEM;
		memset(v, 0, sz);
		*options = v;
	}
	for (i = 0; i < PCP_MAX_OPTIONS; i++)
		if ((*options)[i] == NULL)
			break;
	if (i >= PCP_MAX_OPTIONS)
		return PCP_ERR_TOOMANYOPTS;
	sz = sizeof(pcp_option_t) + PCP_LEN_OPTFLT;
	o = (pcp_option_t *) pcp_malloc(sz);
	if (o == NULL) {
		if (*options == v) {
			*options = NULL;
			free(v);
		}
		return PCP_ERR_NOMEM;
	}
	memset(o, 0, sz);
	o->code = PCP_OPTION_FILTER;
	o->length = PCP_LEN_OPTFLT;
	o->data = (uint8_t *) (o + 1);
	o->data[PCP_OPTDFLT_RSVD] = 0;
	o->data[PCP_OPTDFLT_PLEN] = (uint8_t) prefixlen + 96;
	n16 = htons(port);
	memcpy(o->data + PCP_OPTDFLT_PORT, &n16, 2);
	o->optaddr4 = addr;
	pcp_mapoptaddr(o);
	memcpy(o->data + PCP_OPTDFLT_PREF, o->optaddr6, 16);
	(*options)[i] = o;
	return PCP_OK;
}

/* Stack FILTER IPv6 option */

int
pcp_filter6(pcp_option_t ***options,
	    int prefixlen, uint8_t *addr, uint16_t port)
{
	pcp_option_t *o, **v = NULL;
	size_t sz;
	unsigned int i;
	uint16_t n16;

	if ((options == NULL) || (addr == NULL) ||
	    (prefixlen < 0) || (prefixlen > 128))
		return PCP_ERR_INVAL;
	if (*options == NULL) {
		sz = sizeof(pcp_option_t *) * PCP_VECOPT_SIZE;
		v = (pcp_option_t **) pcp_malloc(sz);
		if (v == NULL)
			return PCP_ERR_NOMEM;
		memset(v, 0, sz);
		*options = v;
	}
	for (i = 0; i < PCP_MAX_OPTIONS; i++)
		if ((*options)[i] == NULL)
			break;
	if (i >= PCP_MAX_OPTIONS)
		return PCP_ERR_TOOMANYOPTS;
	sz = sizeof(pcp_option_t) + PCP_LEN_OPTFLT;
	o = (pcp_option_t *) pcp_malloc(sz);
	if (o == NULL) {
		if (*options == v) {
			*options = NULL;
			free(v);
		}
		return PCP_ERR_NOMEM;
	}
	memset(o, 0, sz);
	o->code = PCP_OPTION_FILTER;
	o->length = PCP_LEN_OPTFLT;
	o->data = (uint8_t *) (o + 1);
	o->data[PCP_OPTDFLT_RSVD] = 0;
	o->data[PCP_OPTDFLT_PLEN] = (uint8_t) prefixlen;
	n16 = htons(port);
	memcpy(o->data + PCP_OPTDFLT_PORT, &n16, 2);
	memcpy(o->optaddr6, addr, 16);
	memcpy(o->data + PCP_OPTDFLT_PREF, o->optaddr6, 16);
	(*options)[i] = o;
	return PCP_OK;
}

static int clearrequest(pcp_t *pcp);
static int createrequest(pcp_t *pcp,
			 pcp_request_t *request,
			 pcp_option_t **options);

/* clear/create a request */

int
pcp_makerequest(pcp_t *pcp, pcp_request_t *request, pcp_option_t **options)
{
	int result;

	if (pcp == NULL)
		return PCP_ERR_INVAL;
	result = clearrequest(pcp);
	if ((request == NULL) || (result != PCP_OK))
		return result;
	return createrequest(pcp, request, options);
}

/* Clear request */

static int
clearrequest(pcp_t *pcp)
{
	if (pcp->request != NULL)
		pcp_free(pcp->request);
	pcp->reqlen = 0;
	pcp->request = NULL;
	pcp->tries = 0;
	pcp->retry.tv_sec = 0;
	pcp->retry.tv_usec = 0;
	return PCP_OK;
}

/* Create request */

static int
createrequest(pcp_t *pcp, pcp_request_t *request, pcp_option_t **options)
{
	unsigned int len, off, i;
	uint32_t n32;
	uint16_t n16;

	if (pcp->autofill) {
		int ret;
		ret = _pcp_fillsrcaddr(pcp, request);
		if (ret != PCP_OK)
			return ret;
	}

	switch (request->opcode) {
	case PCP_OPCODE_ANNOUNCE:
		off = len = PCP_LEN_ANNOUNCE;
		break;
	case PCP_OPCODE_MAP:
		off = len = PCP_LEN_MAP;
		break;
	case PCP_OPCODE_PEER:
		off = len = PCP_LEN_PEER;
		break;
	default:
		return PCP_ERR_INVAL;
	}

	if (options != NULL)
		for (i = 0; options[i] != NULL; i++)
			len += PCP_LEN_OPTHDR + PCP_ALIGN(options[i]->length);
	if (len > 1024)
		return PCP_ERR_INVAL;
	pcp->request = (uint8_t *) pcp_malloc(len);
	if (pcp->request == NULL)
		return PCP_ERR_NOMEM;
	memset(pcp->request, 0, len);
	pcp->reqlen = len;

	pcp->request[PCP_VERSION] = 1;
	pcp->request[PCP_OPCODE] = request->opcode;
	if (request->opcode != PCP_OPCODE_ANNOUNCE) {
		n32 = htonl(request->lifetime);
		memcpy(pcp->request + PCP_LIFETIME, &n32, 4);
	}
	memcpy(pcp->request + PCP_CLIENT_ADDR, request->intaddr6, 16);


	if (request->opcode == PCP_OPCODE_MAP) {
		pcp->request[PCP_MAP_PROTOCOL] = request->protocol;
		n16 = htons(request->intport);
		memcpy(pcp->request + PCP_MAP_INTERNAL_PORT, &n16, 2);
		n16 = htons(request->extport);
		memcpy(pcp->request + PCP_MAP_EXTERNAL_PORT, &n16, 2);
		memcpy(pcp->request + PCP_MAP_EXTERNAL_ADDR,
		       request->extaddr6, 16);
	} else if (request->opcode == PCP_OPCODE_PEER) {
		pcp->request[PCP_PEER_PROTOCOL] = request->protocol;
		n16 = htons(request->intport);
		memcpy(pcp->request + PCP_PEER_INTERNAL_PORT, &n16, 2);
		n16 = htons(request->extport);
		memcpy(pcp->request + PCP_PEER_EXTERNAL_PORT, &n16, 2);
		memcpy(pcp->request + PCP_PEER_EXTERNAL_ADDR,
		       &request->extaddr6, 16);
		n16 = htons(request->peerport);
		memcpy(pcp->request + PCP_PEER_REMOTE_PORT, &n16, 2);
		memcpy(pcp->request + PCP_PEER_REMOTE_ADDR,
		       request->remotepeer6, 16);
	}

	if (options != NULL)
		for (i = 0; options[i] != NULL; i++) {
			pcp_option_t *opt = options[i];
			unsigned int olen = opt->length;
			
			pcp->request[off + PCP_OPT_CODE] = opt->code;
			n16 = htons(olen);
			memcpy(pcp->request + off + PCP_OPT_LEN, &n16, 2);
			memcpy(pcp->request + off + PCP_OPT_DATA,
			       opt->data, olen);
			off += PCP_LEN_OPTHDR + PCP_ALIGN(olen);
		}
	return PCP_OK;
}

/* Send/resend the current request */

int
pcp_sendrequest(pcp_t *pcp)
{
	unsigned int tm;

	if ((pcp == NULL) || (pcp->usable == 0))
		return PCP_ERR_INVAL;
	if (pcp->request == NULL)
		return PCP_ERR_NOREQUEST;
	if (send(pcp->s, pcp->request, pcp->reqlen, 0) < 0)
		return PCP_ERR_SEND;
	if (gettimeofday(&pcp->retry, NULL) < 0)
		return PCP_ERR_SYSCALL;
	if (pcp->tries < 30) {
		tm = PCP_RETRY_TIMER << pcp->tries;
		if (tm > PCP_MAX_TIMEOUT)
			tm = PCP_MAX_TIMEOUT;
	} else
		tm = PCP_MAX_TIMEOUT;
	pcp->retry.tv_sec += tm;
	pcp->tries++;
	return PCP_OK;
}

#define BAD(off)					\
	do {						\
		pcp_debug_offset = off;			\
		pcp_debug_line = __LINE__;		\
		return PCP_ERR_RECVBAD;			\
	} while (0)

static int
decoderesponse(pcp_t *pcp, pcp_response_t *response);

/* Receive the response to the current request */

int
pcp_recvresponse(pcp_t *pcp, pcp_response_t *response)
{
	int cc;

	if ((pcp == NULL) || (pcp->usable == 0) || (response == NULL))
		return PCP_ERR_INVAL;
	memset(response, 0, sizeof(pcp_response_t));
	if (pcp->response == NULL) {
		pcp->response = (uint8_t *) pcp_malloc(PCP_RESPLENGTH);
		if (pcp->response == NULL)
			return PCP_ERR_NOMEM;
	}
	memset(pcp->response, 0, PCP_RESPLENGTH);
	cc = recv(pcp->s, pcp->response, PCP_RESPLENGTH, 0);
	if (cc < 0) {
#ifdef WIN32
		errno = WSAGetLastError();
		if (errno == WSAEWOULDBLOCK)
#else
		if (errno == EWOULDBLOCK)
#endif
		{
			if (pcp->rcusable != 0) {
				cc = recv(pcp->ms, pcp->response,
					  PCP_RESPLENGTH, 0);
				if (cc < 0) {
#ifdef WIN32
					errno = WSAGetLastError();
					if (errno == WSAEWOULDBLOCK)
#else
					if (errno == EWOULDBLOCK)
#endif
						return PCP_TRY_AGAIN;
					return PCP_ERR_RECV;
				}
				pcp->resplen = cc;
				return decoderesponse(pcp, response);
			}
			return PCP_TRY_AGAIN;
		}
		return PCP_ERR_RECV;
	}
	pcp->resplen = cc;
	return decoderesponse(pcp, response);
}

int
decoderesponse(pcp_t *pcp, pcp_response_t *response)
{
	uint32_t n32;
	uint16_t n16;
	unsigned int off = 0, optidx = 0;

	pcp_debug_offset = 0;
	pcp_debug_line = __LINE__;

	/* don't merge: keep the line! */
	if (pcp->resplen < 4) 
		BAD(pcp->resplen);
	else if ((pcp->response[PCP_OPCODE] & PCP_OPCODE_RESPONSE) == 0)
		BAD(PCP_OPCODE);
	else if (pcp->response[PCP_VERSION] != 1)
		BAD(PCP_VERSION);
	else if (pcp->resplen < PCP_OPCODE_HDR)
		BAD(pcp->resplen);
	else if (pcp->resplen > 1024)
		BAD(pcp->resplen);
	else if ((pcp->resplen & 3) != 0)
		BAD(pcp->resplen);
	if (pcp->response[PCP_OPCODE] ==
	    (PCP_OPCODE_ANNOUNCE | PCP_OPCODE_RESPONSE)) {
		response->assigned.opcode = pcp->response[PCP_OPCODE];
		response->result = pcp->response[PCP_RESULT_CODE];
		memcpy(&n32, pcp->response + PCP_LIFETIME, 4);
		response->assigned.lifetime = ntohl(n32);
		memcpy(&n32, pcp->response + PCP_EPOCH, 4);
		response->epoch = ntohl(n32);
		if ((pcp->request == NULL) ||
		    (pcp->request[PCP_OPCODE] != PCP_OPCODE_ANNOUNCE)) 
			return PCP_TRY_AGAIN;
	} else if (pcp->request == NULL)
		BAD(PCP_OPCODE);
	if (pcp->response[PCP_OPCODE] !=
	    (pcp->request[PCP_OPCODE] | PCP_OPCODE_RESPONSE))
		BAD(PCP_OPCODE);
	switch (pcp->request[PCP_OPCODE]) {
	case PCP_OPCODE_ANNOUNCE:
		break;
	case PCP_OPCODE_MAP:
	case PCP_OPCODE_PEER:
		response->assigned.opcode = pcp->response[PCP_OPCODE];
		response->result = pcp->response[PCP_RESULT_CODE];
		memcpy(&n32, pcp->response + PCP_LIFETIME, 4);
		response->assigned.lifetime = ntohl(n32);
		memcpy(&n32, pcp->response + PCP_EPOCH, 4);
		response->epoch = ntohl(n32);
		break;
	default:
		BAD(PCP_OPCODE);
	}
	if (pcp->request[PCP_OPCODE] == PCP_OPCODE_MAP) {
		if (pcp->resplen < PCP_LEN_MAP)
			BAD(pcp->resplen);
	} else if (pcp->request[PCP_OPCODE] == PCP_OPCODE_PEER) {
		if (pcp->resplen < PCP_LEN_PEER)
			BAD(pcp->resplen);
	}
	switch (pcp->request[PCP_OPCODE]) {
	case PCP_OPCODE_ANNOUNCE:
		off = PCP_LEN_ANNOUNCE;
		break;
	case PCP_OPCODE_MAP:
		if (pcp->response[PCP_MAP_PROTOCOL] !=
		    pcp->request[PCP_MAP_PROTOCOL])
			BAD(PCP_MAP_PROTOCOL);
		response->assigned.protocol = pcp->response[PCP_MAP_PROTOCOL];
		if (memcmp(pcp->response + PCP_MAP_INTERNAL_PORT,
			   pcp->request + PCP_MAP_INTERNAL_PORT, 2) != 0)
			BAD(PCP_MAP_INTERNAL_PORT);
		memcpy(&n16, pcp->response + PCP_MAP_INTERNAL_PORT, 2);
		response->assigned.intport = ntohs(n16);
		memcpy(&n16, pcp->response + PCP_MAP_EXTERNAL_PORT, 2);
		response->assigned.extport = ntohs(n16);
		memcpy(&response->assigned.extaddr6,
		       pcp->response + PCP_MAP_EXTERNAL_ADDR, 16);
		off = PCP_LEN_MAP;
		break;
	case PCP_OPCODE_PEER:
		if (pcp->response[PCP_PEER_PROTOCOL] !=
		    pcp->request[PCP_PEER_PROTOCOL])
			BAD(PCP_PEER_PROTOCOL);
		response->assigned.protocol = pcp->response[PCP_PEER_PROTOCOL];
		if (memcmp(pcp->response + PCP_PEER_INTERNAL_PORT,
			   pcp->request + PCP_PEER_INTERNAL_PORT, 2) != 0)
			BAD(PCP_PEER_INTERNAL_PORT);
		memcpy(&n16, pcp->response + PCP_PEER_INTERNAL_PORT, 2);
		response->assigned.intport = ntohs(n16);
		memcpy(&n16, pcp->response + PCP_PEER_EXTERNAL_PORT, 2);
		response->assigned.extport = ntohs(n16);
		memcpy(&response->assigned.extaddr6,
		       pcp->response + PCP_PEER_EXTERNAL_ADDR, 16);
		if (memcmp(pcp->response + PCP_PEER_REMOTE_PORT,
			   pcp->request + PCP_PEER_REMOTE_PORT, 2) != 0)
			BAD(PCP_PEER_REMOTE_PORT);
		memcpy(&n16, pcp->response + PCP_PEER_REMOTE_PORT, 2);
		response->assigned.peerport = ntohs(n16);
		if (memcmp(pcp->response + PCP_PEER_REMOTE_ADDR,
			   pcp->request + PCP_PEER_REMOTE_ADDR, 16) != 0)
			BAD(PCP_PEER_REMOTE_ADDR);
		memcpy(&response->assigned.remotepeer6,
		       pcp->response + PCP_PEER_REMOTE_ADDR, 16);
		off = PCP_LEN_PEER;
		break;
	}

	while (pcp->resplen != (int) off) {
		if ((int) off > pcp->resplen)
			BAD(off);
		if (optidx > PCP_MAX_OPTIONS)
			return PCP_ERR_TOOMANYOPTS;
		response->optoffs[optidx] = off;
		response->options[optidx].code =
			pcp->response[off + PCP_OPT_CODE];
		memcpy(&n16, pcp->response + off + PCP_OPT_LEN, 2);
		response->options[optidx].length = ntohs(n16);
		off += PCP_LEN_OPTHDR;
		response->options[optidx].data = pcp->response + off;
		switch (response->options[optidx].code) {
		case PCP_OPTION_THIRD_PARTY:
			if (response->options[optidx].length
			    != PCP_LEN_OPT3PTY)
				break;
			memcpy(response->options[optidx].optaddr6,
			       pcp->response + off, 16);
			break;
		case PCP_OPTION_FILTER:
			if (response->options[optidx].length
			    != PCP_LEN_OPTFLT)
				break;
			memcpy(response->options[optidx].optaddr6,
			       pcp->response + off + PCP_OPTDFLT_PREF, 16);
			break;
		}
		off += PCP_ALIGN(response->options[optidx].length);
		optidx++;
	}
	response->optcnt = optidx;
	if (response->result == 0)
		return PCP_OK;
	if ((response->assigned.lifetime > PCP_RESET_THRESHOLD) || !pcp->reset)
		return PCP_ERR_PROTOBASE - response->result;
	/* transient error: resetting */
	if (gettimeofday(&pcp->retry, NULL) < 0)
		return PCP_ERR_SYSCALL;
	pcp->tries = 0;
	pcp->retry.tv_sec += response->assigned.lifetime;
	return PCP_TRY_AGAIN;
}

/* Translate an error to its human friendly description */

const char *
pcp_strerror(int error)
{
	switch (error) {
	case PCP_OK:
		return "ok";
	case PCP_TRY_AGAIN:
		return "try again";
	case PCP_ERR_INVAL:
		return "invalid arguments";
	case PCP_ERR_NOMEM:
		return "malloc() failed";
	case PCP_ERR_SOCKET:
		return "socket() syscall failed";
	case PCP_ERR_BIND:
		return "bind() syscall failed";
	case PCP_ERR_CONNECT:
		return "connect() syscall failed";
	case PCP_ERR_SEND:
		return "send() syscall failed";
	case PCP_ERR_RECV:
		return "recv() syscall failed";
	case PCP_ERR_SYSCALL:
		return "syscall failed";
	case PCP_ERR_NOREQUEST:
		return "no current request";
	case PCP_ERR_RECVBAD:
		return "bad response";
	case PCP_ERR_TOOMANYOPTS:
		return "too many options";
	case PCP_ERR_FAILURE:
		return "internal failure";
	case PCP_ERR_APP0:
		return err_app0_msg;
	case PCP_ERR_APP1:
		return err_app1_msg;
	case PCP_ERR_APP2:
		return err_app2_msg;
	case PCP_ERR_PROTOBASE:
		return "base for protocol errors (shouldn't happen)";
	case PCP_ERR_UNSUPVERSION:
		return "unsupported version";
	case PCP_ERR_NOTAUTH:
		return "not authorized";
	case PCP_ERR_BADREQUEST:
		return "malformed request";
	case PCP_ERR_UNSUPOPCODE:
		return "unsupported opcode";
	case PCP_ERR_UNSUPOPTION:
		return "unsupported option";
	case PCP_ERR_BADOPTION:
		return "malformed option";
	case PCP_ERR_NETFAILURE:
		return "network failure";
	case PCP_ERR_NORESOURCES:
		return "out of resources";
	case PCP_ERR_UNSUPPROTO:
		return "unsupported protocol";
	case PCP_ERR_EXQUOTA:
		return "user exceeded quota";
	case PCP_ERR_CANTPROVIDE:
		return "cannot provide external";
	case PCP_ERR_ADDRMISMATCH:
		return "address mismatch";
	case PCP_ERR_TOOMANYPEER:
		return "excessive number of remote peers";
	case PCP_ERR_PROCERROR:
		return "processing error";
	default:
		if (error >= PCP_ERR_PROTOBASE)
			return "unknown protocol error";
		else
			return "unknown error";
	}
}

/* Synthetic getmapping() */

int
pcp_getmapping(pcp_t *pcp,
	       pcp_request_t *request,
	       pcp_option_t **options,
	       pcp_response_t *response)
{
	int ret, fd, tries;
	fd_set set;
	struct timeval tv;

	if ((pcp == NULL) || (response == NULL))
		return PCP_ERR_INVAL;
	ret = pcp_getsocket(pcp, &fd);
	if (ret < PCP_OK)
		return ret;
	if (request != NULL)
		ret = pcp_makerequest(pcp, request, options);
	if (ret < PCP_OK)
		return ret;
	ret = pcp_sendrequest(pcp);
	if (ret < PCP_OK)
		return ret;
	for (;;) {
		ret = pcp_gettries(pcp, &tries);
		if (ret < PCP_OK)
			return ret;
		if ((PCP_MAX_TRIES > 0) && (tries > PCP_MAX_TRIES))
			return PCP_ERR_FAILURE;
		FD_ZERO(&set);
		FD_SET(fd, &set);
		ret = pcp_gettimeout(pcp, &tv, 1);
		if (ret < PCP_OK)
			return ret;
		ret = select(fd + 1, &set, NULL, NULL, &tv);
		if (ret < 0)
			return PCP_ERR_SYSCALL;
		if (FD_ISSET(fd, &set)) {
			memset(response, 0, sizeof(*response));
			ret = pcp_recvresponse(pcp, response);
			if (ret == PCP_TRY_AGAIN)
				continue;
			else
				return ret;
		}
		ret = pcp_sendrequest(pcp);
		if (ret < PCP_OK)
			return ret;
	}
}

/* Synthetix pcp_renewmapping() */

int
pcp_renewmapping(pcp_t *pcp, pcp_response_t *response)
{
	pcp_option_t **options;
	int ret;
	pcp_response_t r;

	if ((pcp == NULL) || (response == NULL))
		return PCP_ERR_INVAL;
	options = NULL;
	ret = pcp_prefer_failure(&options);
	if (ret != PCP_OK)
		return ret;
	memcpy(&r, response, sizeof(r));
	r.result = 0;
	r.epoch = 0;
	/* optimistic! */
	r.assigned.lifetime *= 2;
	ret = pcp_getmapping(pcp, &r.assigned, options, &r);
	pcp_freeoptions(options);
	if (ret == PCP_OK)
		memcpy(response, &r, sizeof(r));
	return ret;
}

/* Synthetic pcp_delmapping() */

int
pcp_delmapping(pcp_t *pcp, pcp_response_t *response)
{
	int ret;
	pcp_response_t r;

	if ((pcp == NULL) || (response == NULL))
		return PCP_ERR_INVAL;
	memcpy(&r, response, sizeof(r));
	r.result = 0;
	r.epoch = 0;
	r.assigned.lifetime = 0;
	r.assigned.extport = 0;
	ret = pcp_getmapping(pcp, &r.assigned, NULL, &r);
	if (ret == PCP_OK)
		memcpy(response, &r, sizeof(r));
	return ret;
}

/* Helper: take a file descriptor as the socket to use */

int
_pcp_fdinit(pcp_t *pcp, int fd)
{
	if (pcp == NULL)
		return PCP_ERR_INVAL;
	pcp->s = fd;
	pcp->usable = 1;
	return PCP_OK;
}

/* Helper: fill the intaddr field from getsockname() */

int
_pcp_fillsrcaddr(pcp_t *pcp, pcp_request_t *request)
{
	struct sockaddr_in *s4;
	struct sockaddr_in6 *s6;
	struct sockaddr_storage ss;
	int fd, ret;
	socklen_t len;
	
	if ((pcp == NULL) || (request == NULL))
		return PCP_ERR_INVAL;
	ret = pcp_getsocket(pcp, &fd);
	if (ret < PCP_OK)
		return ret;
	memset(&ss, 0, sizeof(ss));
#ifdef BSD
	ss.ss_len = sizeof(ss);
#endif
	len = sizeof(ss);

	if (getsockname(fd, (struct sockaddr *) &ss, &len) != 0)
		return PCP_ERR_SYSCALL;
	switch (ss.ss_family) {
	case AF_INET:
		s4 = (struct sockaddr_in *) &ss;
		memcpy(&request->intaddr4, &s4->sin_addr, 4);
		pcp_mapintaddr(request);
		return PCP_OK;
	case AF_INET6:
		s6 = (struct sockaddr_in6 *) &ss;
		memcpy(request->intaddr6, &s6->sin6_addr, 16);
		return PCP_OK;
	default:
		return PCP_ERR_FAILURE;
	}
}

/* Internal set request */

int
_pcp_setrequest(pcp_t *pcp, uint8_t *request, uint16_t reqlen)
{
	int result;

	if ((pcp == NULL) || ((request == NULL) && (reqlen != 0)))
		return PCP_ERR_INVAL;
	result = clearrequest(pcp);
	if ((request == NULL) || (result != PCP_OK))
		return result;
	pcp->request = (uint8_t *) pcp_malloc((size_t) reqlen);
	if (pcp->request == NULL)
		return PCP_ERR_NOMEM;
	memcpy(pcp->request, request, (size_t) reqlen);
	pcp->reqlen = reqlen;
	return PCP_OK;
}

/* Internal get request */

int
_pcp_getrequest(pcp_t *pcp, uint8_t **request, uint16_t *reqlen)
{
	if ((pcp == NULL) || (request == NULL) || (reqlen == NULL))
		return PCP_ERR_INVAL;
	*request = pcp->request;
	*reqlen = pcp->reqlen;
	return PCP_OK;
}

/* Internal set response */

int
_pcp_setresponse(pcp_t *pcp, uint8_t *response, uint16_t resplen)
{
	if ((pcp == NULL) || (resplen > PCP_RESPLENGTH) ||
	    ((response == NULL) && (resplen != 0)))
		return PCP_ERR_INVAL;
	pcp->resplen = resplen;
	if (resplen == 0)
		return PCP_OK;
	if (pcp->response == NULL) {
		pcp->response = (uint8_t *) pcp_malloc(PCP_RESPLENGTH);
		if (pcp->response == NULL) {
			pcp->resplen = 0;
			return PCP_ERR_NOMEM;
		}
	}
	memset(pcp->response, 0, PCP_RESPLENGTH);
	memcpy(pcp->response, response, resplen);
	return PCP_OK;
}

/* Internal get response */

int
_pcp_getresponse(pcp_t *pcp, uint8_t **response, uint16_t *resplen)
{
	if ((pcp == NULL) || (response == NULL) || (resplen == NULL))
		return PCP_ERR_INVAL;
	*response = pcp->response;
	*resplen = pcp->resplen;
	return PCP_OK;
}

/* Internal submit response */

int
_pcp_submitresponse(pcp_t *pcp, pcp_response_t *response)
{
	if ((pcp == NULL) || (response == NULL))
		return PCP_ERR_INVAL;
	memset(response, 0, sizeof(pcp_response_t));
	return decoderesponse(pcp, response);
}

/* Internal set option */

int
_pcp_setoption(pcp_option_t ***options, pcp_option_t *opt)
{
	pcp_option_t *o, **v = NULL;
	size_t sz;
	unsigned int i;

	if ((options == NULL) || (opt == NULL))
		return PCP_ERR_INVAL;
	if (*options == NULL) {
		sz = sizeof(pcp_option_t *) * PCP_VECOPT_SIZE;
		v = (pcp_option_t **) pcp_malloc(sz);
		if (v == NULL)
			return PCP_ERR_NOMEM;
		memset(v, 0, sz);
		*options = v;
	}
	for (i = 0; i < PCP_MAX_OPTIONS; i++)
		if ((*options)[i] == NULL)
			break;
	if (i >= PCP_MAX_OPTIONS)
		return PCP_ERR_TOOMANYOPTS;
	sz = sizeof(pcp_option_t) + PCP_ALIGN(opt->length);
	o = (pcp_option_t *) pcp_malloc(sz);
	if (o == NULL) {
		if (*options == v) {
			*options = NULL;
			free(v);
		}
		return PCP_ERR_NOMEM;
	}
	memset(o, 0, sz);
	o->code = opt->code;
	o->length = opt->length;
	o->data = (uint8_t *) (o + 1);
	memcpy(o->data, opt->data, opt->length);
	(*options)[i] = o;
	return PCP_OK;
}

/* Weak definition of pcp_malloc()/pcp_free() */

#ifndef WIN32
#pragma weak pcp_malloc
#pragma weak pcp_free

void *pcp_malloc(size_t size) { return malloc(size); }
void pcp_free(void *ptr) { free(ptr); }
#endif


/* Weak definition of err_app[012]_msg */

#ifndef WIN32
#pragma weak err_app0_msg
#pragma weak err_app1_msg
#pragma weak err_app2_msg
#endif

const char *err_app0_msg = "user application error 0";
const char *err_app1_msg = "user application error 1";
const char *err_app2_msg = "user application error 2";

/* (Weak) support for a minimal pcp.conf file */

#ifndef WIN32
#pragma weak _pcp_config
#endif

int
_pcp_config(struct sockaddr_storage *server, socklen_t *len)
{
	FILE *f;
	char *p, line[1025];
#ifdef WIN32
	char path[MAX_PATH];
	UINT sd;
#endif

	if ((server == NULL) || (len == NULL))
		return PCP_ERR_INVAL;
	memset(server, 0, sizeof(*server));
	*len = 0;

#ifdef WIN32
	memset(path, 0, sizeof(path));
	sd = GetSystemDirectoryA(path, sizeof(path));
	if (sd == 0)
		return PCP_ERR_SYSCALL;
	if ((sd + strlen(PCP_CONFFILE)) >= sizeof(path))
		return PCP_ERR_NOMEM;
	(void) strncat(path, PCP_CONFFILE, sizeof(path) - (sd + 1));
	f = fopen(path, "r");
#else
	f = fopen(PCP_CONFFILE, "r");
#endif
	if (f == NULL)
		return PCP_NOCONFFILE;
	for (;;) {
		memset(line, 0, sizeof(line));
		if (fgets(line, 1024, f) == NULL)
			goto bad;
		if ((*line == '#') || (*line == '\r') || (*line == '\n'))
			continue;
		p = strtok(line, " \r\n");
		if (p == NULL)
			continue;
		if (strchr(p, ':') != NULL) {
			struct sockaddr_in6 *s6;

			s6 = (struct sockaddr_in6 *) server;
			if (inet_pton(AF_INET6, p, &s6->sin6_addr) != 1)
				goto bad;
			server->ss_family = AF_INET6;
			*len = sizeof(*s6);
			(void) fclose(f);
			return PCP_OK;
		} else {
			struct sockaddr_in *s4;

			s4 = (struct sockaddr_in *) server;
			if (inet_pton(AF_INET, p, &s4->sin_addr) != 1)
				goto bad;
			server->ss_family = AF_INET;
			*len = sizeof(*s4);
			(void) fclose(f);
			return PCP_OK;
		}
	}
    bad:
	(void) fclose(f);
	return PCP_ERR_FAILURE;
}

/* Dynamic library entry point for Windows */

#ifdef WIN32
LIBSPEC BOOL WINAPI
DllMain(HMODULE hModule, DWORD ul_reason_for_call, LPVOID lpvReserved)
{
	switch (ul_reason_for_call) { 
	case DLL_PROCESS_ATTACH:
	case DLL_THREAD_ATTACH: 
	case DLL_THREAD_DETACH: 
	case DLL_PROCESS_DETACH: 
	default:
		break;
	}
	return (TRUE);
}
#endif
