#!/usr/bin/env python

# Copyright (C) 2011-2012  Internet Systems Consortium, Inc. ("ISC")
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND ISC DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
# AND FITNESS.  IN NO EVENT SHALL ISC BE LIABLE FOR ANY SPECIAL, DIRECT,
# INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
# LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
# OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
# PERFORMANCE OF THIS SOFTWARE.
#

"""
Simple PCP PEER client rewritten in Python

Francis_Dupont@isc.org, July 2011

Usage: [-v] <server> <address> <priv> <dst> <port> <pub> <proto> <lifetime>
"""   

__version__ = "$Id: 015dd7d8f6efa2146bba2ba97697a788a10155cf $"


import getopt
import libpcp
import select
import socket
import sys

def print_(*args):
    """print as a function"""
    prntstr = str(args[0])
    for arg in args[1:]:
        prntstr += ' ' + str(arg)
    print(prntstr)

def printoptions(options):
    """print options"""
    for (code, data) in options:
        if code == libpcp.OPTION_FILTER:
            print_('option filter:', repr(data))
        elif code == libpcp.OPTION_PREF_FAIL:
            print_('option prefer_failure:', repr(data))
        elif code == libpcp.OPTION_THIRD_PARTY:
            print_('option third party:', repr(data))
        else:
            print_('unknown option:', code, ':', repr(data))

def loop(pcp, verbose):
    """send/recv loop"""
    pcp.sendrequest()
    while True:
        pcpfd = pcp.getsocket()
        tries = pcp.gettries()
        if tries > libpcp.MAX_TRIES:
            print_('too many retries: giving up')
            return -1
        timeout = pcp.gettimeout(1)
        fds = select.select([pcpfd], [], [], timeout)
        if len(fds[0]) > 0:
            (ret, resp) = pcp.recvresponse()
            if ret == libpcp.TRY_AGAIN:
                if verbose:
                    print_('trying again...')
                continue
            print_('epoch:', resp['epoch'])
            print_('internal port:', resp['intport'])
            print_('remote address:', resp['remotepeer'])
            print_('remote port:', resp['peerport'])
            print_('lifetime:', resp['lifetime'])
            printoptions(resp['options'])
            pcp.close()
            return resp['result']
        else:
            if verbose:
                print_('timeout: retrying...')
            pcp.sendrequest()

def usage(progname):
    """usage"""
    print_('usage:', progname,
          '[-v] <server> <address> <priv> <dst>',
          '<port> <pub> <proto> <lifetime>')

def main(args):
    """main"""
    pcp = None
    verbose = False
    proto = -1

    try:
        opts, argv = getopt.getopt(args[1:], 'v')
    except getopt.GetoptError:
        usage(args[0])
        raise
    for opt, _ in opts:
        if opt == '-v':
            verbose = True
            continue

    if len(argv) != 8:
        usage(args[0])
        return -1
    if argv[6] == 'all':
        proto = 0
    elif argv[6] == 'udp':
        proto = socket.IPPROTO_UDP
    elif argv[6] == 'tcp':
        proto = socket.IPPROTO_TCP
    else:
        print_('unknown protocol:', argv[6])
        return -1

    pcp = libpcp.PCP()
    if pcp is None:
        print_('PCP object creation failure?')
        return -1
    pcp.open(server=argv[0], source=argv[1])
    pcp.makerequest(
        opcode = libpcp.OPCODE_PEER,
        protocol = proto,
        remotepeer = argv[3],
        lifetime = int(argv[7]),
        intport = int(argv[2]),
        extport = int(argv[5]),
        peerport = int(argv[4]),
        extaddr = '0.0.0.0')
    return loop(pcp, verbose)

main(sys.argv)
