/*
 * Copyright (C) 2011-2012  Internet Systems Consortium, Inc. ("ISC")
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND ISC DISCLAIMS ALL WARRANTIES WITH
 * REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS.  IN NO EVENT SHALL ISC BE LIABLE FOR ANY SPECIAL, DIRECT,
 * INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
 * OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */

/* $Id: 44c8cf33aa003e7c49f1a072d8cd2e44389d6086 $ */

/*
 * PCP (port-control-protocol) IPv6 rapid recovery multicast listener
 *
 * Francis_Dupont@isc.org, November 2011
 *
 * ./rclisten6 <interface_index>
 */

#ifdef WIN32
#include <ws2tcpip.h>
#else
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <fcntl.h>
#include <unistd.h>
#endif
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "pcp.h"

int
main(int argc, char *argv[])
{
	pcp_t pcp;
	pcp_response_t response;
#ifdef WIN32
	SOCKET fd;
	u_long ioctlArg;
#else
	int fd, on, flags;
#endif
	int ret, ret2;
	struct sockaddr_in6 service;
	struct ipv6_mreq mr;
	fd_set set;
#ifdef WIN32
	WSADATA wsaData;
#endif

#define USAGE	"usage: %s <interface_index>\n"
	if (argc != 2) {
		fprintf(stderr, USAGE, argv[0]);
		exit(1);
	}
#ifdef WIN32
	ret = WSAStartup(MAKEWORD(2,2), &wsaData);
	if (ret != 0) {
		fprintf(stderr, "WSAStartup() failed.\n");
		exit(1);
	}
#endif
	memset(&service, 0, sizeof(service));
	service.sin6_family = AF_INET6;
#ifdef BSD
	service.sin6_len = sizeof(service);
#endif
	service.sin6_port = htons(PCP_RC_PORT);
	memset(&mr, 0, sizeof(mr));
	if (inet_pton(AF_INET6, PCP_RC_ADDR6, &mr.ipv6mr_multiaddr) <= 0) {
		fprintf(stderr, "inet_ntop(%s) failed\n", PCP_RC_ADDR6);
		exit(1);
	}
	mr.ipv6mr_interface = atoi(argv[1]);
	fd = socket(PF_INET6, SOCK_DGRAM, IPPROTO_UDP);
#ifdef WIN32
	if (fd < 0) {
		fprintf(stderr, "socket: %d\n", WSAGetLastError());
		exit(1);
	}
	if (bind(fd, (struct sockaddr *) &service, sizeof(service)) < 0) {
		fprintf(stderr, "bind: %d\n", WSAGetLastError());
		exit(1);
	}
	ioctlArg = 1;
	if (ioctlsocket(fd, FIONBIO, &ioctlArg) != 0) {
		fprintf(stderr, "ioctl(FIONBIO): %d\n", WSAGetLastError());
		closesocket(fd);
		exit(1);
	}
	if (setsockopt(fd, IPPROTO_IPV6, IPV6_ADD_MEMBERSHIP,
		       (char *) &mr, sizeof(mr)) < 0) {
		fprintf(stderr, "setsockopt(ADD_MEMBERSHIP): %d\n",
			WSAGetLastError());
		closesocket(fd);
		exit(1);
	}
#else
	if (fd < 0) {
		fprintf(stderr, "socket: %s\n", strerror(errno));
		exit(1);
	}
	on = 1;
#ifndef SO_REUSEPORT
#define SO_REUSEPORT	SO_REUSEADDR
#endif
	if (setsockopt(fd, SOL_SOCKET, SO_REUSEPORT, &on, sizeof(on)) < 0) {
		fprintf(stderr, "setsockopt(REUSEPORT): %s\n",
			strerror(errno));
		close(fd);
		exit(1);
	}
	if (bind(fd, (struct sockaddr *) &service, sizeof(service)) < 0) {
		fprintf(stderr, "bind: %s\n", strerror(errno));
		exit(1);
	}
	if ((flags = fcntl(fd, F_GETFL, 0)) < 0) {
		fprintf(stderr, "fcntl(GETFL): %s\n", strerror(errno));
		close(fd);
		exit(1);
	}
	if (fcntl(fd, F_SETFL, flags | O_NONBLOCK) < 0) {
		fprintf(stderr, "fcntl(SETFL): %s\n", strerror(errno));
		close(fd);
		exit(1);
	}
	if (setsockopt(fd, IPPROTO_IPV6, IPV6_JOIN_GROUP,
		       &mr, sizeof(mr)) < 0) {
		fprintf(stderr, "setsockopt(JOIN_GROUP): %s\n",
			strerror(errno));
		close(fd);
		exit(1);
	}
#endif
	memset(&service, 0, sizeof(service));
	service.sin6_family = AF_INET6;
#ifdef BSD
	service.sin6_len = sizeof(service);
#endif
#ifdef WIN32
	if (inet_pton(AF_INET6, "::1", &service.sin6_addr) <= 0) {
		fprintf(stderr, "inet_ntop(loopback) failed\n");
		closesocket(fd);
		exit(1);
	}
#endif
	ret = pcp_init((struct sockaddr *) &service, NULL, &pcp);
	if (ret < PCP_OK) {
		fprintf(stderr, "pcp_init: %s\n", pcp_strerror(ret));
#ifdef WIN32
		closesocket(fd);
#else
		close(fd);
#endif
		exit(1);
	}
	ret = pcp_setrcsocket(&pcp, fd);
	if (ret < PCP_OK) {
		fprintf(stderr, "pcp_setrcsocket(): %s\n", pcp_strerror(ret));
		goto done;
	}
	for (;;) {
		FD_ZERO(&set);
		FD_SET(fd, &set);
		ret = select(fd + 1, &set, NULL, NULL, NULL);
		memset(&response, 0, sizeof(response));
		if (ret < 0) {
#ifdef WIN32
			fprintf(stderr, "select: %d\n", WSAGetLastError());
#else
			fprintf(stderr, "select: %s\n", strerror(errno));
#endif
			continue;
		}
		ret = pcp_recvresponse(&pcp, &response);
		if (ret < PCP_OK) {
			fprintf(stderr,
				"pcp_recvresponse: %s\n",
				pcp_strerror(ret));
			if (ret == PCP_ERR_RECVBAD) {
				fprintf(stderr,
					"at line %d\n",
					pcp_debug_line);
				abort();
			}
			goto done;
		}
		if (ret != PCP_TRY_AGAIN)
			fprintf(stderr,
				"not try again: %s\n",
				pcp_strerror(ret));
		break;
	}
	switch (response.assigned.opcode) {
	case PCP_OPCODE_ANNOUNCE | PCP_OPCODE_RESPONSE:
		printf("opcode: ANNOUNCE (%d)\n",
		       PCP_OPCODE_ANNOUNCE | PCP_OPCODE_RESPONSE);
		break;
	case PCP_OPCODE_MAP | PCP_OPCODE_RESPONSE:
		printf("opcode: MAP (%d)\n",
		       PCP_OPCODE_MAP | PCP_OPCODE_RESPONSE);
		break;
	case PCP_OPCODE_PEER | PCP_OPCODE_RESPONSE:
		printf("opcode: PEER (%d)\n",
		       PCP_OPCODE_PEER | PCP_OPCODE_RESPONSE);
		break;
	default:
		printf("opcode: %d\n", (int) response.assigned.opcode);
	}
	printf("lifetime: %d\n", (int) response.assigned.lifetime);
	printf("epoch: %d\n", response.epoch);
	ret = PCP_OK;
done:
	ret2 = pcp_close(&pcp);
	if (ret2 < PCP_OK)
		fprintf(stderr, "pcp_close: %s\n", pcp_strerror(ret2));
#ifdef WIN32
	closesocket(fd);
	(void) WSACleanup();
#else
	close(fd);
#endif
	exit(ret);
}
