/*
 * Copyright (C) 2011-2012  Internet Systems Consortium, Inc. ("ISC")
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND ISC DISCLAIMS ALL WARRANTIES WITH
 * REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS.  IN NO EVENT SHALL ISC BE LIABLE FOR ANY SPECIAL, DIRECT,
 * INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
 * OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */

/* $Id: 4bebe36bfb0dbfd86ec3833c685b3282e88d7227 $ */

/*
 * PCP (port-control-protocol) server
 *
 * Francis_Dupont@isc.org, August 2011
 *
 * usage: pcpd [-d] <args>, arguments (including the option -d) are
 * given to the CGN code through pcpd_cgn_init().
 *
 * schema: active/master mode
 *
 * +--------------------+.---------------.--------------------+
 * | PCP server library || pcpd_t object | CGN implementation |
 * |                    ||               |                    |
 * |   PCP protocol     || recvrequest   |                    |
 * |      engine        || sendresponse  |   network          |
 * |                    || addmapping    |   interfaces       |
 * |                    || delmapping    |                    |
 * |                    ||     ...       |   mapping          |
 * |     main()         |.---------------.   management       |
 * |                    ||     ||                             |
 * |   mainloop()       ||     ||            etc              |
 * |                   <=======//                             |
 * |                   => cgn_init()                          |
 * |                    ||                                    |
 * +--------------------++------------------------------------+
 *
 * schema: passive/slave mode
 *
 * +--------------------+.---------------.--------------------+
 * | PCP server library || pcpd_t object | CGN implementation |
 * |                    ||               |                    |
 * |   PCP protocol     || recvrequest   |                    |
 * |      engine        || sendresponse  |   network          |
 * |                    || addmapping    |   interfaces       |
 * |                    || delmapping    |                    |
 * |                    ||     ...       |   mapping          |
 * |                    |.---------------.   management       |
 * |                    ||     ^^                             |
 * |                    ||     ||            etc              |
 * |                   ========//                             |
 * |          lib_init() <=                  main()           |
 * |                    ||                                    |
 * |                    ||                   mainloop()       |
 * |                    ||                                    |
 * +--------------------++------------------------------------+
 *
 * in active/master mode, the CGN must provide pcpd_cgn_init(),
 * pcpd_cgn_close() and with PCP subscriber ID caching or MAP update
 * pcpd_cgn_register() (all called from the CGN main())
 *
 * in passive/slave mode, the CGN must call pcpd_lib_init(), and
 * can call pcpd_lib_runonce() (receive PCP request, process, send
 * PCP response), pcpd_lib_close(), optionally pcpd_update() and
 * with PCP subscriber ID caching pcpd_uncache().
 *
 * compilation flags are:
 *  NO_SOCKET		(default undefined)
 *  NO_SYSLOG		(default undefined)
 *  HASHSIZE		(default 128)
 *  CHECKLIB		(default 0)
 *  PASSIVE_ONLY	(default undefined)
 */

#include "pcpd.h"

#ifndef NO_SOCKET
#include <arpa/inet.h>
#include <netinet/in.h>
#endif

#include <stdlib.h>
#ifndef NO_SYSLOG
#include <syslog.h>
#else
#define LOG_NDELAY 0
#define LOG_PERROR 1
#define LOG_DEBUG 2
#define LOG_LOCAL6 3
#endif

/* check stack / buffer sizes */

#if PCPD_SID_SIZE < 16
#error "PCPD_SID_SIZE must be at least 16"
#elif PCPD_SID_SIZE > 256
#error "oversized PCPD_SID_SIZE"
#endif

size_t pcpd_sid_size = PCPD_SID_SIZE;

#if PCPD_BUF_SIZE < 1025
#error "PCPD_BUF_SIZE must be at least 1025"
#elif PCPD_BUF_SIZE > 2048
#error "oversized PCPD_BUF_SIZE"
#endif

size_t pcpd_buf_size = PCPD_BUF_SIZE;

/* Local macros */

#define UNUSED(x)	(x) = (x)

/* used to make error cases not common */

#ifdef __GNUC__
#define likely(expr)	__builtin_expect(expr, 1)
#define unlikely(expr)	__builtin_expect(expr, 0)
#else
#define likely(expr)	(expr)
#define unlikely(expr)	(expr)
#endif

/* double linked list from BSD queue.h */

#define ISC_LIST_HEAD(name, type)					\
	struct name {							\
		struct type *lh_first;					\
	}

#define ISC_LIST_INIT(head)						\
	do {								\
		(head)->lh_first = NULL;				\
	} while (0)

#define ISC_LIST_ENTRY(type)						\
	struct {							\
		struct type *le_next;					\
		struct type **le_prev;					\
	}

#define ISC_LIST_EMPTY(head)						\
	((head)->lh_first == NULL)

#define ISC_LIST_FIRST(head)						\
	((head)->lh_first)

#define ISC_LIST_NEXT(elm, field)					\
	((elm)->field.le_next)

#define ISC_LIST_PREV(elm, field)					\
	((elm)->field.le_prev)

#define ISC_LIST_INSERT(head, elm, field)				\
	do {								\
		(elm)->field.le_next = (head)->lh_first;		\
		if ((head)->lh_first != NULL)				\
			(head)->lh_first->field.le_prev =		\
				&(elm)->field.le_next;			\
		(head)->lh_first = (elm);				\
		(elm)->field.le_prev = &(head)->lh_first;		\
	} while (0)

#define ISC_LIST_REMOVE(elm, field)					\
	do {								\
		if ((elm)->field.le_next != NULL)			\
			(elm)->field.le_next->field.le_prev =		\
				(elm)->field.le_prev;			\
		*(elm)->field.le_prev = (elm)->field.le_next;		\
		(elm)->field.le_next = NULL;				\
		(elm)->field.le_prev = NULL;				\
	} while (0)

#define ISC_LIST_FOREACH(var, head, field)				\
	for ((var) = (head)->lh_first;					\
	     (var) != NULL;						\
	     (var) = (var)->field.le_next)

#define ISC_LIST_FOREACH_SAFE(var, head, field, tvar)			\
	for ((var) = (head)->lh_first;					\
	     ((var) != NULL) &&	(((tvar) = (var)->field.le_next), 1);	\
	     (var) = (tvar))

/* masks */

uint8_t zero[16];

const uint8_t mask[129][16] = {
	{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0x80, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0xc0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0xe0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0xf0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0xf8, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0xfc, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0xfe, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0xff, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0xff, 0x80, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0xff, 0xc0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0xff, 0xe0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0xff, 0xf0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0xff, 0xf8, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0xff, 0xfc, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0xff, 0xfe, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0x80, 0x00, 0x00, 0x00, 0x00, 0x00,
	  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0xc0, 0x00, 0x00, 0x00, 0x00, 0x00,
	  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0xe0, 0x00, 0x00, 0x00, 0x00, 0x00,
	  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0xf0, 0x00, 0x00, 0x00, 0x00, 0x00,
	  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0xf8, 0x00, 0x00, 0x00, 0x00, 0x00,
	  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0xfc, 0x00, 0x00, 0x00, 0x00, 0x00,
	  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0xfe, 0x00, 0x00, 0x00, 0x00, 0x00,
	  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0x00, 0x00, 0x00, 0x00, 0x00,
	  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0x80, 0x00, 0x00, 0x00, 0x00,
	  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0xc0, 0x00, 0x00, 0x00, 0x00,
	  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0xe0, 0x00, 0x00, 0x00, 0x00,
	  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0xf0, 0x00, 0x00, 0x00, 0x00,
	  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0xf8, 0x00, 0x00, 0x00, 0x00,
	  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0xfc, 0x00, 0x00, 0x00, 0x00,
	  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0xfe, 0x00, 0x00, 0x00, 0x00,
	  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff, 0x00, 0x00, 0x00, 0x00,
	  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff, 0x80, 0x00, 0x00, 0x00,
	  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff, 0xc0, 0x00, 0x00, 0x00,
	  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff, 0xe0, 0x00, 0x00, 0x00,
	  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff, 0xf0, 0x00, 0x00, 0x00,
	  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff, 0xf8, 0x00, 0x00, 0x00,
	  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff, 0xfc, 0x00, 0x00, 0x00,
	  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff, 0xfe, 0x00, 0x00, 0x00,
	  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff, 0xff, 0x00, 0x00, 0x00,
	  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff, 0xff, 0x80, 0x00, 0x00,
	  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xc0, 0x00, 0x00,
	  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xe0, 0x00, 0x00,
	  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xf0, 0x00, 0x00,
	  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xf8, 0x00, 0x00,
	  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xfc, 0x00, 0x00,
	  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xfe, 0x00, 0x00,
	  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0x00, 0x00,
	  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0x80, 0x00,
	  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xc0, 0x00,
	  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xe0, 0x00,
	  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xf0, 0x00,
	  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xf8, 0x00,
	  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xfc, 0x00,
	  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xfe, 0x00,
	  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0x00,
	  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0x80,
	  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xc0,
	  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xe0,
	  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xf0,
	  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xf8,
	  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xfc,
	  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xfe,
	  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	  0x80, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	  0xc0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	  0xe0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	  0xf0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	  0xf8, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	  0xfc, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	  0xfe, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	  0xff, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	  0xff, 0x80, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	  0xff, 0xc0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	  0xff, 0xe0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	  0xff, 0xf0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	  0xff, 0xf8, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	  0xff, 0xfc, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	  0xff, 0xfe, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	  0xff, 0xff, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	  0xff, 0xff, 0x80, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	  0xff, 0xff, 0xc0, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	  0xff, 0xff, 0xe0, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	  0xff, 0xff, 0xf0, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	  0xff, 0xff, 0xf8, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	  0xff, 0xff, 0xfc, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	  0xff, 0xff, 0xfe, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	  0xff, 0xff, 0xff, 0x00, 0x00, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	  0xff, 0xff, 0xff, 0x80, 0x00, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	  0xff, 0xff, 0xff, 0xc0, 0x00, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	  0xff, 0xff, 0xff, 0xe0, 0x00, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	  0xff, 0xff, 0xff, 0xf0, 0x00, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	  0xff, 0xff, 0xff, 0xf8, 0x00, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	  0xff, 0xff, 0xff, 0xfc, 0x00, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	  0xff, 0xff, 0xff, 0xfe, 0x00, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	  0xff, 0xff, 0xff, 0xff, 0x00, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	  0xff, 0xff, 0xff, 0xff, 0x80, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	  0xff, 0xff, 0xff, 0xff, 0xc0, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	  0xff, 0xff, 0xff, 0xff, 0xe0, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	  0xff, 0xff, 0xff, 0xff, 0xf0, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	  0xff, 0xff, 0xff, 0xff, 0xf8, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	  0xff, 0xff, 0xff, 0xff, 0xfc, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	  0xff, 0xff, 0xff, 0xff, 0xfe, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	  0xff, 0xff, 0xff, 0xff, 0xff, 0x00, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	  0xff, 0xff, 0xff, 0xff, 0xff, 0x80, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	  0xff, 0xff, 0xff, 0xff, 0xff, 0xc0, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	  0xff, 0xff, 0xff, 0xff, 0xff, 0xe0, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	  0xff, 0xff, 0xff, 0xff, 0xff, 0xf0, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	  0xff, 0xff, 0xff, 0xff, 0xff, 0xf8, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	  0xff, 0xff, 0xff, 0xff, 0xff, 0xfc, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	  0xff, 0xff, 0xff, 0xff, 0xff, 0xfe, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	  0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0x00, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	  0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0x80, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	  0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xc0, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	  0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xe0, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	  0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xf0, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	  0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xf8, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	  0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xfc, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	  0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xfe, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	  0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0x00 },
	{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	  0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0x80 },
	{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	  0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xc0 },
	{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	  0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xe0 },
	{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	  0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xf0 },
	{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	  0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xf8 },
	{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	  0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xfc },
	{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	  0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xfe },
	{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	  0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff }
};

/* global references */

in_addr_t cpe;
int terminate;

/* subscribers */

#ifndef HASHSIZE
#define HASHSIZE	128
#endif

struct subscriber {
	ISC_LIST_ENTRY(subscriber) chain;	/* in bucket chain */
	const pcpd_t *pcpd;			/* pcpd object it belongs to */
	const uint8_t id[PCPD_SID_SIZE];	/* identifier */
	time_t epoch;				/* date of epoch */
};
ISC_LIST_HEAD(bucket, subscriber) subscribers[HASHSIZE];

#ifndef CHECKLIB
#define CHECKLIB 0
#endif
#if CHECKLIB
struct cgn {
	ISC_LIST_ENTRY(cgn) chain;		/* CGN chain */
	uint32_t refcnt;			/* reference counter */
	const void *pointer;			/* pointer we got */
	pcpd_t rovalue;				/* read-only value */
};
ISC_LIST_HEAD(cgns, cgn) cgns;
#endif

/* debug helper */

static void
debug(const pcpd_t *pcpd, const char *msg, ...)
{
	va_list va;

	va_start(va, msg);
	pcpd_vsyslog(pcpd, LOG_DEBUG, msg, va);
	va_end(va);
}

/* check initialized pcpd sanity */

char *checkpb;

static int
checkpcpd(const pcpd_t *pcpd)
{
	if (unlikely((pcpd->opcodes & PCPD_OPCODE_MAP) == 0)) {
		checkpb = "no MAP?";
		return 0;
	}
	if (unlikely((pcpd->flags & PCPD_PORT0) &&
		     ((pcpd->flags & PCPD_OTHERPROTO) == 0))) {
		checkpb = "port == 0 support with only TCP and UDP?";
		return 0;
	}
	if (unlikely((pcpd->flags & PCPD_ONEONE) &&
		     ((pcpd->flags & PCPD_OTHERPROTO) == 0))) {
		checkpb = "NAT 1:1 but not other proto?";
		return 0;
	}
	if (unlikely((pcpd->flags & PCPD_ONEONE) &&
		     ((pcpd->flags & PCPD_PORT0) == 0))) {
		checkpb= "NAT 1:1 but no port == 0 support?";
		return 0;
	}
	if (unlikely((pcpd->flags & PCPD_FIREWALL) &&
		     (pcpd->opcodes & PCPD_OPCODE_PEER))) {
		checkpb = "PEER in firewall?";
		return 0;
	}
	if (unlikely((pcpd->flags & PCPD_FIREWALL) &&
		     ((pcpd->flags & PCPD_MAPALL) != 0) &&
		     ((pcpd->flags & PCPD_MAPALL) != PCPD_MAPALL))) {
		checkpb = "inconsistent mapped in firewall?";
		return 0;
	}
	if (unlikely(((pcpd->flags & PCPD_3RDPARTY) == 0) &&
		     (pcpd->flags & (PCPD_3RDPARTY_CPE | PCPD_3RDPARTY_AO)))) {
		checkpb = "inconsistent third party";
		return 0;
	}
	if (unlikely((pcpd->cgntype == PCPD_TYPE_NAT) &&
		     (pcpd->flags & PCPD_CACHE))) {
		checkpb = "subscriber ID caching for a not CGN";
		return 0;
	}
	if (unlikely(pcpd->pcpd_recvrequest == NULL)) {
		checkpb = "need pcpd_recvrequest";
		return 0;
	}
	if (unlikely(pcpd->pcpd_sendresponse == NULL)) {
		checkpb = "need pcpd_sendresponse";
		return 0;
	}
	if (unlikely(pcpd->pcpd_getsubscriber == NULL)) {
		checkpb = "need pcpd_getsubscriber";
		return 0;
	}
	if (unlikely(pcpd->pcpd_addmapping == NULL)) {
		checkpb = "need pcpd_addmapping";
		return 0;
	}
	if (unlikely(pcpd->pcpd_delmapping == NULL)) {
		checkpb = "need pcpd_delmapping";
		return 0;
	}
	if (unlikely((pcpd->opcodes & PCPD_OPCODE_PEER) &&
		     (pcpd->pcpd_adddynamic == NULL))) {
		checkpb = "inconsistent PEER support";
		return 0;
	}
	return 1;
}

/* Jenkins hash */

#define __jhash_mix(a, b, c) { \
	a -= b; a -= c; a ^= (c >> 13);	\
	b -= c; b -= a; b ^= (a << 8);	\
	c -= a; c -= b; c ^= (b >> 13);	\
	a -= b; a -= c; a ^= (c >> 12);	\
	b -= c; b -= a; b ^= (a << 16);	\
	c -= a; c -= b; c ^= (b >> 5);	\
	a -= b; a -= c; a ^= (c >> 3);	\
	b -= c; b -= a; b ^= (a << 10);	\
	c -= a; c -= b; c ^= (b >> 15);	}

#define JHASH_GOLDEN_RATIO	0x9e3779b9

static inline unsigned short
jhash(const uint8_t *sid)
{
	char *id = (char *) sid;
	uint32_t a, b, c, k;

	memcpy(&a, id, 4);
	a += JHASH_GOLDEN_RATIO;
	memcpy(&b, id + 4, 4);
	b += JHASH_GOLDEN_RATIO;
	memcpy(&c, id + 8, 4);

	__jhash_mix(a, b, c);

	memcpy(&k, id + 12, 4);
	a += k;

	__jhash_mix(a, b, c);

	c = (c >> 16) + (c & 0xffff);
	c &= 0xffff;
	return (unsigned short) c;
}

/* subscribers */

static struct subscriber *
getsubscriber(const pcpd_t *pcpd, const pcpctx_t *ctx, int *error)
{
	struct subscriber *s;
	struct bucket *bucket;
	unsigned short hash;

	if ((pcpd->flags & PCPD_CACHE) == 0) {
		s = (struct subscriber *) pcpd_malloc(pcpd, sizeof(*s));
		if (unlikely(s == NULL)) {
			debug(pcpd, "malloc(subscriber)");
			*error = PCPD_ERR_NORESOURCES;
			return s;
		}
		memset(s, 0, sizeof(*s));
		memcpy((void *) s->id, ctx->sid, PCPD_SID_SIZE);
		*error = pcpd->pcpd_getsubscriber(pcpd, ctx, &s->epoch);
		if (unlikely(*error < PCPD_OK)) {
			debug(pcpd, "getsubscriber: %s",
			      pcpd_strerror(*error));
			pcpd_free(pcpd, s);
			return NULL;
		}
		return s;
	}
	hash = jhash(ctx->sid);
	bucket = &subscribers[hash];
	ISC_LIST_FOREACH(s, bucket, chain)
		if ((s->pcpd == pcpd) &&
		    (memcmp(ctx->sid, &s->id, PCPD_SID_SIZE) == 0))
			break;
	if (s != NULL) {
		if (s != ISC_LIST_FIRST(bucket)) {
			ISC_LIST_REMOVE(s, chain);
			ISC_LIST_INSERT(bucket, s, chain);
		}
		*error = PCPD_OK;
		return s;
	}
	s = (struct subscriber *) pcpd_malloc(pcpd, sizeof(*s));
	if (unlikely(s == NULL)) {
		debug(pcpd, "malloc(subscriber)");
		*error = PCPD_ERR_NORESOURCES;
		return s;
	}
	memset(s, 0, sizeof(*s));
	s->pcpd = pcpd;
	memcpy((void *) s->id, ctx->sid, PCPD_SID_SIZE);
	*error = pcpd->pcpd_getsubscriber(pcpd, ctx, &s->epoch);
	if (unlikely(*error < PCPD_OK)) {
		debug(pcpd, "getsubscriber: %s", pcpd_strerror(*error));
		pcpd_free(pcpd, s);
		return NULL;
	}
	ISC_LIST_INSERT(bucket, s, chain);
	return s;
}

/* seconds since start of epoch */

static inline uint32_t
sssoe(const pcpd_t *pcpd, const struct subscriber *s)
{
	return (uint32_t)(pcpd_time(pcpd) - s->epoch);
}

/* callback on subscriber destroy/reset */

void
pcpd_uncache(const pcpd_t *pcpd, const pcpctx_t *ctx)
{
	struct subscriber *s;
	struct bucket *bucket;
	unsigned short hash;

	if (unlikely((pcpd == NULL) || (ctx == NULL)))
		return;
	if ((pcpd->flags & PCPD_CACHE) == 0) {
		debug(pcpd, "uncache without caching");
		return;
	}
	hash = jhash(ctx->sid);
	bucket = &subscribers[hash];
	ISC_LIST_FOREACH(s, bucket, chain)
		if ((s->pcpd == pcpd) &&
		    (memcmp(ctx->sid, &s->id, PCPD_SID_SIZE) == 0))
			break;
	if (unlikely(s == NULL)) {
		debug(pcpd, "uncache can't find '%s'",
		      pcpd_id2str(pcpd, ctx->sid));
		return;
	}
	ISC_LIST_REMOVE(s, chain);
	debug(pcpd, "uncache free '%s'", pcpd_id2str(pcpd, ctx->sid));
	pcpd_free(pcpd, s);
}

/* set an error response code */

static void
seterr(const pcpd_t *pcpd, const pcpctx_t *ctx,
       uint8_t code, uint32_t holdtime)
{
	uint8_t *resp = ctx->response;

	if (likely(pcpd->pcpd_incrcounter != NULL))
		pcpd->pcpd_incrcounter(pcpd, PCPS_RESULT, code);
	debug(pcpd, "set error %d", (int) code);
	resp[PCP_RESULT_CODE] = code;
	if (holdtime == 0)
		holdtime = errholdtime(pcpd, code);
	holdtime = htonl(holdtime);
	memcpy(resp + PCP_LIFETIME, &holdtime, 4);
}

/* process an ANNOUNCE request */

static size_t
announcerequest(const pcpd_t *pcpd, pcpctx_t *ctx, size_t l)
{
	uint8_t *req = ctx->request, *resp = ctx->response;
	uint8_t option;
	uint16_t optlen, unproc = 0;
	size_t off;
	int result;

	/*
	 * process options
	 */

	off = PCP_LEN_ANNOUNCE;
	ctx->off3rdparty = 0;
	while (off < l) {
		if (unlikely(off + PCP_LEN_OPTHDR > l)) {
			debug(pcpd, "option header overflow");
			return 0;
		}
		option = req[off + PCP_OPT_CODE];
		if (likely(pcpd->pcpd_incrcounter != NULL))
			pcpd->pcpd_incrcounter(pcpd, PCPS_OPTION, option);
		memcpy(&optlen, req + off + PCP_OPT_LEN, 2);
		optlen = ntohs(optlen);
		if (unlikely(off + PCP_LEN_OPTHDR + optlen > l)) {
			debug(pcpd, "option overflow");
			return 0;
		}
		debug(pcpd, "unknown option %d", (int) option);
		/* if mandatory to process: unsupported */
		if ((option & PCP_OPTION_OPTIONAL) == 0)
			return 0;
		else {
			/* mark option for removal from response */
			resp[off + PCP_OPT_CODE] = 0;
			++unproc;
		}
		off += PCP_LEN_OPTHDR + PCP_ALIGN(optlen);
	}

	/* check authorization now */

	if (pcpd->pcpd_authorize != NULL) {
		result = pcpd->pcpd_authorize(pcpd, ctx, l);
		if (unlikely(result != PCPD_OK))
			return 0;
	}

	/* lifetime */
	memset(resp + PCP_LIFETIME, 0, 4);
	if (likely(pcpd->pcpd_incrcounter != NULL))
		pcpd->pcpd_incrcounter(pcpd, PCPS_RESULT, PCP_ERR_SUCCESS);

	if (unlikely(unproc != 0)) {
		/* remove unprocessed options from the response */
		for (off = PCP_LEN_PEER; off < l; ) {
			memcpy(&optlen, resp + off + PCP_OPT_LEN, 2);
			optlen = ntohs(optlen);
			optlen = PCP_LEN_OPTHDR + PCP_ALIGN(optlen);
			if (resp[off + PCP_OPT_CODE] == 0) {
				memmove(resp + off,
					resp + off + optlen,
					optlen);
				l -= optlen;
			}
			else
				off += optlen;
		}
	}

	return l;
}

/* process a MAP request */

static size_t
maprequest(const pcpd_t *pcpd, pcpctx_t *ctx, size_t l)
{
	uint8_t *req = ctx->request, *resp = ctx->response;
	uint8_t proto, option, *optdata, errcode, flags = 0;
	uint8_t intaddr[16], extaddr[16];
	const uint8_t *msk;
	uint16_t sport, nport, optlen, unproc = 0;
	uint32_t lifetime;
	unsigned int i, prefail = 0, filter = 0;
	unsigned char filter0 = 0, newmap = 0;
	size_t off;
	int result;
	filter_t *f, *filters = NULL;

	if (unlikely(l < PCP_LEN_MAP)) {
		debug(pcpd, "underrun %d", l);
		seterr(pcpd, ctx, PCP_ERR_BADREQUEST, pcpd->longerrht);
		return l;
	}

	/* decode */
	memcpy(&lifetime, req + PCP_LIFETIME, 4);
	lifetime = ntohl(lifetime);
	proto = req[PCP_MAP_PROTOCOL];
	memcpy(&sport, req + PCP_MAP_INTERNAL_PORT, 2);
	sport = ntohs(sport);
	memcpy(&nport, req + PCP_MAP_EXTERNAL_PORT, 2);
	nport = ntohs(nport);
	memcpy(extaddr, req + PCP_MAP_EXTERNAL_ADDR, 16);
	if ((pcpd->flags & PCPD_MAPEXT) && !pcpd_unmapaddr(extaddr)) {
		debug(pcpd, "external addr not mapped");
		seterr(pcpd, ctx, PCP_ERR_BADREQUEST, pcpd->longerrht);
		return l;
	}

	/* check ports for a firewall */
	if (unlikely((pcpd->flags & PCPD_FIREWALL) && (sport != nport))) {
		debug(pcpd, "firewall with different ports");
		seterr(pcpd, ctx, PCP_ERR_BADREQUEST, pcpd->longerrht);
		return l;
	}

	/* check protocol */
	if ((pcpd->flags & PCPD_OTHERPROTO) == 0) {
		if (unlikely((proto != IPPROTO_TCP) &&
			     (proto != IPPROTO_UDP) &&
			     (proto != 0))) {
			debug(pcpd, "bad protocol %d", (int) proto);
			seterr(pcpd, ctx, PCP_ERR_UNSUPPROTO,
			       pcpd->longerrht);
			return l;
		}
	}
	if ((pcpd->flags & PCPD_ONEONE) == 0) {
		if (unlikely((proto == 0) && (lifetime != 0))) {
			debug(pcpd, "bad protocol zero");
			seterr(pcpd, ctx, PCP_ERR_UNSUPPROTO,
			       pcpd->longerrht);
			return l;
		}
	}
	if (unlikely((proto == 0) && (sport != 0))) {
		debug(pcpd, "source port != 0 in proto == 0");
		seterr(pcpd, ctx, PCP_ERR_BADREQUEST, pcpd->longerrht);
		return l;
	}

	/*
	 * process options
	 */

	off = PCP_LEN_MAP;
	ctx->off3rdparty = 0;
	while (off < l) {
		if (unlikely(off + PCP_LEN_OPTHDR > l)) {
			debug(pcpd, "option header overflow");
			seterr(pcpd, ctx, PCP_ERR_BADOPTION, pcpd->longerrht);
			goto done;
		}
		option = req[off + PCP_OPT_CODE];
		if (likely(pcpd->pcpd_incrcounter != NULL))
			pcpd->pcpd_incrcounter(pcpd, PCPS_OPTION, option);
		memcpy(&optlen, req + off + PCP_OPT_LEN, 2);
		optlen = ntohs(optlen);
		if (unlikely(off + PCP_LEN_OPTHDR + optlen > l)) {
			debug(pcpd, "option overflow");
			seterr(pcpd, ctx, PCP_ERR_BADOPTION, pcpd->longerrht);
			goto done;
		}
		optdata = req + off + PCP_OPT_DATA;
		switch (option) {
		case PCP_OPTION_THIRD_PARTY:
			/* disallowed */
			if (unlikely((pcpd->flags & PCPD_3RDPARTY) == 0)) {
				debug(pcpd, "3rd party option disallowed");
			unsupp:
				seterr(pcpd, ctx,
				       PCP_ERR_UNSUPOPTION,
				       pcpd->longerrht);
				goto done;
			}
			/* once */
			if (unlikely(ctx->off3rdparty != 0)) {
				debug(pcpd, "multiple 3rd party option");
			malformed:
				seterr(pcpd, ctx,
				       PCP_ERR_BADOPTION,
				       pcpd->longerrht);
				goto done;
			}
			/* length */
			if (unlikely(optlen != PCP_LEN_OPT3PTY)) {
				debug(pcpd, "3rd party bad length %d", optlen);
				goto malformed;
			}
			/* not the same than the source address */
			if (unlikely(memcmp(optdata,
					    req + PCP_CLIENT_ADDR,
					    16) == 0)) {
				debug(pcpd, "3rd party same as source");
				seterr(pcpd, ctx,
				       PCP_ERR_BADREQUEST,
				       pcpd->longerrht);
				goto done;
			}
			ctx->off3rdparty = off;
			/* zero only */
			if (pcpd->flags & PCPD_3RDPARTY_AO) {
				if (unlikely((memcmp(intaddr, zero,
						     sizeof(zero) != 0)) &&
					     (memcmp(intaddr, pcpd_mapped,
						     sizeof(zero) != 0)))) {
					debug(pcpd, "3rd party not zero");
					goto malformed;
				}
			}
			memcpy(intaddr, optdata, 16);
			if ((pcpd->flags & PCPD_MAPINT) &&
			    (memcmp(intaddr, zero, sizeof(zero)) != 0) &&
			    !pcpd_unmapaddr(intaddr)) {
				debug(pcpd, "3rd party not mapped");
				goto malformed;
			}
			/* CPE only */
			if (pcpd->flags & PCPD_3RDPARTY_CPE) {
				if (unlikely((memcmp(req + PCP_CLIENT_ADDR,
						     pcpd_mapped, 12) != 0) ||
					     (memcmp(req +
						     PCP_CLIENT_ADDR + 12,
						     &cpe, 4) != 0))) {
					debug(pcpd, "3rd party not from CPE");
					goto malformed;
				}
			}
			/* the CPE will validate the address */
			break;

		case PCP_OPTION_PREF_FAIL:
			if (unlikely((optlen != PCP_LEN_OPT0) ||
				     prefail || (lifetime == 0))) {
				debug(pcpd, "illegal prefer failure option");
				goto malformed;
			}
			prefail = 1;
			break;

		case PCP_OPTION_FILTER:
			if (unlikely(pcpd->pcpd_filter == NULL)) {
				debug(pcpd, "filter option not supported");
				goto unsupp;
			}
			if (unlikely((optlen != PCP_LEN_OPTFLT) ||
				     (lifetime == 0))) {
				debug(pcpd, "illegal filter option");
				goto malformed;
			}
			if (optdata[PCP_OPTDFLT_PLEN] == 0) {
				if (unlikely(filter != 0)) {
					debug(pcpd, "filter 0 not at head");
					goto malformed;
				}
				filter0 = 1;
				break;
			}
			filter++;
			f = (filter_t *) pcpd_malloc(pcpd,
						     sizeof(*f) * filter);
			if (unlikely(f == NULL)) {
				debug(pcpd, "malloc %d filters", filter);
				seterr(pcpd, ctx,
				       PCP_ERR_NORESOURCES,
				       pcpd->shorterrht);
				goto done;
			}
			if (filter > 1) {
				memcpy(f, filters, sizeof(*f) * (filter - 1));
				pcpd_free(pcpd, filters);
			}
			filters = f;
			f = filters + (filter - 1);
			memset(f, 0, sizeof(*f));
			f->plen = optdata[PCP_OPTDFLT_PLEN];
			memcpy(&f->port, optdata + PCP_OPTDFLT_PORT, 2);
			f->port = ntohs(f->port);
			memcpy(f->prefix, optdata + PCP_OPTDFLT_PREF, 16);
			if (unlikely(f->plen > 128)) {
				debug(pcpd, "bad prefix length in filter");
				goto malformed;
			}
			msk = mask[f->plen];
			for (i = 0; i < 16; i++)
				if (unlikely((f->prefix[i] & ~msk[i]) != 0)) {
					debug(pcpd, "bad prefix in filter");
					goto malformed;
				}
			if (pcpd->flags & PCPD_MAPDST) {
				if (unlikely(f->plen < 96)) {
					debug(pcpd,
					      "bad prefix length in filter");
					goto malformed;
				}
				if (unlikely(!pcpd_unmapaddr(f->prefix))) {
					debug(pcpd,
					      "filter prefix not mapped");
					goto malformed;
				}
				f->plen -= 96;
			}
			break;

		default:
			debug(pcpd, "unknown option %d", (int) option);
			/* if mandatory to process: unsupported */
			if ((option & PCP_OPTION_OPTIONAL) == 0)
				goto unsupp;
			else {
				/* mark option for removal from response */
				resp[off + PCP_OPT_CODE] = 0;
				++unproc;
			}
		}
		off += PCP_LEN_OPTHDR + PCP_ALIGN(optlen);
	}

	/* get the internal address */
	if (ctx->off3rdparty == 0) {
		memcpy(intaddr, req + PCP_CLIENT_ADDR, 16);
		if ((pcpd->flags & PCPD_MAPINT) && !pcpd_unmapaddr(intaddr)) {
			debug(pcpd, "bad source/internal address");
			seterr(pcpd, ctx,
			       PCP_ERR_BADREQUEST,
			       pcpd->longerrht);
			goto done;
		}
		if (unlikely(memcmp(intaddr, zero, 16) == 0)) {
			debug(pcpd, "zero source/internal address");
			seterr(pcpd, ctx,
			       PCP_ERR_BADREQUEST,
			       pcpd->longerrht);
			goto done;
		}
	}

	/* check addresses for a firewall */
	if (unlikely((pcpd->flags & PCPD_FIREWALL) &&
		     (memcmp(intaddr, extaddr, 16) != 0))) {
		debug(pcpd, "firewall with different addresses");
		seterr(pcpd, ctx, PCP_ERR_BADREQUEST, pcpd->longerrht);
		goto done;
	}

	/* SCTP special case */
	if (proto == IPPROTO_SCTP) {
		if (likely(prefail == 0)) {
			nport = sport;
			prefail = 1;
		} else if (unlikely(sport != nport)) {
			debug(pcpd,
			      "SCTP and prefer failure with different ports");
			seterr(pcpd, ctx, PCP_ERR_CANTPROVIDE,
			       pcpd->longerrht);
			return l;
		}
	}

	if (lifetime == 0) {
		/* delete */
		size_t zl;

		if (pcpd->flags & PCPD_MAPINT)
			zl = 4;
		else
			zl = 16;
		if (memcmp(intaddr, zero, zl) == 0) {
			if (unlikely(sport != 0)) {
				debug(pcpd,
				      "no internal address for port != 0");
				seterr(pcpd, ctx, PCP_ERR_BADREQUEST,
				       pcpd->longerrht);
				goto done;
			}
			if (unlikely(proto != 0)) {
				debug(pcpd,
				      "no internal address for proto != 0");
				seterr(pcpd, ctx, PCP_ERR_BADREQUEST,
				       pcpd->longerrht);
				goto done;
			}
		}

		/* check authorization now */

		if (pcpd->pcpd_authorize != NULL) {
			result = pcpd->pcpd_authorize(pcpd, ctx, l);
			if (unlikely(result != PCPD_OK)) {
				if (likely(result < PCPD_ERR_PROTOBASE))
					errcode = PCPD_ERR_PROTOBASE - result;
				else
					errcode = PCP_ERR_PROCESSING;
				seterr(pcpd, ctx, errcode, 0);
				goto done;
			}
		}

		result = pcpd->pcpd_delmapping(pcpd, ctx,
					       proto, intaddr, sport);
		if (unlikely(result != PCPD_OK)) {
			if (likely(result < PCPD_ERR_PROTOBASE))
				errcode = PCPD_ERR_PROTOBASE - result;
			else
				errcode = PCP_ERR_PROCESSING;
			seterr(pcpd, ctx, errcode, 0);
			goto done;
		}
	} else {
		/* (lifetime > 0) */
		if (unlikely(memcmp(intaddr, zero, 16) == 0)) {
			debug(pcpd, "no internal address");
			seterr(pcpd, ctx, PCP_ERR_BADREQUEST, pcpd->longerrht);
			goto done;
		}
		if (unlikely((sport == 0) &&
			     (((pcpd->flags & PCPD_PORT0) == 0) ||
			      (proto == IPPROTO_TCP) ||
			      (proto == IPPROTO_UDP) ||
			      (proto == IPPROTO_SCTP)))) {
			debug(pcpd, "no internal port");
			seterr(pcpd, ctx,
			       PCP_ERR_BADREQUEST,
			       pcpd->longerrht);
			goto done;
		}
		if (prefail) {
			size_t zl;

			flags |= PCPD_PREFER_FAILURE;
			if (pcpd->flags & PCPD_MAPEXT)
				zl = 4;
			else
				zl = 16;
			if (unlikely((nport == 0) &&
				     (memcmp(extaddr, zero, zl) == 0))) {
				debug(pcpd, "prefer failure with nothing");
				seterr(pcpd, ctx,
				       PCP_ERR_BADOPTION,
				       pcpd->longerrht);
				goto done;
			}
		}

		/* check authorization now */

		if (pcpd->pcpd_authorize != NULL) {
			result = pcpd->pcpd_authorize(pcpd, ctx, l);
			if (unlikely(result != PCPD_OK)) {
				if (likely(result < PCPD_ERR_PROTOBASE))
					errcode = PCPD_ERR_PROTOBASE - result;
				else
					errcode = PCP_ERR_PROCESSING;
				seterr(pcpd, ctx, errcode, 0);
				goto done;
			}
		}

		if (filter != 0)
			flags |= PCPD_CREATE_AS_LOCKED;
		result = pcpd->pcpd_addmapping(pcpd, ctx,
					       flags, proto,
					       intaddr, sport,
					       extaddr, &nport,
					       &lifetime);
		if (unlikely(result == PCPD_STATICMAPPING)) {
			/* infinite lifetime */
			lifetime = UINT32_MAX;
			memcpy(resp + PCP_LIFETIME, &lifetime, 4);
			/* external address */
			if (pcpd->flags & PCPD_MAPEXT)
				pcpd_mapaddr(extaddr);
			memcpy(resp + PCP_MAP_EXTERNAL_ADDR, extaddr, 16);
			/* external port */
			nport = htons(nport);
			memcpy(resp + PCP_MAP_EXTERNAL_PORT, &nport, 2);
			if (likely(pcpd->pcpd_incrcounter != NULL))
				pcpd->pcpd_incrcounter(pcpd,
						       PCPS_RESULT,
						       PCP_ERR_SUCCESS);
			goto done;
		} else if (result == PCPD_NEWMAPPING) {
			newmap = 1;
			result = PCPD_OK;
		} else if (unlikely(result < PCPD_ERR_PROTOBASE)) {
			errcode = PCPD_ERR_PROTOBASE - result;
			seterr(pcpd, ctx, errcode, lifetime);
			goto done;
		} else if (unlikely(result < PCPD_OK)) {
			seterr(pcpd, ctx, PCP_ERR_PROCESSING, lifetime);
			goto done;
		}
	}

	/*
	 * filter
	 */

	if (unlikely((filter != 0) || filter0)) {
		flags = 0;
		if (filter0)
			flags |= PCPD_DELETE_FILTERS;
		if (newmap)
			flags |= PCPD_UNLOCK_ON_SUCCESS;
		result = pcpd->pcpd_filter(pcpd, ctx,
					   proto, intaddr, sport,
					   flags, (uint8_t) filter, filters);
		if (unlikely(result < PCPD_ERR_PROTOBASE)) {
			(void) pcpd->pcpd_delmapping(pcpd, ctx,
						     proto, intaddr, sport);
			errcode = PCPD_ERR_PROTOBASE - result;
			seterr(pcpd, ctx, errcode, 0);
			goto done;
		} else if (unlikely(result < PCPD_OK)) {
			(void) pcpd->pcpd_delmapping(pcpd, ctx,
						     proto, intaddr, sport);
			seterr(pcpd, ctx,
			       PCP_ERR_PROCESSING,
			       pcpd->shorterrht);
			goto done;
		}
	}

	/* lifetime */
	lifetime = htonl(lifetime);
	memcpy(resp + PCP_LIFETIME, &lifetime, 4);
	/* external address */
	if (pcpd->flags & PCPD_MAPEXT)
		pcpd_mapaddr(extaddr);
	memcpy(resp + PCP_MAP_EXTERNAL_ADDR, extaddr, 16);
	/* external port */
	nport = htons(nport);
	memcpy(resp + PCP_MAP_EXTERNAL_PORT, &nport, 2);
	if (pcpd->pcpd_incrcounter != NULL)
		pcpd->pcpd_incrcounter(pcpd, PCPS_RESULT, PCP_ERR_SUCCESS);

	if (unlikely(unproc != 0)) {
		/* remove unprocessed options from the response */
		for (off = PCP_LEN_MAP; off < l; ) {
			memcpy(&optlen, resp + off + PCP_OPT_LEN, 2);
			optlen = ntohs(optlen);
			optlen = PCP_LEN_OPTHDR + PCP_ALIGN(optlen);
			if (resp[off + PCP_OPT_CODE] == 0) {
				memmove(resp + off,
					resp + off + optlen,
					optlen);
				l -= optlen;
			}
			else
				off += optlen;
		}
	}

    done:
	if (filters != NULL)
		pcpd_free(pcpd, filters);
	return l;
}

/* process a PEER request */

static size_t
peerrequest(const pcpd_t *pcpd, pcpctx_t *ctx, size_t l)
{
	uint8_t *req = ctx->request, *resp = ctx->response;
	uint8_t proto, option, *optdata, errcode;
	uint8_t intaddr[16], extaddr[16], remote[16];
	uint16_t sport, nport, rport, optlen, unproc = 0;
	uint32_t lifetime;
	size_t off;
	int result;

	if (unlikely(l < PCP_LEN_PEER)) {
		debug(pcpd, "underrun %d", l);
		seterr(pcpd, ctx, PCP_ERR_BADREQUEST, pcpd->longerrht);
		return l;
	}

	/* decode */
	memcpy(&lifetime, req + PCP_LIFETIME, 4);
	lifetime = ntohl(lifetime);
	proto = req[PCP_PEER_PROTOCOL];
	memcpy(&sport, req + PCP_PEER_INTERNAL_PORT, 2);
	sport = ntohs(sport);
	memcpy(&nport, req + PCP_PEER_EXTERNAL_PORT, 2);
	nport = ntohs(nport);
	memcpy(extaddr, req + PCP_PEER_EXTERNAL_ADDR, 16);
	if ((pcpd->flags & PCPD_MAPEXT) && !pcpd_unmapaddr(extaddr)) {
		debug(pcpd, "external addr not mapped");
		seterr(pcpd, ctx, PCP_ERR_BADREQUEST, pcpd->longerrht);
		return l;
	}
	memcpy(&rport, req + PCP_PEER_REMOTE_PORT, 2);
	rport = ntohs(rport);
	memcpy(remote, req + PCP_PEER_REMOTE_ADDR, 16);
	if ((pcpd->flags & PCPD_MAPDST) && !pcpd_unmapaddr(remote)) {
		debug(pcpd, "remote address not mapped");
		seterr(pcpd, ctx, PCP_ERR_BADREQUEST, pcpd->longerrht);
		return l;
	}

	/* check protocol */
	if ((pcpd->flags & PCPD_OTHERPROTO) == 0) {
		if (unlikely((proto != IPPROTO_TCP) &&
			     (proto != IPPROTO_UDP))) {
			debug(pcpd, "bad protocol %d", (int) proto);
			seterr(pcpd, ctx, PCP_ERR_UNSUPPROTO,
			       pcpd->longerrht);
			return l;
		}
	} else if (unlikely(proto == 0)) {
		debug(pcpd, "bad protocol zero");
		seterr(pcpd, ctx, PCP_ERR_UNSUPPROTO, pcpd->longerrht);
		return l;
	}

	/* check internal port */
	if (unlikely(sport == 0)) {
		debug(pcpd, "no internal port");
		seterr(pcpd, ctx, PCP_ERR_BADREQUEST, pcpd->longerrht);
		return l;
	}

	/* check SCTP */
	if (unlikely((proto == IPPROTO_SCTP) && (sport != nport))) {
		debug(pcpd, "SCTP with different ports");
		seterr(pcpd, ctx, PCP_ERR_BADREQUEST, pcpd->longerrht);
		return l;
	}

	/*
	 * process options
	 */

	off = PCP_LEN_PEER;
	ctx->off3rdparty = 0;
	while (off < l) {
		if (unlikely(off + PCP_LEN_OPTHDR > l)) {
			debug(pcpd, "option header overflow");
			seterr(pcpd, ctx, PCP_ERR_BADOPTION, pcpd->longerrht);
			return l;
		}
		option = req[off + PCP_OPT_CODE];
		if (likely(pcpd->pcpd_incrcounter != NULL))
			pcpd->pcpd_incrcounter(pcpd, PCPS_OPTION, option);
		memcpy(&optlen, req + off + PCP_OPT_LEN, 2);
		optlen = ntohs(optlen);
		if (unlikely(off + PCP_LEN_OPTHDR + optlen > l)) {
			debug(pcpd, "option overflow");
			seterr(pcpd, ctx, PCP_ERR_BADOPTION, pcpd->longerrht);
			return l;
		}
		optdata = req + off + PCP_OPT_DATA;
		switch (option) {
		case PCP_OPTION_THIRD_PARTY:
			/* disallowed */
			if (unlikely((pcpd->flags & PCPD_3RDPARTY) == 0)) {
				debug(pcpd, "3rd party option disallowed");
			unsupp:
				seterr(pcpd, ctx,
				       PCP_ERR_UNSUPOPTION,
				       pcpd->longerrht);
				return l;
			}
			/* once */
			if (unlikely(ctx->off3rdparty != 0)) {
				debug(pcpd, "multiple 3rd party option");
			malformed:
				seterr(pcpd, ctx,
				       PCP_ERR_BADOPTION,
				       pcpd->longerrht);
				return l;
			}
			/* length */
			if (unlikely(optlen != PCP_LEN_OPT3PTY)) {
				debug(pcpd, "3rd party bad length %d", optlen);
				goto malformed;
			}
			/* not the same than the source address */
			if (unlikely(memcmp(optdata,
					    req + PCP_CLIENT_ADDR,
					    16) == 0)) {
				debug(pcpd, "3rd party same as source");
				seterr(pcpd, ctx,
				       PCP_ERR_BADREQUEST,
				       pcpd->longerrht);
				return l;
			}
			ctx->off3rdparty = off;
			/* zero only */
			if (pcpd->flags & PCPD_3RDPARTY_AO) {
				if (unlikely((memcmp(intaddr, zero,
						     sizeof(zero) != 0)) &&
					     (memcmp(intaddr, pcpd_mapped,
						     sizeof(zero) != 0)))) {
					debug(pcpd, "3rd party not zero");
					goto malformed;
				}
			}
			memcpy(intaddr, optdata, 16);
			if ((pcpd->flags & PCPD_MAPINT) &&
			    (memcmp(intaddr, zero, sizeof(zero)) != 0) &&
			    !pcpd_unmapaddr(intaddr)) {
				debug(pcpd, "3rd party not mapped");
				goto malformed;
			}
			/* CPE only */
			if (pcpd->flags & PCPD_3RDPARTY_CPE) {
				if (unlikely((memcmp(req + PCP_CLIENT_ADDR,
						     pcpd_mapped, 12) != 0) ||
					     (memcmp(req +
						     PCP_CLIENT_ADDR + 12,
						     &cpe, 4) != 0))) {
					debug(pcpd, "3rd party not from CPE");
					goto malformed;
				}
			}
			/* the CPE will validate the address */
			break;

		case PCP_OPTION_PREF_FAIL:
			debug(pcpd, "prefer failure option in PEER");
			goto malformed;

		case PCP_OPTION_FILTER:
			debug(pcpd, "filter option in PEER");
			goto malformed;

		default:
			debug(pcpd, "unknown option %d", (int) option);
			/* if mandatory to process: unsupported */
			if ((option & PCP_OPTION_OPTIONAL) == 0)
				goto unsupp;
			else {
				/* mark option for removal from response */
				resp[off + PCP_OPT_CODE] = 0;
				++unproc;
			}
		}
		off += PCP_LEN_OPTHDR + PCP_ALIGN(optlen);
	}

	/* get the internal address */
	if (ctx->off3rdparty == 0) {
		memcpy(intaddr, req + PCP_CLIENT_ADDR, 16);
		if ((pcpd->flags & PCPD_MAPINT) && !pcpd_unmapaddr(intaddr)) {
			debug(pcpd, "bad source/internal address");
			seterr(pcpd, ctx, PCP_ERR_BADREQUEST, pcpd->longerrht);
			return l;
		}
		if (unlikely(memcmp(intaddr, zero, 16) == 0)) {
			debug(pcpd, "zero source/internal address");
			seterr(pcpd, ctx, PCP_ERR_BADREQUEST, pcpd->longerrht);
			return l;
		}
	}

	if (unlikely(memcmp(intaddr, zero, 16) == 0)) {
		debug(pcpd, "no internal address");
		seterr(pcpd, ctx, PCP_ERR_BADREQUEST, pcpd->longerrht);
		return l;
	}

	/* check authorization now */

	if (pcpd->pcpd_authorize != NULL) {
		result = pcpd->pcpd_authorize(pcpd, ctx, l);
		if (unlikely(result != PCPD_OK)) {
			if (likely(result < PCPD_ERR_PROTOBASE))
				errcode = PCPD_ERR_PROTOBASE - result;
			else
				errcode = PCP_ERR_PROCESSING;
			seterr(pcpd, ctx, errcode, 0);
			return l;
		}
	}

	result = pcpd->pcpd_adddynamic(pcpd, ctx,
				       proto, intaddr, sport,
				       remote, rport,
				       extaddr, &nport,
				       &lifetime);
	if (unlikely(result < PCPD_ERR_PROTOBASE)) {
		errcode = PCPD_ERR_PROTOBASE - result;
		seterr(pcpd, ctx, errcode, lifetime);
		return l;
	} else if (unlikely(result < PCPD_OK)) {
		seterr(pcpd, ctx, PCP_ERR_PROCESSING, lifetime);
		return l;
	}

	/* lifetime */
	lifetime = htonl(lifetime);
	memcpy(resp + PCP_LIFETIME, &lifetime, 4);
	/* external address */
	if (pcpd->flags & PCPD_MAPEXT)
		pcpd_mapaddr(extaddr);
	memcpy(resp + PCP_PEER_EXTERNAL_ADDR, extaddr, 16);
	/* external port */
	nport = htons(nport);
	memcpy(resp + PCP_PEER_EXTERNAL_PORT, &nport, 2);
	if (likely(pcpd->pcpd_incrcounter != NULL))
		pcpd->pcpd_incrcounter(pcpd, PCPS_RESULT, PCP_ERR_SUCCESS);

	if (unlikely(unproc != 0)) {
		/* remove unprocessed options from the response */
		for (off = PCP_LEN_PEER; off < l; ) {
			memcpy(&optlen, resp + off + PCP_OPT_LEN, 2);
			optlen = ntohs(optlen);
			optlen = PCP_LEN_OPTHDR + PCP_ALIGN(optlen);
			if (resp[off + PCP_OPT_CODE] == 0) {
				memmove(resp + off,
					resp + off + optlen,
					optlen);
				l -= optlen;
			}
			else
				off += optlen;
		}
	}

	return l;
}

/* decode a request */

static size_t
getrequest(const pcpd_t *pcpd, pcpctx_t *ctx, size_t l)
{
	uint8_t *req = ctx->request, *resp = ctx->response;
	struct subscriber *s;
	uint8_t version, opcode, errcode, client[16];
	uint32_t epoch;
	int result;

	/* copy the request buffer for the response */
	memcpy(resp, req, l);

	/* set header */
	resp[PCP_OPCODE] |= PCP_OPCODE_RESPONSE;
	resp[PCP_RESULT_CODE] = PCP_ERR_SUCCESS;

	/* first sanity checks - drop packet if these occur */
	if (unlikely(l < PCP_MINLEN0)) {
		debug(pcpd, "underrun %d", l);
		return 0;
	}

	/* drop responses */
	opcode = req[PCP_OPCODE];
	if (unlikely(opcode & PCP_OPCODE_RESPONSE)) {
		debug(pcpd, "response");
		return 0;
	}

	/* reject bad versions */
	version = req[PCP_VERSION];
	if (version != 1) {
		if (likely(pcpd->pcpd_incrcounter != NULL))
			pcpd->pcpd_incrcounter(pcpd, PCPS_RESULT,
					      PCP_ERR_UNSUPVERSION);
		debug(pcpd, "bad version %d", version);
		resp[PCP_VERSION] = 1;
		resp[PCP_RESULT_CODE] = PCP_ERR_UNSUPVERSION;
		return (l < 4) ? 4 : l;
	}

	/* second case of underrun */
	if (unlikely(l < PCP_MINLEN1)) {
		debug(pcpd, "underrun %d", l);
		return 0;
	}

	/* get subscriber and set the epoch */
	result = PCPD_OK;
	memcpy(client, req + PCP_CLIENT_ADDR, 16);
	memset(resp + PCP_CLIENT_ADDR, 0, 16);
	s = getsubscriber(pcpd, ctx, &result);
	if (likely(s != NULL)) {
		epoch = htonl(sssoe(pcpd, s));
		if ((pcpd->flags & PCPD_CACHE) == 0)
			pcpd_free(pcpd, s);
		memcpy(resp + PCP_EPOCH, &epoch, 4);
	}

	/* this must be as soon as the epoch is set */
	if (unlikely(result != PCPD_OK)) {
		debug(pcpd, "error finding subscriber '%s': %s",
		      pcpd_id2str(pcpd, ctx->sid), pcpd_strerror(result));
		if (result < PCPD_ERR_PROTOBASE)
			errcode = PCPD_ERR_PROTOBASE - result;
		else
			errcode = PCP_ERR_NOTAUTH;
		seterr(pcpd, ctx, errcode, 0);
		/* length checks are below */
		if (unlikely(l > PCP_MAXLEN))
			l = PCP_MAXLEN;
		else if (unlikely((l & 3) != 0))
			l = PCP_ALIGN(l);
		return l;
	}

	/* overrun */
	if (unlikely(l > PCP_MAXLEN)) {
		debug(pcpd, "overrun %d", l);
		seterr(pcpd, ctx, PCP_ERR_BADREQUEST, pcpd->longerrht);
		return PCP_MAXLEN;
	}

	/* bad align */
	if (unlikely((l & 3) != 0)) {
		debug(pcpd, "bad align %d", l);
		seterr(pcpd, ctx, PCP_ERR_BADREQUEST, pcpd->longerrht);
		return PCP_ALIGN(l);
	}

	/* client address */
	if ((pcpd->flags & PCPD_MAPSRC) && !pcpd_unmapaddr(client)) {
		debug(pcpd, "source address not mapped");
		seterr(pcpd, ctx, PCP_ERR_ADDRMISMATCH, pcpd->longerrht);
		return l;
	}
	if (ctx->client->sa_family == AF_INET6) {
		struct sockaddr_in6 *c6 = (struct sockaddr_in6 *) ctx->client;

		if (unlikely(memcmp(client, &c6->sin6_addr, 16) != 0)) {
			debug(pcpd, "source address mismatch");
			seterr(pcpd, ctx,
			       PCP_ERR_ADDRMISMATCH,
			       pcpd->longerrht);
			return l;
		}
	} else if (ctx->client->sa_family == AF_INET) {
		struct sockaddr_in *c4 = (struct sockaddr_in *) ctx->client;

		if (unlikely(memcmp(client, &c4->sin_addr, 4) != 0)) {
			debug(pcpd, "source address mismatch");
			seterr(pcpd, ctx,
			       PCP_ERR_ADDRMISMATCH,
			       pcpd->longerrht);
			return l;
		}
	}

	/* call opcode processing routine */
	if (likely(pcpd->pcpd_incrcounter != NULL))
		pcpd->pcpd_incrcounter(pcpd, PCPS_OPERATION, opcode);
	switch (opcode) {
	case PCP_OPCODE_ANNOUNCE:
		if (unlikely((pcpd->opcodes & PCPD_OPCODE_ANNOUNCE) == 0))
			return 0;
		l = announcerequest(pcpd, ctx, l);
		return l;
	case PCP_OPCODE_MAP:
		if (unlikely((pcpd->opcodes & PCPD_OPCODE_MAP) == 0))
			break;
		l = maprequest(pcpd, ctx, l);
		return l;
	case PCP_OPCODE_PEER:
		if (unlikely((pcpd->opcodes & PCPD_OPCODE_PEER) == 0))
			break;
		l = peerrequest(pcpd, ctx, l);
		return l;
	default:
		break;
	}
	seterr(pcpd, ctx, PCP_ERR_UNSUPOPCODE, pcpd->longerrht);
	return l;
}

/* send MAP updates */

void
pcpd_update(const pcpd_t *pcpd, const pcpctx_t *ctx, const size_t length,
	    const uint8_t *extaddr, const uint16_t extport,
	    const uint32_t lifetime)
{
	uint8_t *resp;
	struct subscriber *s;
	uint32_t epoch, lt;
	uint16_t nport;
	int result;

	if (unlikely((pcpd == NULL) || (ctx == NULL) ||
		     (length < PCP_LEN_MAP) || (length > PCP_MAXLEN) ||
		     (extaddr == NULL)))
		return;
	if ((pcpd->flags & PCPD_UPDATE) == 0)
		return;
	debug(pcpd, "update");
	resp = ctx->response;
	if (resp[PCP_VERSION] == 0)
		memcpy(resp, ctx->request, length);
	resp[PCP_OPCODE] |= PCP_OPCODE_RESPONSE;
	resp[PCP_RESULT_CODE] = PCP_ERR_SUCCESS;
	memset(resp + PCP_CLIENT_ADDR, 0, 16);
	s = getsubscriber(pcpd, ctx, &result);
	if (unlikely(s == NULL)) {
		debug(pcpd, "can't find subscriber '%s'",
		      pcpd_id2str(pcpd, ctx->sid));
		return;
	}
	epoch = htonl(sssoe(pcpd, s));
	if ((pcpd->flags & PCPD_CACHE) == 0)
		pcpd_free(pcpd, s);
	if (result != PCPD_OK) {
		debug(pcpd, "error finding subscriber '%s': %s",
		      pcpd_id2str(pcpd, ctx->sid), pcpd_strerror(result));
		return;
	}
	/* update! */
	memcpy(resp + PCP_EPOCH, &epoch, 4);
	lt = htonl(lifetime);
	memcpy(resp + PCP_LIFETIME, &lt, 4);
	if (pcpd->flags & PCPD_MAPEXT) {
		uint8_t addr[16];

		memcpy(addr, extaddr, 16);
		pcpd_mapaddr(addr);
		memcpy(resp + PCP_MAP_EXTERNAL_ADDR, addr, 16);
	} else
		memcpy(resp + PCP_MAP_EXTERNAL_ADDR, extaddr, 16);
	nport = htons(extport);
	memcpy(resp + PCP_MAP_EXTERNAL_PORT, &nport, 2);
	/* send */
	pcpd->pcpd_sendresponse(pcpd, ctx, length);
	return;
}

/* library init */

int
pcpd_lib_init(pcpd_t *pcpd)
{
#if CHECKLIB
	struct cgn *c = NULL;
#endif
	unsigned int i;

	if (unlikely((pcpd == NULL) || !checkpcpd(pcpd)))
		return PCPD_ERR_INVAL;
	if (pcpd->longerrht == 0)
		pcpd->longerrht = 1800;
	if (pcpd->shorterrht == 0)
		pcpd->shorterrht = 30;

	if (cpe == 0) {
#ifndef NO_SOCKET
		cpe = inet_addr("192.0.0.2");
#else
		uint8_t b[4];

		b[0] = 192;
		b[1] = 0;
		b[2] = 0;
		b[3] = 2;
		memcpy(&cpe, b, 4);
#endif

#if CHECKLIB
		ISC_LIST_INIT(&cgns);
#endif

		for (i = 0; i < HASHSIZE; i++)
			ISC_LIST_INIT(&subscribers[i]);
	}
#if CHECKLIB
	ISC_LIST_FOREACH(c, &cgns, chain)
		if ((c->pointer == pcpd) &&
		    (memcmp(&c->rovalue, pcpd, sizeof(*pcpd)) == 0)) {
			debug(pcpd, "already registered %u times", c->refcnt);
			c->refcnt++;
			return PCPD_OK;
		}
	c = (struct cgn *) pcpd_malloc(pcpd, sizeof(*c));
	if (unlikely(c == NULL))
		return PCPD_ERR_NORESOURCES;
	memset(c, 0, sizeof(*c));
	c->pointer = pcpd;
	memcpy(&c->rovalue, pcpd, sizeof(*pcpd));
	ISC_LIST_INSERT(&cgns, c, chain);
#endif
	return PCPD_OK;
}

/* library close */

int
pcpd_lib_close(const pcpd_t *pcpd)
{
#if CHECKLIB
	struct cgn *c = NULL;
#endif
	unsigned int i;

	if (unlikely(pcpd == NULL))
		return PCPD_ERR_INVAL;
#if CHECKLIB
	ISC_LIST_FOREACH(c, &cgns, chain)
		if ((c->pointer == pcpd) &&
		    (memcmp(&c->rovalue, pcpd, sizeof(*pcpd)) == 0))
			break;
	if (unlikely((c == NULL) || (c->refcnt == 0)))
		return PCPD_ERR_INVAL;
	c->refcnt -= 1;
	if (c->refcnt > 0)
		return (int) c->refcnt;
	ISC_LIST_REMOVE(c, chain);
	pcpd_free(pcpd, c);
#endif
	for (i = 0; i < HASHSIZE; i++) {
		struct subscriber *s, *ts;

		ISC_LIST_FOREACH_SAFE(s, &subscribers[i], chain, ts)
			if (s->pcpd == pcpd) {
				ISC_LIST_REMOVE(s, chain);
				pcpd_free(pcpd, s);
			}
	}
	return PCPD_OK;
}

/* library run once */

static inline void
ctx_free(const pcpd_t *pcpd, pcpctx_t *ctx)
{
	if (ctx->client != NULL)
		pcpd_free(pcpd, ctx->client);
	if (ctx->sid != NULL)
		pcpd_free(pcpd,ctx->sid);
	if (ctx->request != NULL)
		pcpd_free(pcpd,ctx->request);
	if (ctx->response != NULL)
		pcpd_free(pcpd,ctx->response);
}

static inline int
ctx_init(const pcpd_t *pcpd, pcpctx_t *ctx)
{
	memset(ctx, 0, sizeof(*ctx));
	ctx->client = (struct sockaddr *) pcpd_malloc(pcpd,
				sizeof(struct sockaddr_storage));
	if (ctx->client == NULL)
		goto nomem;
	ctx->sid = (uint8_t *) pcpd_malloc(pcpd, PCPD_SID_SIZE);
	if (ctx->sid == NULL)
		goto nomem;
	ctx->request = (uint8_t *) pcpd_malloc(pcpd, PCPD_BUF_SIZE);
	if (ctx->request == NULL)
		goto nomem;
	ctx->response = (uint8_t *) pcpd_malloc(pcpd, PCPD_BUF_SIZE);
	if (ctx->response == NULL)
		goto nomem;
	return PCPD_OK;

    nomem:
	ctx_free(pcpd, ctx);
	return PCPD_ERR_NORESOURCES;
}

int
pcpd_lib_runonce(const pcpd_t *pcpd, const pcpctx_t *ctx0)
{
	pcpctx_t sctx, *ctx;
	int ret;
#if CHECKLIB
	struct cgn *c = NULL;
#endif
	size_t len;

	if (unlikely(pcpd == NULL))
		return PCPD_ERR_INVAL;

#if CHECKLIB
	ISC_LIST_FOREACH(c, &cgns, chain)
		if ((c->pointer == pcpd) &&
		    (memcmp(&c->rovalue, pcpd, sizeof(*pcpd)) == 0))
			break;
	if (unlikely(c == NULL))
		return PCPD_ERR_INVAL;
#endif

	if (ctx0 != NULL)
		/* leave the optional ctxclear to the caller */
		ctx = (pcpctx_t *) ctx0;
	else {
		ret = ctx_init(pcpd, &sctx);
		if (ret != PCPD_OK)
			return ret;
		ctx = &sctx;
		ctxclear(ctx);
	}

	len = pcpd->pcpd_recvrequest(pcpd, ctx);
	debug(pcpd, "get a %d byte long request from '%s'",
	      len, pcpd_id2str(pcpd, ctx->sid));
	if (unlikely(len == 0))
		goto done;
	len = getrequest(pcpd, ctx, len);
	if (likely(len > 0)) {
		debug(pcpd, "send response");
		pcpd->pcpd_sendresponse(pcpd, ctx, len);
	}
    done:
	if (ctx0 == NULL)
		ctx_free(pcpd, ctx);
	return PCPD_OK;
}

#ifndef PASSIVE_ONLY

/* main loop */

static void
mainloop(const pcpd_t *pcpd)
{
	pcpctx_t ctx;
	struct sockaddr_storage ss;
	uint8_t sid[PCPD_SID_SIZE];
	uint8_t request[PCPD_BUF_SIZE];
	uint8_t response[PCPD_BUF_SIZE];
	size_t len;

	ctx.client = (struct sockaddr *) &ss;
	ctx.sid = sid;
	ctx.request = request;
	ctx.response = response;

	while (!terminate) {
		ctxclear(&ctx);
		len = pcpd->pcpd_recvrequest(pcpd, &ctx);
		debug(pcpd, "get a %d byte long request from '%s'",
		      len, pcpd_id2str(pcpd, sid));
		if (unlikely(len == 0))
			continue;
		len = getrequest(pcpd, &ctx, len);
		if (likely(len > 0)) {
			debug(pcpd, "send response");
			pcpd->pcpd_sendresponse(pcpd, &ctx, len);
		}
	}
}

/* main */

#pragma weak main

int
main(int argc, char *argv[])
{
	pcpd_t pcpd;
	int ret, logopt = LOG_NDELAY, i;
#ifdef NO_SOCKET
	uint8_t b[4];
#endif

	memset(&pcpd, 0, sizeof(pcpd));
	pcpd.longerrht = 1800;
	pcpd.shorterrht = 30;
#ifndef NO_SOCKET
	cpe = inet_addr("192.0.0.2");
#else
	b[0] = 192;
	b[1] = 0;
	b[2] = 0;
	b[3] = 2;
	memcpy(&cpe, b, 4);
#endif

	for (i = 0; i < HASHSIZE; i++)
		ISC_LIST_INIT(&subscribers[i]);

	for (i = 1; i < argc; i++)
		if (strcmp(argv[i], "-d") == 0) {
			logopt += LOG_PERROR;
			break;
		}
	
	ret = pcpd_cgn_init(&pcpd, argc, argv);
	if (unlikely((ret < PCPD_OK) || !checkpcpd(&pcpd)))
		exit(ret);

	pcpd_openlog(&pcpd, "pcpd", logopt, LOG_LOCAL6);
	switch (pcpd.flags & (PCPD_UPDATE | PCPD_CACHE)) {
	default:
	case 0:
		ret = PCPD_OK;
		break;
	case PCPD_UPDATE:
		ret = pcpd_cgn_register(&pcpd, &pcpd_update, NULL);
		break;
	case PCPD_CACHE:
		ret = pcpd_cgn_register(&pcpd, NULL, &pcpd_uncache);
		break;
	case PCPD_UPDATE | PCPD_CACHE:
		ret = pcpd_cgn_register(&pcpd, &pcpd_update, &pcpd_uncache);
		break;
	}
	if (unlikely(ret < PCPD_OK)) {
		debug(&pcpd, "register uncache: %s", pcpd_strerror(ret));
		exit(ret);
	}
	debug(&pcpd, "started");

	mainloop(&pcpd);

	debug(&pcpd, "exiting");
	for (i = 0; i < HASHSIZE; i++) {
		struct subscriber *s;

		while ((s = ISC_LIST_FIRST(&subscribers[i])) != NULL) {
			ISC_LIST_REMOVE(s, chain);
			pcpd_free(&pcpd, s);
		}
	}
	pcpd_closelog(&pcpd);
	ret = pcpd_cgn_close(&pcpd);
	exit(ret);
}

#endif

/*
 * Weak definitions
 */

#pragma weak pcpd_strerror

const char *
pcpd_strerror(const int error)
{
	switch (error) {
	case PCPD_STATICMAPPING:
		return "already existing static mapping";
	case PCPD_NEWMAPPING:
		return "a new mapping was created";
	case PCPD_OK:
		return "ok";
	case PCPD_ERR_INVAL:
		return "invalid arguments";
	case PCPD_ERR_PROTOBASE:
		return "base for protocol errors (shouldn't happen)";
	case PCPD_ERR_UNSUPVERSION:
		return "unsupported version";
	case PCPD_ERR_NOTAUTH:
		return "not authorized";
	case PCPD_ERR_BADREQUEST:
		return "malformed request";
	case PCPD_ERR_UNSUPOPCODE:
		return "unsupported opcode";
	case PCPD_ERR_UNSUPOPTION:
		return "unsupported option";
	case PCPD_ERR_BADOPTION:
		return "malformed option";
	case PCPD_ERR_NETFAILURE:
		return "network failure";
	case PCPD_ERR_NORESOURCES:
		return "out of resources";
	case PCPD_ERR_UNSUPPROTO:
		return "unsupported protocol";
	case PCPD_ERR_EXQUOTA:
		return "user exceeded quota";
	case PCPD_ERR_CANTPROVIDE:
		return "cannot provide external";
	case PCPD_ERR_ADDRMISMATCH:
		return "address mismatch";
	case PCPD_ERR_TOOMANYPEER:
		return "excessive number of remote peers";
	case PCPD_ERR_PROCESSING:
		return "processing error";
	default:
		if (error < PCPD_ERR_PROTOBASE)
			return "unknown protocol error";
		else
			return "unknown error";
	}
}

#pragma weak pcpd_malloc

void *
pcpd_malloc(const pcpd_t *p, size_t size)
{
	UNUSED(p);
	return malloc(size);
}

#pragma weak pcpd_free

void
pcpd_free(const pcpd_t *p, void *ptr)
{
	UNUSED(p);
	free(ptr);
}

#ifndef PASSIVE_ONLY

#pragma weak pcpd_openlog

void
pcpd_openlog(const pcpd_t *p, const char *ident, int option, int facility)
{
	UNUSED(p);
#ifndef NO_SYSLOG
	openlog(ident, option, facility);
#else
	UNUSED(ident);
	UNUSED(option);
	UNUSED(facility);
#endif
}

#endif

#pragma weak pcpd_vsyslog

void
pcpd_vsyslog(const pcpd_t *p, int priority, const char *format, va_list ap)
{
	UNUSED(p);
#ifndef NO_SYSLOG
	vsyslog(priority, format, ap);
#else
	UNUSED(priority);
	UNUSED(format);
	UNUSED(ap);
#endif
}

#ifndef PASSIVE_ONLY

#pragma weak pcpd_closelog

void
pcpd_closelog(const pcpd_t *p)
{
	UNUSED(p);
#ifndef NO_SYSLOG
	closelog();
#endif
}

#endif

#pragma weak pcpd_time

time_t
pcpd_time(const pcpd_t *p)
{
	UNUSED(p);
	return time(NULL);
}

#pragma weak pcpd_inet_ntop

const char *
pcpd_inet_ntop(int af, const void *src, char *dst, socklen_t size)
{
#ifndef NO_SOCKET
	return inet_ntop(af, src, dst, size);
#else
	UNUSED(af);
	UNUSED(src);
	UNUSED(dst);
	UNUSED(size);
	return NULL;
#endif
}

#pragma weak pcpd_id2str

const char *
pcpd_id2str(const pcpd_t *pcpd, const uint8_t *id)
{
	static char host[INET6_ADDRSTRLEN];
	char hex[] = "0123456789abcdef";

	memset(host, 0, sizeof(host));
	switch (idsize(pcpd)) {
	case 0:
		break;
	case 16:
		(void) pcpd_inet_ntop(AF_INET6, id, host, sizeof(host));
		break;
	case 6:
		host[0] = hex[id[0] >> 4];
		host[1] = hex[id[0] & 0xf];
		host[2] = ':';
		host[3] = hex[id[1] >> 4];
		host[4] = hex[id[1] & 0xf];
		host[5] = ':';
		host[6] = hex[id[2] >> 4];
		host[7] = hex[id[2] & 0xf];
		host[8] = ':';
		host[9] = hex[id[3] >> 4];
		host[10] = hex[id[3] & 0xf];
		host[11] = ':';
		host[12] = hex[id[4] >> 4];
		host[13] = hex[id[4] & 0xf];
		host[14] = ':';
		host[15] = hex[id[5] >> 4];
		host[16] = hex[id[5] & 0xf];
		break;
	case 4:
	default:
		(void) pcpd_inet_ntop(AF_INET, id, host, sizeof(host));
		break;
	}
	return host;
}
