/*
 * Copyright (C) 2011-2012  Internet Systems Consortium, Inc. ("ISC")
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND ISC DISCLAIMS ALL WARRANTIES WITH
 * REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS.  IN NO EVENT SHALL ISC BE LIABLE FOR ANY SPECIAL, DIRECT,
 * INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
 * OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */

/* $Id: 1cbb1aa9a83366e72675503d2ef0cd6037d6d835 $ */

/* PCP service for miniupnpd */

#include <stdio.h>
#include <stdlib.h>
#include <syslog.h>
#include <time.h>
#include <unistd.h>
#include <arpa/inet.h>
#include "config.h"
#include "pcpd.h"
#include "pcpsrv.h"
#include "upnpglobalvars.h"
#include "getifaddr.h"
#include "upnpredirect.h"
#include "commonrdr.h"

#ifdef ENABLE_PCP

static size_t _recvrequest(const pcpd_t *pcpd,
			   const pcpctx_t *ctx);
static void _sendresponse(const pcpd_t *pcpd,
			  const pcpctx_t *ctx,
			  const size_t length);
static int _getsubscriber(const pcpd_t *pcpd,
			  const pcpctx_t *ctx,
			  time_t *epoch);
static int _addmapping(const pcpd_t *pcpd,
		       const pcpctx_t *ctx,
		       const uint8_t flags,
		       const uint8_t proto,
		       const uint8_t *intaddr,
		       const uint16_t intport,
		       uint8_t *extaddr,
		       uint16_t *extport,
		       uint32_t *lifetime);
static int _delmapping(const pcpd_t *pcpd,
			const pcpctx_t *ctx,
			const uint8_t proto,
			const uint8_t *intaddr,
			const uint16_t intport);

struct pcpsrv_s {
	LIST_ENTRY(pcpsrv_s) list;
	pcpd_t pcpd;
	int sock;
};
LIST_HEAD(pcpsrv_list, pcpsrv_s);
struct pcpsrv_list pcpsrvs;

int pcpsrv_init(struct pcpsrv_s *srv)
{
	srv->pcpd.opcodes = PCPD_OPCODE_ANNOUNCE | PCPD_OPCODE_MAP;
	srv->pcpd.cgntype = PCPD_TYPE_NAT;
	srv->pcpd.flags = PCPD_MAPALL;
	srv->pcpd.pcpd_recvrequest = _recvrequest;
	srv->pcpd.pcpd_sendresponse = _sendresponse;
	srv->pcpd.pcpd_getsubscriber = _getsubscriber;
	srv->pcpd.pcpd_addmapping = _addmapping;
	srv->pcpd.pcpd_delmapping = _delmapping;
	return pcpd_lib_init(&srv->pcpd);
}

void ClosePCPSocket(int s)
{
	struct pcpsrv_s *srv;

	if(s < 0)
		return;
	for(srv = pcpsrvs.lh_first; srv != NULL; srv = srv->list.le_next)
	{
		if(srv->sock == s)
			break;
	}
	if(srv == NULL)
		return;
	pcpd_lib_close(&srv->pcpd);
	LIST_REMOVE(srv, list);
	free(srv);
	close(s);
}

int OpenAndConfPCPSocket(in_addr_t addr)
{
	struct pcpsrv_s *srv;
	int spcp, ret;

	srv = (struct pcpsrv_s *)malloc(sizeof(struct pcpsrv_s));
	if(srv == NULL)
	{
		syslog(LOG_ERR, "malloc(pcpsrv): %m");
		return -1;
	}
	memset(srv, 0, sizeof(struct pcpsrv_s));
	srv->pcpd.opaque = srv;
	spcp = socket(PF_INET, SOCK_DGRAM, 0/*IPPROTO_UDP*/);
	if(spcp<0)
	{
		syslog(LOG_ERR, "socket(pcp): %m");
		free(srv);
		return -1;
	}
	else
	{
		struct sockaddr_in pcp_addr;
		memset(&pcp_addr, 0, sizeof(pcp_addr));
		pcp_addr.sin_family = AF_INET;
		pcp_addr.sin_port = htons(PCP_PORT);
		//pcp_addr.sin_addr.s_addr = INADDR_ANY;
		pcp_addr.sin_addr.s_addr = addr;
		if(bind(spcp, (struct sockaddr *)&pcp_addr, sizeof(pcp_addr)) < 0)
		{
			syslog(LOG_ERR, "bind(pcp): %m");
			close(spcp);
			free(srv);
			return -1;
		}
		if(setsockopt(spcp, IPPROTO_IP, IP_MULTICAST_IF,
			      &pcp_addr.sin_addr, sizeof(struct in_addr)) < 0)
		{
			syslog(LOG_ERR, "MULTICAST_IF(pcp): %m");
			close(spcp);
			return -1;
		}
	}
	srv->sock = spcp;
	ret = pcpsrv_init(srv);
	if(ret != PCPD_OK)
	{
		syslog(LOG_ERR, "init(pcpsrv): %s", pcpd_strerror(ret));
		close(spcp);
		free(srv);
		return -1;
	}
	LIST_INSERT_HEAD(&pcpsrvs, srv, list);
	return spcp;
}

int OpenAndConfPCPSockets(int * sockets)
{
	int i, j;
	struct lan_addr_s * lan_addr;

	LIST_INIT(&pcpsrvs);
	for(i = 0, lan_addr = lan_addrs.lh_first; lan_addr != NULL; lan_addr = lan_addr->list.le_next, i++)
	{
		sockets[i] = OpenAndConfPCPSocket(lan_addr->addr.s_addr);
		if(sockets[i] < 0)
		{
			for(j=0; j<i; j++)
			{
				ClosePCPSocket(sockets[j]);
				sockets[j] = -1;
			}
			return -1;
		}
	}
	return 0;
}

void ProcessIncomingPCPPacket(int s)
{
	struct pcpsrv_s *srv;

	if(s < 0)
		return;
	for(srv = pcpsrvs.lh_first; srv != NULL; srv = srv->list.le_next)
	{
		if(srv->sock == s)
			break;
	}
	if(srv == NULL)
		return;
	pcpd_lib_runonce(&srv->pcpd, NULL);
}

#if 0
/* iterate through the redirection list to find those who came
 * from PCP and select the first to expire */
int ScanPCPforExpiration()
{
	char desc[64];
	unsigned short iport, eport;
	int proto;
	int r, i;
	unsigned timestamp;
	nextpcptoclean_eport = 0;
	nextpcptoclean_timestamp = 0;
	for(i = 0; ; i++) {
		r = get_redirect_rule_by_index(i, 0, &eport, 0, 0,
		                               &iport, &proto, desc, sizeof(desc),
		                               0, 0, &timestamp, 0, 0);
		if(r<0)
			break;
		if(sscanf(desc, "PCP %u", &timestamp) == 1) {
			if( !nextpcptoclean_eport
			  || timestamp < nextpcptoclean_timestamp) {
				nextpcptoclean_eport = eport;
				nextpcptoclean_proto = proto;
				nextpcptoclean_timestamp = timestamp;
				syslog(LOG_DEBUG, "set nextpcptoclean_timestamp to %u", timestamp);
			}
		}
	}
	return 0;
}

/* remove the next redirection that is expired
 */
int CleanExpiredPCP()
{
	char desc[64];
	unsigned timestamp;
	unsigned short iport;
	if(get_redirect_rule(ext_if_name, nextpcptoclean_eport,
	                     (int) nextpcptoclean_proto,
	                     0, 0,
	                     &iport, desc, sizeof(desc),
			     0, 0, &timestamp, 0, 0) < 0)
		return ScanPCPforExpiration();
	/* check desc - this is important since we keep expiration time as part
	 * of the desc.
	 * If the rule is renewed, timestamp and nextpcptoclean_timestamp 
	 * can be different. In that case, the rule must not be removed ! */
	if(sscanf(desc, "PCP %u", &timestamp) == 1) {
		if(timestamp > nextpcptoclean_timestamp)
			return ScanPCPforExpiration();
	}
	/* remove redirection then search for next one:) */
	if(_upnp_delete_redir(nextpcptoclean_eport, nextpcptoclean_proto)<0)
		return -1;
	syslog(LOG_INFO, "Expired PCP mapping port %hu %s removed",
	       nextpcptoclean_eport,
	       nextpcptoclean_proto==IPPROTO_TCP?"TCP":"UDP");
	return ScanPCPforExpiration();
}
#endif

static size_t
_recvrequest(const pcpd_t *pcpd, const pcpctx_t *ctx)
{
	struct pcpsrv_s *srv = (struct pcpsrv_s *) pcpd->opaque;
	struct sockaddr_in *c4 = (struct sockaddr_in *) ctx->client;
	socklen_t cl = sizeof(*c4);
	int cc;

	cc = recvfrom(srv->sock, ctx->request,
		      PCPD_BUF_SIZE, 0,
		      ctx->client, &cl);
	if (cc < 0) {
		syslog(LOG_ERR, "recvfrom(pcp): %m");
		return 0;
	}
	if (!GETFLAG(ENABLEPCPMASK)) {
		syslog(LOG_ERR, "drop spurious pcp");
		return 0;
	}
	return (size_t) cc;
}

static void
_sendresponse(const pcpd_t *pcpd, const pcpctx_t *ctx, const size_t length)
{
	struct pcpsrv_s *srv = (struct pcpsrv_s *) pcpd->opaque;
	socklen_t cl = sizeof(struct sockaddr_in);

	(void) sendto(srv->sock, ctx->response, length, 0, ctx->client, cl);
}

static int
_getsubscriber(const pcpd_t *pcpd, const pcpctx_t *ctx, time_t *epoch)
{
	*epoch = startup_time;
	return PCPD_OK;
}

static int
_addmapping(const pcpd_t *pcpd, const pcpctx_t *ctx,
	    const uint8_t flags, const uint8_t proto,
	    const uint8_t *intaddr, const uint16_t intport,
	    uint8_t *extaddr, uint16_t *extport, uint32_t *lifetime)
{
	char client[16], owner[16], desc[64];
	struct in_addr iaddr, eaddr;
	uint16_t eport, riport;
	int ret;
	unsigned int ts;

	if ((use_ext_ip_addr == NULL) ||
	    (inet_pton(AF_INET, use_ext_ip_addr, &eaddr) <= 0)) {
		*lifetime = pcpd->shorterrht;
		return PCPD_ERR_NETFAILURE;
	}
	if (*lifetime < 120)
		*lifetime = 120;
	else if (*lifetime > 86400)
		*lifetime = 86400;
	if ((proto != IPPROTO_UDP) && (proto != IPPROTO_TCP)) {
		*lifetime = pcpd->longerrht;
		return PCPD_ERR_UNSUPPROTO;
	}
	if (inet_ntop(AF_INET, intaddr, client, sizeof(client)) == NULL) {
		syslog(LOG_ERR, "_addmapping(inet_ntop)");
		*lifetime = pcpd->shorterrht;
		return PCPD_ERR_PROCESSING;
	}
	memcpy(&iaddr, intaddr, 4);
	if (*extport == 0)
		eport = intport;
	else
		eport = *extport;
	if (!check_upnp_rule_against_permissions(upnppermlist, num_upnpperm,
						 eport, iaddr, intport)) {
		*lifetime = pcpd->longerrht;
		return PCPD_ERR_NOTAUTH;
	}
	do {
		ret = get_redirect_rule(ext_if_name, eport, (int) proto,
					owner, sizeof(owner), &riport,
					NULL, 0, NULL, 0,
					&ts, NULL, NULL);
		if (ret == 0) {
			/* mapping already exists */
			if ((strcmp(client, owner) == 0) &&
			    (riport == intport)) {
				syslog(LOG_INFO,
				       "port %hu %s already redirected to "
				       "%s:%hu, replacing",
				       eport,
				       proto == IPPROTO_TCP ? "TCP" : "UDP",
				       owner, riport);
				if (_upnp_delete_redir(eport,
						       (int) proto) < 0) {
					syslog(LOG_ERR,
					       "failed to remove mapping");
					break;
				}
			} else {
				if (flags & PCPD_PREFER_FAILURE) {
					*lifetime = pcpd->shorterrht;
					return PCPD_ERR_CANTPROVIDE;
				}
				eport++;
				continue;
			}
		}
		/* do the mapping */
#if 0
		ts = (unsigned int)(time(NULL) - startup_time) + *lifetime;
		snprintf(desc, sizeof(desc), "PCP %u", ts);
#else
		ts = time(NULL) + *lifetime;
		snprintf(desc, sizeof(desc), "PCP %hu %s",
			 eport, proto == IPPROTO_TCP ? "TCP" : "UDP");
#endif
		if (upnp_redirect_internal(NULL, eport, client, intport,
					   (int) proto, desc, ts) < 0) {
			syslog(LOG_ERR,
			       "Failed to add PCP %hu %s->%s:%hu '%s'",
			       eport, proto == IPPROTO_TCP ? "TCP" : "UDP",
			       client, intport, desc);
			*lifetime = pcpd->shorterrht;
			return PCPD_ERR_PROCESSING;
		}
#if 0
		if (!nextpcptoclean_eport || (ts < nextpcptoclean_timestamp)) {
			nextpcptoclean_timestamp = ts;
			nextpcptoclean_eport = eport;
			nextpcptoclean_proto = proto;
		}
#endif
		break;
	} while (ret == 0);
	memcpy(extaddr, &eaddr, 4);
	*extport = eport;

	return PCPD_OK;
}

static int _delmapping(const pcpd_t *pcpd, const pcpctx_t *ctx,
			const uint8_t proto, const uint8_t *intaddr,
			const uint16_t intport)
{
	char client[16];
	int index = 0, rproto, ret;
	u_short eport, iport;
	char addr[16], desc[64];
	unsigned int ts;

	if (inet_ntop(AF_INET, intaddr, client, sizeof(client)) == NULL) {
		syslog(LOG_ERR, "_delmapping(inet_ntop)");
		return PCPD_ERR_PROCESSING;
	}
	while (get_redirect_rule_by_index(index, 0, &eport, addr,
					  sizeof(addr), &iport, &rproto,
					  desc, sizeof(desc),
					  0, 0, &ts, 0, 0) >= 0) {
		syslog(LOG_DEBUG, "%d %d %hu->'%s':%hu '%s'",
		       index, rproto, eport, addr, iport, desc);
		if (memcmp(desc, "PCP", 3) != 0) {
			index++;
			continue;
		}
		if (strcmp(addr, client) != 0) {
			index++;
			continue;
		}
		if ((proto != 0) && (proto != rproto)) {
			index++;
			continue;
		}
		if ((intport != 0) && (intport != iport)) {
			index++;
			continue;
		}
		ret = _upnp_delete_redir(eport, rproto);
		if (ret < 0) {
			syslog(LOG_ERR, "failed to remove mapping");
			index++;
		} else
			syslog(LOG_INFO, "PCP %s port %hu mapping removed",
			       rproto == IPPROTO_TCP ? "TCP" : "UDP", eport);
	}

	return PCPD_OK;
}

void
SendPCPNotification(int *sockets, int n_sockets)
{
	struct sockaddr_in sockname;
	uint8_t notif[PCP_LEN_ANNOUNCE];
	uint32_t epoch;
	int j, n;

	memset(notif, 0, PCP_LEN_ANNOUNCE);
	/* version (1) */
	notif[PCP_VERSION] = 1;
	/* opcode (128) */
	notif[PCP_OPCODE] = PCP_OPCODE_ANNOUNCE | PCP_OPCODE_RESPONSE;
	/* seconds since "start of epoch" :
	 * time elapsed since the port mapping table was initialized on
	 * startup or reset for any other reason */
	epoch = htonl(time(NULL) - startup_time);
	memcpy(notif + PCP_EPOCH, &epoch, 4);

	memset(&sockname, 0, sizeof(sockname));
	sockname.sin_family = AF_INET;
	if (inet_pton(AF_INET, PCP_RC_ADDR4, &sockname.sin_addr) <= 0) {
		syslog(LOG_WARNING, "%s: inet_pton(%s)",
		       "SendPCPNotification", PCP_RC_ADDR4);
		return;
	}
	for (j = 0; j < n_sockets; j++) {
		if (sockets[j] < 0)
			continue;
		sockname.sin_port = htons(PCP_RC_PORT);
		n = sendto(sockets[j], notif, PCP_LEN_ANNOUNCE, 0,
			   (struct sockaddr *) &sockname, sizeof(sockname));
		if (n < 0) {
			syslog(LOG_ERR, "%s: sendto(%d): %m",
			       "SendPCPNotification", sockets[j]);
			continue;
		}
	}
}

#endif
