#!/usr/bin/env python

# Copyright (C) 20112-2012  Internet Systems Consortium, Inc. ("ISC")
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND ISC DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
# AND FITNESS.  IN NO EVENT SHALL ISC BE LIABLE FOR ANY SPECIAL, DIRECT,
# INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
# LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
# OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
# PERFORMANCE OF THIS SOFTWARE.
#

"""
All in one CGN script for pcpd simulating a plain mode AFTR
(merge between net.py and libpcp/interactive.py)

Francis_Dupont@isc.org, August 2011

Usage: pcpd [-d] aio

schema:

        pcpd.c             python.c             this file
------------------ pcpd ----------------         (aio.py)
+--------------------+------------------+-------------------------+
| PCP server library | python embedding | CGN fake implementation |
|                    |                  |                         |
|                    <= C structure     |    python program       |
|                    |   python object =>                         |
|                    |                  |                         |
|                    |                  | PCP client bindings     |
|                    |                  |                         |
|                    |                  |  interactive client     |
|   PCP protocol     |   recvrequest    | <- submit requests      |
|      engine        |   sendresponse   | -> print responses      |
|                    |                  |                         |
|                    |   addmapping     |    Mapping class        |
|                    |   delmapping     |         ...             |
|                    |       ...        |                         |
+--------------------+------------------+-------------------------+
"""   

__version__ = "$Id: b2d6dc3a3a7257887bf9d108f3c1721c4fc785bd $"

import libpcp
import libpcpd
import readline
import socket
import sys
import time

def totextp(protocol):
    """protocol to text"""
    if protocol == socket.IPPROTO_TCP:
        return 'tcp'
    elif protocol == socket.IPPROTO_UDP:
        return 'udp'
    else:
        return 'proto' + str(protocol)

def totext4(addr):
    """IPv4 address to text"""
    return socket.inet_ntop(socket.AF_INET, addr)

def totext6(addr):
    """IPv6 address to text"""
    return socket.inet_ntop(socket.AF_INET6, addr)

def fromtext4(text):
    """text to IPv4 address"""
    return socket.inet_pton(socket.AF_INET, text)

def fromtext6(text):
    """text to IPv6 address"""
    return socket.inet_pton(socket.AF_INET6, text)

ZERO4 = fromtext4('0.0.0.0')

__PCP__ = libpcp.PCP()
if __PCP__ is None:
    print 'PCP object creation failure?'
    sys.exit(-1)
__PCP__.fdopen(3)
__PCP__.setautofill(0)

def ilhelp():
    """interactive loop help"""

    print '(h)elp: this command'
    print '(r)equest: build and send a request'
    print 'r(a)w: send a raw packet'
    print '(e)val: evaluate an expression'
    print 'e(x)ec: execute a statement'
    print 'exec(f)ile: execute a file'
    print '(l)oadfile: divert the standard input from a file'
    print '(s)tate: show the CGN state'
    print '(q)uit: quit the program'
    print '#...: ignored comment'
    print

STDIN = []

def iloop():
    """interactive loop"""
    while True:
        try:
            command = raw_input('(h)help|' + \
                                    '(r)equest|r(a)w|' + \
                                    '(e)val|e(x)ec|exec(f)ile|' + \
                                    '(l)oadfile|' + \
                                    '(s)tate|' + \
                                    '(q)uit? ')
            if (command == '') or (command[0] == '#'):
                continue
            elif command[0] == 'h':
                ilhelp()
            elif command[0] == 'r':
                try:
                    return buildrequest()
                except Exception as exc:
                    print 'abort on exception:', exc
                except KeyboardInterrupt:
                    print 'interrupted'
            elif command[0] == 'a':
                try:
                    return rawrequest()
                except Exception as exc:
                    print 'abort on exception:', exc
                except KeyboardInterrupt:
                    print 'interrupted'
            elif command == 'e':
                expr = raw_input('expression to eval? ')
                print eval(expr)
            elif command == 'x':
                statement = raw_input('statement to execute? ')
                exec statement
            elif command == 'f':
                filename = raw_input('path of the file to execute? ')
                execfile(filename)
            elif command == 'l':
                filename = raw_input('path of the file to take input from? ')
                try:
                    newin = open(filename, 'r')
                except IOError as exc:
                    print 'got IO error:', exc
                    continue
                STDIN.append(sys.stdin)
                sys.stdin = newin
            elif command == 's':
                __CGN__.showstate()
            elif command == 'q':
                sys.exit(0)
            else:
                print 'unknown command:', command
        except SystemExit:
            raise
        except EOFError:
            if len(STDIN) == 0:
                raise
            print 'at end of file', sys.stdin.name
            sys.stdin.close()
            sys.stdin = STDIN.pop()
        except:
            print 'got unexpected exception:', sys.exc_info()[:2]
            return

def printoptions(options, debug):
    """print options"""
    for (code, data) in options:
        if code == libpcp.OPTION_FILTER:
            if len(data) != 20:
                if debug:
                    print 'option filter: [', repr(data), ']'
                else:
                    print 'option filter'
            else:
                if debug:
                    print 'option filter: [', repr(data), ']', \
                           __PCP__.getoptinfo(code, data)
                else:
                    print 'option filter:', __PCP__.getoptinfo(code, data)
            continue
        elif code == libpcp.OPTION_PREF_FAIL:
            if len(data) != 0:
                if debug:
                    print 'option prefer_failure: [', repr(data), ']'
                else:
                    print 'option prefer_failure'
            else:
                print 'option prefer_failure'
            continue
        elif code == libpcp.OPTION_THIRD_PARTY:
            if (len(data) != 16):
                if debug:
                    print 'option third party: [', repr(data), ']'
                else:
                    print 'option third party'
            else:
                if debug:
                    print 'option third party: [', repr(data), ']', \
                           __PCP__.getoptinfo(code, data)
                else:
                    print 'option third party:', __PCP__.getoptinfo(code, data)
            continue
        else:
            if debug:
                print 'unknown option:', code, '[', repr(data), ']'
            else:
                print 'unknown option:', code
            continue

def parseresponse(address, sid, response):
    """parse a PCP response"""
    debug = address[1] != 1024
    print 'received response from', totext6(sid)
    if debug:
        print repr(response)
    __PCP__.setresponse(response)
    try:
        (_, resp) = __PCP__.submitresponse()
    except Exception as exc:
        print 'got exception:', exc
        return
    if debug:
        print 'result:', resp['result']
        print 'opcode:', resp['opcode']
    print 'epoch:', resp['epoch']
    print 'internal port:', resp['intport']
    print 'external address:', resp['extaddr']
    print 'external port:', resp['extport']
    if 'remotepeer' in resp:
        print 'remote address:', resp['remotepeer']
        print 'remote port:', resp['peerport']
    print 'lifetime:', resp['lifetime']
    printoptions(resp['options'], debug)
    return resp['result']

def buildrequest():
    """build a PCP request"""
    param = raw_input('debug? [y|n] (default no) ')
    debug = param == 'y'

    requestkw = {}
    options = []
    param = raw_input('subscriber ID (IPv6 address)? (no default) ')
    if param == '':
        raise Exception('default subscriber ID is not yet supported')
    source = param
    sid = fromtext6(param)
    param = raw_input('source address? (default is ID)? ')
    if param != '':
        source = param
    peer = False
    param = raw_input('opcode? [map|peer] (default map) ')
    if (param == '') or (param == 'map'):
        requestkw['opcode'] = libpcp.OPCODE_MAP
    elif param == 'peer':
        requestkw['opcode'] = libpcp.OPCODE_PEER
        peer = True
    else:
        raise Exception('unknown opcode: ' + param)
    if debug:
        fake = 1025
    else:
        fake = 1024
    param = raw_input('internal address? (default is source) ')
    if param != '':
        requestkw['intaddr'] = param
    else:
        requestkw['intaddr'] = source
    param = raw_input('internal port? (no default) ')
    if param == '':
        raise Exception('an internal port number is required')
    requestkw['intport'] = int(param)
    param = raw_input('protocol? (no default) ')
    if param == 'all':
        requestkw['protocol'] = 0
    elif param == 'udp':
        requestkw['protocol'] = socket.IPPROTO_UDP
    elif param == 'tcp':
        requestkw['protocol'] = socket.IPPROTO_TCP
    else:
        requestkw['protocol'] = int(param)
    param = raw_input('lifetime? (default is 0) ')
    if param != '':
        requestkw['lifetime'] = int(param)
    if peer:
        param = raw_input('remote peer address? ')
        if param != '':
            requestkw['remotepeer'] = param
            param = raw_input('remote peer port? ')
            if param != '':
                requestkw['peerport'] = int(param)
    param = raw_input('external address? (default 0.0.0.0) ')
    if param == '':
        requestkw['extaddr'] = '0.0.0.0'
    else:
        requestkw['extaddr'] = param
    param = raw_input('external port? (default 0) ')
    if param != '':
        requestkw['extport'] = int(param)
    while True:
        param = raw_input('append an option? [y|n] (default no) ')
        if param != 'y':
            break
        param = raw_input('option code? ' + \
                              '[third_party|prefer_failure|filter] ')
        if param == 'third_party':
            family = socket.AF_INET
            address = raw_input('third party address? (no default) ')
            options.append(__PCP__.third_party(family, address))
            continue
        elif param == 'prefer_failure':
            options.append(__PCP__.prefer_failure())
            continue
        elif param == 'filter':
            family = socket.AF_INET
            address = raw_input('filter address? (no default) ')
            param = raw_input('filter prefix length? (default 32) ')
            if param == '':
                param = '32'
            prefixlen = int(param)
            param = raw_input('filter port? (default is 0) ')
            if param == '':
                param = '0'
            options.append(__PCP__.filter(family, prefixlen,
                                          address, int(param)))
            continue
        else:
            print 'unknown option', param
            continue
    if options:
        requestkw['options'] = options

    if debug:
        print 'calling makerequest method with', requestkw
    __PCP__.makerequest(**requestkw)
    request = __PCP__.getrequest()
    return (socket.AF_INET6, (source, fake), sid, request)

def rawrequest():
    """get and send a raw request"""
    param = raw_input('debug? [y|n] (default no) ')
    debug = param == 'y'
    param = raw_input('subscriber ID (IPv6 address)? (no default) ')
    if param == '':
        raise Exception('default subscriber ID is not yet supported')
    source = param
    sid = fromtext6(param)
    param = raw_input('source address? (default is ID)? ')
    if param != '':
        source = param
    if debug:
        fake = 1025
    else:
        fake = 1024
    param = input('raw packet? (no default) ')
    if len(param) < 2:
        raise Exception('minimal length is 2 bytes')
    __PCP__.setrequest(param)
    return (socket.AF_INET6, (source, fake), sid, param)

class Filter(object):
    """Filter"""
    __slots__ = ['addr', 'plen', 'port']
    __hash__ = None

    def __init__(self, addr, plen, port):
        """initializer"""
        self.addr = addr
        self.plen = plen
        self.port = port

    def __cmp__(self, other):
        """compare"""
        if not isinstance(other, Filter):
            return 1
        if self.plen != other.plen:
            return cmp(self.plen, other.plen)
        if self.addr != other.addr:
            return cmp(self.addr, other.addr)
        return cmp(self.port, other.port)

    def totext(self, indent=''):
        """pretty printer"""
        print (indent + '['), totext4(self.addr), '/', self.plen, \
            ' ', self.port, ']'
        
class Mapping(object):
    """explicit dynamic mapping"""
    __slots__ = ['protocol', 'intaddr', 'intport',
                 'extaddr', 'extport', 'expire', 'filters']
    __hash__ = None

    def __init__(self, protocol, intaddr, intport,
                 extaddr, extport, expire):
        """initializer"""
        self.protocol = protocol
        self.intaddr = intaddr
        self.intport = intport
        self.extaddr = extaddr
        self.extport = extport
        self.expire = expire
        self.filters = []

    def __cmp__(self, other):
        """compare"""
        if not isinstance(other, Mapping):
            return 1
        if self.protocol != other.protocol:
            return cmp(self.protocol, other.protocol)
        if self.intaddr != other.intaddr:
            return cmp(self.intaddr, other.intaddr)
        return cmp(self.intport, other.intport)

    def totext(self, indent=''):
        """pretty print"""
        if self.expire == 0:
            print (indent + 'map'), totextp(self.protocol), \
                totext4(self.intaddr), ':', self.intport, '->', \
                totext4(self.extaddr), ':', self.extport, 'static'
        else:
            print (indent + 'map'), totextp(self.protocol), \
                totext4(self.intaddr), ':', self.intport, '->', \
                totext4(self.extaddr), ':', self.extport, 'in', \
                self.expire - time.time()

class Dynamic(object):
    """implicit dynamic mapping"""
    __slots__ = ['protocol', 'intaddr', 'intport', 'peeraddr',
                 'peerport', 'extaddr', 'extport', 'expire']
    __hash__ = None

    def __init__(self, protocol, intaddr, intport, peeraddr,
                 peerport, extaddr, extport, expire):
        """initializer"""
        self.protocol = protocol
        self.intaddr = intaddr
        self.intport = intport
        self.peeraddr = peeraddr
        self.peerport = peerport
        self.extaddr = extaddr
        self.extport = extport
        self.expire = expire

    def __cmp__(self, other):
        """compare"""
        if not isinstance(other, Dynamic):
            return 1
        if self.protocol != other.protocol:
            return cmp(self.protocol, other.protocol)
        if self.intaddr != other.intaddr:
            return cmp(self.intaddr, other.intaddr)
        return cmp(self.intport, other.intport)

    def totext(self, indent=''):
        """pretty print"""
        if self.expire == 0:
            print (indent + 'dynamic'), totextp(self.protocol), \
                totext4(self.intaddr), ':', self.intport, '->', \
                totext4(self.extaddr), ':', self.extport, 'to', \
                totext4(self.peeraddr), ':', self.peerport, 'static'
        else:
            print (indent + 'dynamic'), totextp(self.protocol), \
                totext4(self.intaddr), ':', self.intport, '->', \
                totext4(self.extaddr), ':', self.extport, 'to', \
                totext4(self.peeraddr), ':', self.peerport, 'in', \
                self.expire - time.time()

class Subscriber(object):
    """Subscriber"""
    __slots__ = ['sid', 'extaddr', 'start', 'mappings', 'dynamics']
    __hash__ = None

    def __init__(self, sid):
        """initializer"""
        self.sid = sid
        self.extaddr = fromtext4('198.18.200.1')
        self.start = time.time()        
        self.mappings = []
        self.dynamics = []

    def totext(self, indent=''):
        """pretty print"""
        print (indent + 'subscriber'), totext6(self.sid), 'since', \
            time.time() - self.start, 'at', totext4(self.extaddr)

    def showstate(self, debug):
        """show a subscriber"""
        self.totext()
        print len(self.mappings), 'explicit dynamic mappings'
        for mapping in self.mappings:
            mapping.totext('\t')
            if debug and (len(mapping.filters) > 0):
                print '\t', len(mapping.filters), 'filters'
                for filt in mapping.filters:
                    filt.totext('\t\t')
        print len(self.dynamics), 'implicit dynamic mappings'
        for dynamic in self.dynamics:
            dynamic.totext('\t')

UINT32_MAX = 4294967295

class CGN(object):
    """Carrier Grade NAT simulator"""
    __slots__ = ['cgntype', 'opcodes', 'flags', 'longerrht', 'shorterrht',
                 'subscribers', 'acl6', 'acl4',
                 'xnextport', 'xbusy', 'inextport', 'ibusy', 'stats']

    def __init__(self):
        """initializer"""
        self.opcodes = libpcpd.OPCODE_ANNOUNCE | \
            libpcpd.OPCODE_MAP | libpcpd.OPCODE_PEER
        self.cgntype = libpcpd.DSLITE
        self.flags =  libpcpd.MAPINT | libpcpd.MAPEXT | \
            libpcpd.MAPDST | libpcpd.THIRDPARTY
        self.longerrht = libpcpd.LONGERRHT
        self.shorterrht = libpcpd.SHORTERRHT
        self.subscribers = {}
        self.acl6 = [('2001::', 48)]
        self.acl4 = [('10.0.0.0', 8), ('172.16.0.0', 16),
                     ('192.168.0.0', 16), ('192.0.0.0', 29)]
        self.xnextport = 32000
        self.xbusy = []
        self.inextport = 1024
        self.ibusy = []
        self.stats = { 'operation' : 257 * [0L],
                       'result' : 257 * [0L],
                       'option' : 257 * [0L] }
        
    def expire(self):
        """expire routine"""
        now = time.time()
        for client in self.subscribers.itervalues():
            for mapping in client.mappings:
                if mapping.expire == 0:
                    continue
                if mapping.expire > now:
                    continue
                print 'expire:', mapping.extport
                client.mappings.remove(mapping)
                self.xbusy.remove(mapping.extport)
            for dynamic in client.dynamics:
                if dynamic.expire == 0:
                    continue
                if dynamic.expire > now:
                    continue
                print 'expire:', dynamic.extport
                client.dynamics.remove(dynamic)
                self.ibusy.remove(dynamic.extport)

    def showstate(self):
        """show the CGN state"""
        what = raw_input('debug? [y|n] (default no) ')
        debug = what == 'y'
        print len(self.subscribers), 'subscribers'
        print
        for client in self.subscribers.itervalues():
            client.showstate(debug)
            print
        if debug:
            print len(self.acl6), 'acl6 entries'
            for acl in self.acl6:
                print '\t' + acl[0] + '/' + str(acl[1])
            print len(self.acl4), 'acl4 entries'
            for acl in self.acl4:
                print '\t' + acl[0] + '/' + str(acl[1])
            print 'explicit next port:', self.xnextport
            print 'busy explicit ports:', self.xbusy
            print 'implicit next port:', self.inextport
            print 'busy implicit ports:', self.ibusy
        print 'requests:', self.stats['operation'][256]
        if debug:
            for idx in xrange(0, 256):
                cnt = self.stats['operation'][idx]
                if cnt != 0L:
                    print ('\topcode['+ str(idx) + ']:'), cnt
        print 'responses:', self.stats['result'][256]
        if debug:
            for idx in xrange(0, 256):
                cnt = self.stats['result'][idx]
                if cnt != 0L:
                    print ('\tresult['+ str(idx) + ']:'), cnt
        print 'options:', self.stats['option'][256]
        if debug:
            for idx in xrange(0, 256):
                cnt = self.stats['option'][idx]
                if cnt != 0L:
                    print ('\toption['+ str(idx) + ']:'), cnt

    def recvrequest(self):
        """API / receive request"""
        return iloop()

    def sendresponse(self, family, address, sid, response):
        """API / send response"""
        parseresponse(address, sid, response)

    def getsubscriber(self, sid):
        """API / get subscriber"""
        if not sid in self.subscribers:
            self.subscribers[sid] = Subscriber(sid)
        client = self.subscribers[sid]
        return (libpcpd.OK, int(client.start))

    def authorize(self, family, address, sid, request, off3rdpty):
        """API / authorize"""
        return libpcpd.OK

    def addmapping(self, flags, sid, protocol, intaddr, intport,
                   extaddr, extport, lifetime):
        """API / add explicit dynamic mapping"""
        client = self.subscribers.get(sid)
        if client is None:
            return (libpcpd.ERR_NOTAUTH, extaddr, 0, 1800)
        if flags & libpcpd.PREFER_FAILURE:
            if (extaddr != ZERO4) and (extaddr != client.extaddr):
                return (libpcpd.ERR_CANTPROVIDE, extaddr, 0, 300)
            if extport == 0:
                flags &= ~libpcpd.PREFER_FAILURE
        if lifetime < 120:
            lifetime = 120
        elif lifetime > 7200:
            lifetime = 7200
        map0 = Mapping(protocol, intaddr, intport,
                       client.extaddr, extport, time.time() + lifetime)
        for mapping in client.mappings:
            if mapping == map0:
                if (flags & libpcpd.PREFER_FAILURE) and \
                        (mapping.extport != extport):
                    return (libpcpd.ERR_CANTPROVIDE, extaddr, 0, 300)
                if mapping.expire == 0:
                    return (libpcpd.STATICMAPPING, mapping.extaddr,
                            mapping.extport, UINT32_MAX)
                extport = mapping.extport
                mapping.expire = time.time() + lifetime
                return (libpcpd.OK, client.extaddr, extport, lifetime)
        if flags & libpcpd.PREFER_FAILURE:
            if extport < 32000:
                return (libpcpd.ERR_CANTPROVIDE, extaddr, 0, 1800)
            if extport in self.xbusy:
                return (libpcpd.ERR_CANTPROVIDE, extaddr, 0, 30)
        elif (extport < 32000) or (extport in self.xbusy):
            while self.xnextport in self.xbusy:
                self.xnextport += 1
                if self.xnextport > 32999:
                    self.xnextport = 32000
            extport = self.xnextport
            self.xnextport += 1
            if self.xnextport > 32999:
                self.xnextport = 32000
        self.xbusy.append(extport)
        map0 = Mapping(protocol, intaddr, intport,
                       client.extaddr, extport, time.time() + lifetime)
        client.mappings.append(map0)
        if flags & libpcpd.CREATE_AS_LOCKED:
            return (libpcpd.NEWMAPPING,
                    client.extaddr, extport, lifetime)
        else:
            return (libpcpd.OK, client.extaddr, extport, lifetime)

    def delmapping(self, sid, protocol, intaddr, intport):
        """API / delete explicit dynamic mapping"""
        client = self.subscribers.get(sid)
        if client is None:
            return libpcpd.OK
        for mapping in client.mappings:
            if intport != 0:
                if mapping.intport != intport:
                    continue
            if intaddr != ZERO4:
                if mapping.intaddr != intaddr:
                    continue
            if protocol != 0:
                if mapping.protocol != protocol:
                    continue
            client.mappings.remove(mapping)
            self.xbusy.remove(mapping.extport)
        return libpcpd.OK

    def adddynamic(self, sid, protocol, intaddr, intport,
                   peeraddr, peerport, extaddr, extport, lifetime):
        """API / add implicit dynamic mapping"""
        client = self.subscribers.get(sid)
        if client is None:
            return (libpcpd.ERR_NOTAUTH, extaddr, 0, 1800)
        if (extaddr != ZERO4) and (extaddr != client.extaddr):
            return (libpcpd.ERR_CANTPROVIDE, extaddr, 0, 300)
        dyn0 = Dynamic(protocol, intaddr, intport, peeraddr, peerport,
                       client.extaddr, extport, time.time() + lifetime)
        for dynamic in client.dynamics:
            if dynamic == dyn0:
                if (extport != 0) and (dynamic.extport != extport):
                    return (libpcpd.ERR_CANTPROVIDE, extaddr, 0, 300)
                if dynamic.expire == 0:
                    return (libpcpd.STATICMAPPING, dynamic.extaddr,
                            dynamic.extport, UINT32_MAX)
                extport = dynamic.extport
                dynamic.expire = time.time() + lifetime
                return (libpcpd.OK, client.extaddr, extport, lifetime)
        if extport == 0:
            while self.inextport in self.ibusy:
                self.inextport += 1
                if self.inextport > 19999:
                    self.inextport = 1024
            extport = self.inextport
            self.inextport += 1
            if self.inextport > 19999:
                self.inextport = 1024
        if (extport < 1024) or ((extport >= 32000) and (extport < 33000)):
            return (libpcpd.ERR_CANTPROVIDE, extaddr, 0, 1800)
        if (extport in self.ibusy) or (extport in self.xbusy):
            return (libpcpd.ERR_CANTPROVIDE, extaddr, 0, 30)
        if Mapping(protocol, intaddr, intport, client.extaddr,
                   extport, 0) in client.mappings:
            return (libpcpd.ERR_PROCESSING, extaddr, 0, 300)
        self.ibusy.append(extport)
        dyn0 = Dynamic(protocol, intaddr, intport, peeraddr, peerport,
                       client.extaddr, extport, time.time() + lifetime)
        client.dynamics.append(dyn0)
        return (libpcpd.OK, client.extaddr, extport, lifetime)

    def filter(self, sid, protocol, intaddr, intport, flags, filters):
        """API / manage explicit dynamic mapping remote filters"""
        client = self.subscribers.get(sid)
        if client is None:
            return libpcpd.ERR_NOTAUTH
        mapping = None
        for mapping in client.mappings:
            if mapping.intport != intport:
                continue
            if mapping.intaddr != intaddr:
                continue
            if mapping.protocol == protocol:
                break
        if mapping is None:
            return libpcpd.ERR_PROCESSING
        saved = list(mapping.filters)
        if flags & libpcpd.DELETE_FILTERS:
            mapping.filters = []
        try:
            for triple in filters:
                (plen, port, addr) = triple
                filt = Filter(addr, plen, port)
                if filt in mapping.filters:
                    continue
                mapping.filters.append(filt)
            if len(mapping.filters) > 10:
                mapping.filters = saved
                return libpcpd.ERR_TOOMANYPEER
            mapping.filters.sort()
            return libpcpd.OK
        except:
            mapping.filters = saved
            return libpcpd.ERR_PROCESSING

    def incrcounter(self, code, idx):
        """API / increment statistic counter"""
        statvec = self.stats.get(code)
        if statvec is not None:
            statvec[idx] += 1
            statvec[256] += 1

__CGN__ = CGN()
