/*
 * Copyright (C) 2011  Internet Systems Consortium, Inc. ("ISC")
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND ISC DISCLAIMS ALL WARRANTIES WITH
 * REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS.  IN NO EVENT SHALL ISC BE LIABLE FOR ANY SPECIAL, DIRECT,
 * INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
 * OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */

/* $Id: 518aea82adf028511c5213a57b04c9957268b1a1 $ */

/* PCP daemon: communication with the AFTR module
 *
 * Francis_Dupont@isc.org, August 2011
 */

#include "internal.h"

/* offsets and interesting constants for IPv4 and UDP headers */
#define IPLEN		2
#define IPID		4
#define IPTTL		8
#define IPPROTO		9
#define IPCKSUM		10
#define IPSRC		12
#define IPDST		16
#define UDPSPORT	20
#define UDPDPORT	22
#define UDPLEN		24
#define UDPCKSUM	26
#define IPHDRLEN	20
#define UDPHDRLEN	8

/* globals */

int transid;
char nextlinebuf[1024];
int nextlinebeg, nextlineend;
char pendingbuf[2048];
int pendingcnt;
int aftr;
int sock6 = -1;
int sock4 = -1;

/* send a command to AFTR daemon */

void
ask(char *cmd)
{
	char buf[128];

	(void) snprintf(buf, sizeof(buf), "%s\n", cmd);
	(void) send(aftr, buf, strlen(buf), 0);
	debug("send to AFTR: %s", cmd);
}

/* get next line from AFTR daemon */

char *
getnextline(int wait)
{
	fd_set set;
	struct timeval tv;
	sigset_t asset;
	char *ret, *eol;

	if (nextlinebeg >= nextlineend) {
		if (wait) {
			FD_ZERO(&set);
			FD_SET(aftr, &set);
			tv.tv_sec = 0;
			tv.tv_usec = 100000;
			sigemptyset(&asset);
			sigaddset(&asset, SIGALRM);
			sigprocmask(SIG_BLOCK, &asset, NULL);
			select(aftr + 1, &set, NULL, NULL, &tv);
			sigprocmask(SIG_UNBLOCK, &asset, NULL);
			if (!FD_ISSET(aftr, &set))
				return NULL;
		}
		memset(nextlinebuf, 0, sizeof(nextlinebuf));
		nextlinebeg = 0;
		nextlineend = recv(aftr, nextlinebuf,
				   sizeof(nextlinebuf) - 1, 0);
	}
	eol = strchr(nextlinebuf + nextlinebeg, '\n');
	if (eol == NULL) {
		if (nextlineend == 0)
			return NULL;
		debug("getnextline: missing EOL");
		eol = nextlinebuf + nextlinebeg + nextlineend;
	}
	*eol++ = '\0';
	ret = nextlinebuf + nextlinebeg;
	nextlinebeg = eol - nextlinebuf;
	debug("received from AFTR: %s", ret);
	return ret;
}
		
/* expect a text from AFTR daemon */

char *
expect(const char *text, int *result)
{
	char *got, buf[64];
	int i, l, coll;

	transid++;
	*result = 0;
	coll = 0;
	(void) snprintf(buf, sizeof(buf), "echo %d\n", transid);
	(void) send(aftr, buf, strlen(buf), 0);
	for (i = 5; i != 0; i--) {
		got = getnextline(1);
		if (got == NULL)
			continue;
		if (strncmp(got, text, strlen(text)) == 0) {
			debug("got expected '%s' from AFTR", text);
			*result = 1;
			return got + strlen(text);
		}
		if (strncmp(got, "collision", 9) == 0) {
			debug("got collision from AFTR");
			coll = 1;
		}
		if (strcmp(got, "command failed") == 0) {
			debug("got failure from AFTR");
			if (coll)
				*result = -1;
			return NULL;
		}
		l = strlen(got);
		if (strncmp(got, buf, l) == 0) {
			debug("got echo from AFTR");
			return NULL;
		}
		memcpy(pendingbuf + pendingcnt, got, l);
		pendingcnt += l;
		pendingbuf[pendingcnt++] = '\0';
	}
	return NULL;
}

/* ask list pcp */

void
asklistpcp(void)
{
	uint32_t addr, previous = 0;
	uint16_t minport[2], maxport[2];
	uint8_t proto;
	char *got, buf[64];
	int i, l;

	transid++;
	minport[0] = minport[1] = 0;
	maxport[0] = maxport[1] = 0;

	ask("list pcp");
	(void) snprintf(buf, sizeof(buf), "echo %d\n", transid);
	(void) send(aftr, buf, strlen(buf), 0);

	for (i = 5; i != 0; i--) {
		got = getnextline(1);
		if (got == NULL)
			continue;
		if (strncmp(got, "pcp ", 4) == 0) {
			char *line, *tok, *ptok;
			int minp, maxp;

			line = strdup(got);
			if (line == NULL) {
				fprintf(stderr,
					"asklistpcp: can't copy '%s': %s\n",
					got, strerror(errno));
				exit(1);
			}
			(void) strtok(line, " ");
			if ((tok = strtok(NULL, " ")) == NULL) {
				fprintf(stderr,
					"asklistpcp: can't get "
					"address in '%s'\n",
					got);
				exit(1);
			}
			if (inet_pton(AF_INET, tok, &addr) != 1) {
				fprintf(stderr,
					"asklistpcp: bad address in '%s'\n",
					got);
				exit(1);
			}
			if ((previous != 0) && (previous != addr)) {
				new_natsrc(previous, minport, maxport);
				minport[0] = minport[1] = 0;
				maxport[0] = maxport[1] = 0;
			}
			previous = addr;
			if ((tok = strtok(NULL, " ")) == NULL) {
				fprintf(stderr,
					"asklistpcp: can't get "
					"protocol in '%s'\n",
					got);
				exit(1);
			}
			if (strcmp(tok, "tcp") == 0)
				proto = IPPROTO_TCP;
			else if (strcmp(tok, "udp") == 0)
				proto = IPPROTO_UDP;
			else {
				fprintf(stderr,
					"asklistpcp: bad protocol in '%s'\n",
					got);
				exit(1);
			}
			if ((ptok = strtok(NULL, " ")) == NULL) {
				fprintf(stderr,
					"asklistpcp: can't get "
					"ports in '%s'\n",
					got);
				exit(1);
			}
			if (strtok(NULL, " ") != NULL) {
				fprintf(stderr,
					"asklistpcp: extra token in '%s'\n",
					got);
				exit(1);
			}
			if ((tok = strtok(ptok, "-")) == NULL) {
				fprintf(stderr,
					"asklistpcp: can't get "
					"minport in '%s'\n",
					got);
				exit(1);
			}
			minp = atoi(tok);
			if ((minp < 0) || (minp > 65535)) {
				fprintf(stderr,
					"asklistpcp: bad minport in '%s'\n",
					got);
				exit(1);
			}
			minport[PROTO2INDEX(proto)] = (uint16_t) minp;
			if ((tok = strtok(NULL, "\n")) == NULL) {
				fprintf(stderr,
					"asklistpcp: can't get "
					"maxport in '%s'\n",
					got);
				exit(1);
			}
			maxp = atoi(tok);
			if ((maxp < 0) || (maxp > 65535) ||
			    (minp > maxp) || ((maxp == 0) && (minp != 0))) {
				fprintf(stderr,
					"asklistpcp: bad maxport in '%s'\n",
					got);
				exit(1);
			}
			maxport[PROTO2INDEX(proto)] = (uint16_t) maxp;
			continue;
		}
		if (strcmp(got, "command failed") == 0) {
			debug("got failure from AFTR");
			return;
		}
		l = strlen(got);
		if (strncmp(got, buf, l) == 0) {
			if (previous != 0)
				new_natsrc(previous, minport, maxport);
			return;
		}
		memcpy(pendingbuf + pendingcnt, got, l);
		pendingcnt += l + 1;
		pendingbuf[pendingcnt] = '\0';
	}
	debug("timeout on 'list pcp'");
}

/* ask tunnel info */

uint32_t
asktunnel(struct in6_addr *addr)
{
	char host[INET6_ADDRSTRLEN];
	char cmd[128], resp[128];
	char *nsrc;
	int ret;
	uint32_t in;

	(void) inet_ntop(AF_INET6, addr, host, INET6_ADDRSTRLEN);
	(void) snprintf(cmd, sizeof(cmd), "try tunnel %s", host);
	(void) snprintf(resp, sizeof(resp), "tunnel %s ", host);
	ask(cmd);
	nsrc = expect(resp, &ret);
	if (ret != 1) {
		debug("tunnel %s failed?", host);
		return 0;
	} else {
		if (inet_pton(AF_INET, nsrc, &in) <= 0) {
			debug("inet_pton(%s) failed?", nsrc);
			return 0;
		}
		debug("tunnel %s %s", host, nsrc);
		return in;
	}
}

/* ask tunnel add */

int
askaddtunnel(struct client *client)
{
	char host6[INET6_ADDRSTRLEN];
	char host4[INET_ADDRSTRLEN];
	char text[128];
	uint32_t addr;
	int ret;

	addr = client->natsrc->addr;
	(void) inet_ntop(AF_INET6, &client->ipv6, host6, INET6_ADDRSTRLEN);
	(void) inet_ntop(AF_INET, &addr, host4, INET_ADDRSTRLEN);
	(void) snprintf(text, sizeof(text), "try tunnel %s %s", host6, host4);
	ask(text);
	(void) expect(text + 4, &ret);
	if (ret != 1) {
		debug("add '%s' failed?", text + 4);
		return PCPD_ERR_NOTAUTH;
	}
	syslog(LOG_NOTICE, "%s", text + 4);
	return PCPD_OK;
}

/* ask static NAT delete */

void
askdelmap(struct mapping *mapping)
{
	struct client *client;
	char host6[INET6_ADDRSTRLEN];
	char host4[INET_ADDRSTRLEN];
	char cmd[256];
	char *proto;

	client = mapping->client;
	if (mapping->proto == IPPROTO_TCP)
		proto = " tcp ";
	else
		proto = " udp ";
	(void) inet_ntop(AF_INET6, &client->ipv6, host6, INET6_ADDRSTRLEN);
	(void) inet_ntop(AF_INET, &mapping->src, host4, INET_ADDRSTRLEN);
	(void) snprintf(cmd, sizeof(cmd), "delete nat %s%s%s %hu",
			host6, proto, host4, mapping->sport);
	ask(cmd);
	syslog(LOG_NOTICE, "del %s", cmd + 7);
}

/* ask static NAT add */

int
askaddmap(struct mapping *mapping, int *staticport)
{
	struct client *client;
	char host6[INET6_ADDRSTRLEN];
	char src4[INET_ADDRSTRLEN];
	char nsrc4[INET_ADDRSTRLEN];
	char text[256], resp[512];
	char *got, *tok;
	char *proto;
	int ret;

	client = mapping->client;
	if (mapping->proto == IPPROTO_TCP)
		proto = " tcp ";
	else
		proto = " udp ";
	(void) inet_ntop(AF_INET6, &client->ipv6, host6, INET6_ADDRSTRLEN);
	(void) inet_ntop(AF_INET, &mapping->src, src4, INET_ADDRSTRLEN);
	(void) inet_ntop(AF_INET, &client->natsrc->addr,
			 nsrc4, INET_ADDRSTRLEN);
	(void) snprintf(text, sizeof(text), "try map %s%s%s %hu %s %hu",
			host6, proto, src4, mapping->sport, nsrc4,
			mapping->port->port);
	(void) snprintf(resp, sizeof(resp) / 2, "map %s%s%s %hu ",
			host6, proto, src4, mapping->sport);
	ask(text);
	got = expect(resp, &ret);
	if (ret != 1) {
		debug("add '%s' failed?", text + 4);
		return PCPD_ERR_PROCESSING;
	}
	(void) strncat(resp, got, sizeof(resp) / 2);
	syslog(LOG_NOTICE, "add %s", resp);
	if ((tok = strtok(got, " \t")) == NULL) {
		debug("can't get back natted address %s", nsrc4);
		return PCPD_ERR_PROCESSING;
	}
	if (strcmp(tok, "static") == 0) {
		if (((tok = strtok(NULL, " \t")) == NULL) ||
		    (strcmp(tok, nsrc4) != 0)) {
			debug("can't get back static natted address %s",
			      nsrc4);
			return PCPD_ERR_PROCESSING;
		}
		tok = strtok(NULL, " \t");
		if ((tok == NULL) || (strtok(NULL, " \t") != NULL)) {
			debug("can't get static natted port");
			return PCPD_ERR_PROCESSING;
		}
		*staticport = atoi(tok);
		if ((*staticport <= 0) || (*staticport > 65535)) {
			debug("bad static natted port %s???", tok);
			return PCPD_ERR_PROCESSING;
		}
		return PCPD_STATICMAPPING;
	}
	if (strcmp(tok, nsrc4) != 0) {
		debug("can't get back natted address %s", nsrc4);
		return PCPD_ERR_PROCESSING;
	}
	tok = strtok(NULL, " \t");
	if ((tok == NULL) || (strtok(NULL, " \t") != NULL)) {
		debug("can't get natted port");
		return PCPD_ERR_PROCESSING;
	}
	if (atoi(tok) != mapping->port->port) {
		debug("bad natted port %s???", tok);
		return PCPD_ERR_PROCESSING;
	}
	return PCPD_OK;
}

/* ask implicit dynamic NAT */

int
askdynamic(struct client *client, uint8_t protocol,
	   uint32_t src, uint32_t nsrc, uint32_t dst,
	   uint16_t sport, uint16_t *nport, uint16_t dport,
	   u_int *lifetime)
{
	char host6[INET6_ADDRSTRLEN];
	char src4[INET_ADDRSTRLEN];
	char nsrc4[INET_ADDRSTRLEN];
	char dst4[INET_ADDRSTRLEN];
	char text[256], resp[256];
	char *got;
	char *proto;
	int ret;

	if (protocol == IPPROTO_TCP)
		proto = " tcp ";
	else
		proto = " udp ";
	(void) inet_ntop(AF_INET6, &client->ipv6, host6, INET6_ADDRSTRLEN);
	(void) inet_ntop(AF_INET, &src, src4, INET_ADDRSTRLEN);
	(void) inet_ntop(AF_INET, &nsrc, nsrc4, INET_ADDRSTRLEN);
	(void) inet_ntop(AF_INET, &dst, dst4, INET_ADDRSTRLEN);
	(void) snprintf(text, sizeof(text),
			"try dynamic %s%s%s %hu %s %hu %s %hu %u",
			host6, proto, src4, sport, dst4, dport, 
			nsrc4, *nport, *lifetime);
	(void) snprintf(resp, sizeof(resp),
			"dynamic %s%s%s %hu %s %hu %s ",
			host6, proto, src4, sport, dst4, dport, nsrc4);
	ask(text);
	got = expect(resp, &ret);
	if (ret != 1) {
		debug("%s failed?", text + 4);
		*lifetime = LONGERRHT;
		return PCPD_ERR_PROCESSING;
	}
	if (sscanf(got, "%hu %u", nport, lifetime) != 2) {
		debug("can't get nport and lifetime from %s", got);
		*lifetime = LONGERRHT;
		return PCPD_ERR_PROCESSING;
	}
	debug("%s%s", resp, got);
	return PCPD_OK;
}

/* ask filter delete */

void
askdelfilter(struct client *client,
	     uint32_t src, uint16_t port, uint8_t protocol)
{
	char host6[INET6_ADDRSTRLEN];
	char host4[INET_ADDRSTRLEN];
	char cmd[256];
	char *proto;

	if (protocol == IPPROTO_TCP)
		proto = " tcp ";
	else
		proto = " udp ";
	(void) inet_ntop(AF_INET6, &client->ipv6, host6, INET6_ADDRSTRLEN);
	(void) inet_ntop(AF_INET, &src, host4, INET_ADDRSTRLEN);
	(void) snprintf(cmd, sizeof(cmd), "delete filter %s%s%s %hu",
			host6, proto, host4, port);
	ask(cmd);
	debug("del %s", cmd + 7);
}

/* ask filter replace */

int
askreplacefilter(struct client *client,
		 uint32_t src, uint16_t sport, uint8_t protocol,
		 uint8_t prefixlen, uint32_t dst, uint16_t dport)
{
	char host6[INET6_ADDRSTRLEN];
	char host4[INET_ADDRSTRLEN];
	char dst4[INET_ADDRSTRLEN];
	char cmd[256];
	char resp[128];
	char *proto;
	int ret;

	if (protocol == IPPROTO_TCP)
		proto = " tcp ";
	else
		proto = " udp ";
	(void) inet_ntop(AF_INET6, &client->ipv6, host6, INET6_ADDRSTRLEN);
	(void) inet_ntop(AF_INET, &src, host4, INET_ADDRSTRLEN);
	(void) inet_ntop(AF_INET, &dst, dst4, INET_ADDRSTRLEN);
	(void) snprintf(cmd, sizeof(cmd),
			"filter replace %s%s%s %hu %s/%d %hu",
			host6, proto, host4, sport,
			dst4, (int) prefixlen, dport);
	(void) snprintf(resp, sizeof(resp), "replaced filter %s%s%s %hu",
			host6, proto, host4, sport);
	ask(cmd);
	(void) expect(resp, &ret);
	if (ret < 0)
		return PCPD_ERR_TOOMANYPEER;
	if (ret != 1)
		return PCPD_ERR_PROCESSING;
	return PCPD_OK;
}

/* ask filter add */

int
askaddfilter(struct client *client,
	     uint32_t src, uint16_t sport, uint8_t protocol,
	     uint8_t prefixlen, uint32_t dst, uint16_t dport)
{
	char host6[INET6_ADDRSTRLEN];
	char host4[INET_ADDRSTRLEN];
	char dst4[INET_ADDRSTRLEN];
	char cmd[256];
	char resp[128];
	char *proto;
	int ret;

	if (protocol == IPPROTO_TCP)
		proto = " tcp ";
	else
		proto = " udp ";
	(void) inet_ntop(AF_INET6, &client->ipv6, host6, INET6_ADDRSTRLEN);
	(void) inet_ntop(AF_INET, &src, host4, INET_ADDRSTRLEN);
	(void) inet_ntop(AF_INET, &dst, dst4, INET_ADDRSTRLEN);
	(void) snprintf(cmd, sizeof(cmd), "filter %s%s%s %hu %s/%d %hu",
			host6, proto, host4, sport,
			dst4, (int) prefixlen, dport);
	(void) snprintf(resp, sizeof(resp), "added filter %s%s%s %hu",
			host6, proto, host4, sport);
	ask(cmd);
	(void) expect(resp, &ret);
	if (ret < 0)
		return PCPD_ERR_TOOMANYPEER;
	if (ret != 1)
		return PCPD_ERR_PROCESSING;
	return PCPD_OK;
}

/* check for a 'tunnel del ' notification */

void
checkdeltunnel(const char *got)
{
	struct client *client;
	struct in6_addr ipv6;

	if (strncmp(got, "tunnel del ", 11) != 0)
		return;
	if (inet_pton(AF_INET6, got + 11, &ipv6) != 1)
		return;
	client = getclient(&ipv6, 0);
	if (client == NULL)
		return;
	syslog(LOG_NOTICE, "delete client %s", got + 11);
	delclient(client);
}

/* Compute checksum */

int
in_cksum(u_char *p, u_int l)
{
	int sum = 0;

	while (l > 1) {
		sum += *p++ << 8;
		sum += *p++;
		l -= 2;
	}
	if (l == 1)
		sum += *p << 8;
	sum = ((sum >> 16) & 0xffff) + (sum & 0xffff);
	sum += sum >> 16;
	return (0xffff & ~sum);
}

int
udp_cksum(u_char *pkt, u_int plen)
{
	int sum;
	u_char saved[4];

	memcpy(saved, pkt + 8, 4);
	pkt[8] = 0;
	memcpy(pkt + 10, pkt + UDPLEN, 2);
	sum = in_cksum(pkt + 8, plen - 8);
	memcpy(pkt + 8, saved, 4);
	return sum;
}

/* receive request */

u_char packet[IPHDRLEN + UDPHDRLEN + PCPD_BUF_SIZE];
struct sockaddr_in aftr_in;

size_t
recvrequest(const pcpd_t *pcpd, const pcpctx_t *ctx)
{
	fd_set set;
	struct timeval tv;
	int maxfd, cc, plen;
	char *got;

	UNUSED(pcpd);
	for(;;) {
		expire();
		while (pendingcnt > 0) {
			int len;

			checkdeltunnel(pendingbuf);
			len = strlen(pendingbuf) + 1;
			memmove(pendingbuf, pendingbuf + len,
				pendingcnt - len);
			pendingcnt -= len;
		}
		pendingcnt = 0;
		servedb();

		FD_ZERO(&set);
		FD_SET(aftr, &set);
		maxfd = aftr;
		if (sock6 != -1) {
			FD_SET(sock6, &set);
			if (sock6 > maxfd)
				maxfd = sock6;
		}
		if (sock4 != -1) {
			FD_SET(sock4, &set);
			if (sock4 > maxfd)
				maxfd = sock4;
		}
		tv.tv_sec = 1;
		tv.tv_usec = 0;
		cc = select(maxfd + 1, &set, NULL, NULL, &tv);
		if ((cc < 0) && (errno != EINTR)) {
			syslog(LOG_CRIT, "select fatal error: %m");
			exit(0);
		}
		if (cc <= 0)
			continue;

		if (FD_ISSET(aftr, &set)) {
			got = getnextline(0);
			if (got == NULL) {
				syslog(LOG_CRIT, 
				       "AFTR daemon connection closed");
				exit(0);
			}
			checkdeltunnel(got);
		}
		if ((sock6 != -1) && FD_ISSET(sock6, &set)) {
			struct sockaddr_in6 *c6;
			size_t length = PCPD_BUF_SIZE;
			socklen_t sl;

			c6 = (struct sockaddr_in6 *) ctx->client;
			sl = sizeof(*c6);
			plen = recvfrom(sock6, ctx->request, length, 0,
					ctx->client, &sl);
			if (plen < 0) {
				syslog(LOG_CRIT, "PCP socket error: %m");
				continue;
			}
			memcpy(ctx->sid, &c6->sin6_addr, 16);
			return (size_t) plen;
		}
		if ((sock4 != -1) && FD_ISSET(sock4, &set)) {
			struct sockaddr_in *c4;
			struct msghdr msg;
			struct iovec iov[2];
			size_t ulen;

			c4 = (struct sockaddr_in *) ctx->client;
			memset(packet, 0, sizeof(packet));
			memset(&msg, 0, sizeof(msg));
			iov[0].iov_base = ctx->sid;
			iov[0].iov_len = 16;
			iov[1].iov_base = packet;
			iov[1].iov_len = sizeof(packet);
			msg.msg_iov = iov;
			msg.msg_iovlen = 2;
			msg.msg_name = &aftr_in;
			msg.msg_namelen = sizeof(aftr_in);
			plen = recvmsg(sock4, &msg, 0);
			if (plen < 0) {
				syslog(LOG_CRIT, "encap socket error: %m");
				continue;
			}
			ulen = packet[UDPLEN] << 8;
			ulen |= packet[UDPLEN + 1];
			if ((ulen < UDPHDRLEN) ||
			    ((int) ulen + IPHDRLEN + 16 > plen)) {
				syslog(LOG_CRIT, "underflow recvmsg()");
				continue;
			}
			c4->sin_family = AF_INET;
			memcpy(&c4->sin_addr, packet + IPSRC, 4);
			memcpy(&c4->sin_port, packet + UDPSPORT, 2);
			ulen -= UDPHDRLEN;
			if (ulen > PCPD_BUF_SIZE)
				ulen = PCPD_BUF_SIZE;
			memcpy(ctx->request,
			       packet + IPHDRLEN + UDPHDRLEN,
			       ulen);
			return ulen;
		}
	}
}

/* send response */

void
sendresponse(const pcpd_t *pcpd, const pcpctx_t *ctx, const size_t length)
{
	UNUSED(pcpd);
	debug("send response");

	/* plain mode */
	if (ctx->client->sa_family == AF_INET6) {
		(void) sendto(sock6, ctx->response, length, 0,
			      ctx->client, sizeof(struct sockaddr_in6));
		return;
	}
	/* encapsulation mode */
	if (ctx->client->sa_family == AF_INET) {
		struct msghdr msg;
		struct iovec iov[2];
		u_char tmp[4];
		size_t len = length;
		int cksum;

		memset(&msg, 0, sizeof(msg));
		memcpy(packet + IPHDRLEN + UDPHDRLEN, ctx->response, len);
		iov[0].iov_base = (void *) ctx->sid;
		iov[0].iov_len = 16;
		iov[1].iov_base = packet;
		iov[1].iov_len = IPHDRLEN + UDPHDRLEN + len;
		msg.msg_iov = iov;
		msg.msg_iovlen = 2;
		msg.msg_name = &aftr_in;
		msg.msg_namelen = sizeof(aftr_in);

		/* adjust v4 headers, send back to aftr */
		memcpy(tmp, packet + IPSRC, 4);
		memcpy(packet + IPSRC, packet + IPDST, 4);
		memcpy(packet + IPDST, tmp, 4);
		memcpy(tmp, packet + UDPSPORT, 2);
		memcpy(packet + UDPSPORT, packet + UDPDPORT, 2);
		memcpy(packet + UDPDPORT, tmp, 2);
		len += UDPHDRLEN;
		packet[UDPLEN] = len >> 8;
		packet[UDPLEN + 1] = len & 0xff;
		len += IPHDRLEN;
		packet[IPLEN] = len >> 8;
		packet[IPLEN + 1] = len & 0xff;
		memset(packet + IPID, 0, 2);
		packet[IPTTL] = 64;
		memset(packet + IPCKSUM, 0, 2);
		cksum = in_cksum(packet, IPHDRLEN);
		packet[IPCKSUM] = cksum >> 8;
		packet[IPCKSUM + 1] = cksum & 0xff;
		memset(packet + UDPCKSUM, 0, 2);
		cksum = udp_cksum(packet, len);
		packet[UDPCKSUM] = cksum >> 8;
		packet[UDPCKSUM + 1] = cksum & 0xff;

		(void) sendmsg(sock4, &msg, 0);
	}
}
			
/* initcom */

void
initcom(int encap, char *address)
{
	int on;
	struct sockaddr_in sa, se;
	struct sockaddr_in6 sa6;
	extern char *optarg;
	extern int optind;

#ifndef ENCAP_TO_PCPD
#define ENCAP_TO_PCPD 10151
#endif

	/* connect to aftr control interface */
	aftr = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
	if (aftr < 0) {
		syslog(LOG_ERR, "socket(inet): %m");
		exit(1);
	}
	memset(&sa, 0, sizeof(sa));
	sa.sin_family = AF_INET;
#ifndef __linux__
	sa.sin_len = sizeof(sa);
#endif
	(void) inet_pton(AF_INET, "127.0.0.1", &sa.sin_addr);
	sa.sin_port = htons(1015);
	if (connect(aftr, (struct sockaddr *) &sa, sizeof(sa)) < 0) {
		syslog(LOG_ERR, "connect(inet): %m");
		exit(1);
	}
	ask("session log off");
	ask("session notify on");
	ask("session config on");
	ask("session name pcpd");
	asklistpcp();

	if (ISC_SLIST_EMPTY(&natsrcs)) {
		fprintf(stderr, "list pcp is empty\n");
		exit(1);
	}

	/* create the listening sockets */
	if (encap == 0) {
		sock6 = socket(PF_INET6, SOCK_DGRAM, IPPROTO_UDP);
		if (sock6 < 0) {
			syslog(LOG_ERR, "socket(inet6): %m");
			exit(1);
		}
		on = 1;
		if (setsockopt(sock6, IPPROTO_IPV6, IPV6_V6ONLY,
			       &on, sizeof(on)) < 0) {
			syslog(LOG_ERR, "setsockopt(V6ONLY): %m");
			exit(1);
		}
		memset(&sa6, 0, sizeof(sa6));
		sa6.sin6_family = AF_INET6;
#ifndef __linux__
		sa6.sin6_len = sizeof(sa6);
#endif
		if ((address != NULL) &&
		    (inet_pton(AF_INET6, address, &sa6.sin6_addr) <= 0)) {
			fprintf(stderr, "bad address \"%s\"\n", address);
			exit(1);
		}
		sa6.sin6_port = htons(PCP_PORT);
		if (bind(sock6, (struct sockaddr *) &sa6, sizeof(sa6)) < 0) {
			syslog(LOG_ERR, "bind(inet6): %m");
			exit(1);
		}
	} else {
		sock4 = socket(PF_INET, SOCK_DGRAM, 0);
		if (sock4 < 0) {
			syslog(LOG_ERR, "socket(encap): %m");
			exit(1);
		}
		memset(&se, 0, sizeof(se));
		se.sin_family = AF_INET;
#ifndef __linux__
		se.sin_len = sizeof(se);
#endif
		(void) inet_pton(AF_INET, "127.0.0.1", &se.sin_addr);
		se.sin_port = htons(ENCAP_TO_PCPD);
		if (bind(sock4, (struct sockaddr *) &se, sizeof(se)) < 0) {
			syslog(LOG_ERR, "bind(encap): %m");
			exit(1);
		}
	}
}
