/*
 * Copyright (C) 2011  Internet Systems Consortium, Inc. ("ISC")
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND ISC DISCLAIMS ALL WARRANTIES WITH
 * REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS.  IN NO EVENT SHALL ISC BE LIABLE FOR ANY SPECIAL, DIRECT,
 * INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
 * OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */

/* $Id: dc6448595913bf4728b90532b49fbfe9c1f1c866 $ */

/* PCP daemon: subscriber and explicit dynamic mapping database module
 *
 * Francis_Dupont@isc.org, August 2011
 */

#include "internal.h"

#include <db.h>

/* globals */

DB_ENV *dbenv;
DB *dbtun;
DB *dbmap;

/* db_errcall helper */

void
db_errcall(const DB_ENV *_dbenv, const char *_pfx, const char *msg)
{
	_dbenv = _dbenv;
	_pfx = _pfx;
	syslog(LOG_ERR, "db %s", msg);
}

/* atexist helpers */

void
dbenv_close(void)
{
	(void) dbenv->close(dbenv, 0);
}

void
dbtun_close(void)
{
	(void) dbtun->close(dbtun, 0);
}

void
dbmap_close(void)
{
	(void) dbmap->close(dbmap, 0);
}

/* db tunnels: key is IPv6, data are natted IPv4 | start */

/* db tunnel add */

void
dbaddtunnel(struct client *client)
{
	DBT key, data;
	char kdata[8];

	memset(&key, 0, sizeof(key));
	memset(&data, 0, sizeof(data));
	key.size = 16;
	key.data = &client->ipv6;
	data.size = 8;
	data.data = kdata;
	memcpy(kdata, &client->natsrc->addr, 4);
	memcpy(kdata + 4, &client->start, 4);
	if (dbtun->put(dbtun, NULL, &key, &data, 0) != 0)
		debug("db->put(tun) failed");
}

/* db tunnel del */

void
dbdeltunnel(struct client *client)
{
	DBT key;

	memset(&key, 0, sizeof(key));
	key.size = 16;
	key.data = &client->ipv6;
	if (dbtun->del(dbtun, NULL, &key, 0) != 0)
		debug("db->del(tun) failed");
}

/* db maps:
 *   key is natted IPv4 | nport | proto,
 *   data are:
 *      peer IPv6 address [0:15]
 *    | source IPv4 address [16:19]
 *    | filter IPv4 address [20:23]
 *    | source port [24:25]
 *    | filter port [26:27]
 *    | expire [28:31]
 *    | filter prefix length [32]
 *    | filter state [33]
 *    | down flag [34]
 */

/* db map add */

void
dbaddmap(struct mapping *mapping)
{
	DBT key, data;
	char kbuf[7];
	char kdata[35];

	memset(&key, 0, sizeof(key));
	memset(&data, 0, sizeof(data));
	key.size = 7;
	key.data = kbuf;
	data.size = 35;
	data.data = kdata;
	memcpy(kbuf, &mapping->client->natsrc->addr, 4);
	memcpy(kbuf + 4, &mapping->port->port, 2);
	memcpy(kbuf + 6, &mapping->proto, 1);
	memcpy(kdata, &mapping->client->ipv6, 16);
	memcpy(kdata + 16, &mapping->src, 4);
	memcpy(kdata + 20, &mapping->faddr, 4);
	memcpy(kdata + 24, &mapping->sport, 2);
	memcpy(kdata + 26, &mapping->fport, 2);
	memcpy(kdata + 28, &mapping->expire, 4);
	memcpy(kdata + 32, &mapping->fplen, 1);
	memcpy(kdata + 33, &mapping->fstate, 1);
	memcpy(kdata + 34, &mapping->down, 1);
	if (dbmap->put(dbmap, NULL, &key, &data, 0) != 0)
		debug("db->put(map) failed");
}

/* db map del */

void
dbdelmap(struct mapping *mapping)
{
	DBT key;
	char kbuf[7];

	memset(&key, 0, sizeof(key));
	key.size = 7;
	key.data = kbuf;
	memcpy(kbuf, &mapping->client->natsrc->addr, 4);
	memcpy(kbuf + 4, &mapping->port->port, 2);
	memcpy(kbuf + 6, &mapping->proto, 1);
	if (dbmap->del(dbmap, NULL, &key, 0) != 0)
		debug("db->del(map) failed");
}

/* reload state databases */

void
reload(int dbclean)
{
	DBC *dbc;
	DBT key, data;
	char host6[INET6_ADDRSTRLEN];
	char host4[INET_ADDRSTRLEN];
	struct client *client;
	struct natsrc *natsrc;
	struct mapping *mapping;
	uint32_t addr;
	int ret;

	dbc = NULL;
	memset(&key, 0, sizeof(key));
	memset(&data, 0, sizeof(data));

	/* tunnels */
	if (dbtun->cursor(dbtun, NULL, &dbc, 0) != 0) {
		syslog(LOG_CRIT, "db->cursor(tun)");
		exit(1);
	}
	for (;;) {
		struct bucket *bucket;
		uint32_t start;
		u_short hash;

		ret = dbc->get(dbc, &key, &data, DB_NEXT);
		if (ret != 0) {
			if (ret != DB_NOTFOUND)
				syslog(LOG_ERR, "dbc->get(tun)");
			(void) dbc->close(dbc);
			break;
		}
		if ((key.size != 16) || (data.size != 8)) {
			syslog(LOG_ERR, "db(tunnel): bad");
			if (dbclean)
				(void) dbtun->del(dbtun, NULL, &key, 0);
			continue;
		}
		memcpy(&addr, (caddr_t) data.data, 4);
		memcpy(&start, (caddr_t) data.data + 4, 4);
		ISC_SLIST_FOREACH(natsrc, &natsrcs, chain)
			if (natsrc->addr == addr)
				break;
		if (natsrc == NULL) {
			(void) inet_ntop(AF_INET, &addr,
					 host4, INET_ADDRSTRLEN);
			debug("db(tunnel): can't find %s", host4);
			if (dbclean)
				(void) dbtun->del(dbtun, NULL, &key, 0);
			continue;
		}
		client = new_client((struct in6_addr *) key.data, natsrc);
		if (client == NULL) {
			debug("new_client failed");
			continue;
		}
		if (askaddtunnel(client) < PCPD_OK) {
			free(client);
			continue;
		}
		hash = jhash((struct in6_addr *) key.data) % HASHSIZE;
		bucket = &clients[hash];
		ISC_LIST_INSERT(bucket, client, chain);
		client->start = start;
	}
	dbc = NULL;
	
	/* static nats */
	if (dbmap->cursor(dbmap, NULL, &dbc, 0) != 0) {
		syslog(LOG_CRIT, "db->cursor(map)");
		exit(1);
	}
	for (;;) {
		uint32_t src, faddr, expire;
		uint16_t sport, nport, fport;
		uint8_t proto, fplen, fstate, down;
		uint32_t lifetime;
		int locked, result;
		time_t now;
		
		ret = dbc->get(dbc, &key, &data, DB_NEXT);
		if (ret != 0) {
			if (ret != DB_NOTFOUND)
				syslog(LOG_ERR, "dbc->get(map)");
			(void) dbc->close(dbc);
			break;
		}
		if ((key.size != 7) || (data.size != 35)) {
			syslog(LOG_ERR, "db(map): bad");
			if (dbclean)
				(void) dbmap->del(dbmap, NULL, &key, 0);
			continue;
		}
		memcpy(&addr, key.data, 4);
		memcpy(&nport, (caddr_t) key.data + 4, 2);
		memcpy(&proto, (caddr_t) key.data + 6, 1);
		memcpy(&src, (caddr_t) data.data + 16, 4);
		memcpy(&faddr, (caddr_t) data.data + 20, 4);
		memcpy(&sport, (caddr_t) data.data + 24, 2);
		memcpy(&fport, (caddr_t) data.data + 26, 2);
		memcpy(&expire, (caddr_t) data.data + 28, 4);
		memcpy(&fplen, (caddr_t) data.data + 32, 1);
		memcpy(&fstate, (caddr_t) data.data + 33, 1);
		memcpy(&down, (caddr_t) data.data + 34, 1);
		if ((proto != IPPROTO_TCP) && (proto != IPPROTO_UDP)) {
			debug("db(map): bad proto %d", (int) proto);
			if (dbclean)
				(void) dbmap->del(dbmap, NULL, &key, 0);
			continue;
		}
		now = time(NULL);
		if ((expire != 0) && (now >= (time_t) expire)) {
			if (dbclean)
				(void) dbmap->del(dbmap, NULL, &key, 0);
			continue;
		}
		client = getclient((struct in6_addr *) data.data, 0);
		if (client == NULL) {
			(void) inet_ntop(AF_INET6, data.data,
					 host6, INET6_ADDRSTRLEN);
			debug("db(map): can't find %s", host6);
			if (dbclean)
				(void) dbmap->del(dbmap, NULL, &key, 0);
			continue;
		}
		natsrc = client->natsrc;
		if (natsrc->addr != addr) {
			(void) inet_ntop(AF_INET6, data.data,
					 host6, INET6_ADDRSTRLEN);
			debug("db(map): address mismatch on %s", host6);
			if (dbclean)
				(void) dbmap->del(dbmap, NULL, &key, 0);
			continue;
		}
		if (down) {
			mapping = setmappinghold(client, src, sport,
						 nport, expire, proto);
			if (mapping != NULL) {
				mapping->faddr = faddr;
				mapping->fport = fport;
				mapping->fplen = fplen;
				mapping->fstate = fstate;
				continue;
			}
			debug("db(map): setmappinghold failed");
			if (dbclean)
				(void) dbmap->del(dbmap, NULL, &key, 0);
			continue;
		}
		lifetime = expire - now;
		locked = 1;
		mapping = setmapping(client, src, sport, &nport,
				     locked, &lifetime, proto, &result);
		if (mapping != NULL) {
			mapping->faddr = faddr;
			mapping->fport = fport;
			mapping->fplen = fplen;
			mapping->fstate = fstate;
			continue;
		}
		debug("db(map): setmapping failed %d", result);
		if (dbclean)
			(void) dbmap->del(dbmap, NULL, &key, 0);
	}
}

/* dump databases */

int
dump_client(FILE *ft, struct client *client)
{
	uint8_t *p;
	int i;

	/* key */
	fputc(' ', ft);
	p = (uint8_t *) &client->ipv6;
	for (i = 0; i < 16; i++)
		fprintf(ft, "%02x", (u_int) *p++);
	fputc('\n', ft);

	/* data */
	fputc(' ', ft);
	p = (uint8_t *) &client->natsrc->addr;
	for (i = 0; i < 4; i++)
		fprintf(ft, "%02x", (u_int) *p++);
	p = (uint8_t *) &client->start;
	for (i = 0; i < 4; i++)
		fprintf(ft, "%02x", (u_int) *p++);
	fputc('\n', ft);
	return -ferror(ft);
}

int
dump_mapping(FILE *fn, struct mapping *mapping)
{
	uint8_t *p;
	int i;

	/* key */
	fputc(' ', fn);
	p = (uint8_t *) &mapping->client->natsrc->addr;
	for (i = 0; i < 4; i++)
		fprintf(fn, "%02x", (u_int) *p++);
	p = (uint8_t *) &mapping->port->port;
	fprintf(fn, "%02x", (u_int) *p++);
	fprintf(fn, "%02x", (u_int) *p);
	fprintf(fn, "%02x\n", (u_int) mapping->proto);

	/* data */
	fputc(' ', fn);
	p = (uint8_t *) &mapping->client->ipv6;
	for (i = 0; i < 16; i++)
		fprintf(fn, "%02x", (u_int) *p++);
	p = (uint8_t *) &mapping->src;
	for (i = 0; i < 4; i++)
		fprintf(fn, "%02x", (u_int) *p++);
	p = (uint8_t *) &mapping->faddr;
	for (i = 0; i < 4; i++)
		fprintf(fn, "%02x", (u_int) *p++);
	p = (uint8_t *) &mapping->sport;
	fprintf(fn, "%02x", (u_int) *p++);
	fprintf(fn, "%02x", (u_int) *p);
	p = (uint8_t *) &mapping->fport;
        fprintf(fn, "%02x", (u_int) *p++);
        fprintf(fn, "%02x", (u_int) *p);
	p = (uint8_t *) &mapping->expire;
	for (i = 0; i < 4; i++)
		fprintf(fn, "%02x", (u_int) *p++);
	fprintf(fn, "%02x", (u_int) mapping->fplen);
	fprintf(fn, "%02x", (u_int) mapping->fstate);
	fprintf(fn, "%02x\n", (u_int) mapping->down);
	return -ferror(fn);
}

void
dump(void)
{
	struct client *client;
	struct bucket *bucket;
	struct mapping *mapping;
	FILE *ft, *fn;
	int i, ret;

	ft = fopen("tunnels.dump", "w+");
	if (ft == NULL) {
		debug("dump(tunnel): can't open file: %s", strerror(errno));
		return;
	}
	fn = fopen("maps.dump", "w+");
	if (fn == NULL) {
		debug("dump(map): can't open file: %s", strerror(errno));
		(void) fclose(ft);
		return;
	}

	ret = 0;
	for (i = 0; i < HASHSIZE; i++) {
		bucket = &clients[i];
		ISC_LIST_FOREACH(client, bucket, chain) {
			ret = dump_client(ft, client);
			if (ret < 0)
				break;
			ISC_LIST_FOREACH(mapping, &client->mappings, chain) {
				ret = dump_mapping(fn, mapping);
				if (ret < 0)
					break;
			}
			if (ret < 0)
				break;
		}
		if (ret < 0)
			break;
	}
	(void) fclose(ft);
	(void) fclose(fn);
}

/* signal handler (just set a flag) */

int todump;

void
handler(int dummy)
{
	UNUSED(dummy);
	todump = 1;
}

/* servedb */

void
servedb(void)
{
	if (todump)
		dump();
	todump = 0;
}

/* initdb */

void
initdb(int dbclean)
{
	int ret;

	ret = db_env_create(&dbenv, 0);
	if (ret != 0) {
		syslog(LOG_ERR, "db_env_create: %s", db_strerror(ret));
		exit(1);
	}
	dbenv->set_errcall(dbenv, db_errcall);
#define DBENV_FLAGS	(DB_CREATE | DB_INIT_MPOOL | DB_INIT_TXN)
	ret = dbenv->open(dbenv, NULL, DBENV_FLAGS, 0);
	if (ret != 0) {
		syslog(LOG_ERR, "dbenv->open: %s", db_strerror(ret));
		exit(1);
	}
	(void) atexit(dbenv_close);
	ret = db_create(&dbtun, dbenv, 0);
	if (ret != 0) {
		syslog(LOG_ERR, "db_create(tun): %s", db_strerror(ret));
		exit(1);
	}
	ret = dbtun->open(dbtun, NULL, "tunnels.db", NULL,
			  DB_HASH, DB_CREATE | DB_AUTO_COMMIT, 0);
	if (ret != 0) {
		syslog(LOG_ERR, "db->open(tunnels.db): %s", db_strerror(ret));
		exit(1);
	}
	(void) atexit(dbtun_close);
	ret = db_create(&dbmap, dbenv, 0);
	if (ret != 0) {
		syslog(LOG_ERR, "db_create(map): %s", db_strerror(ret));
		exit(1);
	}
	ret = dbmap->open(dbmap, NULL, "maps.db", NULL,
			  DB_HASH, DB_CREATE | DB_AUTO_COMMIT, 0);
	if (ret != 0) {
		syslog(LOG_ERR, "db->open(maps.db): %s", db_strerror(ret));
		exit(1);
	}
	(void) atexit(dbmap_close);
	reload(dbclean);
}
