/*
 * Copyright (C) 2011  Internet Systems Consortium, Inc. ("ISC")
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND ISC DISCLAIMS ALL WARRANTIES WITH
 * REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS.  IN NO EVENT SHALL ISC BE LIABLE FOR ANY SPECIAL, DIRECT,
 * INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
 * OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */

/* $Id: c3d0b07967206c45969649a4aca64ead7c3441a6 $ */

/* PCP daemon: core include
 *
 * Francis_Dupont@isc.org, August 2011
 */

#define _GNU_SOURCE

#include "pcpd.h"

#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/select.h>
#include <sys/socket.h>
#include <sys/time.h>

#include <arpa/inet.h>
#include <netinet/in.h>
#include <sys/un.h>

#include <ctype.h>
#include <errno.h>
#include <netdb.h>
#include <signal.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include <syslog.h>
#include <time.h>
#include <unistd.h>

/* Local macros */

#define UNUSED(x)	(x) = (x)

#define ISC_SLIST_HEAD(name, type)					\
	struct name {							\
		struct type *slh_first;					\
	}

#define ISC_SLIST_INIT(head)						\
	do {								\
		(head)->slh_first = NULL;				\
	} while (0)

#define ISC_SLIST_ENTRY(type)						\
	struct {							\
		struct type *sle_next;					\
	}

#define ISC_SLIST_EMPTY(head)						\
	((head)->slh_first == NULL)

#define ISC_SLIST_FIRST(head)						\
	((head)->slh_first)

#define ISC_SLIST_NEXT(elm, field)					\
	((elm)->field.sle_next)

#define ISC_SLIST_INSERT_HEAD(head, elm, field)				\
	do {								\
		(elm)->field.sle_next = (head)->slh_first;		\
		(head)->slh_first = (elm);				\
	} while (0)

#define ISC_SLIST_INSERT_AFTER(elm0, elm, field)			\
	do {								\
		(elm)->field.sle_next = (elm0)->field.sle_next;		\
		(elm0)->field.sle_next = (elm);				\
	} while (0)

#define ISC_SLIST_REMOVE_HEAD(head, field)				\
	do {								\
		(head)->slh_first = (head)->slh_first->field.sle_next;	\
	} while (0)

#define ISC_SLIST_REMOVE(head, elm, type, field)			\
	do {								\
		if ((head)->slh_first == (elm)) {			\
			ISC_SLIST_REMOVE_HEAD(head, field);		\
		} else {						\
			struct type *cur = (head)->slh_first;		\
			while (cur->field.sle_next != (elm))		\
				cur = cur->field.sle_next;		\
			cur->field.sle_next = (elm)->field.sle_next;	\
		}							\
	} while (0)

#define ISC_SLIST_FOREACH(var, head, field)				\
	for ((var) = (head)->slh_first;					\
	     (var) != NULL;						\
	     (var) = (var)->field.sle_next)

#define ISC_SLIST_FOREACH_SAFE(var, head, field, tvar)			\
	for ((var) = (head)->slh_first;					\
	     ((var) != NULL) &&	(((tvar) = (var)->field.sle_next), 1);	\
	     (var) = (tvar))

#define ISC_LIST_HEAD(name, type)					\
	struct name {							\
		struct type *lh_first;					\
	}

#define ISC_LIST_INIT(head)						\
	do {								\
		(head)->lh_first = NULL;				\
	} while (0)

#define ISC_LIST_ENTRY(type)						\
	struct {							\
		struct type *le_next;					\
		struct type **le_prev;					\
	}

#define ISC_LIST_EMPTY(head)						\
	((head)->lh_first == NULL)

#define ISC_LIST_FIRST(head)						\
	((head)->lh_first)

#define ISC_LIST_NEXT(elm, field)					\
	((elm)->field.le_next)

#define ISC_LIST_PREV(elm, field)					\
	((elm)->field.le_prev)

#define ISC_LIST_INSERT(head, elm, field)				\
	do {								\
		(elm)->field.le_next = (head)->lh_first;		\
		if ((head)->lh_first != NULL)				\
			(head)->lh_first->field.le_prev =		\
				&(elm)->field.le_next;			\
		(head)->lh_first = (elm);				\
		(elm)->field.le_prev = &(head)->lh_first;		\
	} while (0)

#define ISC_LIST_REMOVE(elm, field)					\
	do {								\
		if ((elm)->field.le_next != NULL)			\
			(elm)->field.le_next->field.le_prev =		\
				(elm)->field.le_prev;			\
		*(elm)->field.le_prev = (elm)->field.le_next;		\
		(elm)->field.le_next = NULL;				\
		(elm)->field.le_prev = NULL;				\
	} while (0)

#define ISC_LIST_FOREACH(var, head, field)				\
	for ((var) = (head)->lh_first;					\
	     (var) != NULL;						\
	     (var) = (var)->field.le_next)

#define ISC_LIST_FOREACH_SAFE(var, head, field, tvar)			\
	for ((var) = (head)->lh_first;					\
	     ((var) != NULL) &&	(((tvar) = (var)->field.le_next), 1);	\
	     (var) = (tvar))

#define PROTO2INDEX(x)	((x) == IPPROTO_TCP ? 0 : 1)

#define INDEX2PROTO(x)	((x) == 0 ? IPPROTO_TCP : IPPROTO_UDP)

/* defines */

#ifndef MAXPCP
#define MAXPCP		64
#endif
#ifndef LONGERRHT
#define LONGERRHT	1800
#endif
#ifndef SHORTERRHT
#define SHORTERRHT	30
#endif

/* XXX should be made configurable */
#define HOLD		120
#define HASHSIZE	128

/* port mapping */

struct mapping {
	ISC_LIST_ENTRY(mapping) chain;
	struct client *client;
	struct port *port;
	uint32_t src, faddr;
	u_int heap_index;
	time_t expire;
	uint16_t sport, fport;
	uint8_t proto;
	uint8_t fplen;
	uint8_t fstate;
	uint8_t down;
};

/* client */

struct client {
	ISC_LIST_ENTRY(client) chain;
	struct in6_addr ipv6;
	struct natsrc *natsrc;
	time_t start;
	u_int mapcnt;
	ISC_LIST_HEAD(, mapping) mappings;
};

ISC_LIST_HEAD(bucket, client) clients[HASHSIZE];

/* per natted address object, clients per IPv6, IPv6 per port */

struct port {
	struct client *client;
	uint16_t port;
};

struct natsrc {
	ISC_SLIST_ENTRY(natsrc) chain;
	uint32_t addr;
	uint16_t minport[2], maxport[2];
	struct port *ports[2];
};

ISC_SLIST_HEAD(, natsrc) natsrcs;

extern void ask(char *cmd);
extern void asklistpcp(void);
extern uint32_t asktunnel(struct in6_addr *addr);
extern int askaddtunnel(struct client *client);
extern void askdelmap(struct mapping *mapping);
extern int askaddmap(struct mapping *mapping, int *staticport);
extern int askdynamic(struct client *client, uint8_t protocol,
		      uint32_t src, uint32_t nsrc, uint32_t dst,
		      uint16_t sport, uint16_t *nport, uint16_t dport,
		      uint32_t *lifetime);
extern void askdelfilter(struct client *client, uint32_t src,
			 uint16_t port, uint8_t protocol);
extern int askreplacefilter(struct client *client,
			    uint32_t src, uint16_t sport, uint8_t protocol,
			    uint8_t prefixlen, uint32_t dst, uint16_t dport);
extern int askaddfilter(struct client *client,
			uint32_t src, uint16_t sport, uint8_t protocol,
			uint8_t prefixlen, uint32_t dst, uint16_t dport);
extern void checkdeltunnel(const char *got);
extern void initcom(int encap, char *address);
extern size_t recvrequest(const pcpd_t *pcpd, const pcpctx_t *ctx);
extern void sendresponse(const pcpd_t *pcpd, const pcpctx_t *ctx,
			 const size_t length);

extern void dbaddtunnel(struct client *client);
extern void dbdeltunnel(struct client *client);
extern void dbaddmap(struct mapping *mapping);
extern void dbdelmap(struct mapping *mapping);
extern void initdb(int dbclean);
extern void servedb(void);

extern void debug(const char *msg, ...);
extern u_short jhash(struct in6_addr *ipv6);
extern struct client *new_client(struct in6_addr *ipv6, struct natsrc *natsrc);
extern struct client *getclient(struct in6_addr *ipv6, int prompt);
extern void new_natsrc(uint32_t addr, uint16_t *minport, uint16_t *maxport);
extern void delclient(struct client *client);
extern struct mapping *setmapping(struct client *client, uint32_t src,
				  uint16_t sport, uint16_t *nport,
				  int locked, u_int *lifetime,
				  uint8_t proto, int *result);
extern struct mapping *setmappinghold(struct client *client,
				      uint32_t src, uint16_t sport,
				      uint16_t nport, u_int expire,
				      uint8_t proto);
extern void expire(void);
