/*
 * Copyright (C) 2011-2012  Internet Systems Consortium, Inc. ("ISC")
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND ISC DISCLAIMS ALL WARRANTIES WITH
 * REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS.  IN NO EVENT SHALL ISC BE LIABLE FOR ANY SPECIAL, DIRECT,
 * INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
 * OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */

/* $Id: 2273f7778daa82987c04bba987a3bdd5fae4a124 $ */

/* PCP daemon: core module using library (vs. CGN) calls
 *
 * Francis_Dupont@isc.org, August 2011
 *
 * Usage: [-c] [-d] [-e | address]
 */

#include "internal.h"

/* XXX should be made configurable */
#define HOLD		120
#define HASHSIZE	128

/* heap */

u_int map_heap_size;
u_int map_heap_last;
struct mapping **map_heap;

/* debug helper */

void
debug(const char *msg, ...)
{
	va_list va;

	va_start(va, msg);

	vsyslog(LOG_DEBUG, msg, va);
	va_end(va);
}

/* Jenkins hash */

#define __jhash_mix(a, b, c) { \
	a -= b; a -= c; a ^= (c >> 13);	\
	b -= c; b -= a; b ^= (a << 8);	\
	c -= a; c -= b; c ^= (b >> 13);	\
	a -= b; a -= c; a ^= (c >> 12);	\
	b -= c; b -= a; b ^= (a << 16);	\
	c -= a; c -= b; c ^= (b >> 5);	\
	a -= b; a -= c; a ^= (c >> 3);	\
	b -= c; b -= a; b ^= (a << 10);	\
	c -= a; c -= b; c ^= (b >> 15);	}

#define JHASH_GOLDEN_RATIO	0x9e3779b9

u_short
jhash(struct in6_addr *ipv6)
{
	char *addr = (char *) ipv6;
	uint32_t a, b, c, k;

	memcpy(&a, addr, 4);
	a += JHASH_GOLDEN_RATIO;
	memcpy(&b, addr + 4, 4);
	b += JHASH_GOLDEN_RATIO;
	memcpy(&c, addr + 8, 4);

	__jhash_mix(a, b, c);

	memcpy(&k, addr + 12, 4);
	a += k;

	__jhash_mix(a, b, c);

	c = (c >> 16) + (c & 0xffff);
	c &= 0xffff;
	return (u_short) c;
}

/* mapping heap */

/* resize */

int
map_heap_resize(void)
{
	struct mapping **mh;
	u_int ms = map_heap_size + 1024;

	mh = (struct mapping **) malloc(ms * sizeof(*mh));
	if (mh == NULL) {
		syslog(LOG_ERR, "malloc(heap): %m");
		return 0;
	}
	if (map_heap != NULL) {
		memcpy(mh, map_heap, map_heap_size * sizeof(*mh));
		free(map_heap);
	}
	map_heap = mh;
	map_heap_size = ms;
	return 1;
}

/* parent is i/2, left child is i*2 and right child is i*2+1 */

void
map_heap_float_up(u_int i, struct mapping *m)
{
	u_int p;

	for (p = i >> 1;
	     (i > 1) && (m->expire < map_heap[p]->expire);
	     i = p, p = i >> 1) {
		map_heap[i] = map_heap[p];
		map_heap[i]->heap_index = i;
	}
	map_heap[i] = m;
	map_heap[i]->heap_index = i;

	if ((i > 1) && (map_heap[i]->expire < map_heap[i >> 1]->expire))
		syslog(LOG_CRIT, "map_heap_float_up[%u]", i);
}

void
map_heap_sink_down(u_int i, struct mapping *m)
{
	u_int j, size, half;

	size = map_heap_last;
	half = size / 2;
	while (i <= half) {
		/* Find the smallest of the (at most) two children */
		j = i << 1;
		if ((j < size) &&
		    (map_heap[j + 1]->expire < map_heap[j]->expire))
			j++;
		if (m->expire < map_heap[j]->expire)
			break;
		map_heap[i] = map_heap[j];
		map_heap[i]->heap_index = i;
		i = j;
	}
	map_heap[i] = m;
	map_heap[i]->heap_index = i;

	if ((i > 1) && (map_heap[i]->expire < map_heap[i >> 1]->expire))
		syslog(LOG_CRIT, "map_heap_sink_down[%u]", i);
}

int
map_heap_insert(struct mapping *m)
{
	u_int new_last;

	new_last = map_heap_last + 1;
	if ((new_last >= map_heap_size) && !map_heap_resize())
		return 0;
	map_heap_last = new_last;
	map_heap_float_up(new_last, m);
	return 1;
}

void
map_heap_delete(u_int i)
{
	struct mapping *m;
	int less;

	if ((i < 1) || (i > map_heap_last)) {
		syslog(LOG_CRIT, "map_heap_delete[%u]", i);
		return;
	}

	if (i == map_heap_last) {
		map_heap[map_heap_last] = NULL;
		map_heap_last -= 1;
		return;
	}
	m = map_heap[map_heap_last];
	map_heap[map_heap_last] = NULL;
	map_heap_last -= 1;

	less = (m->expire < map_heap[i]->expire);
	map_heap[i] = m;
	if (less)
		map_heap_float_up(i, map_heap[i]);
	else
		map_heap_sink_down(i, map_heap[i]);
}

void
map_heap_increased(u_int i)
{
	if ((i < 1) || (i > map_heap_last)) {
		syslog(LOG_CRIT, "map_heap_increased[%u]", i);
		return;
	}

	map_heap_float_up(i, map_heap[i]);
}

void
map_heap_decreased(u_int i)
{
	if ((i < 1) || (i > map_heap_last)) {
		syslog(LOG_CRIT, "map_heap_decreased[%u]", i);
		return;
	}

	map_heap_sink_down(i, map_heap[i]);
}

struct mapping *
map_heap_element(u_int i)
{
	if (i < 1) {
		syslog(LOG_CRIT, "map_heap_element[%u]", i);
		return NULL;
	}

	if (i <= map_heap_last)
		return map_heap[i];
	return NULL;
}

/* create a new mapping */

struct mapping *
new_mapping(struct client *client, uint16_t nport, uint8_t proto)
{
	struct mapping *mapping;
	struct natsrc *natsrc;
	struct port *ports;

	mapping = (struct mapping *) malloc(sizeof(*mapping));
	if (mapping == NULL)
		return mapping;
	memset(mapping, 0, sizeof(*mapping));
	mapping->client = client;
	mapping->proto = proto;
	natsrc = client->natsrc;
	ports = natsrc->ports[PROTO2INDEX(proto)];
	mapping->port = ports + (nport - natsrc->minport[PROTO2INDEX(proto)]);
	if (mapping->port->port != nport)
		syslog(LOG_CRIT, "port mismatch");
	return mapping;
}

/* hold down a mapping */

void
hold(struct mapping *mapping)
{
	mapping->down = 1;
	mapping->expire = time(NULL) + HOLD;
	if (!map_heap_insert(mapping))
		syslog(LOG_CRIT, "map_heap_insert: %m");
	dbaddmap(mapping);
}

/* create a new client */

struct client *
new_client(struct in6_addr *ipv6, struct natsrc *natsrc)
{
	struct client *client;

	client = (struct client *) malloc(sizeof(*client));
	if (client == NULL)
		return client;
	memset(client, 0, sizeof(*client));
	memcpy(&client->ipv6, ipv6, sizeof(*ipv6));
	client->natsrc = natsrc;
	client->start = time(NULL);
	ISC_LIST_INIT(&client->mappings);
	return client;
}

/* get the client with an IPv6 address */

struct client *
getclient(struct in6_addr *ipv6, int prompt)
{
	struct client *client;
	struct bucket *bucket;
	u_short hash;

	hash = jhash(ipv6) % HASHSIZE;
	bucket = &clients[hash];
	ISC_LIST_FOREACH(client, bucket, chain)
		if (memcmp(ipv6, &client->ipv6, sizeof(*ipv6)) == 0)
			break;
	if (prompt && (client != NULL)) {
		ISC_LIST_REMOVE(client, chain);
		ISC_LIST_INSERT(bucket, client, chain);
	}
	if ((client != NULL) && (client->start == 0))
		client->start = time(NULL);
	return client;
}

/* search the mapping with nport/proto */

struct mapping *
getmap(struct client *client, uint16_t nport, uint8_t proto)
{
	struct mapping *mapping;

	ISC_LIST_FOREACH(mapping, &client->mappings, chain)
		if ((mapping->port->port == nport) &&
		    (mapping->proto == proto))
			break;
	return mapping;
}

/* search the mapping with src/sport/proto */

struct mapping *
getsrc(struct client *client, uint32_t src, uint16_t sport, uint8_t proto)
{
	struct mapping *mapping = NULL;

	ISC_LIST_FOREACH(mapping, &client->mappings, chain) {
		if ((mapping->src != src) || (mapping->proto != proto))
			continue;
		if (mapping->sport == sport)
			return mapping;
	}
	return NULL;
}

/* create a new natsrc structure */

void
new_natsrc(uint32_t addr, uint16_t *minport, uint16_t *maxport)
{
	char host[INET_ADDRSTRLEN];
	struct natsrc *natsrc;
	struct port *port;
	size_t size[2];
	u_int i;

	(void) inet_ntop(AF_INET, &addr, host, INET_ADDRSTRLEN);
	debug("new_natsrc:\n\t%s tcp %d-%d\n\t%s udp %d-%d",
	      host, (int) minport[0], (int) maxport[0],
	      host, (int) minport[1], (int) maxport[1]);

	if ((minport[0] == 0) || (minport[1] == 0)) {
		fprintf(stderr,
			"new_natsrc: unsupported void pcp range for %s\n",
			host);
		return;
	}

	size[0] = (1 + maxport[0] - minport[0]) * sizeof(*port);
	size[1] = (1 + maxport[1] - minport[1]) * sizeof(*port);
	natsrc = (struct natsrc *) malloc(sizeof(*natsrc));
	if (natsrc == NULL) {
		fprintf(stderr, "new_natsrc failed: %s\n", strerror(errno));
		exit(1);
	}
	memset(natsrc, 0, sizeof(*natsrc));
	natsrc->addr = addr;
	natsrc->minport[0] = minport[0];
	natsrc->minport[1] = minport[1];
	natsrc->maxport[0] = maxport[0];
	natsrc->maxport[1] = maxport[1];
	port = (struct port *) malloc(size[0]);
	if (port == NULL) {
		fprintf(stderr, "new_natsrc failed0: %s\n", strerror(errno));
		exit(1);
	}
	memset(port, 0, size[0]);
	natsrc->ports[0] = port;
	for (i = minport[0]; i <= maxport[0]; i++, port++)
		port->port = i;
	port = (struct port *) malloc(size[1]);
	if (port == NULL) {
		fprintf(stderr, "new_natsrc failed1: %s\n", strerror(errno));
		exit(1);
	}
	memset(port, 0, size[1]);
	natsrc->ports[1] = port;
	for (i = minport[1]; i <= maxport[1]; i++, port++)
		port->port = i;
	ISC_SLIST_INSERT_HEAD(&natsrcs, natsrc, chain);
}

/* get first free port */

int
getfreeport(struct natsrc *natsrc, uint8_t proto)
{
	struct port *port;
	u_int i;

	port = natsrc->ports[PROTO2INDEX(proto)];
	for (i = natsrc->minport[PROTO2INDEX(proto)];
	     i <= natsrc->maxport[PROTO2INDEX(proto)];
	     i++, port++)
		if (port->client == NULL)
			return (int) i;
	syslog(LOG_WARNING, "can't get a free port");
	return PCPD_ERR_NORESOURCES;
}

/* delete a client (direct removal of mappings) */

void
delclient(struct client *client)
{
	struct mapping *mapping;

	ISC_LIST_FOREACH(mapping, &client->mappings, chain) {
		if (mapping->expire != 0)
			map_heap_delete(mapping->heap_index);
		mapping->port->client = NULL;
		dbdelmap(mapping);
		ISC_LIST_REMOVE(mapping, chain);
		free(mapping);
		client->mapcnt -= 1;
	}
	ISC_LIST_REMOVE(client, chain);
	dbdeltunnel(client);
	free(client);
}

/* get external address */

struct client *
getextaddr(struct in6_addr *ipv6)
{
	struct client *client;
	struct natsrc *natsrc;
	uint32_t nsrc;

	client = getclient(ipv6, 1);
	if (client != NULL)
		return client;
	natsrc = NULL;
	nsrc = asktunnel(ipv6);
	if (nsrc != 0) {
		ISC_SLIST_FOREACH(natsrc, &natsrcs, chain)
			if (natsrc->addr == nsrc)
				break;
	}
	if (natsrc != NULL) {
		struct bucket *bucket;
		u_short hash;

		client = new_client(ipv6, natsrc);
		if (client == NULL)
			return NULL;
		hash = jhash(ipv6) % HASHSIZE;
		bucket = &clients[hash];
		ISC_LIST_INSERT(bucket, client, chain);
		dbaddtunnel(client);
		return client;
	} else {
		char host[INET6_ADDRSTRLEN];

		(void) inet_ntop(AF_INET6, ipv6, host, INET6_ADDRSTRLEN);
		debug("can't get external address for %s", host);
		return NULL;
	}
}

/* set a mapping */

#define RETERR(x, lt)		\
	do {			\
		*result = x;	\
		*lifetime = lt;	\
		return NULL;	\
	} while (0)

struct mapping *
setmapping(struct client *client, uint32_t src,
	   uint16_t sport, uint16_t *nport,
	   int flags, uint32_t *lifetime,
	   uint8_t proto, int *result)
{
	struct client *holder;
	struct natsrc *natsrc;
	struct mapping *mapping;
	int outofrange, staticport;
	time_t previous;

	natsrc = client->natsrc;
	mapping = getsrc(client, src, sport, proto);
	if ((mapping != NULL) && (mapping->port->port != *nport)) {
		if (flags & PCPD_PREFER_FAILURE)
			RETERR(PCPD_ERR_CANTPROVIDE, SHORTERRHT);
		*nport = mapping->port->port;
		return setmapping(client, src, sport, nport,
				  0, lifetime, proto, result);
	}
	if (*nport == 0) {
		int candidate;

		if (flags & PCPD_PREFER_FAILURE)
			RETERR(PCPD_ERR_CANTPROVIDE, LONGERRHT);
		candidate = getfreeport(natsrc, proto);
		if (candidate < 0)
			RETERR(candidate, SHORTERRHT);
		*nport = (uint16_t) candidate;
		return setmapping(client, src, sport, nport,
				  0, lifetime, proto, result);
	}
	holder = NULL;
	outofrange = 0;
	if ((*nport < natsrc->minport[PROTO2INDEX(proto)]) ||
	    (*nport > natsrc->maxport[PROTO2INDEX(proto)])) {
		outofrange = 1;
	} else {
		struct port *ports;

		ports = natsrc->ports[PROTO2INDEX(proto)];
		ports += *nport - natsrc->minport[PROTO2INDEX(proto)];
		holder = ports->client;
	}
	if (outofrange || ((holder != NULL) && (holder != client))) {
		if (flags & PCPD_PREFER_FAILURE) {
			if (outofrange)
				RETERR(PCPD_ERR_CANTPROVIDE, LONGERRHT);
			else
				RETERR(PCPD_ERR_CANTPROVIDE, SHORTERRHT);
		}
		*nport = 0;
		return setmapping(client, src, sport, nport,
				  flags, lifetime, proto, result);
	}
		
	if (holder == NULL) {
		if (client->mapcnt >= MAXPCP)
			RETERR(PCPD_ERR_EXQUOTA, SHORTERRHT);
		mapping = new_mapping(client, *nport, proto);
		if (mapping == NULL)
			RETERR(PCPD_ERR_NORESOURCES, SHORTERRHT);
		mapping->src = src;
		mapping->sport = sport;
		staticport = 0;
		*result = askaddmap(mapping, &staticport);
		if (*result == PCPD_STATICMAPPING) {
			free(mapping);
			if ((flags & PCPD_PREFER_FAILURE) &&
			    (staticport != *nport))
				RETERR(PCPD_ERR_CANTPROVIDE, UINT32_MAX);
			debug("setmapping: static");
			*nport = staticport;
			RETERR(PCPD_STATICMAPPING, UINT32_MAX);
		} else if (*result >= PCPD_OK) {
			mapping->expire = time(NULL) + *lifetime;
			if (!map_heap_insert(mapping))
				syslog(LOG_CRIT, "map_heap_insert: %m");
			dbaddmap(mapping);
			ISC_LIST_INSERT(&client->mappings, mapping, chain);
			mapping->port->client = client;
			client->mapcnt += 1;
			debug("setmapping: new");
			if (flags & PCPD_CREATE_AS_LOCKED)
				*result = PCPD_NEWMAPPING;
			return mapping;
		} else {
			debug("setmapping: failed (new)");
			free(mapping);
			RETERR(*result, SHORTERRHT);
		}
	}

	/* else (holder == client) */
	mapping = getmap(client, *nport, proto);
	if (mapping->src != src) {
		if (flags & PCPD_PREFER_FAILURE)
			RETERR(PCPD_ERR_CANTPROVIDE, SHORTERRHT);
		*nport = 0;
		return setmapping(client, src, sport, nport,
				  flags, lifetime, proto, result);
	}
	if (mapping->down) {
		mapping->down = 0;
		mapping->sport = sport;
	} else if (mapping->sport != sport) {
		if (flags & PCPD_PREFER_FAILURE)
			RETERR(PCPD_ERR_CANTPROVIDE, SHORTERRHT);
		*nport = 0;
		return setmapping(client, src, sport, nport,
				  flags, lifetime, proto, result);
	}
	staticport = 0;
	*result = askaddmap(mapping, &staticport);
	if (*result < PCPD_OK) {
		debug("setmapping: failed (renew)");
		*lifetime = SHORTERRHT;
		goto freemapping;
	}
	if (*result == PCPD_STATICMAPPING) {
		debug("setmapping: static & renew");
		if ((flags & PCPD_PREFER_FAILURE) && (staticport != *nport))
			*result = PCPD_ERR_CANTPROVIDE;
		else
			*nport = staticport;
		*lifetime = UINT32_MAX;
		goto freemapping;
	}		
	previous = mapping->expire;
	mapping->expire = time(NULL) + *lifetime;
	if (previous == 0) {
		if (!map_heap_insert(mapping))
			syslog(LOG_CRIT, "map_heap_insert: %m");
	} else if (previous > mapping->expire)
		map_heap_increased(mapping->heap_index);
	else
		map_heap_decreased(mapping->heap_index);
	dbaddmap(mapping);
	debug("setmapping: renew");
	return mapping;

    freemapping:
	if (mapping->expire != 0)
		map_heap_delete(mapping->heap_index);
	mapping->heap_index = 0;
	ISC_LIST_REMOVE(mapping, chain);
	mapping->port->client = NULL;
	mapping->client->mapcnt -= 1;
	free(mapping);
	return NULL;
}

/* send a mapping directly in hold down state */

struct mapping *
setmappinghold(struct client *client, uint32_t src, uint16_t sport,
	       uint16_t nport, u_int expire, uint8_t proto)
{
	struct natsrc *natsrc;
	struct port *ports;
	struct mapping *mapping;

	natsrc = client->natsrc;
	if ((nport < natsrc->minport[PROTO2INDEX(proto)]) ||
	    (nport > natsrc->maxport[PROTO2INDEX(proto)]))
		return NULL;
	ports = natsrc->ports[PROTO2INDEX(proto)];
	ports += nport - natsrc->minport[PROTO2INDEX(proto)];
	if (ports->client != NULL)
		return NULL;
	if (client->mapcnt >= 64)
		return NULL;
	mapping = new_mapping(client, nport, proto);
	if (mapping == NULL)
		return NULL;
	mapping->down = 1;
	mapping->src = src;
	mapping->sport = sport;
	mapping->expire = expire;
	if (!map_heap_insert(mapping))
		syslog(LOG_CRIT, "map_heap_insert: %m");
	dbaddmap(mapping);
	ISC_LIST_INSERT(&client->mappings, mapping, chain);
	mapping->port->client = client;
	client->mapcnt += 1;
	debug("setmappinghold: new");
	return mapping;
}

/* expire routine */

void
expire(void)
{
	struct mapping *mapping;
	time_t now;

	now = time(NULL);
	for (;;) {
		mapping = map_heap_element(1);
		if ((mapping == NULL) || (mapping->expire >= now))
			return;
		/* expire one mapping */
		if (mapping->expire != 0)
			map_heap_delete(mapping->heap_index);
		mapping->heap_index = 0;
		debug("expire");
		if (mapping->down) {
			/* final */
			dbdelmap(mapping);
			ISC_LIST_REMOVE(mapping, chain);
			mapping->port->client = NULL;
			mapping->client->mapcnt -= 1;
			free(mapping);
		} else {
			/* hold it */
			askdelmap(mapping);
			hold(mapping);
		}
	}
}

/* rapid recovery */

pcpd_t *rcpcpd;
int rctrmt;

void
rchandler(int sig)
{
	uint8_t rc[24];
	struct sockaddr_in6 addr;
	u_short hash;
	struct bucket *bucket;
	struct client *client;
	struct itimerval it;
	pcpctx_t ctx;

	sig = sig;
	ctx.response = rc;
	ctx.client = (struct sockaddr *)&addr;
	ctx.sid = (uint8_t *)&addr.sin6_addr;
	memset(&addr, 0, sizeof(addr));
	addr.sin6_family = AF_INET6;
#ifndef __linux__
	addr.sin6_len = sizeof(addr);
#endif
	addr.sin6_port = htons(PCP_RC_PORT);
	memset(&it, 0, sizeof(it));
	/* send an ANNOUNCE response to all plain mode subscribers */
	for (hash = 0; hash < HASHSIZE; hash++) {
		bucket = &clients[hash];
		ISC_LIST_FOREACH(client, bucket, chain) {
			memcpy(&addr.sin6_addr, &client->ipv6, 16);
			pcpd_announce(rcpcpd, &ctx, client->start);
		}
	}
	++rctrmt;
	if (rctrmt == 1)
		it.it_value.tv_usec = PCP_RC_INITTRMT;
	else if (rctrmt == 2)
		it.it_value.tv_usec = PCP_RC_INITTRMT * 2;
	else if (rctrmt >= PCP_RC_MAXTRMT)
		return;
	else
		it.it_value.tv_sec = 1 << (rctrmt - 3);
	(void) setitimer(ITIMER_REAL, &it, NULL);
}

void
rcsetup(pcpd_t *pcpd)
{
	rcpcpd = pcpd;
	(void) signal(SIGALRM, rchandler);
	rchandler(0);
}

/* get subscriber */

int
getsubscriber(const pcpd_t *pcpd, const pcpctx_t *ctx, time_t *epoch)
{
	struct client *client;

	UNUSED(pcpd);
	client = getextaddr((struct in6_addr *) ctx->sid);
	if (client == NULL)
		return PCPD_ERR_NOTAUTH;
	*epoch = client->start;
	return PCPD_OK;
}

/* add explicit dynamic mapping */

int
addmapping(const pcpd_t *pcpd, const pcpctx_t *ctx,
	   const uint8_t flags, const uint8_t proto,
	   const uint8_t *intaddr, const uint16_t intport,
	   uint8_t *extaddr, uint16_t *extport, uint32_t *lifetime)
{
	struct client *client;
	uint32_t src;
	uint8_t cflags = flags;
	int result;

	UNUSED(pcpd);
	client = getclient((struct in6_addr *) ctx->sid, 1);
	if (client == NULL)
		return PCPD_ERR_INVAL;
	if (flags & PCPD_PREFER_FAILURE) {
		uint32_t required;

		/* handle the required external address */
		memcpy(&required, extaddr, 4);
		if ((required != 0) && (required != client->natsrc->addr))
			return PCPD_ERR_CANTPROVIDE;
		if (*extport == 0)
			cflags &= ~PCPD_PREFER_FAILURE;
	}
	memcpy(extaddr, &client->natsrc->addr, 4);
	memcpy(&src, intaddr, 4);
	(void) setmapping(client, src, intport, extport,
			  cflags, lifetime, proto, &result);
	return result;
}

/* delete explicit dynamic mapping */

int
delmapping(const pcpd_t *pcpd, const pcpctx_t *ctx,
	   const uint8_t proto, const uint8_t *intaddr, const uint16_t intport)
{
	struct client *client;
	struct mapping *mapping, *next;
	uint32_t src;

	UNUSED(pcpd);
	client = getclient((struct in6_addr *) ctx->sid, 1);
	if (client == NULL)
		return PCPD_OK;
	memcpy(&src, intaddr, 4);
	ISC_LIST_FOREACH_SAFE(mapping, &client->mappings, chain, next) {
		if (mapping->down)
			continue;
		if ((src != 0) && (mapping->src != src))
			continue;
		if ((proto != 0) && (mapping->proto != proto))
			continue;
		if ((intport != 0) && (mapping->sport != intport))
			continue;
		askdelmap(mapping);
		if (mapping->expire != 0)
			map_heap_delete(mapping->heap_index);
		mapping->heap_index = 0;
		hold(mapping);
		debug("delmapping: hold");
	}
	return PCPD_OK;
}

/* add implicit dynamic mapping */

int
adddynamic(const pcpd_t *pcpd, const pcpctx_t *ctx,
	   const uint8_t proto,
	   const uint8_t *intaddr, const uint16_t intport,
	   const uint8_t *peeraddr, const uint16_t peerport,
	   uint8_t *extaddr, uint16_t *extport,
	   uint32_t *lifetime)
{
	struct client *client;
	uint32_t src, remote, required;
	int result;

	UNUSED(pcpd);
	client = getclient((struct in6_addr *) ctx->sid, 1);
	if (client == NULL)
		return PCPD_ERR_INVAL;
	memcpy(&required, extaddr, 4);
	if ((required != 0) && (required != client->natsrc->addr))
		return PCPD_ERR_CANTPROVIDE;
	memcpy(extaddr, &client->natsrc->addr, 4);
	memcpy(&src, intaddr, 4);
	memcpy(&remote, peeraddr, 4);
	result = askdynamic(client, proto, src, client->natsrc->addr,
			    remote, intport, extport, peerport, lifetime);
	return result;
}

/* filter management */

int
filter(const pcpd_t *pcpd, const pcpctx_t *ctx,
       const uint8_t proto, const uint8_t *intaddr, const uint16_t intport,
       const uint8_t flags, const uint8_t count, const filter_t *filters)
{
	struct client *client;
	struct mapping *mapping;
	uint32_t src, addr;
	int result;

	UNUSED(pcpd);
	client = getclient((struct in6_addr *) ctx->sid, 1);
	if (client == NULL)
		return PCPD_ERR_INVAL;
	memcpy(&src, intaddr, 4);
	mapping = getsrc(client, src, intport, proto);
	if (mapping == NULL)
		return PCPD_ERR_PROCESSING;
	if ((flags & PCPD_DELETE_FILTERS) && (count == 0)) {
		askdelfilter(client, src, intport, proto);
		mapping->fstate = 0;
		mapping->fplen = 0;
		mapping->faddr = 0;
		mapping->fport = 0;
		dbaddmap(mapping);
		return PCPD_OK;
	}
	if (count == 0)
		return PCPD_OK;
	if (count > 1)
		return PCPD_ERR_TOOMANYPEER;
	memcpy(&addr, filters->prefix, 4);
	if (flags & PCPD_DELETE_FILTERS)
		result = askreplacefilter(client, src, intport, proto,
					  filters->plen, addr, filters->port);
	else
		result = askaddfilter(client, src, intport, proto,
				      filters->plen, addr, filters->port);
	if (result >= PCPD_OK) {
		mapping->fplen = filters->plen;
		mapping->faddr = addr;
		mapping->fport = filters->port;
		mapping->fstate = 1;
		dbaddmap(mapping);
	}
	return result;
}

/* statistic counters */

uint64_t stats_operation[256], stats_result[256], stats_option[256];
uint64_t stats_operation_total, stats_result_total, stats_option_total;

/* on multithread & co, declare counters as volatile and
   use GCC __sync_fetch_and_add() atomic primitive */

void
incrcounter(const pcpd_t *pcpd, const pcpctx_t *ctx,
	    const pcpstat_t code, const uint8_t idx)
{
	UNUSED(pcpd);
	UNUSED(ctx);
	switch (code) {
	case PCPS_OPERATION:
		stats_operation[idx] += 1;
		stats_operation_total += 1;
		return;
	case PCPS_RESULT:
		stats_result[idx] += 1;
		stats_result_total += 1;
		return;
	case PCPS_OPTION:
		stats_option[idx] += 1;
		stats_option_total += 1;
		return;
	}
}

/* main (note in passive mode open/close log is CGN responsibility) */

pcpd_t pcpd;

int
main(int argc, char *argv[])
{
	int encap = 0, dbclean = 0;
	int opt, ret, logopt = LOG_NDELAY;
	u_int i;
	char *address;
	extern char *optarg;
	extern int optind;

	for (i = 0; i < HASHSIZE; i++)
		ISC_LIST_INIT(&clients[i]);
	ISC_SLIST_INIT(&natsrcs);

	while ((opt = getopt(argc, argv, "cde")) != -1)
		switch (opt) {
		case 'c':
			dbclean = 1;
			break;
		case 'd':
			logopt += LOG_PERROR;
			break;
		case 'e':
			encap = 1;
			break;
		default:
		usage:
#define USAGE	"Usage: %s [-c] [-d] [-e | address]\n"
			fprintf(stderr, USAGE, argv[0]);
			exit(1);
		}
	if (optind + 1 == argc) {
		address = argv[optind];
		if (encap)
			fprintf(stderr,
				"encapsulation [-e] and address are"
				" incompatible: address \"%s\" is ignored\n",
				address);
	} else if (optind == argc)
		address = NULL;
	else {
		fprintf(stderr, "extra arguments\n");
		goto usage;
	}

	pcpd.opcodes = PCPD_OPCODE_MAP | PCPD_OPCODE_PEER;
	pcpd.cgntype = PCPD_TYPE_DSLITE;
	pcpd.flags = PCPD_MAPINT | PCPD_MAPEXT | PCPD_MAPDST;
	pcpd.flags |= PCPD_3RDPARTY;
	if (encap)
		pcpd.flags |= PCPD_MAPSRC | PCPD_3RDPARTY_CPE;
	pcpd.pcpd_recvrequest = recvrequest;
	pcpd.pcpd_sendresponse = sendresponse;
	pcpd.pcpd_getsubscriber = getsubscriber;
	pcpd.pcpd_addmapping = addmapping;
	pcpd.pcpd_delmapping = delmapping;
	pcpd.pcpd_adddynamic = adddynamic;
	pcpd.pcpd_filter = filter;
	pcpd.pcpd_incrcounter = incrcounter;

	pcpd_openlog(&pcpd, "pcpd", logopt, LOG_LOCAL6);
	initcom(encap, address);
	initdb(dbclean);

	ret = pcpd_lib_init(&pcpd);
	if (ret != PCPD_OK) {
		fprintf(stderr, "pcpd_lib_init fails: %s\n",
			pcpd_strerror(ret));
		exit(ret);
	}

	ask("session dynamic on");

	if (encap == 0)
		rcsetup(&pcpd);

	for (;;) {
		ret = pcpd_lib_runonce(&pcpd, NULL);
		if (ret != PCPD_OK) {
			fprintf(stderr, "pcpd_lib_runonce fails: %s\n",
				pcpd_strerror(ret));
			break;
		}
	}
	ret = pcpd_lib_close(&pcpd);
	pcpd_closelog(&pcpd);
	exit(ret);
}

#ifndef PASSIVE_ONLY

/* define them to get no undef references */

int
pcpd_cgn_init(pcpd_t *pcpd, int argc, char *argv[])
{
	UNUSED(pcpd);
	UNUSED(argc);
	UNUSED(argv);
	abort();
}
	
int
pcpd_cgn_close(pcpd_t *pcpd)
{
	UNUSED(pcpd);
	abort();
}

int
pcpd_cgn_register(pcpd_t *pcpd,
		  void (*announce_callback) (const pcpd_t *pcpd,
					     const pcpctx_t *ctx,
					     const time_t epoch),
		  void (*update_callback) (const pcpd_t *pcpd,
					   const pcpctx_t *ctx,
					   const uint8_t protocol,
					   const uint8_t *intaddr,
					   const uint16_t intport,
					   const uint8_t *extaddr,
					   const uint16_t extport,
					   const uint32_t lifetime),
		  void (*uncache_callback) (const pcpd_t *pcpd,
					    const pcpctx_t *ctx))
{
	UNUSED(pcpd);
	UNUSED(announce_callback);
	UNUSED(update_callback);
	UNUSED(uncache_callback);
	return PCPD_ERR_INVAL;
}

#endif
