#!/usr/bin/env python

# Copyright (C) 2011-2012  Internet Systems Consortium, Inc. ("ISC")
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND ISC DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
# AND FITNESS.  IN NO EVENT SHALL ISC BE LIABLE FOR ANY SPECIAL, DIRECT,
# INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
# LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
# OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
# PERFORMANCE OF THIS SOFTWARE.
#

"""
Network based CGN script for pcpd simulating a NAT444

Francis_Dupont@isc.org, August 2011

Usage: pcpd [-d] net [address]

schema:

        pcpd.c             python.c             this file
------------------ pcpd ----------------         (net.py)
+--------------------+------------------+-------------------------+
| PCP server library | python embedding | CGN fake implementation |
|                    |                  |                         |
|                    <= C structure     |    python program       |
|                    |   python object =>                         |
|                    |                  |                         |
|                    |                  | socket python module    |
|   PCP protocol     |   recvrequest    | <-------------------    |
|      engine        |   sendresponse   | -----------------> ^    |
|                    |                  |                  | |    |
|                    |   addmapping     | Mapping class    | |    |
|                    |   delmapping     |       ...        | |    |
|                    |       ...        |                  | |    |
+--------------------+------------------+------------------|-|----+
                                                           V |
                                                         network
                                                           | ^
                                                           V |
                                                        PCP client
"""   

__version__ = "$Id: cabf395f615a5f7ecd84f67a0c2d0eb02357204e $"

import libpcpd
import socket
import sys
import time

def totextp(protocol):
    """protocol to text"""
    if protocol == socket.IPPROTO_TCP:
        return 'tcp'
    elif protocol == socket.IPPROTO_UDP:
        return 'udp'
    else:
        return 'proto' + str(protocol)

def totexta(addr):
    """IPaddress to text"""
    return socket.inet_ntop(socket.AF_INET, addr)

def fromtext(text):
    """text to IP address"""
    return socket.inet_pton(socket.AF_INET, text)

ZERO4 = fromtext('0.0.0.0')

class Filter(object):
    """Filter"""
    __slots__ = ['addr', 'plen', 'port']
    __hash__ = None

    def __init__(self, addr, plen, port):
        """initializer"""
        self.addr = addr
        self.plen = plen
        self.port = port

    def __cmp__(self, other):
        """compare"""
        if not isinstance(other, Filter):
            return 1
        if self.plen != other.plen:
            return cmp(self.plen, other.plen)
        if self.addr != other.addr:
            return cmp(self.addr, other.addr)
        return cmp(self.port, other.port)

    def totext(self):
        """pretty printer"""
        print '[', totexta(self.addr), '/', self.plen, ' ', self.port, ']'
        
class Mapping(object):
    """explicit dynamic mapping"""
    __slots__ = ['protocol', 'intaddr', 'intport',
                 'extaddr', 'extport', 'expire', 'filters']
    __hash__ = None

    def __init__(self, protocol, intaddr, intport,
                 extaddr, extport, expire):
        """initializer"""
        self.protocol = protocol
        self.intaddr = intaddr
        self.intport = intport
        self.extaddr = extaddr
        self.extport = extport
        self.expire = expire
        self.filters = []

    def __cmp__(self, other):
        """compare"""
        if not isinstance(other, Mapping):
            return 1
        if self.protocol != other.protocol:
            return cmp(self.protocol, other.protocol)
        if self.intaddr != other.intaddr:
            return cmp(self.intaddr, other.intaddr)
        return cmp(self.intport, other.intport)

    def totext(self):
        """pretty print"""
        if self.expire == 0:
            print 'map', totextp(self.protocol), \
                totexta(self.intaddr), ':', self.intport, '->', \
                totexta(self.extaddr), ':', self.extport, 'static'
        else:
            print 'map', totextp(self.protocol), \
                totexta(self.intaddr), ':', self.intport, '->', \
                totexta(self.extaddr), ':', self.extport, 'in', \
                self.expire - time.time()

class Dynamic(object):
    """implicit dynamic mapping"""
    __slots__ = ['protocol', 'intaddr', 'intport', 'peeraddr',
                 'peerport', 'extaddr', 'extport', 'expire']
    __hash__ = None

    def __init__(self, protocol, intaddr, intport, peeraddr,
                 peerport, extaddr, extport, expire):
        """initializer"""
        self.protocol = protocol
        self.intaddr = intaddr
        self.intport = intport
        self.peeraddr = peeraddr
        self.peerport = peerport
        self.extaddr = extaddr
        self.extport = extport
        self.expire = expire

    def __cmp__(self, other):
        """compare"""
        if not isinstance(other, Dynamic):
            return 1
        if self.protocol != other.protocol:
            return cmp(self.protocol, other.protocol)
        if self.intaddr != other.intaddr:
            return cmp(self.intaddr, other.intaddr)
        return cmp(self.intport, other.intport)

    def totext(self):
        """pretty print"""
        if self.expire == 0:
            print 'dyn', totextp(self.protocol), \
                totexta(self.intaddr), ':', self.intport, '->', \
                totexta(self.extaddr), ':', self.extport, 'to', \
                totexta(self.peeraddr), ':', self.peerport, 'static'
        else:
            print 'dyn', totextp(self.protocol), \
                totexta(self.intaddr), ':', self.intport, '->', \
                totexta(self.extaddr), ':', self.extport, 'to', \
                totexta(self.peeraddr), ':', self.peerport, 'in', \
                self.expire - time.time()

class Subscriber(object):
    """Subscriber"""
    __slots__ = ['sid', 'extaddr', 'start', 'mappings', 'dynamics']
    __hash__ = None

    def __init__(self, sid):
        """initializer"""
        self.sid = sid
        self.extaddr = fromtext('198.18.200.1')
        self.start = time.time()        
        self.mappings = []
        self.dynamics = []

    def totext(self):
        """pretty print"""
        print totexta(self.sid), 'since', time.time() - self.start, \
            'at', totexta(self.extaddr)

UINT32_MAX = 4294967295

class CGN(object):
    """Carrier Grade NAT simulator"""
    __slots__ = ['cgntype', 'opcodes', 'flags', 'longerrht', 'shorterrht',
                 'subscribers', 'acl',
                 'xnextport', 'xbusy', 'inextport', 'ibusy',
                 'socket', 'stats']

    def __init__(self):
        """initializer"""
        self.opcodes = libpcpd.OPCODE_ANNOUNCE | \
            libpcpd.OPCODE_MAP | libpcpd.OPCODE_PEER
        self.cgntype = libpcpd.NAT444
        self.flags = libpcpd.MAPSRC | libpcpd.MAPINT | \
            libpcpd.MAPEXT | libpcpd.MAPDST | \
            libpcpd.THIRDPARTY | libpcpd.THIRDPARTY_AO
        self.longerrht = libpcpd.LONGERRHT
        self.shorterrht = libpcpd.SHORTERRHT
        self.subscribers = {}
        self.acl = [('10.0.0.0', 8), ('172.16.0.0', 16),
                    ('192.168.0.0', 16), ('192.0.0.0', 29)]
        self.xnextport = 32000
        self.xbusy = []
        self.inextport = 1024
        self.ibusy = []
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.stats = { 'operation' : 257 * [0L],
                       'result' : 257 * [0L],
                       'option' : 257 * [0L] }
        
    def expire(self):
        """expire routine"""
        now = time.time()
        for client in self.subscribers.itervalues():
            for mapping in client.mappings:
                if mapping.expire == 0:
                    continue
                if mapping.expire > now:
                    continue
                print 'expire:', mapping.extport
                client.mappings.remove(mapping)
                self.xbusy.remove(mapping.extport)
            for dynamic in client.dynamics:
                if dynamic.expire == 0:
                    continue
                if dynamic.expire > now:
                    continue
                print 'expire:', dynamic.extport
                client.dynamics.remove(dynamic)
                self.ibusy.remove(dynamic.extport)

    def recvrequest(self):
        """API / receive request"""
        while True:
            try:
                (request, address) = self.socket.recvfrom(1080)
                break
            except socket.timeout:
                self.expire()
        sid = fromtext(address[0])
        return (socket.AF_INET, (address[0], address[1]), sid, request)

    def sendresponse(self, family, address, sid, response):
        """API / send response"""
        self.socket.sendto(response, address)

    def getsubscriber(self, sid):
        """API / get subscriber"""
        if not sid in self.subscribers:
            self.subscribers[sid] = Subscriber(sid)
        client = self.subscribers[sid]
        return (libpcpd.OK, int(client.start))

    def authorize(self, family, address, sid, request, off3rdpty):
        """API / authorize"""
        return libpcpd.OK

    def addmapping(self, flags, sid, protocol, intaddr, intport,
                   extaddr, extport, lifetime):
        """API / add explicit dynamic mapping"""
        client = self.subscribers.get(sid)
        if client is None:
            return (libpcpd.ERR_NOTAUTH, extaddr, 0, 1800)
        if flags & libpcpd.PREFER_FAILURE:
            if (extaddr != ZERO4) and (extaddr != client.extaddr):
                return (libpcpd.ERR_CANTPROVIDE, extaddr, 0, 300)
            if extport == 0:
                flags &= ~libpcpd.PREFER_FAILURE
        if lifetime < 120:
            lifetime = 120
        elif lifetime > 7200:
            lifetime = 7200
        map0 = Mapping(protocol, intaddr, intport,
                       client.extaddr, extport, time.time() + lifetime)
        for mapping in client.mappings:
            if mapping == map0:
                if (flags & libpcpd.PREFER_FAILURE) and \
                        (mapping.extport != extport):
                    return (libpcpd.ERR_CANTPROVIDE, extaddr, 0, 300)
                if mapping.expire == 0:
                    return (libpcpd.STATICMAPPING, mapping.extaddr,
                            mapping.extport, UINT32_MAX)
                extport = mapping.extport
                mapping.expire = time.time() + lifetime
                return (libpcpd.OK, client.extaddr, extport, lifetime)
        if flags & libpcpd.PREFER_FAILURE:
            if extport < 32000:
                return (libpcpd.ERR_CANTPROVIDE, extaddr, 0, 1800)
            if extport in self.xbusy:
                return (libpcpd.ERR_CANTPROVIDE, extaddr, 0, 30)
        elif (extport < 32000) or (extport in self.xbusy):
            while extport in self.xbusy:
                self.xnextport += 1
                if self.xnextport > 32999:
                    self.xnextport = 32000
            extport = self.xnextport
            self.xnextport += 1
            if self.xnextport > 32999:
                self.xnextport = 32000
        self.xbusy.append(extport)
        map0 = Mapping(protocol, intaddr, intport,
                       client.extaddr, extport, time.time() + lifetime)
        client.mappings.append(map0)
        if flags & libpcpd.CREATE_AS_LOCKED:
            return (libpcpd.NEWMAPPING,
                    client.extaddr, extport, lifetime)
        else:
            return (libpcpd.OK, client.extaddr, extport, lifetime)

    def delmapping(self, sid, protocol, intaddr, intport):
        """API / delete explicit dynamic mapping"""
        client = self.subscribers.get(sid)
        if client is None:
            return libpcpd.OK
        for mapping in client.mappings:
            if mapping.expire == 0:
                continue
            if intport != 0:
                if mapping.intport != intport:
                    continue
            if intaddr != ZERO4:
                if mapping.intaddr != intaddr:
                    continue
            if protocol != 0:
                if mapping.protocol != protocol:
                    continue
            client.mappings.remove(mapping)
            self.xbusy.remove(mapping.extport)
        if (intport == 0) and (protocol == 0):
            for dynamic in client.dynamics:
                if dynamic.expire == 0:
                    continue
                if intaddr != ZERO4:
                    if dynamic.intaddr != intaddr:
                        continue
                client.dynamics.remove(dynamic)
                self.ibusy.remove(dynamic.extport)
        return libpcpd.OK

    def adddynamic(self, sid, protocol, intaddr, intport,
                   peeraddr, peerport, extaddr, extport, lifetime):
        """API / add implicit dynamic mapping"""
        client = self.subscribers.get(sid)
        if client is None:
            return (libpcpd.ERR_NOTAUTH, extaddr, 0, 1800)
        if (extaddr != ZERO4) and (extaddr != client.extaddr):
            return (libpcpd.ERR_CANTPROVIDE, extaddr, 0, 300)
        dyn0 = Dynamic(protocol, intaddr, intport, peeraddr, peerport,
                       client.extaddr, extport, time.time() + lifetime)
        for dynamic in client.dynamics:
            if dynamic == dyn0:
                if (extport != 0) and (dynamic.extport != extport):
                    return (libpcpd.ERR_CANTPROVIDE, extaddr, 0, 300)
                if dynamic.expire == 0:
                    return (libpcpd.STATICMAPPING, dynamic.extaddr,
                            dynamic.extport, UINT32_MAX)
                extport = dynamic.extport
                dynamic.expire = time.time() + lifetime
                return (libpcpd.OK, client.extaddr, extport, lifetime)
        if extport == 0:
            while self.inextport in self.ibusy:
                self.inextport += 1
                if self.inextport > 19999:
                    self.inextport = 1024
            extport = self.inextport
            self.inextport += 1
            if self.inextport > 19999:
                self.inextport = 1024
        if (extport < 1024) or ((extport >= 32000) and (extport < 33000)):
            return (libpcpd.ERR_CANTPROVIDE, extaddr, 0, 1800)
        if (extport in self.ibusy) or (extport in self.xbusy):
            return (libpcpd.ERR_CANTPROVIDE, extaddr, 0, 30)
        if Mapping(protocol, intaddr, intport, client.extaddr,
                   extport, 0) in client.mappings:
            return (libpcpd.ERR_PROCESSING, extaddr, 0, 300)
        self.ibusy.append(extport)
        dyn0 = Dynamic(protocol, intaddr, intport, peeraddr, peerport,
                       client.extaddr, extport, time.time() + lifetime)
        client.dynamics.append(dyn0)
        return (libpcpd.OK, client.extaddr, extport, lifetime)

    def filter(self, sid, protocol, intaddr, intport, flags, filters):
        """API / manage explicit dynamic mapping remote filters"""
        client = self.subscribers.get(sid)
        if client is None:
            return libpcpd.ERR_NOTAUTH
        mapping = None
        for mapping in client.mappings:
            if mapping.intport != intport:
                continue
            if mapping.intaddr != intaddr:
                continue
            if mapping.protocol == protocol:
                break
        if mapping is None:
            return libpcpd.ERR_PROCESSING
        saved = list(mapping.filters)
        if flags & libpcpd.DELETE_FILTERS:
            mapping.filters = []
        try:
            for triple in filters:
                (plen, port, addr) = triple
                filt = Filter(addr, plen, port)
                if filt in mapping.filters:
                    continue
                mapping.filters.append(filt)
            if len(mapping.filters) > 50:
                mapping.filters = saved
                return libpcpd.ERR_TOOMANYPEER
            mapping.filters.sort()
            return libpcpd.OK
        except:
            mapping.filters = saved
            return libpcpd.ERR_PROCESSING

    def incrcounter(self, code, idx):
        """API / increment statistic counter"""
        statvec = self.stats.get(code)
        if statvec is not None:
            statvec[idx] += 1
            statvec[256] += 1

__CGN__ = CGN()

def main(args):
    """main"""
    if len(args) == 0:
        print 'usage: pcpd script [address]'
        return -1
    if len(args) > 1:
        address = args[1]
    else:
        address = '0.0.0.0'
    __CGN__.socket.settimeout(10.0)
    __CGN__.socket.bind((address, libpcpd.PORT))

main(sys.argv)
