/*
 * Copyright (C) 2011-2012  Internet Systems Consortium, Inc. ("ISC")
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND ISC DISCLAIMS ALL WARRANTIES WITH
 * REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS.  IN NO EVENT SHALL ISC BE LIABLE FOR ANY SPECIAL, DIRECT,
 * INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
 * OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */

/* $Id: 68f8b72557054e35ddef86650851d93bb045d871 $ */

/*
 * PCP (port-control-protocol) server: python stub
 *
 * Francis_Dupont@isc.org, August 2011
 *
 * usage: pcpd [-d] <python-file> <extra arguments>
 *
 * schema:
 *
 * +--------------------+------------------+--------------------+
 * | PCP server library | python embedding | CGN implementation |
 * |                    |                  |                    |
 * |                    <= C structure     |                    |
 * |                    |   python object =>                    |
 * |                    |                  |                    |
 * |   PCP protocol     |   recvrequest    |      python        |
 * |      engine        |   sendresponse   |      program       |
 * |                    |   addmapping     |                    |
 * |                    |   delmapping     |                    |
 * |                    |       ...        |                    |
 * +--------------------+------------------+--------------------+
 *        pcpd.c              python.c            xxx.py
 */

#include <Python.h>

#include "pcpd.h"

#include <netinet/in.h>
#include <arpa/inet.h>
#include <signal.h>
#include <unistd.h>

#define UNUSED(x)	(x) = (x)

size_t
PCPD_recvrequest(const pcpd_t *pcpd, const pcpctx_t *ctx)
{
	PyObject *func, *result;
	struct sockaddr_in *c4;
	struct sockaddr_in6 *c6;
	int family, port;
	char *host = NULL;
	uint8_t *id = NULL, *req = NULL;
	int ilen, reqlen;

	if ((pcpd == NULL) || (pcpd->opaque == NULL)) {
		fprintf(stderr, "bad call to recvrequest\n");
		return 0;
	}
	func = PyObject_GetAttrString(pcpd->opaque, "recvrequest");
	if ((func == NULL) || !PyCallable_Check(func)) {
		if (PyErr_Occurred())
			PyErr_Print();
		Py_XDECREF(func);
		fprintf(stderr,
			"PyObject_GetAttrString(recvrequest) failed\n");
		return 0;
	}
	result = PyObject_CallObject(func, NULL);
	Py_DECREF(func);
	if (result == NULL) {
		PyErr_Print();
		fprintf(stderr, "PyObject_CallObject(recvrequest) failed\n");
		return 0;
	}
	if (!PyArg_ParseTuple(result, "i(si)s#s#", &family,
			      &host, &port, &id, &ilen, &req, &reqlen)) {
		PyErr_Print();
		Py_DECREF(result);
		fprintf(stderr, "PyArg_ParseTuple(recvrequest) failed\n");
		return 0;
	}
	Py_DECREF(result);
	switch (family) {
	case AF_INET:
		c4 = (struct sockaddr_in *) ctx->client;
		c4->sin_family = AF_INET;
#ifdef BSD
		c4->sin_len = sizeof(*c4);
#endif
		(void) inet_pton(AF_INET, host, &c4->sin_addr);
		c4->sin_port = htons((u_short) port);
		break;

	case AF_INET6:
		c6 = (struct sockaddr_in6 *) ctx->client;
		c6->sin6_family = AF_INET6;
#ifdef BSD
		c6->sin6_len = AF_INET6;
#endif
		c6->sin6_port = htons((u_short) port);
		(void) inet_pton(AF_INET6, host, &c6->sin6_addr);
		break;
	}
	if (ilen > PCPD_SID_SIZE)
		ilen = PCPD_SID_SIZE;
	memcpy(ctx->sid, id, ilen);
	if (reqlen > PCPD_BUF_SIZE)
		reqlen = PCPD_BUF_SIZE;
	memcpy(ctx->request, req, reqlen);

	return (size_t) reqlen;
}

void
PCPD_sendresponse(const pcpd_t *pcpd, const pcpctx_t *ctx, const size_t length)
{
	PyObject *func, *args = NULL, *result;
	struct sockaddr_in *c4;
	struct sockaddr_in6 *c6;
	char host[INET6_ADDRSTRLEN];
	int port, ilen;
	
	if ((pcpd == NULL) || (pcpd->opaque == NULL)) {
		fprintf(stderr, "bad call to sendresponse\n");
		return;
	}
	func = PyObject_GetAttrString(pcpd->opaque, "sendresponse");
	if ((func == NULL) || !PyCallable_Check(func)) {
		if (PyErr_Occurred())
			PyErr_Print();
		Py_XDECREF(func);
		fprintf(stderr,
			"PyObject_GetAttrString(sendresponse) failed\n");
		return;
	}
	ilen = idsize(pcpd);
	switch (ctx->client->sa_family) {
	case AF_INET:
		c4 = (struct sockaddr_in *) ctx->client;
		(void) inet_ntop(AF_INET, &c4->sin_addr, host, sizeof(host));
		port = (int) ntohs(c4->sin_port);
		args = Py_BuildValue("i(si)s#s#", AF_INET, host, port,
				     ctx->sid, ilen, ctx->response,
				     (int) length);
		break;
	case AF_INET6:
		c6 = (struct sockaddr_in6 *) ctx->client;
		(void) inet_ntop(AF_INET6, &c6->sin6_addr, host, sizeof(host));
		port = (int) ntohs(c6->sin6_port);
		args = Py_BuildValue("i(si)s#s#", AF_INET6, host, port,
				     ctx->sid, ilen, ctx->response,
				     (int) length);
		break;
	}
	if (args == NULL) {
		fprintf(stderr, "Py_BuildValue(sendresponse) failed\n");
		Py_DECREF(func);
		return;
	}
	result = PyObject_CallObject(func, args);
	Py_DECREF(func);
	Py_DECREF(args);
	if (result == NULL) {
		PyErr_Print();
		fprintf(stderr, "PyObject_CallObject(sendresponse) failed\n");
		return;
	}
	Py_DECREF(result);
}

int
PCPD_getsubscriber(const pcpd_t *pcpd, const pcpctx_t *ctx, time_t *epoch)
{
	PyObject *func, *args, *result;
	int ilen, ret, lt;

	if ((pcpd == NULL) || (pcpd->opaque == NULL)) {
		fprintf(stderr, "bad call to getsubscriber\n");
		return PCPD_ERR_INVAL;
	}
	func = PyObject_GetAttrString(pcpd->opaque, "getsubscriber");
	if ((func == NULL) || !PyCallable_Check(func)) {
		if (PyErr_Occurred())
			PyErr_Print();
		Py_XDECREF(func);
		fprintf(stderr,
			"PyObject_GetAttrString(getsubscriber) failed\n");
		return PCPD_ERR_INVAL;
	}
	ilen = idsize(pcpd);
	args = Py_BuildValue("(s#)", ctx->sid, ilen);
	if (args == NULL) {
		PyErr_Print();
		fprintf(stderr, "Py_BuildValue(getsubscriber) failed\n");
		Py_DECREF(func);
		return PCPD_ERR_INVAL;
	}
	result = PyObject_CallObject(func, args);
	Py_DECREF(func);
	Py_DECREF(args);
	if (result == NULL) {
		PyErr_Print();
		fprintf(stderr, "PyObject_CallObject(getsubscriber) failed\n");
		return PCPD_ERR_INVAL;
	}
	if (!PyArg_ParseTuple(result, "ii", &ret, &lt)) {
		PyErr_Print();
		Py_DECREF(result);
		fprintf(stderr, "PyArg_ParseTuple(getsubscriber) failed\n");
		return PCPD_ERR_INVAL;
	}
	Py_DECREF(result);
	*epoch = (time_t) lt;
	return ret;
}

int
PCPD_authorize(const pcpd_t *pcpd, const pcpctx_t *ctx, const size_t length)
{
	PyObject *func, *args = NULL, *result;
	struct sockaddr_in *c4;
	struct sockaddr_in6 *c6;
	char host[INET6_ADDRSTRLEN];
	int port, ilen, ret;

	if ((pcpd == NULL) || (pcpd->opaque == NULL)) {
		fprintf(stderr, "bad call to authorize\n");
		return PCPD_ERR_INVAL;
	}
	func = PyObject_GetAttrString(pcpd->opaque, "authorize");
	if ((func == NULL) || !PyCallable_Check(func)) {
		if (PyErr_Occurred())
			PyErr_Print();
		Py_XDECREF(func);
		fprintf(stderr,
			"PyObject_GetAttrString(authorize) failed\n");
		return PCPD_ERR_INVAL;
	}
	ilen = idsize(pcpd);
	switch (ctx->client->sa_family) {
	case AF_INET:
		c4 = (struct sockaddr_in *) ctx->client;
		(void) inet_ntop(AF_INET, &c4->sin_addr, host, sizeof(host));
		port = (int) ntohs(c4->sin_port);
		args = Py_BuildValue("i(si)s#s#i", AF_INET, host, port,
				     ctx->sid, ilen, ctx->request,
				     (int) length, (int) ctx->off3rdparty);
		break;
	case AF_INET6:
		c6 = (struct sockaddr_in6 *) ctx->client;
		(void) inet_ntop(AF_INET6, &c6->sin6_addr, host, sizeof(host));
		port = (int) ntohs(c6->sin6_port);
		args = Py_BuildValue("i(si)s#s#i", AF_INET6, host, port,
				     ctx->sid, ilen, ctx->request,
				     (int) length, (int) ctx->off3rdparty);
		break;
	}
	if (args == NULL) {
		fprintf(stderr, "Py_BuildValue(authorize) failed\n");
		Py_DECREF(func);
		return PCPD_ERR_INVAL;
	}
	result = PyObject_CallObject(func, args);
	Py_DECREF(func);
	Py_DECREF(args);
	if ((result == NULL) || !PyInt_Check(result)) {
		if (PyErr_Occurred())
			PyErr_Print();
		Py_XDECREF(result);
		fprintf(stderr, "PyObject_CallObject(authorize) failed\n");
		return PCPD_ERR_INVAL;
	}
	ret = PyInt_AsLong(result);
	Py_DECREF(result);
	return ret;
}

int
PCPD_addmapping(const pcpd_t *pcpd, const pcpctx_t *ctx,
		const uint8_t flags, const uint8_t proto,
		const uint8_t *intaddr, const uint16_t intport,
		uint8_t *extaddr, uint16_t *extport, uint32_t *lifetime)
{
	PyObject *func, *args, *result;
	int ilen, ret, port, lt, ial, eal;
	uint8_t *addr = NULL;

	if ((pcpd == NULL) || (pcpd->opaque == NULL)) {
		fprintf(stderr, "bad call to addmapping\n");
		return PCPD_ERR_INVAL;
	}
	func = PyObject_GetAttrString(pcpd->opaque, "addmapping");
	if ((func == NULL) || !PyCallable_Check(func)) {
		if (PyErr_Occurred())
			PyErr_Print();
		Py_XDECREF(func);
		fprintf(stderr, "PyObject_GetAttrString(addmapping) failed\n");
		return PCPD_ERR_INVAL;
	}
	ilen = idsize(pcpd);
	port = (int) *extport;
	lt = (int) *lifetime;
	if (pcpd->flags & PCPD_MAPINT)
		ial = 4;
	else
		ial = 16;
	if (pcpd->flags & PCPD_MAPEXT)
		eal = 4;
	else
		eal = 16;
	args = Py_BuildValue("is#is#is#ii", (int) flags, ctx->sid, ilen,
			     (int) proto, intaddr, ial, (int) intport,
			     extaddr, eal, port, lt);
	if (args == NULL) {
		PyErr_Print();
		fprintf(stderr, "Py_BuildValue(addmapping) failed\n");
		Py_DECREF(func);
		return PCPD_ERR_INVAL;
	}
	result = PyObject_CallObject(func, args);
	Py_DECREF(func);
	Py_DECREF(args);
	if (result == NULL) {
		PyErr_Print();
		fprintf(stderr, "PyObject_CallObject(addmapping) failed\n");
		return PCPD_ERR_INVAL;
	}
	if (!PyArg_ParseTuple(result, "is#ii",
			      &ret, &addr, &eal, &port, &lt)) {
		PyErr_Print();
		Py_DECREF(result);
		fprintf(stderr, "PyArg_ParseTuple(addmapping) failed\n");
		return PCPD_ERR_INVAL;
	}
	Py_DECREF(result);
	if ((eal != 4) && (eal != 16)) {
		fprintf(stderr, "bad external address length %d\n", eal);
		return PCPD_ERR_INVAL;
	}
	memcpy(extaddr, addr, eal);
	*extport = (uint16_t) port;
	*lifetime = (uint32_t) lt;
	return ret;
}

int
PCPD_delmapping(const pcpd_t *pcpd, const pcpctx_t *ctx,
		const uint8_t proto,
		const uint8_t *intaddr, const uint16_t intport)
{
	PyObject *func, *args, *result;
	int ilen, al, ret;

	if ((pcpd == NULL) || (pcpd->opaque == NULL)) {
		fprintf(stderr, "bad call to delmapping\n");
		return PCPD_ERR_INVAL;
	}
	func = PyObject_GetAttrString(pcpd->opaque, "delmapping");
	if ((func == NULL) || !PyCallable_Check(func)) {
		if (PyErr_Occurred())
			PyErr_Print();
		Py_XDECREF(func);
		fprintf(stderr, "PyObject_GetAttrString(delmapping) failed\n");
		return PCPD_ERR_INVAL;
	}
	ilen = idsize(pcpd);
	if (pcpd->flags & PCPD_MAPINT)
		al = 4;
	else
		al = 16;
	args = Py_BuildValue("s#is#i", ctx->sid, ilen,
			     (int) proto, intaddr, al, (int) intport);
	if (args == NULL) {
		PyErr_Print();
		fprintf(stderr, "Py_BuildValue(delmapping) failed\n");
		Py_DECREF(func);
		return PCPD_ERR_INVAL;
	}
	result = PyObject_CallObject(func, args);
	Py_DECREF(func);
	Py_DECREF(args);
	if ((result == NULL) || !PyInt_Check(result)) {
		if (PyErr_Occurred())
			PyErr_Print();
		Py_XDECREF(result);
		fprintf(stderr, "PyObject_CallObject(delmapping) failed\n");
		return PCPD_ERR_INVAL;
	}
	ret = PyInt_AsLong(result);
	Py_DECREF(result);
	return ret;
}

int
PCPD_adddynamic(const pcpd_t *pcpd, const pcpctx_t *ctx,
		const uint8_t proto,
		const uint8_t *intaddr, const uint16_t intport,
		const uint8_t *peeraddr, const uint16_t peerport,
		uint8_t *extaddr, uint16_t *extport,
		uint32_t *lifetime)
{
	PyObject *func, *args, *result;
	int ilen, ret, port, lt, ial, pal, eal;
	uint8_t *addr = NULL;

	if ((pcpd == NULL) || (pcpd->opaque == NULL)) {
		fprintf(stderr, "bad call to adddynamic\n");
		return PCPD_ERR_INVAL;
	}
	func = PyObject_GetAttrString(pcpd->opaque, "adddynamic");
	if ((func == NULL) || !PyCallable_Check(func)) {
		pcpd_t *p = (pcpd_t *) pcpd;

		if (PyErr_Occurred())
			PyErr_Print();
		Py_XDECREF(func);
		fprintf(stderr, "PyObject_GetAttrString(adddynamic) failed\n");
		p->pcpd_adddynamic = NULL;
		return PCPD_ERR_INVAL;
	}
	ilen = idsize(pcpd);
	port = (int) *extport;
	lt = (int) *lifetime;
	if (pcpd->flags & PCPD_MAPINT)
		ial = 4;
	else
		ial = 16;
	if (pcpd->flags & PCPD_MAPDST)
		pal = 4;
	else
		pal = 16;
	if (pcpd->flags & PCPD_MAPEXT)
		eal = 4;
	else
		eal = 16;
	args = Py_BuildValue("s#is#is#is#ii", ctx->sid, ilen,
			     (int) proto, intaddr, ial, (int) intport,
			     peeraddr, pal, (int) peerport,
			     extaddr, eal, port, lt);
	if (args == NULL) {
		PyErr_Print();
		fprintf(stderr, "Py_BuildValue(adddynamic) failed\n");
		Py_DECREF(func);
		return PCPD_ERR_INVAL;
	}
	result = PyObject_CallObject(func, args);
	Py_DECREF(func);
	Py_DECREF(args);
	if (result == NULL) {
		PyErr_Print();
		fprintf(stderr, "PyObject_CallObject(adddynamic) failed\n");
		return PCPD_ERR_INVAL;
	}
	if (!PyArg_ParseTuple(result, "is#ii",
			      &ret, &addr, &eal, &port, &lt)) {
		PyErr_Print();
		Py_DECREF(result);
		fprintf(stderr, "PyArg_ParseTuple(adddynamic) failed\n");
		return PCPD_ERR_INVAL;
	}
	Py_DECREF(result);
	if ((eal != 4) && (eal != 16)) {
		fprintf(stderr, "bad external address length %d\n", eal);
		return PCPD_ERR_INVAL;
	}
	memcpy(extaddr, addr, eal);
	*extport = (uint16_t) port;
	*lifetime = (uint32_t) lt;
	return ret;
}

int
PCPD_filter(const pcpd_t *pcpd, const pcpctx_t *ctx,
	    const uint8_t proto,
	    const uint8_t *intaddr, const uint16_t intport,
	    const uint8_t flags, const uint8_t count, const filter_t *filters)
{
	PyObject *func, *el, *list, *args, *result;
	const filter_t *f;
	int ilen, ret, al, fal;
	u_int i;

	if ((pcpd == NULL) || (pcpd->opaque == NULL)) {
		fprintf(stderr, "bad call to filter\n");
		return PCPD_ERR_INVAL;
	}
	func = PyObject_GetAttrString(pcpd->opaque, "filter");
	if ((func == NULL) || !PyCallable_Check(func)) {
		pcpd_t *p = (pcpd_t *) pcpd;

		if (PyErr_Occurred())
			PyErr_Print();
		Py_XDECREF(func);
		fprintf(stderr, "PyObject_GetAttrString(filter) failed\n");
		p->pcpd_filter = NULL;
		return PCPD_ERR_INVAL;
	}
	ilen = idsize(pcpd);
	if (pcpd->flags & PCPD_MAPINT)
		al = 4;
	else
		al = 16;
	if (pcpd->flags & PCPD_MAPDST)
		fal = 4;
	else
		fal = 16;
	list = PyList_New((int) count);
	if (list == NULL) {
		PyErr_Print();
		fprintf(stderr, "PyList_New(%d filter) failed\n", (int) count);
		Py_DECREF(func);
		return PCPD_ERR_INVAL;
	}
	for (i = 0, f = filters; i < count; i++, f++) {
		el = Py_BuildValue("iis#", (int) f->plen,
				   (int) f->port, f->prefix, fal);
		if (el == NULL) {
			PyErr_Print();
			fprintf(stderr, "Py_BuildValue(filter%u) failed\n", i);
			Py_DECREF(func);
			Py_DECREF(list);
			return PCPD_ERR_INVAL;
		}
		if (PyList_SetItem(list, (int) i, el) < 0) {
			PyErr_Print();
			fprintf(stderr,
				"PyList_SetItem(filter%u) failed\n", i);
			Py_DECREF(func);
			Py_DECREF(list);
			return PCPD_ERR_INVAL;
		}
	}
	args = Py_BuildValue("s#is#iiO", ctx->sid, ilen,
			     (int) proto, intaddr, al, (int) intport,
			     (int) flags, list);
	if (args == NULL) {
		PyErr_Print();
		fprintf(stderr, "Py_BuildValue(filter) failed\n");
		Py_DECREF(func);
		Py_DECREF(list);
		return PCPD_ERR_INVAL;
	}
	result = PyObject_CallObject(func, args);
	Py_DECREF(func);
	Py_DECREF(args);
	if ((result == NULL) || !PyInt_Check(result)) {
		if (PyErr_Occurred())
			PyErr_Print();
		Py_XDECREF(result);
		fprintf(stderr, "PyObject_CallObject(filter) failed\n");
		return PCPD_ERR_INVAL;
	}
	ret = PyInt_AsLong(result);
	Py_DECREF(result);
	return ret;
}

void
PCPD_incrcounter(const pcpd_t *pcpd, const pcpctx_t *ctx,
		 const pcpstat_t code, const uint8_t idx)
{
	PyObject *func, *args, *result;
	char *cname;

	UNUSED(ctx);
	if ((pcpd == NULL) || (pcpd->opaque == NULL)) {
		fprintf(stderr, "bad call to incrcounter\n");
		return;
	}
	switch (code) {
	case PCPS_OPERATION:
		cname = "operation";
		break;
	case PCPS_RESULT:
		cname = "result";
		break;
	case PCPS_OPTION:
		cname = "option";
		break;
	default:
		fprintf(stderr, "bad code %d in incrcounter\n", (int) code);
		return;
	}
	func = PyObject_GetAttrString(pcpd->opaque, "incrcounter");
	if ((func == NULL) || !PyCallable_Check(func)) {
		if (PyErr_Occurred())
			PyErr_Print();
		Py_XDECREF(func);
		fprintf(stderr,
			"PyObject_GetAttrString(incrcounter) failed\n");
		return;
	}
	args = Py_BuildValue("sB", cname, idx);
	if (args == NULL) {
		fprintf(stderr, "Py_BuildValue(incrcounter) failed\n");
		Py_DECREF(func);
		return;
	}
	result = PyObject_CallObject(func, args);
	Py_DECREF(func);
	Py_DECREF(args);
	if (result == NULL) {
		PyErr_Print();
		fprintf(stderr, "PyObject_CallObject(incrcounter) failed\n");
		return;
	}
	Py_DECREF(result);
}

u_char initialized;
struct sigaction oldsa, quitsa;

void
onsig(int sig)
{
	fprintf(stderr, "got signal %d\n", sig);
	terminate = 1;
}

int
pcpd_cgn_init(pcpd_t *pcpd, int argc, char *argv[])
{
	PyObject *module, *entry, *attr;
	int opt;
	extern int optind;

	if (pcpd == NULL)
		return PCPD_ERR_INVAL;
	memset(pcpd, 0, sizeof(*pcpd));

	if (initialized == 0) {
		memset(&oldsa, 0, sizeof(oldsa));
		memset(&quitsa, 0, sizeof(quitsa));
		quitsa.sa_handler = onsig;
		/* don't set SA_RESTART */
		(void) sigaction(SIGQUIT, &quitsa, &oldsa);
		Py_Initialize();
	}
	initialized++;

	while ((opt = getopt(argc, argv, "d")) != -1)
		switch (opt) {
		case 'd':
			break;
		default:
		usage:
#define USAGE	"Usage: %s [-d] script [args]\n"
			fprintf(stderr, USAGE, argv[0]);
			return PCPD_ERR_INVAL;
		}
	if (optind == argc)
		goto usage;
	PySys_SetArgv(argc - optind, argv + optind);
	module = PyImport_ImportModule(argv[optind]);
	if (module == NULL) {
		PyErr_Print();
		fprintf(stderr,
			"PyImport_ImportModule(%s) failed\n",
			argv[optind]);
		goto usage;
	}
	entry = PyObject_GetAttrString(module, "__CGN__");
	Py_DECREF(module);
	if (entry == NULL) {
		if (PyErr_Occurred())
			PyErr_Print();
		fprintf(stderr, "PyObject_GetAttrString(__CGN__) failed\n");
		goto usage;
	}
	pcpd->opaque = entry;

	attr = PyObject_GetAttrString(entry, "cgntype");
	if ((attr == NULL) || !PyInt_Check(attr)) {
		if (PyErr_Occurred())
			PyErr_Print();
		pcpd->opaque = NULL;
		Py_XDECREF(attr);
		Py_DECREF(entry);
		fprintf(stderr,
			"PyObject_GetAttrString(cgntype as int) failed\n");
		goto usage;
	}
	pcpd->cgntype = (uint32_t) PyInt_AsLong(attr);
	Py_DECREF(attr);

	attr = PyObject_GetAttrString(entry, "opcodes");
	if ((attr == NULL) || !PyInt_Check(attr)) {
		if (PyErr_Occurred())
			PyErr_Print();
		pcpd->opaque = NULL;
		Py_XDECREF(attr);
		Py_DECREF(entry);
		fprintf(stderr,
			"PyObject_GetAttrString(opcodes as int) failed\n");
		goto usage;
	}
	pcpd->opcodes = (uint32_t) PyInt_AsLong(attr);
	Py_DECREF(attr);

	attr = PyObject_GetAttrString(entry, "flags");
	if ((attr == NULL) || !PyInt_Check(attr)) {
		if (PyErr_Occurred())
			PyErr_Print();
		pcpd->opaque = NULL;
		Py_XDECREF(attr);
		Py_DECREF(entry);
		fprintf(stderr,
			"PyObject_GetAttrString(flags as int) failed\n");
		goto usage;
	}
	pcpd->flags = (uint32_t) PyInt_AsLong(attr);
	Py_DECREF(attr);

	attr = PyObject_GetAttrString(entry, "longerrht");
	if ((attr == NULL) || !PyInt_Check(attr)) {
		if (PyErr_Occurred())
			PyErr_Print();
		pcpd->opaque = NULL;
		Py_XDECREF(attr);
		Py_DECREF(entry);
		fprintf(stderr,
			"PyObject_GetAttrString(longerrht as int) failed\n");
		goto usage;
	}
	pcpd->longerrht = (uint32_t) PyInt_AsLong(attr);
	Py_DECREF(attr);

	attr = PyObject_GetAttrString(entry, "shorterrht");
	if ((attr == NULL) || !PyInt_Check(attr)) {
		if (PyErr_Occurred())
			PyErr_Print();
		pcpd->opaque = NULL;
		Py_XDECREF(attr);
		Py_DECREF(entry);
		fprintf(stderr,
			"PyObject_GetAttrString(shorterrht as int) failed\n");
		goto usage;
	}
	pcpd->shorterrht = (uint32_t) PyInt_AsLong(attr);
	Py_DECREF(attr);

	pcpd->pcpd_recvrequest = PCPD_recvrequest;
	pcpd->pcpd_sendresponse = PCPD_sendresponse;
	pcpd->pcpd_getsubscriber = PCPD_getsubscriber;
	pcpd->pcpd_authorize = PCPD_authorize;
	pcpd->pcpd_addmapping = PCPD_addmapping;
	pcpd->pcpd_delmapping = PCPD_delmapping;
	pcpd->pcpd_adddynamic = PCPD_adddynamic;
	pcpd->pcpd_filter = PCPD_filter;
	pcpd->pcpd_incrcounter = PCPD_incrcounter;

	return PCPD_OK;
}

int
pcpd_cgn_close(pcpd_t *pcpd)
{
	if ((pcpd == NULL) || (pcpd->opaque == NULL) || (initialized == 0))
		return PCPD_OK;
	--initialized;
	if (initialized == 0) {
		Py_DECREF(pcpd->opaque);
		(void) sigaction(SIGQUIT, &oldsa, NULL);
		Py_Finalize();
	}
	memset(pcpd, 0, sizeof(pcpd));
	return PCPD_OK;
}

int
pcpd_cgn_register(pcpd_t *pcpd,
		  void (*announce_callback) (const pcpd_t *pcpd,
					     const pcpctx_t *ctx,
					     const time_t epoch),
		  void (*update_callback) (const pcpd_t *pcpd,
					   const pcpctx_t *ctx,
					   const uint8_t protocol,
					   const uint8_t *intaddr,
					   const uint16_t intport,
					   const uint8_t *extaddr,
					   const uint16_t extport,
					   const uint32_t lifetime),
		  void (*uncache_callback) (const pcpd_t *pcpd,
					    const pcpctx_t *ctx))
{
	UNUSED(pcpd);
	UNUSED(announce_callback);
	UNUSED(update_callback);
	UNUSED(uncache_callback);
	return PCPD_ERR_INVAL;
}
