/*
 * Copyright (C) 2011-2012  Internet Systems Consortium, Inc. ("ISC")
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND ISC DISCLAIMS ALL WARRANTIES WITH
 * REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS.  IN NO EVENT SHALL ISC BE LIABLE FOR ANY SPECIAL, DIRECT,
 * INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
 * OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */

/* $Id: bf2f4c5c955fad1037442f34d75347554e04580b $ */

/*
 * PCP (port-control-protocol) server include
 *
 * Francis_Dupont@isc.org, August 2011
 *
 * types: pcpd_t, filter_t
 */

#ifndef __PCPD_H__
#define __PCPD_H__

/* API version */

#define __PCPD_API_VERSION__		0

#ifdef PCPD_PRIVATE_INCLUDE
#include PCPD_PRIVATE_INCLUDE
#endif
#ifndef NO_SOCKET
#include <sys/socket.h>
#else
#include "socket.h"
#endif
#include <stdarg.h>
#ifndef NO_STDINT
#include <stdint.h>
#endif
#include <string.h>
#include <time.h>

#ifdef __cplusplus
extern "C" {
#endif

/*
 * Types
 */

typedef struct _pcpd_ pcpd_t;
typedef struct _pcpctx_ pcpctx_t;
typedef struct _filter_ filter_t;
typedef enum _pcpstat_ pcpstat_t;

struct _pcpd_ {				/* PCP server */
	void *opaque;			/* opaque handler */
	uint32_t cgntype;		/* CGN type */
	uint32_t opcodes;		/* supported opcodes */
	uint32_t flags;			/* flags */
	uint32_t longerrht, shorterrht;	/* long and short error hold time */
	/* methods */
	size_t (*pcpd_recvrequest)(const pcpd_t *pcpd,
				   const pcpctx_t *ctx);
	void (*pcpd_sendresponse)(const pcpd_t *pcpd,
				  const pcpctx_t *ctx,
				  const size_t length);
	int (*pcpd_getsubscriber)(const pcpd_t *pcpd,
				  const pcpctx_t *ctx,
				  time_t *epoch);
	int (*pcpd_authorize) (const pcpd_t *pcpd,
			       const pcpctx_t *ctx,
			       const size_t length);
	int (*pcpd_addmapping)(const pcpd_t *pcpd,
			       const pcpctx_t *ctx,
			       const uint8_t flags,
			       const uint8_t proto,
			       const uint8_t *intaddr,
			       const uint16_t intport,
			       uint8_t *extaddr,
			       uint16_t *extport,
			       uint32_t *lifetime);
	int (*pcpd_delmapping)(const pcpd_t *pcpd,
			       const pcpctx_t *ctx,
			       const uint8_t proto,
			       const uint8_t *intaddr,
			       const uint16_t intport);
	int (*pcpd_adddynamic)(const pcpd_t *pcpd,
			       const pcpctx_t *ctx,
			       const uint8_t proto,
			       const uint8_t *intaddr,
			       const uint16_t intport,
			       const uint8_t *peeraddr,
			       const uint16_t peerport,
			       uint8_t *extaddr,
			       uint16_t *extport,
			       uint32_t *lifetime);
	int (*pcpd_filter)(const pcpd_t *pcpd,
			   const pcpctx_t *ctx,
			   const uint8_t proto,
			   const uint8_t *intaddr,
			   const uint16_t intport,
			   const uint8_t flags,
			   const uint8_t count,
			   const filter_t *filters);
	void (*pcpd_incrcounter)(const pcpd_t *pcpd,
				 const pcpctx_t *ctx,
				 const pcpstat_t code,
				 const uint8_t idx);
};

/*
 * pcpctx_t field minimum storage constraints:
 *	client: sizeof(struct sockaddr_storage)
 *	sid: PCPD_SID_SIZE
 *	request: PCPD_BUF_SIZE
 *	response: PCPD_BUF_SIZE
 * the PCPD_XXX_SIZE defines are stored in the corresponding lower case
 * globals to make the values used to compile a library object available.
 */

struct _pcpctx_ {
	void *opaque;			/* pointer to private stuff */
	struct sockaddr *client;	/* client address */
	uint8_t *sid;			/* subscriber ID */
	uint8_t *request;		/* request buffer */
	size_t off3rdparty;		/* offset to THIRD_PARTY, if present */
	uint8_t *response;		/* response buffer */
};

struct _filter_ {			/* FILTER */
	uint8_t plen;			/* prefix length */
	uint16_t port;			/* port (or zero) */
	uint8_t prefix[16];		/* prefix */
};	

enum _pcpstat_ {			/* PCP statistics */
	PCPS_OPERATION,			/* operation codes */
	PCPS_RESULT,			/* result codes */
	PCPS_OPTION			/* option codes */
};

/*
 * Routines
 */

extern int pcpd_cgn_init(pcpd_t *pcpd, int argc, char *argv[]);
extern int pcpd_cgn_close(pcpd_t *pcpd);
extern int pcpd_cgn_register(pcpd_t *pcpd,
			     void (*announce_callback) (const pcpd_t *pcpd,
							const pcpctx_t *ctx,
							const time_t epoch),
			     void (*update_callback) (const pcpd_t *pcpd,
						      const pcpctx_t *ctx,
						      const uint8_t protocol,
						      const uint8_t *intaddr,
						      const uint16_t intport,
						      const uint8_t *extaddr,
						      const uint16_t extport,
						      const uint32_t lifetime),
			     void (*uncache_callback) (const pcpd_t *pcpd,
						       const pcpctx_t *ctx));
extern int pcpd_lib_init(pcpd_t *pcpd);
extern int pcpd_lib_close(const pcpd_t *pcpd);
extern int pcpd_lib_runonce(const pcpd_t *pcpd, const pcpctx_t *ctx);
extern void pcpd_update_recycle(const pcpd_t *pcpd, const pcpctx_t *ctx,
				const size_t length, const uint8_t *extaddr,
				const uint16_t extport,
				const uint32_t lifetime);
extern void pcpd_update(const pcpd_t *pcpd, const pcpctx_t *ctx,
			const uint8_t protocol, const uint8_t *intaddr,
			const uint16_t intport, const uint8_t *extaddr,
			const uint16_t extport,  const uint32_t lifetime);
extern void pcpd_uncache(const pcpd_t *pcpd, const pcpctx_t *ctx);
extern void pcpd_announce(const pcpd_t *pcpd, const pcpctx_t *ctx,
			  const time_t epoch);

/*
 * Mapped addresses
 */

static const uint8_t pcpd_mapped[16] =
	{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	  0x00, 0x00, 0xff, 0xff, 0x00, 0x00, 0x00, 0x00 };

static inline void pcpd_mapaddr(uint8_t *addr)
{
	memcpy(addr + 12, addr, 4);
	memcpy(addr, pcpd_mapped, 12);
}

static inline int pcpd_unmapaddr(uint8_t *addr)
{
#ifdef __GNUC__
	if (__builtin_expect(memcmp(addr, pcpd_mapped, 12) == 0, 1)) {
#else
	if (memcmp(addr, pcpd_mapped, 12) == 0) {
#endif
		memcpy(addr, addr + 12, 4);
		memset(addr + 4, 0, 12);
		return 1;
	} else
		return 0;
}

/*
 * Protocol defines
 */

/* opcode values */

#define PCP_OPCODE_RESPONSE	128	/* response bit */
#define PCP_OPCODE_ANNOUNCE	0
#define PCP_OPCODE_MAP		1
#define PCP_OPCODE_PEER		2

/* options */

#define PCP_OPTION_THIRD_PARTY	1	/* third-party request */
#define PCP_OPTION_PREF_FAIL	2	/* prefer failure */
#define PCP_OPTION_FILTER	3	/* remote peer filter */
#define PCP_OPTION_OPTIONAL	128	/* flag bit */

/* offsets in the common header */
#define PCP_VERSION		0
#define PCP_OPCODE		1
#define PCP_RESULT_CODE		3
#define PCP_LIFETIME		4
#define PCP_EPOCH		8
#define PCP_CLIENT_ADDR		8
#define PCP_OPCODE_HDR		24

/* offsets in the MAP header */
#define PCP_MAP_PROTOCOL	(PCP_OPCODE_HDR + 0)
#define PCP_MAP_INTERNAL_PORT	(PCP_OPCODE_HDR + 4)
#define PCP_MAP_EXTERNAL_PORT	(PCP_OPCODE_HDR + 6)
#define PCP_MAP_EXTERNAL_ADDR	(PCP_OPCODE_HDR + 8)

/* offsets in the PEER header */
#define PCP_PEER_PROTOCOL	(PCP_OPCODE_HDR + 0)
#define PCP_PEER_INTERNAL_PORT	(PCP_OPCODE_HDR + 4)
#define PCP_PEER_EXTERNAL_PORT	(PCP_OPCODE_HDR + 6)
#define PCP_PEER_EXTERNAL_ADDR	(PCP_OPCODE_HDR + 8)
#define PCP_PEER_REMOTE_PORT	(PCP_OPCODE_HDR + 24)
#define PCP_PEER_REMOTE_ADDR	(PCP_OPCODE_HDR + 28)


/* offsets in option */
#define PCP_OPT_CODE		0
#define PCP_OPT_RSVD		1
#define PCP_OPT_LEN		2
#define PCP_OPT_DATA		4
#define PCP_OPTDFLT_PLEN	1
#define PCP_OPTDFLT_PORT	2
#define PCP_OPTDFLT_PREF	4

/* lengths */
#define PCP_LEN_ANNOUNCE	24
#define PCP_LEN_MAP		48
#define PCP_LEN_PEER		68
#define PCP_LEN_OPTHDR		4
#define PCP_LEN_OPT0		0
#define PCP_LEN_OPT3PTY		16
#define PCP_LEN_OPTFLT		20
#define PCP_MINLEN0		2
#define PCP_MINLEN1		24
#define PCP_MAXLEN		1024

#define PCP_ALIGN(x)		(((x) + 3) & ~3)

/* errors */

#define PCP_ERR_SUCCESS		0	/* success */
#define PCP_ERR_UNSUPVERSION	1	/* unsupported version */
#define PCP_ERR_NOTAUTH		2	/* not authorized */
#define PCP_ERR_BADREQUEST	3	/* malformed request */
#define PCP_ERR_UNSUPOPCODE	4	/* unsupported opcode */
#define PCP_ERR_UNSUPOPTION	5	/* unsupported option */
#define PCP_ERR_BADOPTION	6	/* malformed option */
#define PCP_ERR_NETFAILURE	7	/* network failure */
#define PCP_ERR_NORESOURCES	8	/* out of resources */
#define PCP_ERR_UNSUPPROTO	9	/* unsupported protocol */
#define PCP_ERR_EXQUOTA		10	/* user exceeded quota */
#define PCP_ERR_CANTPROVIDE	11	/* cannot provide external */
#define PCP_ERR_ADDRMISMATCH	12	/* address mismatch */
#define PCP_ERR_TOOMANYPEER	13	/* excessive number of remote peers */
#define PCP_ERR_PROCESSING	192	/* processing error */

/* others */

#define PCP_PORT		5351	/* default port of PCP service */

#define PCP_RC_PORT		5350
#define PCP_RC_ADDR4		"224.0.0.1"
#define PCP_RC_ADDR6		"FF02::1"
#define PCP_RC_INITTRMT		250000
#define PCP_RC_MAXTRMT		10

/*
 * API defines
 */

/* CGN types */

#define PCPD_TYPE_NAT		0		/* simple (not CGN) NAT44 */
#define PCPD_TYPE_DSLITE	1		/* DS-Lite AFTR */
#define PCPD_TYPE_NAT444	2		/* NAT444 CGN */
#define PCPD_TYPE_NAT64		3		/* NAT64 CGN */
#define PCPD_TYPE_NAT46		4		/* NAT46 CGN */
#define PCPD_TYPE_NPTv6		5		/* NPTv6 (aka NAT66) CGN */
#define PCPD_TYPE_FW4		6		/* IPv4 firewall */
#define PCPD_TYPE_FW6		7		/* IPv4 firewall */
#define PCPD_TYPE_L2NAT		8		/* layer2 (IPv4) CGN */

/* opcodes */

#define PCPD_OPCODE_ANNOUNCE	(1 << PCP_OPCODE_ANNOUNCE)
#define PCPD_OPCODE_MAP		(1 << PCP_OPCODE_MAP)
#define PCPD_OPCODE_PEER	(1 << PCP_OPCODE_PEER)

/* server flags */

#define PCPD_MAPSRC		0x1		/* mapped source */
#define PCPD_MAPINT		0x2		/* mapped internal */
#define PCPD_MAPEXT		0x4		/* mapped external */
#define PCPD_MAPDST		0x8		/* mapped destinations */
#define PCPD_OTHERPROTO		0x10		/* not only UDP and TCP */
#define PCPD_PORT0		0x20		/* port=0 is valid */
#define PCPD_ONEONE		0x40		/* support NAT 1:1 */
#define PCPD_FIREWALL		0x80		/* the CGN is a firewall */
#define PCPD_3RDPARTY		0x100		/* allow third party */
#define PCPD_3RDPARTY_CPE	0x200		/* ... only from 192.0.0.2 */
#define PCPD_3RDPARTY_AO	0x400		/* ... only when zero */
#define PCPD_ANNOUNCE		0x20000000	/* send unsolicited ANNOUNCE */
#define PCPD_UPDATE		0x40000000	/* send map updates */
#define PCPD_CACHE		0x80000000	/* cache subscribers */

#define PCPD_MAPALL	(PCPD_MAPSRC | PCPD_MAPINT | PCPD_MAPEXT | PCPD_MAPDST)

/* 16 <= PCPD_SID_SIZE <= 256 */

#ifndef PCPD_SID_SIZE
#define PCPD_SID_SIZE		128		/* size of subscriber ids */
#endif

extern size_t pcpd_sid_size;			/* as a global */

/* 1025 <= PCPD_BUF_SIZE <= 2048 */

#ifndef PCPD_BUF_SIZE
#define PCPD_BUF_SIZE		1080		/* size of packet buffers */
#endif

extern size_t pcpd_buf_size;			/* as a global */

/* clear a context */

static inline void ctxclear(const pcpctx_t *ctx)
{
	memset(ctx->client, 0, sizeof(struct sockaddr_storage));
	memset(ctx->sid, 0, pcpd_sid_size);
	memset(ctx->request, 0, pcpd_buf_size);
	memset(ctx->response, 0, pcpd_buf_size);
}

/* size (in octet) of the subscriber ID */

static inline int idsize(const pcpd_t *pcpd)
{
	switch (pcpd->cgntype) {
	case PCPD_TYPE_NAT:
		return 0;			/* no ID */
	case PCPD_TYPE_DSLITE:
	case PCPD_TYPE_NAT64:
	case PCPD_TYPE_NPTv6:
	case PCPD_TYPE_FW6:
		return 16;			/* IPv6 address */
	case PCPD_TYPE_L2NAT:
		return 6;			/* MAC address */
	case PCPD_TYPE_NAT444:
	case PCPD_TYPE_NAT46:
	case PCPD_TYPE_FW4:
	default:
		return 4;			/* IPv4 address */
	}
}

/* mapping flags */

#define PCPD_PREFER_FAILURE	1		/* prefer failure */
#define PCPD_CREATE_AS_LOCKED	2		/* create as locked */

/* filter flags */

#define PCPD_DELETE_FILTERS	1		/* delete filters */
#define PCPD_UNLOCK_ON_SUCCESS	2		/* unlock on success */

/* return codes */

#define PCPD_STATICMAPPING	(2)	/* already existing static mapping */
#define PCPD_NEWMAPPING		(1)	/* a new mapping was created */
#define PCPD_OK			(0)	/* OK */
#define PCPD_ERR_INVAL		(-1)	/* invalid arguments */
#define PCPD_ERR_PROTOBASE	(-100)	/* base for protocol errors */
#define PCPD_ERR_UNSUPVERSION	(-101)	/* unsupported version */
#define PCPD_ERR_NOTAUTH	(-102)	/* not authorized */
#define PCPD_ERR_BADREQUEST	(-103)	/* malformed request */
#define PCPD_ERR_UNSUPOPCODE	(-104)	/* unsupported opcode */
#define PCPD_ERR_UNSUPOPTION	(-105)	/* unsupported option */
#define PCPD_ERR_BADOPTION	(-106)	/* malformed option */
#define PCPD_ERR_NETFAILURE	(-107)	/* network failure */
#define PCPD_ERR_NORESOURCES	(-108)	/* out of resources */
#define PCPD_ERR_UNSUPPROTO	(-109)	/* unsupported protocol */
#define PCPD_ERR_EXQUOTA	(-110)	/* user exceeded quota */
#define PCPD_ERR_CANTPROVIDE	(-111)	/* cannot provide external */
#define PCPD_ERR_ADDRMISMATCH	(-112)	/* address mismatch */
#define PCPD_ERR_TOOMANYPEER	(-113)	/* excessive number of remote peers */
#define PCPD_ERR_PROCESSING	(-292)	/* processing error */

/* hold time from error */

static inline uint32_t errholdtime(const pcpd_t *pcpd, const uint8_t code)
{
	switch (code) {
	case PCP_ERR_NETFAILURE:
	case PCP_ERR_NORESOURCES:
	case PCP_ERR_EXQUOTA:
	case PCP_ERR_CANTPROVIDE:
	case PCP_ERR_PROCESSING:
		return pcpd->shorterrht;
	default:
		return pcpd->longerrht;
	}
}

/*
 * Tunables & co
 */

/* set to one to exit from the next main loop interation */

extern int terminate;

/* weak definition so you can overwrite them */

extern const char *pcpd_strerror(const int error);

extern void *pcpd_malloc(const pcpd_t *pcpd, size_t size);
extern void pcpd_free(const pcpd_t *pcpd, void *ptr);

extern void pcpd_openlog(const pcpd_t *pcpd,
			 const char *ident, int option, int facility);
extern void pcpd_vsyslog(const pcpd_t *pcpd,
			 int priority, const char *format, va_list ap);
extern void pcpd_closelog(const pcpd_t *pcpd);

extern time_t pcpd_time(const pcpd_t *pcpd);

extern const char *pcpd_inet_ntop(int af, const void *src,
				  char *dst, socklen_t size);

extern const char *pcpd_id2str(const pcpd_t *pcpd, const uint8_t *sid);

#ifdef __cplusplus
}
#endif

#endif
