/*
 * Copyright (C) 2012  Internet Systems Consortium, Inc. ("ISC")
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND ISC DISCLAIMS ALL WARRANTIES WITH
 * REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS.  IN NO EVENT SHALL ISC BE LIABLE FOR ANY SPECIAL, DIRECT,
 * INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
 * OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */

/* $Id$ */

/*
 * Stateless Deterministic (SD) NAT support code
 *
 * Francis_Dupont@isc.org, March 2012
 *
 * private 3-10 algorithm is:
 *  0bHHHHHHHNNNNNNLLL <-> base + 0b000000HHHHHHHLLL
 */

#include <stdio.h>
#include <string.h>
#include <netinet/in.h>
#include "sd.h"

enum sdalgo { SDALGO_NONE = 0, SDALGO_DIRECT, SDALGO_PRIV310 };

struct sdparams {
	enum sdalgo alg;		/* algorithm */
	unsigned short
		number,			/* subscriber number */
		tcpbase,		/* base for TCP */
		tcpmax,			/* maximum for TCP */
		udpbase,		/* base for UDP */
		udpmax;			/* maximum for UDP */
} sdparams;

/*
 * Parse SD parameters
 *
 * valid formats are:
 * "none"
 * "direct/tcp:<base>-<max>/udp:<base>-<max>/"
 * "p310/<number>/tcp:<base>-<max>/udp:<base>-<max>/"
 */

int
parsesd(const char *params)
{
	memset(&sdparams, 0, sizeof(sdparams));
	if ((params == NULL) || (params[0] == '\0') ||
	    (strcmp(params, "none") == 0) ||
	    (strcmp(params, "direct") == 0)) {
		sdparams.alg = SDALGO_NONE;
		return 0;
	} else if (sscanf(params,
			  "direct/tcp:%hu-%hu/udp:%hu-%hu/",
			  &sdparams.tcpbase, &sdparams.tcpmax,
			  &sdparams.udpbase, &sdparams.udpmax) == 4) {
		if ((sdparams.tcpbase < 1) || (sdparams.tcpbase > 65534) ||
		    (sdparams.udpbase < 1) || (sdparams.udpbase > 65534) ||
		    (sdparams.tcpmax < sdparams.tcpbase + 1) ||
		    (sdparams.udpmax < sdparams.udpbase + 1)) {
			fprintf(stderr,
				"direct parameter ranges are\n"
				"\tbase 1..65534\n"
				"\tmax base+1 .. 65535\n");
			return -1;
		}
		sdparams.alg = SDALGO_DIRECT;
		goto check;
	} else if (sscanf(params,
			  "p310/%hu/tcp:%hu-%hu/udp:%hu-%hu/",
			  &sdparams.number,
			  &sdparams.tcpbase, &sdparams.tcpmax,
			  &sdparams.udpbase, &sdparams.udpmax) == 5) {
		if ((sdparams.number < 1) || (sdparams.number > 63) ||
		    (sdparams.tcpbase < 1) || (sdparams.tcpbase > 64512) ||
		    (sdparams.udpbase < 1) || (sdparams.udpbase > 64512) ||
		    (sdparams.tcpmax < sdparams.tcpbase + 1) ||
		    (sdparams.udpmax < sdparams.udpbase + 1) ||
		    (sdparams.tcpmax > sdparams.tcpbase + 1024) ||
		    (sdparams.udpmax > sdparams.udpbase + 1024)) {
			fprintf(stderr,
				"p310 parameter ranges are\n"
				"\tnumber 1..63\n"
				"\tbase 1..64512\n"
				"\tmax base+1 .. base+1024\n");
			return -1;
		}
		sdparams.alg = SDALGO_PRIV310;
		goto check;
	}
	return -1;

    check:
	/* TODO: check compatibility with permissions */
	return 0;
}

/* Get the next valid local external port */

unsigned short
nexteport(unsigned short eport, int protocol)
{
	switch (sdparams.alg) {
	case SDALGO_DIRECT:
	case SDALGO_PRIV310:
		if (protocol == IPPROTO_TCP) {
			if (eport == 0)
				return sdparams.tcpbase;
			eport++;
			if (eport > sdparams.tcpmax)
				return sdparams.tcpbase;
		} else {
			if (eport == 0)
				return sdparams.udpbase;
			eport++;
			if (eport > sdparams.udpmax)
				return sdparams.udpbase;
		}
		break;
	default:
		eport++;
		if (eport == 0)
			return 1;
		break;
	}
	return eport;
}

/* Map a local external port to its Internet value */

unsigned short
sdmap(unsigned short eport, int protocol)
{
	unsigned short local;

	if (eport == 0)
		return eport;
	switch (sdparams.alg) {
	case SDALGO_PRIV310:
		local = eport;
		if (protocol == IPPROTO_TCP) {
			if ((local < sdparams.tcpbase) ||
			    (local > sdparams.tcpmax))
				break;
			local -= sdparams.tcpbase;
			eport = local & 0x0007;
			eport |= sdparams.number << 3;
			eport |= (local & 0xfff8) << 9;
		} else {
			if ((local < sdparams.udpbase) ||
			    (local > sdparams.udpmax))
				break;
			local -= sdparams.udpbase;
			eport = local & 0x0007;
			eport |= sdparams.number << 3;
			eport |= (local & 0xfff8) << 9;
		}
		break;
	default:
		break;
	}
	return eport;
}

/* Unmap an Internet external port to its local value */

unsigned short
sdunmap(unsigned short eport, int protocol, int *outofrange)
{
	unsigned short remote;

	if (outofrange)
		*outofrange = 0;
	if (eport == 0)
		return eport;
	switch (sdparams.alg) {
	case SDALGO_DIRECT:
		if (protocol == IPPROTO_TCP) {
			if ((eport < sdparams.tcpbase) ||
			    (eport > sdparams.tcpmax)) {
				if (outofrange)
					*outofrange = 1;
			}
		} else {
			if ((eport < sdparams.udpbase) ||
			    (eport > sdparams.udpmax)) {
				if (outofrange)
					*outofrange = 1;
			}
		}
		break;
	case SDALGO_PRIV310:
		if (((eport & 0x01f8) >> 3) != sdparams.number) {
			if (outofrange)
				*outofrange = 1;
			break;
		}
		remote = eport & 0x0007;
		remote |= (eport & 0xfe00) >> 9;
		if (protocol == IPPROTO_TCP) {
			remote += sdparams.tcpbase;
			if (remote > sdparams.tcpmax) {
				if (outofrange)
					*outofrange = 1;
				break;
			}
		} else {
			remote += sdparams.udpbase;
			if (remote > sdparams.udpmax) {
				if (outofrange)
					*outofrange = 1;
				break;
			}
		}
		eport = remote;
		break;
	default:
		break;
	}
	return eport;
}
