/* Shared library add-on to iptables to add ICMPv4 echo ID support. */
#include <stdio.h>
#include <netdb.h>
#include <string.h>
#include <stdlib.h>
#include <stddef.h>
#include <getopt.h>
#include <xtables.h>
#include "xt_echo.h"

#define ICMP_ECHO	8
#define ICMP_ECHOREPLY	0

static void echo_help(void)
{
	printf(
"echo match options:\n"
"[!] --id min[:max]             match ICMPv4 echo ID\n"
"    [--request|--reply]        apply to requests (default) or replies\n");
}

static const struct option echo_opts[] = {
	{ "id", 1, NULL, '1' },
	{ "request", 0, NULL, '2' },
	{ "reply", 0, NULL, '3' },
	{ .name = NULL }
};

static void echo_init(struct xt_entry_match *match)
{
	struct xt_echo_info *info = (void *)match->data;

	info->type = ICMP_ECHO;
	info->id[1] = 0xFFFF;
}

static u_int16_t parse_id(const char *id)
{
	unsigned int idnum;

	if (xtables_strtoui(id, NULL, &idnum, 0, 65535))
		return (u_int16_t)idnum;

	xtables_error(PARAMETER_PROBLEM,
		      "invalid ID `%s' specified", id);
}

static void parse_ids(const char *idstring, u_int16_t *ids)
{
	char *buffer;
	char *cp;

	buffer = strdup(idstring);
	if ((cp = strchr(buffer, ':')) == NULL)
		ids[0] = ids[1] = parse_id(buffer);
	else {
		*cp = '\0';
		cp++;

		ids[0] = buffer[0] ? parse_id(buffer) : 0;
		ids[1] = cp[0] ? parse_id(cp) : 0xFFFF;

		if (ids[0] > ids[1])
			xtables_error(PARAMETER_PROBLEM,
				      "invalid idrange (min > max)");
	}
	free(buffer);
}

static int echo_parse(int c, char **argv,
		      int invert, unsigned int *flags,
		      const void *entry,
		      struct xt_entry_match **match)
{
	struct xt_echo_info *info = (void *)(*match)->data;

	switch (c) {
	case '1':
		if (*flags & 0x1)
			xtables_error(PARAMETER_PROBLEM,
			      "--id may be given only once");
		parse_ids(optarg, info->id);
		if (invert)
			info->invflags |= XT_ECHO_INV_ID;
		*flags |= 0x1;
		break;
	case '2':
		if (*flags & 0x2)
			xtables_error(PARAMETER_PROBLEM,
			      "--request|reply may be given only once");
		info->type = ICMP_ECHO;
		*flags |= 0x2;
		break;
	case '3':
		if (*flags & 0x2)
			xtables_error(PARAMETER_PROBLEM,
			      "--request|reply may be given only once");
		info->type = ICMP_ECHOREPLY;
		*flags |= 0x2;
		break;
	default:
		return 0;
	}

	return 1;
}

static void echo_check(unsigned int flags)
{
	if (!(flags & 0x1))
		xtables_error(PARAMETER_PROBLEM,
			      "You must specify \"--id\"");
}

static void echo_print(const void *ip,
		       const struct xt_entry_match *match, int numeric)
{
	const struct xt_echo_info *info = (const void *)match->data;

	printf("echo ");
	if (info->type == ICMP_ECHO)
		printf("request ");
	else
		printf("reply ");
	if (info->invflags & XT_ECHO_INV_ID)
		printf("!");
	printf("ids:%u:%u ", info->id[0], info->id[1]);
	if (info->invflags & ~XT_ECHO_INV_MASK)
		printf("Unknown invflags: 0x%X ",
		       info->invflags & ~XT_ECHO_INV_MASK);
}

static void echo_save(const void *ip, const struct xt_entry_match *match)
{
	const struct xt_echo_info *info = (const void *)match->data;

	printf("%s--%s -id %u:%u ",
	       info->invflags & XT_ECHO_INV_ID ? "! " : "",
	       info->type == ICMP_ECHO ? "request" : "reply",
	       info->id[0], info->id[1]);
}

static struct xtables_match echo_match = {
	.name          = "echo",
	.family        = AF_INET,
	.version       = XTABLES_VERSION,
	.size          = XT_ALIGN(sizeof (struct xt_echo_info)),
	.userspacesize = XT_ALIGN(sizeof (struct xt_echo_info)),
	.help          = echo_help,
	.init          = echo_init,
	.parse         = echo_parse,
	.final_check   = echo_check,
	.print         = echo_print,
	.save          = echo_save,
	.extra_opts    = echo_opts,
};

void __attribute((constructor)) libxt_echo_init(void)
{
	xtables_register_match(&echo_match);
}
