/*
 * Copyright (C) 2012  Internet Systems Consortium, Inc. ("ISC")
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND ISC DISCLAIMS ALL WARRANTIES WITH
 * REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS.  IN NO EVENT SHALL ISC BE LIABLE FOR ANY SPECIAL, DIRECT,
 * INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
 * OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */

/* $Id$ */

#include <pcap/pcap.h>

#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <netinet/udp.h>

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

char errbuf[PCAP_ERRBUF_SIZE];
char filter[128];
pcap_t *pcap;
struct bpf_program prog;
struct pcap_pkthdr *phdr;
const u_char *pkt;
u_short prmin, prmax;

#define FILTER	"ip src %s and icmp[icmptype] = 3 and icmp[icmpcode] = 13"

void decode(void);

int
main(int argc, char *argv[])
{
	int cc;

	if (argc != 5) {
		fprintf(stderr,
			"usage: %s <if> <addr> <min> <max>\n",
			argv[0]);
		exit(-1);
	}
	prmin = (u_short) atoi(argv[3]);
	prmax = (u_short) atoi(argv[4]);

	pcap = pcap_create(argv[1], errbuf);
	if (pcap == NULL) {
		fprintf(stderr, "pcap_create: %s\n", errbuf);
		exit(-1);
	}
	if (pcap_set_snaplen(pcap, 68) != 0) {
		fprintf(stderr, "pcap_set_snaplen(68)\n");
		exit(-1);
	}
	if (pcap_set_promisc(pcap, 0) != 0) {
		fprintf(stderr, "pcap_set_promisc(0)\n");
		exit(-1);
	}
	if (pcap_activate(pcap) != 0) {
		fprintf(stderr, "pcap_activate: %s\n", pcap_geterr(pcap));
		exit(-1);
	}
	cc = pcap_datalink(pcap);
	if (cc != DLT_LINUX_SLL) {
		fprintf(stderr,
			"pcap_datalink: %d (!= %d)\n",
			cc,
			DLT_LINUX_SLL);
		exit(-1);
	}
	cc = snprintf(filter, sizeof(filter), FILTER, argv[2]);
	if (cc < 0) {
		fprintf(stderr, "snprintf: failed\n");
		exit(-1);
	}
	if (cc >= (int) sizeof(filter)) {
		fprintf(stderr, "snprintf: truncated\n");
		exit(-1);
	}
	if (pcap_compile(pcap, &prog, filter, 1, PCAP_NETMASK_UNKNOWN) != 0) {
		fprintf(stderr, "pcap_compile: %s\n", pcap_geterr(pcap));
		exit(-1);
	}
	if (pcap_setfilter(pcap, &prog) != 0) {
		fprintf(stderr, "pcap_setfilter: %s\n", pcap_geterr(pcap));
		exit(-1);
	}
	pcap_freecode(&prog);
	if (pcap_setdirection(pcap, PCAP_D_IN) != 0) {
		fprintf(stderr, "pcap_setdirection: %s\n", pcap_geterr(pcap));
		exit(-1);
	}

	for (;;) {
		cc = pcap_next_ex(pcap, &phdr, &pkt);
		switch (cc) {
		case 1:
			/* good */
			decode();
			break;
		case 0:
			/* timeout */
			break;
		default:
			/* error */
			fprintf(stderr,
				"pcap_next_ex: %s\n",
				pcap_geterr(pcap));
			exit(-1);
		}
	}

	pcap_close(pcap);
	exit(0);
}

void
decode(void)
{
	const struct iphdr *i, *t;
	const struct udphdr *u;
	size_t l;

///
	fprintf(stderr, "got\n");
	if ((phdr->len < 68) || (phdr->caplen < 68))
		return;
	i = (const struct iphdr *) (pkt + 16);
	l = (i->ihl << 2) + sizeof(struct icmphdr);
	t = (const struct iphdr *) ((u_char *) i + l);
	if ((t->version != 4) ||
	    (t->ihl != 5) ||
	    ((t->frag_off & htons(0x1fff)) != 0) ||
	    ((t->protocol != IPPROTO_TCP) && (t->protocol != IPPROTO_UDP)))
		return;
	u = (const struct udphdr *) (t + 1);
	if ((ntohs(u->source) < prmin) || (ntohs(u->source) > prmax))
		return;
	fprintf(stderr, "OUT OF RANGE!!!\n");
	exit(1);
}
