/*
 * Copyright (C) 2010-2012  Internet Systems Consortium, Inc. ("ISC")
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND ISC DISCLAIMS ALL WARRANTIES WITH
 * REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS.  IN NO EVENT SHALL ISC BE LIABLE FOR ANY SPECIAL, DIRECT,
 * INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
 * OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 * Routines for Port Control Protocol packet disassembly.
 * draft-ietf-pcp-base-21.txt
 */

/* $Id$ */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <tcpdump-stdinc.h>

#include <stdio.h>

#include "pcp.h"
#include "interface.h"
#include "addrtoname.h"
#include "extract.h"
#include "ipproto.h"

static const uint8_t mapped[12] = {
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xff, 0xff
};

static const uint8_t zero[16];

const char *
nat_pmp_errcode(u_int16_t errcode)
{
	static char buf[64];

	switch (errcode) {
	case NAT_PMP_SUCCESS:
		return "success";
	case NAT_PMP_BADVERSION:
		return "unsupported version";
	case NAT_PMP_NOTAUTH:
		return "not authorized/refused";
	case NAT_PMP_NETFAIL:
		return "network failure";
	case NAT_PMP_RESOURCES:
		return "out of resources";
	case NAT_PMP_BADOPCODE:
		return "unsupported opcode";
	default:
		(void) snprintf(buf, sizeof(buf),
				"unknown(%u)", (u_int) errcode);
		return (const char *)buf;
	}
}

void
nat_pmp_print(const u_char *dat, u_int length)
{
	const u_char *ptr = dat;
	u_int16_t errcode;

	TCHECK2(*ptr, 2);
	printf(" nat-pmp:");
	ptr++;
	switch (*ptr) {
	case NAT_PMP_ADDRESS:
		printf(" address request");
		break;
	case NAT_PMP_UDP_PORT:
	case NAT_PMP_TCP_PORT:
		if (*ptr == NAT_PMP_UDP_PORT)
			printf(" UDP");
		else
			printf(" TCP");
		printf(" port request [");
		ptr++;
		TCHECK2(*ptr, 2);
		ptr += 2;
		TCHECK2(*ptr, 2);
		printf("priv=%u", EXTRACT_16BITS(ptr));
		ptr += 2;
		TCHECK2(*ptr, 2);
		printf(",pub=%u", EXTRACT_16BITS(ptr));
		ptr += 2;
		TCHECK2(*ptr, 4);
		printf(",lifetime=%u]", EXTRACT_32BITS(ptr));
		ptr += 4;
		break;
	case NAT_PMP_ADDRESS | NAT_PMP_RESPONSE:
		printf(" address response [");
		ptr++;
		TCHECK2(*ptr, 2);
		errcode = EXTRACT_16BITS(ptr);
		if (errcode != 0) {
			printf(" error: %s]", nat_pmp_errcode(errcode));
			return;
		}
		ptr += 2;
		TCHECK2(*ptr, 4);
		printf("sssoe=%u", EXTRACT_32BITS(ptr));
		ptr += 4;
		TCHECK2(*ptr, 4);
		printf(",external=%s]", intoa(EXTRACT_32BITS(ptr)));
		ptr += 4;
		break;
	case NAT_PMP_UDP_PORT | NAT_PMP_RESPONSE:
	case NAT_PMP_TCP_PORT | NAT_PMP_RESPONSE:
		if (*ptr == NAT_PMP_UDP_PORT)
			printf(" UDP");
		else
			printf(" TCP");
		printf(" port response [");
		ptr++;
		TCHECK2(*ptr, errcode);
		errcode = EXTRACT_16BITS(ptr);
		if (errcode != 0) {
			printf(" error: %s]", nat_pmp_errcode(errcode));
			return;
		}
		ptr += 2;
		TCHECK2(*ptr, 4);
		printf("sssoe=%u", EXTRACT_32BITS(ptr));
		ptr += 4;
		TCHECK2(*ptr, 2);
		printf(",priv=%u", EXTRACT_16BITS(ptr));
		ptr += 2;
		TCHECK2(*ptr, 2);
		printf(",pub=%u", EXTRACT_16BITS(ptr));
		ptr += 2;
		TCHECK2(*ptr, 4);
		printf(",lifetime=%u]", EXTRACT_32BITS(ptr));
		ptr += 4;
		break;
	default:
		printf("unknown opcode(%u)", (u_int) *ptr);
		return;
	}
	return;

 trunc:
	printf("[|nat-pmp]");
}

const char *
pcp_errcode(u_int8_t errcode)
{
	static char buf[64];

	switch (errcode) {
	case PCP_RC_SUCCESS:
		return "success";
	case PCP_RC_UNSUPP_VERSION:
		return "unsupported version";
	case PCP_RC_NOT_AUTHORIZED:
		return "not authorized";
	case PCP_RC_MALFORMED_REQUEST:
		return "malformed request";
	case PCP_RC_UNSUPP_OPCODE:
		return "unsupported opcode";
	case PCP_RC_UNSUPP_OPTION:
		return "unsupported option";
	case PCP_RC_MALFORMED_OPTION:
		return "malformed option";
	case PCP_RC_NETWORK_FAILURE:
		return "network failure";
	case PCP_RC_NO_RESOURCES:
		return "out of resources";
	case PCP_RC_UNSUPP_PROTOCOL:
		return "unsupported protocol";
	case PCP_RC_USER_EX_QUOTA:
		return "user exceeded quota";
	case PCP_RC_CANT_PROVIDE_EXT:
		return "cannot provide external";
	case PCP_RC_ADDRESS_MISMATCH:
		return "address mismatch";
	case PCP_RC_EXCESSIVE_REMOTE_PEERS:
		return "excessive remote peers";
	default:
		(void) snprintf(buf, sizeof(buf),
				"unknown(%u)", (u_int) errcode);
		return (const char *)buf;
	}
}

void
pcp_print(const u_char *dat, u_int length)
{
	const u_char *ptr = dat;
	const u_char *endptr = dat + length;
	u_char *p;
	u_int8_t opcode, proto, opt;
	u_int16_t pport;
	u_int olen;
	int announce = 0, peer = 0;

	TCHECK2(*ptr, 1);

	switch (*ptr) {
	case NAT_PMP_VERSION:
		printf(" NAT-PMP");
		nat_pmp_print(ptr, length);
		return;
	case PCP_VERSION:
		printf(" PCP");
		break;
	default:
		printf(" PCP: unknown version(%u)", (u_int) *ptr);
		return;
	}
	ptr++;

	TCHECK2(*ptr, 4);
	opcode = *ptr;
	if ((opcode & PCP_RESPONSE) == 0)
		printf(" request");
	else
		printf(" response");
	switch (opcode & ~PCP_RESPONSE) {
	case PCP_OC_ANNOUNCE:
		printf(" announce");
		announce = 1;
		break;
	case PCP_OC_MAP:
		printf(" map");
		break;
	case PCP_OC_PEER:
		printf(" peer");
		peer = 1;
		break;
	default:
		printf(" unknown(%u)", (u_int) opcode);
		return;
	}
	ptr++;

	if ((opcode & PCP_RESPONSE) && (*ptr != PCP_RC_SUCCESS))
		printf(",error=%s", pcp_errcode(ptr[1]));
	ptr += 2;

	TCHECK2(*ptr, 4);
	if ((announce == 0) || vflag)
		printf(",lifetime=%u", EXTRACT_32BITS(ptr));
	ptr += 4;

	if ((opcode & PCP_RESPONSE) == 0) {
		TCHECK2(*ptr, 16);
		if (memcmp(ptr, mapped, 12) == 0)
			printf(",internal=%s", ipaddr_string(ptr + 12));
		else
			printf(",internal=%s", ip6addr_string(ptr));
	} else {
		TCHECK2(*ptr, 4);
		printf(",epoch=%u", EXTRACT_32BITS(ptr));
	}
	ptr += 16;

	if (announce != 0)
		goto options;

	TCHECK(*ptr);
	proto = *ptr;
	if (proto == PCP_PROTO_ALL)
		printf(",proto=all");
	else
		printf(",proto=%s",
		       tok2str(ipproto_values, "unknown", (int) proto));
	ptr += 4;
	TCHECK2(*ptr, 4);
	printf(",priv=%u", EXTRACT_16BITS(ptr));
	ptr += 2;
	printf(",pub=%u", EXTRACT_16BITS(ptr));
	ptr += 2;
	TCHECK2(*ptr, 16);
	if (memcmp(ptr, mapped, 12) == 0)
		printf(",external=%s", ipaddr_string(ptr + 12));
	else if (memcmp(ptr, zero, 16) != 0)
		printf(",external=%s", ip6addr_string(ptr));
	ptr += 16;

	if (peer) {
		TCHECK2(*ptr, 2);
		pport = EXTRACT_16BITS(ptr);
		ptr += 4;
		TCHECK2(*ptr, 16);
		if (memcmp(ptr, mapped, 12) == 0)
		  printf(",peer=%s:%u", ipaddr_string(ptr + 12), pport);
		else
		  printf(",peer=%s:%u", ip6addr_string(ptr), pport);
		ptr += 16;
	}

    options:
	while (ptr < endptr) {
		TCHECK2(*ptr, 4);
		opt = *ptr;
		ptr += 2;
		olen = EXTRACT_16BITS(ptr);
		ptr += 2;
		printf("\n\toption:");
		switch (opt) {
		case PCP_OPT_FILTER:
			printf(" filter");
			break;
		case PCP_OPT_PREFER_FAILURE:
			printf(" prefer failure");
			break;
		case PCP_OPT_THIRD_PARTY:
			printf(" third party");
			break;
		default:
			printf(" unknown(%u)", (u_int) opt);
			break;
		}

		TCHECK2(*ptr, (olen + 3) & ~3);
		switch (opt) {
		case PCP_OPT_FILTER:
			printf(" pfxlen=%u", (int) *(ptr + 1));
			printf(" port=%u", EXTRACT_16BITS(ptr + 2));
			if (olen != 20) {
				printf(" data? %u", olen);
				break;
			}
			if (memcmp(ptr + 4, mapped, 12) == 0)
				printf(" addr=%s", ipaddr_string(ptr + 16));
			else
				printf(" addr=%s", ip6addr_string(ptr + 4));
			break;
		case PCP_OPT_PREFER_FAILURE:
			break;
		case PCP_OPT_THIRD_PARTY:
			if (olen != 16) {
				printf(" data? %u", olen);
				break;
			}
			if (memcmp(ptr, mapped, 12) == 0)
				printf(" addr=%s", ipaddr_string(ptr + 12));
			else
				printf(" addr=%s", ip6addr_string(ptr));
			break;
		default:
			if (olen > 0)
				printf(" data? %u", olen);
			break;
		}
		ptr += (olen + 3) & ~3;
	}
	return;

 trunc:
	printf("[|pcp]");
}
