/*
 * Copyright (C) 2011-2012  Internet Systems Consortium, Inc. ("ISC")
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND ISC DISCLAIMS ALL WARRANTIES WITH
 * REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS.  IN NO EVENT SHALL ISC BE LIABLE FOR ANY SPECIAL, DIRECT,
 * INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
 * OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */

/*
 * C/C++ PCP client library unit tests
 *
 * Francis_Dupont@isc.org, November 2011
 *
 * Usage examples:
 *	libpcp_test
 *	libpcp_test --help
 *	libpcp_test --gtest_list_tests
 *	libpcp_test --gtest_filter="*MAP*"
 *	...
 *
 * when compiled with CHECK_DECODE_LINE defined, matches the decoding
 * error cases with the corresponding line in the pcp.c code
 * (i.e., checking the pcp_debug_line value)
 *
 * Defines:
 *	SERVER4		"192.0.0.1"
 *	SERVER6		"2001:0:0:1::1"
 *	ADDR6		"2a01:e35:2ec7:83d0:cb3a:14ff:fe1f:3b06"	
 *	BADADDR4	"1.22.77.199"
 *	BADADDR6	"1::bad"
 *	INTSAMPLE	"2001::1234:fedc"
 *	EXTSAMPLE	"2123::fedc:1234"
 *	PEERSAMPLE	"2321::4321:cdef"
 */

/* $Id$ */

#include <string>
#include <stdint.h>
#ifdef WIN32
#include <winsock2.h>
#include <ws2tcpip.h>
#include <process.h>
#else
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#endif
#include <pcp.h>
#include <gtest/gtest.h>

#ifndef SERVER4
#define SERVER4		"192.0.0.1"
#endif
#ifndef SERVER6
#define SERVER6		"2001:0:0:1::1"
#endif
#ifndef ADDR6
#define ADDR6		"2a01:e35:2ec7:83d0:cb3a:14ff:fe1f:3b06"
#endif
#ifndef BADADDR4
#define BADADDR4	"1.22.77.199"
#endif
#ifndef BADADDR6
#define BADADDR6	"1::bad"
#endif
#ifndef INTSAMPLE
#define INTSAMPLE	"2001::1234:fedc"
#endif
#ifndef EXTSAMPLE
#define EXTSAMPLE	"2123::fedc:1234"
#endif
#ifndef PEERSAMPLE
#define PEERSAMPLE	"2321::4321:cdef"
#endif

uint8_t zero[16];
uint8_t buf[1080];

#ifdef WIN32
int
gettimeofday(struct timeval* tp, void* tzp)
{
	DWORD t;

	t = timeGetTime();
	tp->tv_sec = t / 1000;
	tp->tv_usec = t % 1000;
	return 0;
}
#endif

namespace {

class InitTest : public ::testing::Test {
    protected:
	void SetUp() {
		memset(&pcp, 0, sizeof(pcp));
	}

	void TearDown() {
		(void) pcp_close(&pcp);
	}

	pcp_t pcp;
};

TEST_F(InitTest, NoPCP) {
	int ret;

	ret = pcp_init(NULL, NULL, NULL);
	EXPECT_EQ(ret, PCP_ERR_INVAL);
}

TEST_F(InitTest, NoServer) {
	int ret;

	ret = pcp_init(NULL, NULL, &pcp);
	EXPECT_EQ(ret, PCP_ERR_INVAL);
}

TEST_F(InitTest, FamilyMismatch) {
	struct sockaddr_storage server, source;
	struct sockaddr *srv, *src;
	int ret;

	srv = (struct sockaddr *) &server;
	src = (struct sockaddr *) &source;

	memset(&server, 0, sizeof(server));
	memset(&source, 0, sizeof(source));
	server.ss_family = 1;
	source.ss_family = 2;
	ret = pcp_init(srv, src, &pcp);
	EXPECT_EQ(ret, PCP_ERR_INVAL);
}

TEST_F(InitTest, IPv4Server) {
	struct sockaddr_storage server;
	struct sockaddr *srv;
	struct sockaddr_in *srv4;
	int ret;

	srv = (struct sockaddr *) &server;
	srv4 = (struct sockaddr_in *) &server;

	memset(&server, 0, sizeof(server));
	srv4->sin_family = AF_INET;
	ret = inet_pton(AF_INET, SERVER4, &srv4->sin_addr);
	ASSERT_EQ(ret, 1);
	ret = pcp_init(srv, NULL, &pcp);
	EXPECT_EQ(ret, PCP_OK);
}

TEST_F(InitTest, IPv6Server) {
	struct sockaddr_storage server;
	struct sockaddr *srv;
	struct sockaddr_in6 *srv6;
	int ret;

	srv = (struct sockaddr *) &server;
	srv6 = (struct sockaddr_in6 *) &server;

	memset(&server, 0, sizeof(server));
	srv6->sin6_family = AF_INET6;
#ifdef BSD
	srv6->sin6_len = sizeof(*srv6);
#endif
	ret = inet_pton(AF_INET6, SERVER6, &srv6->sin6_addr);
	ASSERT_EQ(ret, 1);
	ret = pcp_init(srv, NULL, &pcp);
	EXPECT_EQ(ret, PCP_OK);
}

TEST_F(InitTest, BadIPv4Source) {
	struct sockaddr_storage server, source;
	struct sockaddr *srv, *src;
	struct sockaddr_in *srv4, *src4;
	int ret;

	srv = (struct sockaddr *) &server;
	src = (struct sockaddr *) &source;
	srv4 = (struct sockaddr_in *) &server;
	src4 = (struct sockaddr_in *) &source;

	memset(&server, 0, sizeof(server));
	srv4->sin_family = AF_INET;
	ret = inet_pton(AF_INET, SERVER4, &srv4->sin_addr);
	ASSERT_EQ(ret, 1);
	memset(&source, 0, sizeof(source));
	src4->sin_family = AF_INET;
	// expect the addres of the Node-Under-Test is not 1.22.77.199!
	ret = inet_pton(AF_INET, BADADDR4, &src4->sin_addr);
	ASSERT_EQ(ret, 1);
	ret = pcp_init(srv, src, &pcp);
	EXPECT_EQ(ret, PCP_ERR_BIND);
}

TEST_F(InitTest, BadIPv6Source) {
	struct sockaddr_storage server, source;
	struct sockaddr *srv, *src;
	struct sockaddr_in6 *srv6, *src6;
	int ret;

	srv = (struct sockaddr *) &server;
	src = (struct sockaddr *) &source;
	srv6 = (struct sockaddr_in6 *) &server;
	src6 = (struct sockaddr_in6 *) &source;

	memset(&server, 0, sizeof(server));
	srv6->sin6_family = AF_INET6;
#ifdef BSD
	srv6->sin6_len = sizeof(*srv6);
#endif
	ret = inet_pton(AF_INET6, SERVER6, &srv6->sin6_addr);
	ASSERT_EQ(ret, 1);
	memset(&source, 0, sizeof(source));
	src6->sin6_family = AF_INET6;
#ifdef BSD
	src6->sin6_len = sizeof(*src6);
#endif
	// expect the address of the Node-Under-Test is not 1::bad
	ret = inet_pton(AF_INET6, BADADDR6, &src6->sin6_addr);
	ASSERT_EQ(ret, 1);
	ret = pcp_init(srv, src, &pcp);
	EXPECT_EQ(ret, PCP_ERR_BIND);
}

TEST_F(InitTest, IPv4Source) {
	struct sockaddr_storage server, source;
	struct sockaddr *srv, *src;
	struct sockaddr_in *srv4, *src4;
	int ret;

	srv = (struct sockaddr *) &server;
	src = (struct sockaddr *) &source;
	srv4 = (struct sockaddr_in *) &server;
	src4 = (struct sockaddr_in *) &source;

	memset(&server, 0, sizeof(server));
	srv4->sin_family = AF_INET;
	ret = inet_pton(AF_INET, SERVER4, &srv4->sin_addr);
	ASSERT_EQ(ret, 1);
	memset(&source, 0, sizeof(source));
	src4->sin_family = AF_INET;
	ret = inet_pton(AF_INET, "0.0.0.0", &src4->sin_addr);
	ASSERT_EQ(ret, 1);
	ret = pcp_init(srv, src, &pcp);
	EXPECT_EQ(ret, PCP_OK);
}

TEST_F(InitTest, IPv6Source) {
	struct sockaddr_storage server, source;
	struct sockaddr *srv, *src;
	struct sockaddr_in6 *srv6, *src6;
	int ret;

	srv = (struct sockaddr *) &server;
	src = (struct sockaddr *) &source;
	srv6 = (struct sockaddr_in6 *) &server;
	src6 = (struct sockaddr_in6 *) &source;

	memset(&server, 0, sizeof(server));
	srv6->sin6_family = AF_INET6;
#ifdef BSD
	srv6->sin6_len = sizeof(*srv6);
#endif
	ret = inet_pton(AF_INET6, SERVER6, &srv6->sin6_addr);
	ASSERT_EQ(ret, 1);
	memset(&source, 0, sizeof(source));
	src6->sin6_family = AF_INET6;
#ifdef BSD
	src6->sin6_len = sizeof(*src6);
#endif
	ret = inet_pton(AF_INET6, "::", &src6->sin6_addr);
	ASSERT_EQ(ret, 1);
	ret = pcp_init(srv, src, &pcp);
	EXPECT_EQ(ret, PCP_OK);
}

}

namespace {

class CloseTest : public ::testing::Test {
    protected:
	void SetUp() {
		memset(&pcp, 0, sizeof(pcp));
	}

	void TearDown() {
		(void) pcp_close(&pcp);
	}

	pcp_t pcp;
};

TEST_F(CloseTest, NoPCP) {
	int ret;

	ret = pcp_close(NULL);
	EXPECT_EQ(ret, PCP_ERR_INVAL);
}

TEST_F(CloseTest, NotOpened) {
	int ret;

	ret = pcp_close(&pcp);
	EXPECT_EQ(ret, PCP_OK);
}

TEST_F(CloseTest, DoubleClose) {
	int ret;

	(void) pcp_close(&pcp);
	ret = pcp_close(&pcp);
	EXPECT_EQ(ret, PCP_OK);
}

TEST_F(CloseTest, AlreadyClosedSocket) {
	struct sockaddr_storage server;
	struct sockaddr *srv;
	struct sockaddr_in *srv4;
	int ret, fd;

	srv = (struct sockaddr *) &server;
	srv4 = (struct sockaddr_in *) &server;

	memset(&server, 0, sizeof(server));
	srv4->sin_family = AF_INET;
	ret = inet_pton(AF_INET, SERVER4, &srv4->sin_addr);
	ASSERT_EQ(ret, 1);
	ret = pcp_init(srv, NULL, &pcp);
	ASSERT_EQ(ret, PCP_OK);
#ifdef WIN32
	fd = INVALID_SOCKET;
#else
	fd = -1;
#endif
	ret = pcp_getsocket(&pcp, &fd);
	ASSERT_EQ(ret, PCP_OK);
#ifdef WIN32
	ASSERT_NE(fd, INVALID_SOCKET);
	(void) closesocket(fd);
#else
	ASSERT_GE(fd, 0);
	(void) close(fd);
#endif
	ret = pcp_close(&pcp);
	EXPECT_EQ(ret, PCP_ERR_SYSCALL);
}

}

namespace {

class GetSocketTest : public ::testing::Test {
    protected:
	void SetUp() {
		memset(&pcp, 0, sizeof(pcp));
	}

	void TearDown() {
		(void) pcp_close(&pcp);
	}

	pcp_t pcp;
};

TEST_F(GetSocketTest, NoPCP) {
	int ret, fd;

	ret = pcp_getsocket(NULL, &fd);
	EXPECT_EQ(ret, PCP_ERR_INVAL);
}

TEST_F(GetSocketTest, NoSocket) {
	int ret;

	ret = pcp_getsocket(&pcp, NULL);
	EXPECT_EQ(ret, PCP_ERR_INVAL);
}

TEST_F(GetSocketTest, NotInitialized) {
	int ret, fd;

	ret = pcp_getsocket(&pcp, &fd);
	EXPECT_EQ(ret, PCP_ERR_INVAL);
}

TEST_F(GetSocketTest, Closed) {
	struct sockaddr_storage server;
	struct sockaddr *srv;
	struct sockaddr_in *srv4;
	int ret, fd;

	srv = (struct sockaddr *) &server;
	srv4 = (struct sockaddr_in *) &server;

	memset(&server, 0, sizeof(server));
	srv4->sin_family = AF_INET;
	ret = inet_pton(AF_INET, SERVER4, &srv4->sin_addr);
	ASSERT_EQ(ret, 1);
	ret = pcp_init(srv, NULL, &pcp);
	ASSERT_EQ(ret, PCP_OK);
	ret = pcp_close(&pcp);
	EXPECT_EQ(ret, PCP_OK);
	ret = pcp_getsocket(&pcp, &fd);
	EXPECT_EQ(ret, PCP_ERR_INVAL);
}

TEST_F(GetSocketTest, Get) {
	struct sockaddr_storage server;
	struct sockaddr *srv;
	struct sockaddr_in *srv4;
	int ret, fd;

	srv = (struct sockaddr *) &server;
	srv4 = (struct sockaddr_in *) &server;

	memset(&server, 0, sizeof(server));
	srv4->sin_family = AF_INET;
	ret = inet_pton(AF_INET, SERVER4, &srv4->sin_addr);
	ASSERT_EQ(ret, 1);
	ret = pcp_init(srv, NULL, &pcp);
	ASSERT_EQ(ret, PCP_OK);
#ifdef WIN32
	fd = INVALID_SOCKET;
#else
	fd = -1;
#endif
	ret = pcp_getsocket(&pcp, &fd);
	EXPECT_EQ(ret, PCP_OK);
#ifdef WIN32
	ASSERT_NE(fd, INVALID_SOCKET);
#else
	ASSERT_GE(fd, 0);
#endif
}

}

namespace {

class ServerAddressTest : public ::testing::Test {
    protected:
	void SetUp() {
		memset(&pcp, 0, sizeof(pcp));
	}

	void TearDown() {
		(void) pcp_close(&pcp);
	}

	pcp_t pcp;
};

TEST_F(ServerAddressTest, IPv4Address) {
	struct sockaddr_storage server, addr;
	struct sockaddr *srv, *a;
	struct sockaddr_in *srv4, *a4;
	int ret, fd;
	socklen_t len;
	char host[INET_ADDRSTRLEN];

	srv = (struct sockaddr *) &server;
	a = (struct sockaddr *) &addr;
	srv4 = (struct sockaddr_in *) &server;
	a4 = (struct sockaddr_in *) &addr;

	memset(&server, 0, sizeof(server));
	srv4->sin_family = AF_INET;
	ret = inet_pton(AF_INET, SERVER4, &srv4->sin_addr);
	ASSERT_EQ(ret, 1);
	ret = pcp_init(srv, NULL, &pcp);
	ASSERT_EQ(ret, PCP_OK);
#ifdef WIN32
	fd = INVALID_SOCKET;
#else
	fd = -1;
#endif
	ret = pcp_getsocket(&pcp, &fd);
	EXPECT_EQ(ret, PCP_OK);
#ifdef WIN32
	ASSERT_NE(fd, INVALID_SOCKET);
#else
	ASSERT_GE(fd, 0);
#endif
	memset(&addr, 0, sizeof(addr));
	len = sizeof(addr);
	ret = getpeername(fd, a, &len);
	ASSERT_EQ(ret, 0);
	ASSERT_EQ(a4->sin_family, AF_INET);
	EXPECT_EQ(len, sizeof(*a4));
	memset(host, 0, sizeof(host));
	const char *n = inet_ntop(AF_INET, &a4->sin_addr, host, sizeof(host));
	ASSERT_TRUE(n != NULL);
	ret = strcmp(n, SERVER4);
	EXPECT_EQ(ret, 0);
}

TEST_F(ServerAddressTest, IPv4Port) {
	struct sockaddr_storage server, addr;
	struct sockaddr *srv, *a;
	struct sockaddr_in *srv4, *a4;
	int ret, fd;
	socklen_t len;

	srv = (struct sockaddr *) &server;
	a = (struct sockaddr *) &addr;
	srv4 = (struct sockaddr_in *) &server;
	a4 = (struct sockaddr_in *) &addr;

	memset(&server, 0, sizeof(server));
	srv4->sin_family = AF_INET;
	ret = inet_pton(AF_INET, SERVER4, &srv4->sin_addr);
	ASSERT_EQ(ret, 1);
	srv4->sin_port = htons(1234);
	ret = pcp_init(srv, NULL, &pcp);
	ASSERT_EQ(ret, PCP_OK);
#ifdef WIN32
	fd = INVALID_SOCKET;
#else
	fd = -1;
#endif
	ret = pcp_getsocket(&pcp, &fd);
	EXPECT_EQ(ret, PCP_OK);
#ifdef WIN32
	ASSERT_NE(fd, INVALID_SOCKET);
#else
	ASSERT_GE(fd, 0);
#endif
	memset(&addr, 0, sizeof(addr));
	len = sizeof(addr);
	ret = getpeername(fd, a, &len);
	ASSERT_EQ(ret, 0);
	ASSERT_EQ(a4->sin_family, AF_INET);
	EXPECT_EQ(len, sizeof(*a4));
	EXPECT_EQ(ntohs(a4->sin_port), 1234);
}

TEST_F(ServerAddressTest, IPv4DefaultPort) {
	struct sockaddr_storage server, addr;
	struct sockaddr *srv, *a;
	struct sockaddr_in *srv4, *a4;
	int ret, fd;
	socklen_t len;

	srv = (struct sockaddr *) &server;
	a = (struct sockaddr *) &addr;
	srv4 = (struct sockaddr_in *) &server;
	a4 = (struct sockaddr_in *) &addr;

	memset(&server, 0, sizeof(server));
	srv4->sin_family = AF_INET;
	ret = inet_pton(AF_INET, SERVER4, &srv4->sin_addr);
	ASSERT_EQ(ret, 1);
	ret = pcp_init(srv, NULL, &pcp);
	ASSERT_EQ(ret, PCP_OK);
#ifdef WIN32
	fd = INVALID_SOCKET;
#else
	fd = -1;
#endif
	ret = pcp_getsocket(&pcp, &fd);
	EXPECT_EQ(ret, PCP_OK);
#ifdef WIN32
	ASSERT_NE(fd, INVALID_SOCKET);
#else
	ASSERT_GE(fd, 0);
#endif
	memset(&addr, 0, sizeof(addr));
	len = sizeof(addr);
	ret = getpeername(fd, a, &len);
	ASSERT_EQ(ret, 0);
	ASSERT_EQ(a4->sin_family, AF_INET);
	EXPECT_EQ(len, sizeof(*a4));
	EXPECT_EQ(ntohs(a4->sin_port), PCP_PORT);
}

TEST_F(ServerAddressTest, IPv6Address) {
	struct sockaddr_storage server, addr;
	struct sockaddr *srv, *a;
	struct sockaddr_in6 *srv6, *a6;
	int ret, fd;
	socklen_t len;
	char host[INET6_ADDRSTRLEN];

	srv = (struct sockaddr *) &server;
	a = (struct sockaddr *) &addr;
	srv6 = (struct sockaddr_in6 *) &server;
	a6 = (struct sockaddr_in6 *) &addr;

	memset(&server, 0, sizeof(server));
	srv6->sin6_family = AF_INET6;
#ifdef BSD
	srv6->sin6_len = sizeof(*srv6);
#endif
	ret = inet_pton(AF_INET6, SERVER6, &srv6->sin6_addr);
	ASSERT_EQ(ret, 1);
	ret = pcp_init(srv, NULL, &pcp);
	ASSERT_EQ(ret, PCP_OK);
#ifdef WIN32
	fd = INVALID_SOCKET;
#else
	fd = -1;
#endif
	ret = pcp_getsocket(&pcp, &fd);
	EXPECT_EQ(ret, PCP_OK);
#ifdef WIN32
	ASSERT_NE(fd, INVALID_SOCKET);
#else
	ASSERT_GE(fd, 0);
#endif
	memset(&addr, 0, sizeof(addr));
	len = sizeof(addr);
	ret = getpeername(fd, a, &len);
	ASSERT_EQ(ret, 0);
	ASSERT_EQ(a6->sin6_family, AF_INET6);
	EXPECT_EQ(len, sizeof(*a6));
	memset(host, 0, sizeof(host));
	const char *n = inet_ntop(AF_INET6, &a6->sin6_addr,
				  host, sizeof(host));
	ASSERT_TRUE(n != NULL);
	ret = strcmp(n, SERVER6);
	EXPECT_EQ(ret, 0);
}

TEST_F(ServerAddressTest, IPv6Port) {
	struct sockaddr_storage server, addr;
	struct sockaddr *srv, *a;
	struct sockaddr_in6 *srv6, *a6;
	int ret, fd;
	socklen_t len;

	srv = (struct sockaddr *) &server;
	a = (struct sockaddr *) &addr;
	srv6 = (struct sockaddr_in6 *) &server;
	a6 = (struct sockaddr_in6 *) &addr;

	memset(&server, 0, sizeof(server));
	srv6->sin6_family = AF_INET6;
#ifdef BSD
	srv6->sin6_len = sizeof(*srv6);
#endif
	ret = inet_pton(AF_INET6, SERVER6, &srv6->sin6_addr);
	ASSERT_EQ(ret, 1);
	srv6->sin6_port = htons(1234);
	ret = pcp_init(srv, NULL, &pcp);
	ASSERT_EQ(ret, PCP_OK);
#ifdef WIN32
	fd = INVALID_SOCKET;
#else
	fd = -1;
#endif
	ret = pcp_getsocket(&pcp, &fd);
	EXPECT_EQ(ret, PCP_OK);
#ifdef WIN32
	ASSERT_NE(fd, INVALID_SOCKET);
#else
	ASSERT_GE(fd, 0);
#endif
	memset(&addr, 0, sizeof(addr));
	len = sizeof(addr);
	ret = getpeername(fd, a, &len);
	ASSERT_EQ(ret, 0);
	ASSERT_EQ(a6->sin6_family, AF_INET6);
	EXPECT_EQ(len, sizeof(*a6));
	EXPECT_EQ(ntohs(a6->sin6_port), 1234);
}

TEST_F(ServerAddressTest, IPv6DefaultPort) {
	struct sockaddr_storage server, addr;
	struct sockaddr *srv, *a;
	struct sockaddr_in6 *srv6, *a6;
	int ret, fd;
	socklen_t len;

	srv = (struct sockaddr *) &server;
	a = (struct sockaddr *) &addr;
	srv6 = (struct sockaddr_in6 *) &server;
	a6 = (struct sockaddr_in6 *) &addr;

	memset(&server, 0, sizeof(server));
	srv6->sin6_family = AF_INET6;
#ifdef BSD
	srv6->sin6_len = sizeof(*srv6);
#endif
	ret = inet_pton(AF_INET6, SERVER6, &srv6->sin6_addr);
	ASSERT_EQ(ret, 1);
	ret = pcp_init(srv, NULL, &pcp);
	ASSERT_EQ(ret, PCP_OK);
#ifdef WIN32
	fd = INVALID_SOCKET;
#else
	fd = -1;
#endif
	ret = pcp_getsocket(&pcp, &fd);
	EXPECT_EQ(ret, PCP_OK);
#ifdef WIN32
	ASSERT_NE(fd, INVALID_SOCKET);
#else
	ASSERT_GE(fd, 0);
#endif
	memset(&addr, 0, sizeof(addr));
	len = sizeof(addr);
	ret = getpeername(fd, a, &len);
	ASSERT_EQ(ret, 0);
	ASSERT_EQ(a6->sin6_family, AF_INET6);
	EXPECT_EQ(len, sizeof(*a6));
	EXPECT_EQ(ntohs(a6->sin6_port), PCP_PORT);
}

}

namespace {

class SourceAddressTest : public ::testing::Test {
    protected:
	void SetUp() {
		memset(&pcp, 0, sizeof(pcp));
	}

	void TearDown() {
		(void) pcp_close(&pcp);
	}

	pcp_t pcp;
};

TEST_F(SourceAddressTest, IPv4Address) {
	struct sockaddr_storage server, source, destination, addr;
	struct sockaddr *srv, *src, *dst, *a;
	struct sockaddr_in *srv4, *src4, *dst4, *a4;
	int ret, fd;
	socklen_t len;

	srv = (struct sockaddr *) &server;
	src = (struct sockaddr *) &source;
	dst = (struct sockaddr *) &destination;
	a = (struct sockaddr *) &addr;
	srv4 = (struct sockaddr_in *) &server;
	src4 = (struct sockaddr_in *) &source;
	dst4 = (struct sockaddr_in *) &destination;
	a4 = (struct sockaddr_in *) &addr;

	memset(&server, 0, sizeof(server));
	srv4->sin_family = AF_INET;
	ret = inet_pton(AF_INET, SERVER4, &srv4->sin_addr);
	ASSERT_EQ(ret, 1);
	fd = socket(PF_INET, SOCK_DGRAM, 0);
#ifdef WIN32
	ASSERT_NE(fd, INVALID_SOCKET);
#else
	ASSERT_GE(fd, 0);
#endif
	memcpy(&destination, &server, sizeof(addr));
	dst4->sin_port = htons(PCP_PORT);
	ret = connect(fd, dst, sizeof(*a4));
	ASSERT_EQ(ret, 0);
	memset(&addr, 0, sizeof(addr));
	len = sizeof(addr);
	ret = getsockname(fd, a, &len);
	ASSERT_EQ(ret, 0);
	ASSERT_EQ(a4->sin_family, AF_INET);
	EXPECT_EQ(len, sizeof(*a4));
#ifdef WIN32
	ret = closesocket(fd);
#else
	ret = close(fd);
#endif
	EXPECT_EQ(ret, 0);
	memset(&source, 0, sizeof(source));
	src4->sin_family = AF_INET;
	memcpy(&src4->sin_addr, &a4->sin_addr, 4);
	ret = pcp_init(srv, src, &pcp);
	ASSERT_EQ(ret, PCP_OK);
#ifdef WIN32
	fd = INVALID_SOCKET;
#else
	fd = -1;
#endif
	ret = pcp_getsocket(&pcp, &fd);
	EXPECT_EQ(ret, PCP_OK);
#ifdef WIN32
	ASSERT_NE(fd, INVALID_SOCKET);
#else
	ASSERT_GE(fd, 0);
#endif
	memset(&addr, 0, sizeof(addr));
	len = sizeof(addr);
	ret = getsockname(fd, a, &len);
	ASSERT_EQ(ret, 0);
	ASSERT_EQ(a4->sin_family, AF_INET);
	EXPECT_EQ(len, sizeof(*a4));
	ret = memcmp(&src4->sin_addr, &a4->sin_addr, 4);
	EXPECT_EQ(ret, 0);
}

TEST_F(SourceAddressTest, IPv6Address) {
	struct sockaddr_storage server, source, destination, addr;
	struct sockaddr *srv, *src, *dst, *a;
	struct sockaddr_in6 *srv6, *src6, *dst6, *a6;
	int ret, fd;
	socklen_t len;

	srv = (struct sockaddr *) &server;
	src = (struct sockaddr *) &source;
	dst = (struct sockaddr *) &destination;
	a = (struct sockaddr *) &addr;
	srv6 = (struct sockaddr_in6 *) &server;
	src6 = (struct sockaddr_in6 *) &source;
	dst6 = (struct sockaddr_in6 *) &destination;
	a6 = (struct sockaddr_in6 *) &addr;

	memset(&server, 0, sizeof(server));
	srv6->sin6_family = AF_INET6;
#ifdef BSD
	srv6->sin6_len = sizeof(*srv6);
#endif
	ret = inet_pton(AF_INET6, SERVER6, &srv6->sin6_addr);
	ASSERT_EQ(ret, 1);
	fd = socket(PF_INET6, SOCK_DGRAM, 0);
#ifdef WIN32
	ASSERT_NE(fd, INVALID_SOCKET);
#else
	ASSERT_GE(fd, 0);
#endif
	memcpy(&destination, &server, sizeof(addr));
	dst6->sin6_port = htons(PCP_PORT);
	ret = connect(fd, dst, sizeof(*a6));
	ASSERT_EQ(ret, 0);
	memset(&addr, 0, sizeof(addr));
	len = sizeof(addr);
	ret = getsockname(fd, a, &len);
	ASSERT_EQ(ret, 0);
	ASSERT_EQ(a6->sin6_family, AF_INET6);
	EXPECT_EQ(len, sizeof(*a6));
#ifdef WIN32
	ret = closesocket(fd);
#else
	ret = close(fd);
#endif
	EXPECT_EQ(ret, 0);
	memset(&source, 0, sizeof(source));
	src6->sin6_family = AF_INET6;
#ifdef BSD
	srv6->sin6_len = sizeof(*srv6);
#endif
	memcpy(&src6->sin6_addr, &a6->sin6_addr, 16);
	ret = pcp_init(srv, src, &pcp);
	ASSERT_EQ(ret, PCP_OK);
#ifdef WIN32
	fd = INVALID_SOCKET;
#else
	fd = -1;
#endif
	ret = pcp_getsocket(&pcp, &fd);
	EXPECT_EQ(ret, PCP_OK);
#ifdef WIN32
	ASSERT_NE(fd, INVALID_SOCKET);
#else
	ASSERT_GE(fd, 0);
#endif
	memset(&addr, 0, sizeof(addr));
	len = sizeof(addr);
	ret = getsockname(fd, a, &len);
	ASSERT_EQ(ret, 0);
	ASSERT_EQ(a6->sin6_family, AF_INET6);
	EXPECT_EQ(len, sizeof(*a6));
	ret = memcmp(&src6->sin6_addr, &a6->sin6_addr, 16);
	EXPECT_EQ(ret, 0);
}

}

namespace {

class SetRCSocketTest : public ::testing::Test {
    protected:
	void SetUp() {
		memset(&pcp, 0, sizeof(pcp));
	}

	void TearDown() {
		(void) pcp_close(&pcp);
	}

	pcp_t pcp;
};

TEST_F(SetRCSocketTest, NoPCP) {
	int ret, fd;

	fd = 123;
	ret = pcp_setrcsocket(NULL, fd);
	EXPECT_EQ(ret, PCP_ERR_INVAL);
}

TEST_F(SetRCSocketTest, NotInitialized) {
	int ret, fd;

	fd = 123;
	ret = pcp_setrcsocket(&pcp, fd);
	EXPECT_EQ(ret, PCP_ERR_INVAL);
}

TEST_F(SetRCSocketTest, Closed) {
	struct sockaddr_storage server;
	struct sockaddr *srv;
	struct sockaddr_in *srv4;
	int ret, fd;

	srv = (struct sockaddr *) &server;
	srv4 = (struct sockaddr_in *) &server;

	memset(&server, 0, sizeof(server));
	srv4->sin_family = AF_INET;
	ret = inet_pton(AF_INET, SERVER4, &srv4->sin_addr);
	ASSERT_EQ(ret, 1);
	ret = pcp_init(srv, NULL, &pcp);
	ASSERT_EQ(ret, PCP_OK);
	ret = pcp_close(&pcp);
	EXPECT_EQ(ret, PCP_OK);
	fd = 123;
	ret = pcp_setrcsocket(&pcp, fd);
	EXPECT_EQ(ret, PCP_ERR_INVAL);
}

TEST_F(SetRCSocketTest, Set) {
	struct sockaddr_storage server;
	struct sockaddr *srv;
	struct sockaddr_in *srv4;
	int ret, fd;

	srv = (struct sockaddr *) &server;
	srv4 = (struct sockaddr_in *) &server;

	memset(&server, 0, sizeof(server));
	srv4->sin_family = AF_INET;
	ret = inet_pton(AF_INET, SERVER4, &srv4->sin_addr);
	ASSERT_EQ(ret, 1);
	ret = pcp_init(srv, NULL, &pcp);
	ASSERT_EQ(ret, PCP_OK);
	fd = 123;
	ret = pcp_setrcsocket(&pcp, fd);
	EXPECT_EQ(ret, PCP_OK);
}

TEST_F(SetRCSocketTest, Reset) {
	struct sockaddr_storage server;
	struct sockaddr *srv;
	struct sockaddr_in *srv4;
	int ret, fd;

	srv = (struct sockaddr *) &server;
	srv4 = (struct sockaddr_in *) &server;

	memset(&server, 0, sizeof(server));
	srv4->sin_family = AF_INET;
	ret = inet_pton(AF_INET, SERVER4, &srv4->sin_addr);
	ASSERT_EQ(ret, 1);
	ret = pcp_init(srv, NULL, &pcp);
	ASSERT_EQ(ret, PCP_OK);
#ifdef WIN32
	fd = INVALID_SOCKET;
#else
	fd = -1;
#endif
	ret = pcp_setrcsocket(&pcp, fd);
	EXPECT_EQ(ret, PCP_OK);
}

}

namespace {

class GetTriesTest : public ::testing::Test {
    protected:
	void SetUp() {
		memset(&pcp, 0, sizeof(pcp));
	}

	void TearDown() {
		(void) pcp_close(&pcp);
	}

	pcp_t pcp;
};

TEST_F(GetTriesTest, NoPCP) {
	int ret, tries;

	ret = pcp_gettries(NULL, &tries);
	EXPECT_EQ(ret, PCP_ERR_INVAL);
}

TEST_F(GetTriesTest, NoTries) {
	int ret;

	ret = pcp_gettries(&pcp, NULL);
	EXPECT_EQ(ret, PCP_ERR_INVAL);
}

TEST_F(GetTriesTest, GetInitial) {
	int ret, tries;

	tries = -1;
	ret = pcp_gettries(&pcp, &tries);
	EXPECT_EQ(ret, PCP_OK);
	EXPECT_EQ(tries, 0);
}

TEST_F(GetTriesTest, Get) {
	struct sockaddr_storage server;
	struct sockaddr *srv;
	struct sockaddr_in *srv4;
	pcp_request_t req;
	int ret, tries;

	srv = (struct sockaddr *) &server;
	srv4 = (struct sockaddr_in *) &server;

	memset(&server, 0, sizeof(server));
	srv4->sin_family = AF_INET;
	ret = inet_pton(AF_INET, SERVER4, &srv4->sin_addr);
	ASSERT_EQ(ret, 1);
	ret = pcp_init(srv, NULL, &pcp);
	ASSERT_EQ(ret, PCP_OK);
	memset(&req, 0, sizeof(req));
	req.opcode = PCP_OPCODE_MAP;
	ret = pcp_makerequest(&pcp, &req, NULL);
	EXPECT_EQ(ret, PCP_OK);
	ret = pcp_sendrequest(&pcp);
	if (ret == PCP_ERR_SEND) {
		RecordProperty("SendFailure",
			       "connected send() without "
			       "a responder can fail: aborting");
		return;
	}
	EXPECT_EQ(ret, PCP_OK);
	ret = pcp_sendrequest(&pcp);
	if (ret == PCP_ERR_SEND) {
		RecordProperty("SendFailure",
			       "connected send() without "
			       "a responder can fail: aborting");
		return;
	}
	EXPECT_EQ(ret, PCP_OK);
	tries = -1;
	ret = pcp_gettries(&pcp, &tries);
	EXPECT_EQ(ret, PCP_OK);
	EXPECT_EQ(tries, 2);
}

TEST_F(GetTriesTest, Reset) {
	struct sockaddr_storage server;
	struct sockaddr *srv;
	struct sockaddr_in *srv4;
	pcp_request_t req;
	int ret, tries;

	srv = (struct sockaddr *) &server;
	srv4 = (struct sockaddr_in *) &server;

	memset(&server, 0, sizeof(server));
	srv4->sin_family = AF_INET;
	ret = inet_pton(AF_INET, SERVER4, &srv4->sin_addr);
	ASSERT_EQ(ret, 1);
	ret = pcp_init(srv, NULL, &pcp);
	ASSERT_EQ(ret, PCP_OK);
	memset(&req, 0, sizeof(req));
	req.opcode = PCP_OPCODE_MAP;
	ret = pcp_makerequest(&pcp, &req, NULL);
	EXPECT_EQ(ret, PCP_OK);
	ret = pcp_sendrequest(&pcp);
	if (ret == PCP_ERR_SEND) {
		RecordProperty("SendFailure",
			       "connected send() without "
			       "a responder can fail: aborting");
		return;
	}
	EXPECT_EQ(ret, PCP_OK);
	ret = pcp_sendrequest(&pcp);
	if (ret == PCP_ERR_SEND) {
		RecordProperty("SendFailure",
			       "connected send() without "
			       "a responder can fail: aborting");
		return;
	}
	EXPECT_EQ(ret, PCP_OK);
	tries = -1;
	ret = pcp_gettries(&pcp, &tries);
	EXPECT_EQ(ret, PCP_OK);
	EXPECT_EQ(tries, 2);
	ret = pcp_makerequest(&pcp, &req, NULL);
	EXPECT_EQ(ret, PCP_OK);
	ret = pcp_gettries(&pcp, &tries);
	EXPECT_EQ(ret, PCP_OK);
	EXPECT_EQ(tries, 0);
}

}

namespace {

class GetTimeoutTest : public ::testing::Test {
    protected:
	void SetUp() {
		memset(&pcp, 0, sizeof(pcp));
	}

	void TearDown() {
		(void) pcp_close(&pcp);
	}

	pcp_t pcp;
};

TEST_F(GetTimeoutTest, NoPCP) {
	struct timeval tm;
	int ret;

	ret = pcp_gettimeout(NULL, &tm, 0);
	EXPECT_EQ(ret, PCP_ERR_INVAL);
	ret = pcp_gettimeout(NULL, &tm, 1);
	EXPECT_EQ(ret, PCP_ERR_INVAL);
}

TEST_F(GetTimeoutTest, NoTimeout) {
	int ret;

	ret = pcp_gettimeout(&pcp, NULL, 0);
	EXPECT_EQ(ret, PCP_ERR_INVAL);
	ret = pcp_gettimeout(&pcp, NULL, 1);
	EXPECT_EQ(ret, PCP_ERR_INVAL);
}

TEST_F(GetTimeoutTest, NoRequest) {
	struct timeval tm;
	int ret;

	ret = pcp_gettimeout(&pcp, &tm, 0);
	EXPECT_EQ(ret, PCP_ERR_NOREQUEST);
	ret = pcp_gettimeout(&pcp, &tm, 1);
	EXPECT_EQ(ret, PCP_ERR_NOREQUEST);
}

TEST_F(GetTimeoutTest, GetInitial) {
	pcp_request_t req;
	struct timeval tm, now;
	int ret;

	ret = pcp_setautofill(&pcp, 0, NULL);
	EXPECT_EQ(ret, PCP_OK);
	memset(&req, 0, sizeof(req));
	req.opcode = PCP_OPCODE_MAP;
	ret = pcp_makerequest(&pcp, &req, NULL);
	EXPECT_EQ(ret, PCP_OK);
	memset(&tm, 0, sizeof(tm));
	ret = pcp_gettimeout(&pcp, &tm, 0);
	EXPECT_EQ(ret, PCP_OK);
	EXPECT_EQ(tm.tv_sec, 0);
	EXPECT_EQ(tm.tv_usec, 0);
	ret = gettimeofday(&now, NULL);
	ASSERT_EQ(ret, 0);
	ret = pcp_gettimeout(&pcp, &tm, 1);
	EXPECT_EQ(ret, PCP_OK);
	EXPECT_GE(tm.tv_sec, now.tv_sec);
	if (tm.tv_sec != now.tv_sec)
		tm.tv_usec += 1000000;
	EXPECT_GE(tm.tv_usec, now.tv_usec);
	tm.tv_usec -= now.tv_usec;
	EXPECT_LT(tm.tv_usec, 1500);
}

TEST_F(GetTimeoutTest, Get) {
	struct sockaddr_storage server;
	struct sockaddr *srv;
	struct sockaddr_in *srv4;
	pcp_request_t req;
	struct timeval tm, now;
	int ret;
	double delta;

	srv = (struct sockaddr *) &server;
	srv4 = (struct sockaddr_in *) &server;

	memset(&server, 0, sizeof(server));
	srv4->sin_family = AF_INET;
	ret = inet_pton(AF_INET, SERVER4, &srv4->sin_addr);
	ASSERT_EQ(ret, 1);
	ret = pcp_init(srv, NULL, &pcp);
	ASSERT_EQ(ret, PCP_OK);
	memset(&req, 0, sizeof(req));
	req.opcode = PCP_OPCODE_MAP;
	ret = pcp_makerequest(&pcp, &req, NULL);
	EXPECT_EQ(ret, PCP_OK);
	ret = pcp_sendrequest(&pcp);
	if (ret == PCP_ERR_SEND) {
		RecordProperty("SendFailure",
			       "connected send() without "
			       "a responder can fail: aborting");
		return;
	}
	EXPECT_EQ(ret, PCP_OK);
	ret = pcp_sendrequest(&pcp);
	if (ret == PCP_ERR_SEND) {
		RecordProperty("SendFailure",
			       "connected send() without "
			       "a responder can fail: aborting");
		return;
	}
	EXPECT_EQ(ret, PCP_OK);
	memset(&tm, 0, sizeof(tm));
	ret = pcp_gettimeout(&pcp, &tm, 1);
	EXPECT_EQ(ret, PCP_OK);
	delta = (double) tm.tv_sec;
	delta += (double) tm.tv_usec / 1000000.;
	EXPECT_TRUE((delta > 3.8) && (delta < 4.2));
	ret = gettimeofday(&now, NULL);
	ASSERT_EQ(ret, 0);
	// remove 40 years
	now.tv_sec -= 1261440000;
	ret = pcp_gettimeout(&pcp, &tm, 0);
	EXPECT_EQ(ret, PCP_OK);
	tm.tv_sec -= 1261440000;
	delta = (double) tm.tv_sec;
	delta -= (double) now.tv_sec;
	delta += (double) tm.tv_usec / 1000000.;
	delta -= (double) now.tv_usec /1000000.;
	EXPECT_TRUE((delta > 3.8) && (delta < 4.2));
}

TEST_F(GetTimeoutTest, Reset) {
	struct sockaddr_storage server;
	struct sockaddr *srv;
	struct sockaddr_in *srv4;
	pcp_request_t req;
	struct timeval tm, now;
	int ret;

	srv = (struct sockaddr *) &server;
	srv4 = (struct sockaddr_in *) &server;

	memset(&server, 0, sizeof(server));
	srv4->sin_family = AF_INET;
	ret = inet_pton(AF_INET, SERVER4, &srv4->sin_addr);
	ASSERT_EQ(ret, 1);
	ret = pcp_init(srv, NULL, &pcp);
	ASSERT_EQ(ret, PCP_OK);
	memset(&req, 0, sizeof(req));
	req.opcode = PCP_OPCODE_MAP;
	ret = pcp_makerequest(&pcp, &req, NULL);
	EXPECT_EQ(ret, PCP_OK);
	ret = pcp_sendrequest(&pcp);
	if (ret == PCP_ERR_SEND) {
		RecordProperty("SendFailure",
			       "connected send() without "
			       "a responder can fail: aborting");
		return;
	}
	EXPECT_EQ(ret, PCP_OK);
	ret = pcp_sendrequest(&pcp);
	if (ret == PCP_ERR_SEND) {
		RecordProperty("SendFailure",
			       "connected send() without "
			       "a responder can fail: aborting");
		return;
	}
	EXPECT_EQ(ret, PCP_OK);
	ret = pcp_makerequest(&pcp, &req, NULL);
	EXPECT_EQ(ret, PCP_OK);
	memset(&tm, 0, sizeof(tm));
	ret = pcp_gettimeout(&pcp, &tm, 0);
	EXPECT_EQ(ret, PCP_OK);
	EXPECT_EQ(tm.tv_sec, 0);
	EXPECT_EQ(tm.tv_usec, 0);
	ret = gettimeofday(&now, NULL);
	ASSERT_EQ(ret, 0);
	ret = pcp_gettimeout(&pcp, &tm, 1);
	EXPECT_EQ(ret, PCP_OK);
	EXPECT_GE(tm.tv_sec, now.tv_sec);
	if (tm.tv_sec != now.tv_sec)
		tm.tv_usec += 1000000;
	EXPECT_GE(tm.tv_usec, now.tv_usec);
	tm.tv_usec -= now.tv_usec;
	EXPECT_LT(tm.tv_usec, 1500);
}

}

namespace {

class SetResetTest : public ::testing::Test {
    protected:
	void SetUp() {
		memset(&pcp, 0, sizeof(pcp));
	}

	void TearDown() {
		(void) pcp_close(&pcp);
	}

	pcp_t pcp;
};

TEST_F(SetResetTest, NoPCP) {
	int ret, old;

	ret = pcp_setreset(NULL, 0, NULL);
	EXPECT_EQ(ret, PCP_ERR_INVAL);
	ret = pcp_setreset(NULL, 1, &old);
	EXPECT_EQ(ret, PCP_ERR_INVAL);
}

TEST_F(SetResetTest, Initial) {
	int ret, old;

	old = -1;
	ret = pcp_setreset(&pcp, 0, &old);
	EXPECT_EQ(ret, PCP_OK);
	EXPECT_EQ(old, 0);
}

TEST_F(SetResetTest, Set) {
	int ret, old;

	ret = pcp_setreset(&pcp, 1, &old);
	EXPECT_EQ(ret, PCP_OK);
	EXPECT_EQ(old, 0);
	ret = pcp_setreset(&pcp, old, NULL);
	EXPECT_EQ(ret, PCP_OK);
	ret = pcp_setreset(&pcp, 0, &old);
	EXPECT_EQ(ret, PCP_OK);
	EXPECT_EQ(old, 0);
}

TEST_F(SetResetTest, Reset0) {
	pcp_request_t req;
	uint8_t resp[PCP_LEN_MAP], *in;
	pcp_response_t out;
	uint16_t len;
	uint32_t n32;
	int ret;

	ret = pcp_setautofill(&pcp, 0, NULL);
	EXPECT_EQ(ret, PCP_OK);
	memset(&req, 0, sizeof(req));
	req.opcode = PCP_OPCODE_MAP;
	ret = pcp_makerequest(&pcp, &req, NULL);
	ASSERT_EQ(ret, PCP_OK);
	in = NULL;
	len = 0;
	ret = _pcp_getrequest(&pcp, &in, &len);
	ASSERT_EQ(ret, PCP_OK);
	ASSERT_TRUE(in != NULL);
	ASSERT_EQ(len, sizeof(resp));
	memcpy(resp, in, sizeof(resp));
	resp[PCP_OPCODE] |= PCP_OPCODE_RESPONSE;
	resp[PCP_RESULT_CODE] = 1;
	n32 = htonl(10);
	memcpy(resp + PCP_LIFETIME, &n32, 4);
	memset(resp + PCP_CLIENT_ADDR, 0, 16);
	ret = _pcp_setresponse(&pcp, resp, sizeof(resp));
	ASSERT_EQ(ret, PCP_OK);
	ret = pcp_setreset(&pcp, 0, NULL);
	memset(&out, 0, sizeof(out));
	ret = _pcp_submitresponse(&pcp, &out);
	EXPECT_EQ(ret, PCP_ERR_PROTOBASE - 1);
}

TEST_F(SetResetTest, Reset1) {
	pcp_request_t req;
	uint8_t resp[PCP_LEN_MAP], *in;
	pcp_response_t out;
	uint16_t len;
	uint32_t n32;
	struct timeval tm, now;
	int ret;
	double delta;

	ret = pcp_setautofill(&pcp, 0, NULL);
	EXPECT_EQ(ret, PCP_OK);
	memset(&req, 0, sizeof(req));
	req.opcode = PCP_OPCODE_MAP;
	ret = pcp_makerequest(&pcp, &req, NULL);
	ASSERT_EQ(ret, PCP_OK);
	in = NULL;
	len = 0;
	ret = _pcp_getrequest(&pcp, &in, &len);
	ASSERT_EQ(ret, PCP_OK);
	ASSERT_TRUE(in != NULL);
	ASSERT_EQ(len, sizeof(resp));
	memcpy(resp, in, sizeof(resp));
	resp[PCP_OPCODE] |= PCP_OPCODE_RESPONSE;
	resp[PCP_RESULT_CODE] = 1;
	n32 = htonl(10);
	memcpy(resp + PCP_LIFETIME, &n32, 4);
	memset(resp + PCP_CLIENT_ADDR, 0, 16);
	ret = _pcp_setresponse(&pcp, resp, sizeof(resp));
	ASSERT_EQ(ret, PCP_OK);
	ret = pcp_setreset(&pcp, 1, NULL);
	ASSERT_EQ(ret, PCP_OK);
	memset(&out, 0, sizeof(out));
	ret = _pcp_submitresponse(&pcp, &out);
	EXPECT_EQ(ret, PCP_TRY_AGAIN);
	memset(&tm, 0, sizeof(tm));
	ret = gettimeofday(&now, NULL);
	EXPECT_EQ(ret, 0);
	// remove 40 years
	now.tv_sec -= 1261440000;
	ret = pcp_gettimeout(&pcp, &tm, 0);
	EXPECT_EQ(ret, PCP_OK);
	tm.tv_sec -= 1261440000;
	delta = (double) tm.tv_sec;
	delta -= (double) now.tv_sec;
	delta += (double) tm.tv_usec / 1000000.;
	delta -= (double) now.tv_usec /1000000.;
	EXPECT_TRUE((delta > 9.8) && (delta < 10.2));
}

}

namespace {

class SetAutoFillTest : public ::testing::Test {
    protected:
	void SetUp() {
		memset(&pcp, 0, sizeof(pcp));
	}

	void TearDown() {
		(void) pcp_close(&pcp);
	}

	pcp_t pcp;
};

TEST_F(SetAutoFillTest, NoPCP) {
	int ret, old;

	ret = pcp_setreset(NULL, 0, NULL);
	EXPECT_EQ(ret, PCP_ERR_INVAL);
	ret = pcp_setreset(NULL, 1, &old);
	EXPECT_EQ(ret, PCP_ERR_INVAL);
}

TEST_F(SetAutoFillTest, Initial) {
	int ret, old;

	old = -1;
	ret = pcp_setreset(&pcp, 0, &old);
	EXPECT_EQ(ret, PCP_OK);
	EXPECT_EQ(old, 0);
}

TEST_F(SetAutoFillTest, Set) {
	int ret, old;

	ret = pcp_setreset(&pcp, 1, &old);
	EXPECT_EQ(ret, PCP_OK);
	EXPECT_EQ(old, 0);
	ret = pcp_setreset(&pcp, old, NULL);
	EXPECT_EQ(ret, PCP_OK);
	ret = pcp_setreset(&pcp, 0, &old);
	EXPECT_EQ(ret, PCP_OK);
	EXPECT_EQ(old, 0);
}

TEST_F(SetResetTest, NoAutoFill) {
	struct sockaddr_storage server;
	struct sockaddr *srv;
	struct sockaddr_in *srv4;
	pcp_request_t req;
	int ret;

	srv = (struct sockaddr *) &server;
	srv4 = (struct sockaddr_in *) &server;

	memset(&server, 0, sizeof(server));
	srv4->sin_family = AF_INET;
	ret = inet_pton(AF_INET, SERVER4, &srv4->sin_addr);
	ASSERT_EQ(ret, 1);
	ret = pcp_init(srv, NULL, &pcp);
	EXPECT_EQ(ret, PCP_OK);
	ret = pcp_setautofill(&pcp, 0, NULL);
	EXPECT_EQ(ret, PCP_OK);
	memset(&req, 0, sizeof(req));
	req.opcode = PCP_OPCODE_MAP;
	ret = pcp_makerequest(&pcp, &req, NULL);
	EXPECT_EQ(ret, PCP_OK);
	EXPECT_TRUE(memcmp(req.intaddr6, &zero, 16) == 0);
}

TEST_F(SetResetTest, BadAutoFill) {
	pcp_request_t req;
	int ret;

	ret = pcp_setautofill(&pcp, 1, NULL);
	EXPECT_EQ(ret, PCP_OK);
	memset(&req, 0, sizeof(req));
	req.opcode = PCP_OPCODE_MAP;
	ret = pcp_makerequest(&pcp, &req, NULL);
	EXPECT_EQ(ret, PCP_ERR_INVAL);
}

TEST_F(SetResetTest, AutoFill) {
	struct sockaddr_storage server;
	struct sockaddr *srv;
	struct sockaddr_in *srv4;
	pcp_request_t req;
	int ret;

	srv = (struct sockaddr *) &server;
	srv4 = (struct sockaddr_in *) &server;

	memset(&server, 0, sizeof(server));
	srv4->sin_family = AF_INET;
	ret = inet_pton(AF_INET, SERVER4, &srv4->sin_addr);
	ASSERT_EQ(ret, 1);
	ret = pcp_init(srv, NULL, &pcp);
	EXPECT_EQ(ret, PCP_OK);
	ret = pcp_setautofill(&pcp, 1, NULL);
	EXPECT_EQ(ret, PCP_OK);
	memset(&req, 0, sizeof(req));
	req.opcode = PCP_OPCODE_MAP;
	ret = pcp_makerequest(&pcp, &req, NULL);
	EXPECT_EQ(ret, PCP_OK);
	EXPECT_TRUE(memcmp(req.intaddr6, &zero, 16) != 0);
}

TEST_F(SetResetTest, IPv4AutoFill) {
	struct sockaddr_storage server, source;
	struct sockaddr *srv, *src;
	struct sockaddr_in *srv4, *src4;
	pcp_request_t req;
	socklen_t len;
	int ret, fd;

	srv = (struct sockaddr *) &server;
	src = (struct sockaddr *) &source;
	srv4 = (struct sockaddr_in *) &server;
	src4 = (struct sockaddr_in *) &source;

	memset(&server, 0, sizeof(server));
	srv4->sin_family = AF_INET;
	ret = inet_pton(AF_INET, SERVER4, &srv4->sin_addr);
	ASSERT_EQ(ret, 1);
	ret = pcp_init(srv, NULL, &pcp);
	EXPECT_EQ(ret, PCP_OK);
#ifdef WIN32
	fd = INVALID_SOCKET;
#else
	fd = -1;
#endif
	ret = pcp_getsocket(&pcp, &fd);
	ASSERT_EQ(ret, PCP_OK);
#ifdef WIN32
	ASSERT_NE(fd, INVALID_SOCKET);
#else
	ASSERT_GE(fd, 0);
#endif
	memset(&source, 0, sizeof(source));
	len = sizeof(source);
	ret = getsockname(fd, src, &len);
	EXPECT_EQ(ret, 0);
	EXPECT_EQ(src4->sin_family, AF_INET);
	EXPECT_EQ(len, sizeof(*src4));
	ret = pcp_setautofill(&pcp, 1, NULL);
	EXPECT_EQ(ret, PCP_OK);
	memset(&req, 0, sizeof(req));
	req.opcode = PCP_OPCODE_MAP;
	ret = pcp_makerequest(&pcp, &req, NULL);
	EXPECT_EQ(ret, PCP_OK);
	EXPECT_TRUE(memcmp(req.intaddr6, pcp_mapped, 12) == 0);
	EXPECT_TRUE(memcmp(req.intaddr6 + 12, &src4->sin_addr, 4) == 0);
}

TEST_F(SetResetTest, IPv6AutoFill) {
	struct sockaddr_storage server, source;
	struct sockaddr *srv, *src;
	struct sockaddr_in6 *srv6, *src6;
	pcp_request_t req;
	socklen_t len;
	int ret, fd;

	srv = (struct sockaddr *) &server;
	src = (struct sockaddr *) &source;
	srv6 = (struct sockaddr_in6 *) &server;
	src6 = (struct sockaddr_in6 *) &source;

	memset(&server, 0, sizeof(server));
	srv6->sin6_family = AF_INET6;
#ifdef BSD
	srv6->sin6_len = sizeof(*srv6);
#endif
	ret = inet_pton(AF_INET6, SERVER6, &srv6->sin6_addr);
	ASSERT_EQ(ret, 1);
	ret = pcp_init(srv, NULL, &pcp);
	EXPECT_EQ(ret, PCP_OK);
#ifdef WIN32
	fd = INVALID_SOCKET;
#else
	fd = -1;
#endif
	ret = pcp_getsocket(&pcp, &fd);
	ASSERT_EQ(ret, PCP_OK);
#ifdef WIN32
	ASSERT_NE(fd, INVALID_SOCKET);
#else
	ASSERT_GE(fd, 0);
#endif
	memset(&source, 0, sizeof(source));
	len = sizeof(source);
	ret = getsockname(fd, src, &len);
	EXPECT_EQ(ret, 0);
	EXPECT_EQ(src6->sin6_family, AF_INET6);
	EXPECT_EQ(len, sizeof(*src6));
	ret = pcp_setautofill(&pcp, 1, NULL);
	EXPECT_EQ(ret, PCP_OK);
	memset(&req, 0, sizeof(req));
	req.opcode = PCP_OPCODE_MAP;
	ret = pcp_makerequest(&pcp, &req, NULL);
	EXPECT_EQ(ret, PCP_OK);
	EXPECT_TRUE(memcmp(req.intaddr6, &src6->sin6_addr, 16) == 0);
}

}

namespace {

class ThirdParty4Test : public ::testing::Test {
    protected:
	void SetUp() {
		options = NULL;
	}

	void TearDown() {
		pcp_freeoptions(options);
	}

	pcp_t pcp;
	pcp_option_t **options;
};

TEST_F(ThirdParty4Test, NoOptions) {
	int ret;

	ret = pcp_third_party4(NULL, 0);
	EXPECT_EQ(ret, PCP_ERR_INVAL);
}

TEST_F(ThirdParty4Test, TooManyOptions) {
	int ret;

	ret = pcp_third_party4(&options, 1);
	EXPECT_EQ(ret, PCP_OK);
	ret = pcp_third_party4(&options, 2);
	EXPECT_EQ(ret, PCP_OK);
	ret = pcp_third_party4(&options, 3);
	EXPECT_EQ(ret, PCP_OK);
	ret = pcp_third_party4(&options, 4);
	EXPECT_EQ(ret, PCP_OK);
	ret = pcp_third_party4(&options, 5);
	EXPECT_EQ(ret, PCP_ERR_TOOMANYOPTS);
}

TEST_F(ThirdParty4Test, Set) {
	uint32_t addr;
	int ret;

	addr = 0x1234fedc;
	ret = pcp_third_party4(&options, addr);
	EXPECT_EQ(ret, PCP_OK);
	EXPECT_TRUE(options[0] != NULL);
	EXPECT_TRUE(options[1] == NULL);
	EXPECT_EQ(options[0]->code, PCP_OPTION_THIRD_PARTY);
	EXPECT_EQ(options[0]->length, PCP_LEN_OPT3PTY);
	EXPECT_TRUE(options[0]->data != NULL);
	EXPECT_TRUE(memcmp(options[0]->data, pcp_mapped, 12) == 0);
	EXPECT_TRUE(memcmp(options[0]->data + 12, &addr, 4) == 0);
}

}

namespace {

class ThirdParty6Test : public ::testing::Test {
    protected:
	void SetUp() {
		options = NULL;
	}

	void TearDown() {
		pcp_freeoptions(options);
	}

	pcp_t pcp;
	pcp_option_t **options;
};

TEST_F(ThirdParty6Test, NoOptions) {
	uint8_t addr[16];
	int ret;

	ret = pcp_third_party6(NULL, addr);
	EXPECT_EQ(ret, PCP_ERR_INVAL);
}

TEST_F(ThirdParty6Test, NoAddr) {
	int ret;

	ret = pcp_third_party6(&options, NULL);
	EXPECT_EQ(ret, PCP_ERR_INVAL);
}

TEST_F(ThirdParty6Test, TooManyOptions) {
	uint8_t addr[16];
	int ret;

	addr[15] = 1;
	ret = pcp_third_party6(&options, addr);
	EXPECT_EQ(ret, PCP_OK);
	addr[15] = 2;
	ret = pcp_third_party6(&options, addr);
	EXPECT_EQ(ret, PCP_OK);
	addr[15] = 3;
	ret = pcp_third_party6(&options, addr);
	EXPECT_EQ(ret, PCP_OK);
	addr[15] = 4;
	ret = pcp_third_party6(&options, addr);
	EXPECT_EQ(ret, PCP_OK);
	addr[15] = 5;
	ret = pcp_third_party6(&options, addr);
	EXPECT_EQ(ret, PCP_ERR_TOOMANYOPTS);
}

TEST_F(ThirdParty6Test, Set) {
	uint8_t addr[16];
	int ret;

	addr[0] = 0x20;
	addr[1] = 0x01;
	addr[15] = 0x01;
	ret = pcp_third_party6(&options, addr);
	EXPECT_EQ(ret, PCP_OK);
	EXPECT_TRUE(options[0] != NULL);
	EXPECT_TRUE(options[1] == NULL);
	EXPECT_EQ(options[0]->code, PCP_OPTION_THIRD_PARTY);
	EXPECT_EQ(options[0]->length, PCP_LEN_OPT3PTY);
	EXPECT_TRUE(options[0]->data != NULL);
	EXPECT_TRUE(memcmp(options[0]->data, addr, 16) == 0);
}

}

namespace {

class PreferFailureTest : public ::testing::Test {
    protected:
	void SetUp() {
		options = NULL;
	}

	void TearDown() {
		pcp_freeoptions(options);
	}

	pcp_t pcp;
	pcp_option_t **options;
};

TEST_F(PreferFailureTest, NoOptions) {
	int ret;

	ret = pcp_prefer_failure(NULL);
	EXPECT_EQ(ret, PCP_ERR_INVAL);
}

TEST_F(PreferFailureTest, TooManyOptions) {
	int ret;

	ret = pcp_prefer_failure(&options);
	EXPECT_EQ(ret, PCP_OK);
	ret = pcp_prefer_failure(&options);
	EXPECT_EQ(ret, PCP_OK);
	ret = pcp_prefer_failure(&options);
	EXPECT_EQ(ret, PCP_OK);
	ret = pcp_prefer_failure(&options);
	EXPECT_EQ(ret, PCP_OK);
	ret = pcp_prefer_failure(&options);
	EXPECT_EQ(ret, PCP_ERR_TOOMANYOPTS);
}

TEST_F(PreferFailureTest, Set) {
	int ret;

	ret = pcp_prefer_failure(&options);
	EXPECT_EQ(ret, PCP_OK);
	EXPECT_TRUE(options[0] != NULL);
	EXPECT_TRUE(options[1] == NULL);
	EXPECT_EQ(options[0]->code, PCP_OPTION_PREF_FAIL);
	EXPECT_EQ(options[0]->length, PCP_LEN_OPT0);
	EXPECT_TRUE(options[0]->data == NULL);
}

}

namespace {

class Filter4Test : public ::testing::Test {
    protected:
	void SetUp() {
		options = NULL;
	}

	void TearDown() {
		pcp_freeoptions(options);
	}

	pcp_t pcp;
	pcp_option_t **options;
};

TEST_F(Filter4Test, NoOptions) {
	int ret;

	ret = pcp_filter4(NULL, 0, 0, 0);
	EXPECT_EQ(ret, PCP_ERR_INVAL);
}

TEST_F(Filter4Test, PrefixLength) {
	int ret;

	ret = pcp_filter4(&options, -1, 0x1234fedc, 0);
	EXPECT_EQ(ret, PCP_ERR_INVAL);
	ret = pcp_filter4(&options, 33, 0x1234fedc, 0);
	EXPECT_EQ(ret, PCP_ERR_INVAL);
	ret = pcp_filter4(&options, 0, 0x1234fedc, 0);
	EXPECT_EQ(ret, PCP_OK);
	ret = pcp_filter4(&options, 32, 0x1234fedc, 0);
	EXPECT_EQ(ret, PCP_OK);
}

TEST_F(Filter4Test, TooManyOptions) {
	int ret;

	ret = pcp_filter4(&options, 32, 0x1234fedc, 1);
	EXPECT_EQ(ret, PCP_OK);
	ret = pcp_filter4(&options, 32, 0x1234fedc, 2);
	EXPECT_EQ(ret, PCP_OK);
	ret = pcp_filter4(&options, 32, 0x1234fedc, 3);
	EXPECT_EQ(ret, PCP_OK);
	ret = pcp_filter4(&options, 32, 0x1234fedc, 4);
	EXPECT_EQ(ret, PCP_OK);
	ret = pcp_filter4(&options, 32, 0x1234fedc, 5);
	EXPECT_EQ(ret, PCP_ERR_TOOMANYOPTS);
}

TEST_F(Filter4Test, Set) {
	uint32_t addr;
	uint16_t port, n16;
	int ret;

	addr = 0x1234fedc;
	port = 1234;
	ret = pcp_filter4(&options, 24, addr, port);
	EXPECT_EQ(ret, PCP_OK);
	EXPECT_TRUE(options[0] != NULL);
	EXPECT_TRUE(options[1] == NULL);
	EXPECT_EQ(options[0]->code, PCP_OPTION_FILTER);
	EXPECT_EQ(options[0]->length, PCP_LEN_OPTFLT);
	EXPECT_TRUE(options[0]->data != NULL);
	EXPECT_EQ(options[0]->data[PCP_OPTDFLT_RSVD], 0);
	EXPECT_EQ(options[0]->data[PCP_OPTDFLT_PLEN], 24);
	memcpy(&n16, options[0]->data + PCP_OPTDFLT_PORT, 2);
	EXPECT_EQ(htons(n16), port);
	EXPECT_TRUE(memcmp(options[0]->data + PCP_OPTDFLT_PREF,
			   pcp_mapped, 12) == 0);
	EXPECT_TRUE(memcmp(options[0]->data + PCP_OPTDFLT_PREF + 12,
			   &addr, 4) == 0);
}

}

namespace {

class Filter6Test : public ::testing::Test {
    protected:
	void SetUp() {
		options = NULL;
	}

	void TearDown() {
		pcp_freeoptions(options);
	}

	pcp_t pcp;
	pcp_option_t **options;
};

TEST_F(Filter6Test, NoOptions) {
	uint8_t addr[16];
	int ret;

	ret = pcp_filter6(NULL, 0, addr, 0);
	EXPECT_EQ(ret, PCP_ERR_INVAL);
}

TEST_F(Filter6Test, NoAddr) {
	int ret;

	ret = pcp_filter6(&options, 0, NULL, 0);
	EXPECT_EQ(ret, PCP_ERR_INVAL);
}

TEST_F(Filter6Test, PrefixLength) {
	uint8_t addr[16];
	int ret;

	ret = pcp_filter6(&options, -1, addr, 0);
	EXPECT_EQ(ret, PCP_ERR_INVAL);
	ret = pcp_filter6(&options, 129, addr, 0);
	EXPECT_EQ(ret, PCP_ERR_INVAL);
	ret = pcp_filter6(&options, 0, addr, 0);
	EXPECT_EQ(ret, PCP_OK);
	ret = pcp_filter6(&options, 128, addr, 0);
	EXPECT_EQ(ret, PCP_OK);
}

TEST_F(Filter6Test, TooManyOptions) {
	uint8_t addr[16];
	int ret;

	ret = pcp_filter6(&options, 1, addr, 1);
	EXPECT_EQ(ret, PCP_OK);
	ret = pcp_filter6(&options, 2, addr, 2);
	EXPECT_EQ(ret, PCP_OK);
	ret = pcp_filter6(&options, 3, addr, 3);
	EXPECT_EQ(ret, PCP_OK);
	ret = pcp_filter6(&options, 4, addr, 4);
	EXPECT_EQ(ret, PCP_OK);
	ret = pcp_filter6(&options, 5, addr, 5);
	EXPECT_EQ(ret, PCP_ERR_TOOMANYOPTS);
}

TEST_F(Filter6Test, Set) {
	uint8_t addr[16];
	uint16_t port, n16;
	int ret;

	addr[0] = 0x20;
	addr[1] = 0x01;
	addr[15] = 0x01;
	port = 1234;
	ret = pcp_filter6(&options, 80, addr, port);
	EXPECT_EQ(ret, PCP_OK);
	EXPECT_TRUE(options[0] != NULL);
	EXPECT_TRUE(options[1] == NULL);
	EXPECT_EQ(options[0]->code, PCP_OPTION_FILTER);
	EXPECT_EQ(options[0]->length, PCP_LEN_OPTFLT);
	EXPECT_TRUE(options[0]->data != NULL);
	EXPECT_EQ(options[0]->data[PCP_OPTDFLT_RSVD], 0);
	EXPECT_EQ(options[0]->data[PCP_OPTDFLT_PLEN], 80);
	memcpy(&n16, options[0]->data + PCP_OPTDFLT_PORT, 2);
	EXPECT_EQ(htons(n16), port);
	EXPECT_TRUE(memcmp(options[0]->data + PCP_OPTDFLT_PREF,
			   addr, 16) == 0);
}

}

namespace {

class MakeRequestTest : public ::testing::Test {
    protected:
	void SetUp() {
		memset(&pcp, 0, sizeof(pcp));
	}

	void TearDown() {
		(void) pcp_close(&pcp);
	}

	pcp_t pcp;
};

TEST_F(MakeRequestTest, NoPCP) {
	int ret;

	ret = pcp_makerequest(NULL, NULL, NULL);
	EXPECT_EQ(ret, PCP_ERR_INVAL);
}

TEST_F(MakeRequestTest, Reset) {
	pcp_request_t req;
	uint8_t *buf;
	uint16_t len;
	int ret;

	ret = pcp_makerequest(&pcp, NULL, NULL);
	EXPECT_EQ(ret, PCP_OK);
	ret = _pcp_getrequest(&pcp, &buf, &len);
	EXPECT_TRUE(buf == NULL);
	EXPECT_EQ(len, 0);
	memset(&req, 0, sizeof(req));
	req.opcode = PCP_OPCODE_MAP;
	ret = pcp_makerequest(&pcp, &req, NULL);
	EXPECT_EQ(ret, PCP_OK);
	ret = _pcp_getrequest(&pcp, &buf, &len);
	EXPECT_TRUE(buf != NULL);
	EXPECT_EQ(len, PCP_LEN_MAP);
	ret = pcp_makerequest(&pcp, NULL, NULL);
	EXPECT_EQ(ret, PCP_OK);
	ret = _pcp_getrequest(&pcp, &buf, &len);
	EXPECT_TRUE(buf == NULL);
	EXPECT_EQ(len, 0);
}

TEST_F(MakeRequestTest, Opcodes) {
	pcp_request_t req;
	int ret;

	memset(&req, 0, sizeof(req));
	req.opcode = PCP_OPCODE_MAP;
	ret = pcp_makerequest(&pcp, &req, NULL);
	EXPECT_EQ(ret, PCP_OK);
	memset(&req, 0, sizeof(req));
	req.opcode = PCP_OPCODE_PEER;
	ret = pcp_makerequest(&pcp, &req, NULL);
	EXPECT_EQ(ret, PCP_OK);
	memset(&req, 0, sizeof(req));
	req.opcode = 17;
	ret = pcp_makerequest(&pcp, &req, NULL);
	EXPECT_EQ(ret, PCP_ERR_INVAL);
	memset(&req, 0, sizeof(req));
	ret = pcp_makerequest(&pcp, &req, NULL);
	EXPECT_EQ(ret, PCP_ERR_INVAL);
}

TEST_F(MakeRequestTest, TooLong) {
	pcp_option_t opt, *options[2];
	pcp_request_t req;
	int ret;

	memset(&req, 0, sizeof(req));
	req.opcode = PCP_OPCODE_MAP;
	memset(&opt, 0, sizeof(opt));
	opt.code = 17;
	opt.length = 972;
	opt.data = buf;
	options[0] = &opt;
	options[1] = NULL;
	ret = pcp_makerequest(&pcp, &req, options);
	EXPECT_EQ(ret, PCP_OK);
	opt.length = 973;
	ret = pcp_makerequest(&pcp, &req, options);
	EXPECT_EQ(ret, PCP_ERR_INVAL);
}

}

namespace {

class EncodeRequestTest : public ::testing::Test {
    protected:
	void SetUp() {
		memset(&pcp, 0, sizeof(pcp));
		options = NULL;
	}

	void TearDown() {
		pcp_freeoptions(options);
		(void) pcp_close(&pcp);
	}

	pcp_t pcp;
	pcp_option_t **options;
};

TEST_F(EncodeRequestTest, BasicMAP) {
	pcp_request_t req;
	uint8_t *in;
	uint16_t len;
	int ret;

	memset(&req, 0, sizeof(req));
	req.opcode = PCP_OPCODE_MAP;
	ret = pcp_makerequest(&pcp, &req, NULL);
	ASSERT_EQ(ret, PCP_OK);
	in = NULL;
	len = 0;
	ret = _pcp_getrequest(&pcp, &in, &len);
	ASSERT_EQ(ret, PCP_OK);
	ASSERT_TRUE(in != NULL);
	EXPECT_EQ(len, PCP_LEN_MAP);
	EXPECT_EQ(in[PCP_VERSION], 1);
	EXPECT_EQ(in[PCP_OPCODE], PCP_OPCODE_MAP);
	EXPECT_EQ(in[PCP_OPCODE + 1], 0);
	EXPECT_EQ(in[PCP_RESULT_CODE], 0);
}

TEST_F(EncodeRequestTest, BasicPEER) {
	pcp_request_t req;
	uint8_t *in;
	uint16_t len;
	int ret;

	memset(&req, 0, sizeof(req));
	req.opcode = PCP_OPCODE_PEER;
	ret = pcp_makerequest(&pcp, &req, NULL);
	ASSERT_EQ(ret, PCP_OK);
	in = NULL;
	len = 0;
	ret = _pcp_getrequest(&pcp, &in, &len);
	ASSERT_EQ(ret, PCP_OK);
	ASSERT_TRUE(in != NULL);
	EXPECT_EQ(len, PCP_LEN_PEER);
	EXPECT_EQ(in[PCP_VERSION], 1);
	EXPECT_EQ(in[PCP_OPCODE], PCP_OPCODE_PEER);
	EXPECT_EQ(in[PCP_OPCODE + 1], 0);
	EXPECT_EQ(in[PCP_RESULT_CODE], 0);
}

TEST_F(EncodeRequestTest, Lifetime) {
	pcp_request_t req;
	uint8_t *in;
	uint16_t len;
	uint32_t n32;
	int ret;

	memset(&req, 0, sizeof(req));
	req.opcode = PCP_OPCODE_MAP;
	req.lifetime = 123456789;
	ret = pcp_makerequest(&pcp, &req, NULL);
	ASSERT_EQ(ret, PCP_OK);
	in = NULL;
	len = 0;
	ret = _pcp_getrequest(&pcp, &in, &len);
	ASSERT_EQ(ret, PCP_OK);
	ASSERT_TRUE(in != NULL);
	EXPECT_EQ(len, PCP_LEN_MAP);
	n32 = htonl(123456789);
	EXPECT_TRUE(memcmp(in + PCP_LIFETIME, &n32, 4) == 0);
}

TEST_F(EncodeRequestTest, InternalAddress) {
	pcp_request_t req;
	uint8_t *in, addr[16];
	uint16_t len;
	int ret;

	memset(&req, 0, sizeof(req));
	req.opcode = PCP_OPCODE_PEER;
	ret = inet_pton(AF_INET6, ADDR6, addr);
	ASSERT_EQ(ret, 1);
	memcpy(req.intaddr6, addr, 16);
	ret = pcp_makerequest(&pcp, &req, NULL);
	ASSERT_EQ(ret, PCP_OK);
	in = NULL;
	len = 0;
	ret = _pcp_getrequest(&pcp, &in, &len);
	ASSERT_EQ(ret, PCP_OK);
	ASSERT_TRUE(in != NULL);
	EXPECT_EQ(len, PCP_LEN_PEER);
	EXPECT_TRUE(memcmp(in + PCP_CLIENT_ADDR, addr, 16) == 0);
}

TEST_F(EncodeRequestTest, ProtocolMAP) {
	pcp_request_t req;
	uint8_t *in;
	uint16_t len;
	int ret;

	memset(&req, 0, sizeof(req));
	req.opcode = PCP_OPCODE_MAP;
	req.protocol = IPPROTO_UDP;
	ret = pcp_makerequest(&pcp, &req, NULL);
	ASSERT_EQ(ret, PCP_OK);
	in = NULL;
	len = 0;
	ret = _pcp_getrequest(&pcp, &in, &len);
	ASSERT_EQ(ret, PCP_OK);
	ASSERT_TRUE(in != NULL);
	EXPECT_EQ(len, PCP_LEN_MAP);
	EXPECT_EQ(in[PCP_MAP_PROTOCOL], IPPROTO_UDP);
}

TEST_F(EncodeRequestTest, ProtocolPEER) {
	pcp_request_t req;
	uint8_t *in;
	uint16_t len;
	int ret;

	memset(&req, 0, sizeof(req));
	req.opcode = PCP_OPCODE_PEER;
	req.protocol = IPPROTO_TCP;
	ret = pcp_makerequest(&pcp, &req, NULL);
	ASSERT_EQ(ret, PCP_OK);
	in = NULL;
	len = 0;
	ret = _pcp_getrequest(&pcp, &in, &len);
	ASSERT_EQ(ret, PCP_OK);
	ASSERT_TRUE(in != NULL);
	EXPECT_EQ(len, PCP_LEN_PEER);
	EXPECT_EQ(in[PCP_PEER_PROTOCOL], IPPROTO_TCP);
}

TEST_F(EncodeRequestTest, InternalPortMAP) {
	pcp_request_t req;
	uint8_t *in;
	uint16_t len, n16;
	int ret;

	memset(&req, 0, sizeof(req));
	req.opcode = PCP_OPCODE_MAP;
	req.intport = 1234;
	ret = pcp_makerequest(&pcp, &req, NULL);
	ASSERT_EQ(ret, PCP_OK);
	in = NULL;
	len = 0;
	ret = _pcp_getrequest(&pcp, &in, &len);
	ASSERT_EQ(ret, PCP_OK);
	ASSERT_TRUE(in != NULL);
	EXPECT_EQ(len, PCP_LEN_MAP);
	n16 = htons(1234);
	EXPECT_TRUE(memcmp(in + PCP_MAP_INTERNAL_PORT, &n16, 2) == 0);
}

TEST_F(EncodeRequestTest, InternalPortPEER) {
	pcp_request_t req;
	uint8_t *in;
	uint16_t len, n16;
	int ret;

	memset(&req, 0, sizeof(req));
	req.opcode = PCP_OPCODE_PEER;
	req.intport = 1234;
	ret = pcp_makerequest(&pcp, &req, NULL);
	ASSERT_EQ(ret, PCP_OK);
	in = NULL;
	len = 0;
	ret = _pcp_getrequest(&pcp, &in, &len);
	ASSERT_EQ(ret, PCP_OK);
	ASSERT_TRUE(in != NULL);
	EXPECT_EQ(len, PCP_LEN_PEER);
	n16 = htons(1234);
	EXPECT_TRUE(memcmp(in + PCP_PEER_INTERNAL_PORT, &n16, 2) == 0);
}

TEST_F(EncodeRequestTest, ExternalPortMAP) {
	pcp_request_t req;
	uint8_t *in;
	uint16_t len, n16;
	int ret;

	memset(&req, 0, sizeof(req));
	req.opcode = PCP_OPCODE_MAP;
	req.extport = 1234;
	ret = pcp_makerequest(&pcp, &req, NULL);
	ASSERT_EQ(ret, PCP_OK);
	in = NULL;
	len = 0;
	ret = _pcp_getrequest(&pcp, &in, &len);
	ASSERT_EQ(ret, PCP_OK);
	ASSERT_TRUE(in != NULL);
	EXPECT_EQ(len, PCP_LEN_MAP);
	n16 = htons(1234);
	EXPECT_TRUE(memcmp(in + PCP_MAP_EXTERNAL_PORT, &n16, 2) == 0);
}

TEST_F(EncodeRequestTest, ExternalPortPEER) {
	pcp_request_t req;
	uint8_t *in;
	uint16_t len, n16;
	int ret;

	memset(&req, 0, sizeof(req));
	req.opcode = PCP_OPCODE_PEER;
	req.extport = 1234;
	ret = pcp_makerequest(&pcp, &req, NULL);
	ASSERT_EQ(ret, PCP_OK);
	in = NULL;
	len = 0;
	ret = _pcp_getrequest(&pcp, &in, &len);
	ASSERT_EQ(ret, PCP_OK);
	ASSERT_TRUE(in != NULL);
	EXPECT_EQ(len, PCP_LEN_PEER);
	n16 = htons(1234);
	EXPECT_TRUE(memcmp(in + PCP_PEER_EXTERNAL_PORT, &n16, 2) == 0);
}

TEST_F(EncodeRequestTest, ExternalAddressMAP) {
	pcp_request_t req;
	uint8_t *in, addr[16];
	uint16_t len;
	int ret;

	memset(&req, 0, sizeof(req));
	req.opcode = PCP_OPCODE_MAP;
	ret = inet_pton(AF_INET6, ADDR6, addr);
	ASSERT_EQ(ret, 1);
	memcpy(req.extaddr6, addr, 16);
	ret = pcp_makerequest(&pcp, &req, NULL);
	ASSERT_EQ(ret, PCP_OK);
	in = NULL;
	len = 0;
	ret = _pcp_getrequest(&pcp, &in, &len);
	ASSERT_EQ(ret, PCP_OK);
	ASSERT_TRUE(in != NULL);
	EXPECT_EQ(len, PCP_LEN_MAP);
	EXPECT_TRUE(memcmp(in + PCP_MAP_EXTERNAL_ADDR, addr, 16) == 0);
}

TEST_F(EncodeRequestTest, ExternalAddressPEER) {
	pcp_request_t req;
	uint8_t *in, addr[16];
	uint16_t len;
	int ret;

	memset(&req, 0, sizeof(req));
	req.opcode = PCP_OPCODE_PEER;
	ret = inet_pton(AF_INET6, ADDR6, addr);
	ASSERT_EQ(ret, 1);
	memcpy(req.extaddr6, addr, 16);
	ret = pcp_makerequest(&pcp, &req, NULL);
	ASSERT_EQ(ret, PCP_OK);
	in = NULL;
	len = 0;
	ret = _pcp_getrequest(&pcp, &in, &len);
	ASSERT_EQ(ret, PCP_OK);
	ASSERT_TRUE(in != NULL);
	EXPECT_EQ(len, PCP_LEN_PEER);
	EXPECT_TRUE(memcmp(in + PCP_PEER_EXTERNAL_ADDR, addr, 16) == 0);
}

TEST_F(EncodeRequestTest, RemotePeerPort) {
	pcp_request_t req;
	uint8_t *in;
	uint16_t len, n16;
	int ret;

	memset(&req, 0, sizeof(req));
	req.opcode = PCP_OPCODE_PEER;
	req.peerport = 1234;
	ret = pcp_makerequest(&pcp, &req, NULL);
	ASSERT_EQ(ret, PCP_OK);
	in = NULL;
	len = 0;
	ret = _pcp_getrequest(&pcp, &in, &len);
	ASSERT_EQ(ret, PCP_OK);
	ASSERT_TRUE(in != NULL);
	EXPECT_EQ(len, PCP_LEN_PEER);
	n16 = htons(1234);
	EXPECT_TRUE(memcmp(in + PCP_PEER_REMOTE_PORT, &n16, 2) == 0);
}

TEST_F(EncodeRequestTest, RemotePeerAddress) {
	pcp_request_t req;
	uint8_t *in, addr[16];
	uint16_t len;
	int ret;

	memset(&req, 0, sizeof(req));
	req.opcode = PCP_OPCODE_PEER;
	ret = inet_pton(AF_INET6, ADDR6, addr);
	ASSERT_EQ(ret, 1);
	memcpy(req.remotepeer6, addr, 16);
	ret = pcp_makerequest(&pcp, &req, NULL);
	ASSERT_EQ(ret, PCP_OK);
	in = NULL;
	len = 0;
	ret = _pcp_getrequest(&pcp, &in, &len);
	ASSERT_EQ(ret, PCP_OK);
	ASSERT_TRUE(in != NULL);
	EXPECT_EQ(len, PCP_LEN_PEER);
	EXPECT_TRUE(memcmp(in + PCP_PEER_REMOTE_ADDR, addr, 16) == 0);
}

TEST_F(EncodeRequestTest, ReservedMAP) {
	pcp_request_t req;
	uint8_t *in;
	uint16_t len;
	int ret;

	memset(&req, 0, sizeof(req));
	req.opcode = PCP_OPCODE_MAP;
	req.lifetime = 123456789;
	req.protocol = IPPROTO_UDP;
	req.intport = 1234;
	ret = inet_pton(AF_INET6, INTSAMPLE, req.intaddr6);
	ASSERT_EQ(ret, 1);
	req.extport = 5678;
	ret = inet_pton(AF_INET6, EXTSAMPLE, req.extaddr6);
	ASSERT_EQ(ret, 1);
	ret = pcp_makerequest(&pcp, &req, NULL);
	ASSERT_EQ(ret, PCP_OK);
	in = NULL;
	len = 0;
	ret = _pcp_getrequest(&pcp, &in, &len);
	ASSERT_EQ(ret, PCP_OK);
	ASSERT_TRUE(in != NULL);
	EXPECT_EQ(len, PCP_LEN_MAP);
	EXPECT_EQ(in[PCP_RESULT_CODE - 1], 0);
	EXPECT_EQ(in[PCP_RESULT_CODE], 0);
	EXPECT_EQ(in[PCP_MAP_PROTOCOL + 1], 0);
	EXPECT_EQ(in[PCP_MAP_PROTOCOL + 2], 0);
	EXPECT_EQ(in[PCP_MAP_PROTOCOL + 3], 0);
}

TEST_F(EncodeRequestTest, ReservedPEER) {
	pcp_request_t req;
	uint8_t *in;
	uint16_t len;
	int ret;

	memset(&req, 0, sizeof(req));
	req.opcode = PCP_OPCODE_PEER;
	req.lifetime = 123456789;
	req.protocol = IPPROTO_UDP;
	req.intport = 1234;
	ret = inet_pton(AF_INET6, INTSAMPLE, req.intaddr6);
	ASSERT_EQ(ret, 1);
	req.extport = 5678;
	ret = inet_pton(AF_INET6, EXTSAMPLE, req.extaddr6);
	ASSERT_EQ(ret, 1);
	req.peerport = 9876;
	ret = inet_pton(AF_INET6, PEERSAMPLE, req.remotepeer6);
	ASSERT_EQ(ret, 1);
	ret = pcp_makerequest(&pcp, &req, NULL);
	ASSERT_EQ(ret, PCP_OK);
	in = NULL;
	len = 0;
	ret = _pcp_getrequest(&pcp, &in, &len);
	ASSERT_EQ(ret, PCP_OK);
	ASSERT_TRUE(in != NULL);
	EXPECT_EQ(len, PCP_LEN_PEER);
	EXPECT_EQ(in[PCP_RESULT_CODE - 1], 0);
	EXPECT_EQ(in[PCP_RESULT_CODE], 0);
	EXPECT_EQ(in[PCP_PEER_PROTOCOL + 1], 0);
	EXPECT_EQ(in[PCP_PEER_PROTOCOL + 2], 0);
	EXPECT_EQ(in[PCP_PEER_PROTOCOL + 3], 0);
	EXPECT_EQ(in[PCP_PEER_REMOTE_PORT + 2], 0);
	EXPECT_EQ(in[PCP_PEER_REMOTE_PORT + 3], 0);
}

TEST_F(EncodeRequestTest, OptionsMAP) {
	pcp_request_t req;
	pcp_option_t opt1, opt2, opt3;
	uint8_t *in;
	uint16_t len;
	int ret;
	uint8_t data1[2] = { 'a', 'b' };
	uint8_t data2[5] = { 'f', 'e', 'd', 'c', 'b' };
	uint8_t data3[8] = { '0', '1', '2', '3', '4', '5', '6', '7' };
	const uint8_t boptions[32] = {
		0x11, 0x00, 0x00, 0x02,
		0x61, 0x62, 0x00, 0x00,
		0xd9, 0x00, 0x00, 0x05,
		0x66, 0x65, 0x64, 0x63, 0x62, 0x00, 0x00, 0x00,
		0x75, 0x00, 0x00, 0x08,
		0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37 };

	memset(&req, 0, sizeof(req));
	req.opcode = PCP_OPCODE_MAP;
	memset(&opt1, 0, sizeof(opt1));
	opt1.code = 17;
	opt1.length = sizeof(data1);
	opt1.data = data1;
	ret = _pcp_setoption(&options, &opt1);
	ASSERT_EQ(ret, PCP_OK);
	memset(&opt2, 0, sizeof(opt1));
	opt2.code = 217;
	opt2.length = sizeof(data2);
	opt2.data = data2;
	ret = _pcp_setoption(&options, &opt2);
	ASSERT_EQ(ret, PCP_OK);
	memset(&opt3, 0, sizeof(opt1));
	opt3.code = 117;
	opt3.length = sizeof(data3);
	opt3.data = data3;
	ret = _pcp_setoption(&options, &opt3);
	ASSERT_EQ(ret, PCP_OK);
	ret = pcp_makerequest(&pcp, &req, options);
	ASSERT_EQ(ret, PCP_OK);
	in = NULL;
	len = 0;
	ret = _pcp_getrequest(&pcp, &in, &len);
	ASSERT_EQ(ret, PCP_OK);
	ASSERT_TRUE(in != NULL);
	EXPECT_EQ(len, PCP_LEN_MAP + sizeof(boptions));
	EXPECT_TRUE(memcmp(in + PCP_LEN_MAP, boptions, sizeof(boptions)) == 0);
}

TEST_F(EncodeRequestTest, OptionsPEER) {
	pcp_request_t req;
	pcp_option_t opt1, opt2, opt3;
	uint8_t *in;
	uint16_t len;
	int ret;
	uint8_t data1[2] = { 'a', 'b' };
	uint8_t data2[5] = { 'f', 'e', 'd', 'c', 'b' };
	uint8_t data3[8] = { '0', '1', '2', '3', '4', '5', '6', '7' };
	const uint8_t boptions[] = {
		0x11, 0x00, 0x00, 0x02,
		0x61, 0x62, 0x00, 0x00,
		0xd9, 0x00, 0x00, 0x05,
		0x66, 0x65, 0x64, 0x63, 0x62, 0x00, 0x00, 0x00,
		0x75, 0x00, 0x00, 0x08,
		0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37 };

	memset(&req, 0, sizeof(req));
	req.opcode = PCP_OPCODE_PEER;
	memset(&opt1, 0, sizeof(opt1));
	opt1.code = 17;
	opt1.length = sizeof(data1);
	opt1.data = data1;
	ret = _pcp_setoption(&options, &opt1);
	ASSERT_EQ(ret, PCP_OK);
	memset(&opt2, 0, sizeof(opt1));
	opt2.code = 217;
	opt2.length = sizeof(data2);
	opt2.data = data2;
	ret = _pcp_setoption(&options, &opt2);
	ASSERT_EQ(ret, PCP_OK);
	memset(&opt3, 0, sizeof(opt1));
	opt3.code = 117;
	opt3.length = sizeof(data3);
	opt3.data = data3;
	ret = _pcp_setoption(&options, &opt3);
	ASSERT_EQ(ret, PCP_OK);
	ret = pcp_makerequest(&pcp, &req, options);
	ASSERT_EQ(ret, PCP_OK);
	in = NULL;
	len = 0;
	ret = _pcp_getrequest(&pcp, &in, &len);
	ASSERT_EQ(ret, PCP_OK);
	ASSERT_TRUE(in != NULL);
	EXPECT_EQ(len, PCP_LEN_PEER + sizeof(boptions));
	EXPECT_TRUE(memcmp(in + PCP_LEN_PEER, boptions,
			   sizeof(boptions)) == 0);
}

TEST_F(EncodeRequestTest, LargeOption) {
	pcp_request_t req;
	pcp_option_t opt;
	uint8_t *in;
	uint16_t len, n16;
	int ret;
	uint8_t data[499];
	uint8_t boptions[504];

	memset(&req, 0, sizeof(req));
	req.opcode = PCP_OPCODE_MAP;
	memset(&opt, 0, sizeof(opt));
	memset(data, 0xa4, sizeof(data));
	opt.code = 111;
	opt.length = sizeof(data);
	opt.data = data;
	ret = _pcp_setoption(&options, &opt);
	ASSERT_EQ(ret, PCP_OK);
	memset(boptions, 0, sizeof(boptions));
	boptions[PCP_OPT_CODE] = opt.code;
	n16 = htons(opt.length);
	memcpy(boptions + PCP_OPT_LEN, &n16, 2);
	memcpy(boptions + PCP_OPT_DATA, opt.data, opt.length);
	ret = pcp_makerequest(&pcp, &req, options);
	ASSERT_EQ(ret, PCP_OK);
	in = NULL;
	len = 0;
	ret = _pcp_getrequest(&pcp, &in, &len);
	ASSERT_EQ(ret, PCP_OK);
	ASSERT_TRUE(in != NULL);
	EXPECT_EQ(len, PCP_LEN_MAP + sizeof(boptions));
	EXPECT_TRUE(memcmp(in + PCP_LEN_MAP, boptions, sizeof(boptions)) == 0);
}

}

namespace {

class SendRequestTest : public ::testing::Test {
    protected:
	void SetUp() {
		memset(&pcp, 0, sizeof(pcp));
	}

	void TearDown() {
		(void) pcp_close(&pcp);
	}

	pcp_t pcp;
};

TEST_F(SendRequestTest, NoPCP) {
	int ret;

	ret = pcp_sendrequest(NULL);
	EXPECT_EQ(ret, PCP_ERR_INVAL);
}

TEST_F(SendRequestTest, NotInitialized) {
	int ret;

	ret = pcp_sendrequest(&pcp);
	EXPECT_EQ(ret, PCP_ERR_INVAL);
}

TEST_F(SendRequestTest, Closed) {
	struct sockaddr_storage server;
	struct sockaddr *srv;
	struct sockaddr_in *srv4;
	int ret;

	srv = (struct sockaddr *) &server;
	srv4 = (struct sockaddr_in *) &server;

	memset(&server, 0, sizeof(server));
	srv4->sin_family = AF_INET;
	ret = inet_pton(AF_INET, SERVER4, &srv4->sin_addr);
	ASSERT_EQ(ret, 1);
	ret = pcp_init(srv, NULL, &pcp);
	ASSERT_EQ(ret, PCP_OK);
	ret = pcp_close(&pcp);
	EXPECT_EQ(ret, PCP_OK);
	ret = pcp_sendrequest(&pcp);
	EXPECT_EQ(ret, PCP_ERR_INVAL);
}

TEST_F(SendRequestTest, NoRequest) {
	struct sockaddr_storage server;
	struct sockaddr *srv;
	struct sockaddr_in *srv4;
	int ret;

	srv = (struct sockaddr *) &server;
	srv4 = (struct sockaddr_in *) &server;

	memset(&server, 0, sizeof(server));
	srv4->sin_family = AF_INET;
	ret = inet_pton(AF_INET, SERVER4, &srv4->sin_addr);
	ASSERT_EQ(ret, 1);
	ret = pcp_init(srv, NULL, &pcp);
	ASSERT_EQ(ret, PCP_OK);
	ret = pcp_sendrequest(&pcp);
	EXPECT_EQ(ret, PCP_ERR_NOREQUEST);
}

TEST_F(SendRequestTest, BadSend) {
	struct sockaddr_storage server;
	struct sockaddr *srv;
	struct sockaddr_in *srv4;
	pcp_request_t req;
	int ret, fd;

	srv = (struct sockaddr *) &server;
	srv4 = (struct sockaddr_in *) &server;

	memset(&server, 0, sizeof(server));
	srv4->sin_family = AF_INET;
	ret = inet_pton(AF_INET, SERVER4, &srv4->sin_addr);
	ASSERT_EQ(ret, 1);
	ret = pcp_init(srv, NULL, &pcp);
	ASSERT_EQ(ret, PCP_OK);
	memset(&req, 0, sizeof(req));
	req.opcode = PCP_OPCODE_MAP;
	ret = pcp_makerequest(&pcp, &req, NULL);
	EXPECT_EQ(ret, PCP_OK);
#ifdef WIN32
	fd = INVALID_SOCKET;
#else
	fd = -1;
#endif
	ret = pcp_getsocket(&pcp, &fd);
	ASSERT_EQ(ret, PCP_OK);
#ifdef WIN32
	ASSERT_NE(fd, INVALID_SOCKET);
	(void) closesocket(fd);
#else
	ASSERT_GE(fd, 0);
	(void) close(fd);
#endif
	ret = pcp_sendrequest(&pcp);
	EXPECT_EQ(ret, PCP_ERR_SEND);
}

}

namespace {

class RecvResponseTest : public ::testing::Test {
    protected:
	void SetUp() {
		memset(&pcp, 0, sizeof(pcp));
	}

	void TearDown() {
		(void) pcp_close(&pcp);
	}

	pcp_t pcp;
};

TEST_F(RecvResponseTest, NoPCP) {
	pcp_response_t resp;
	int ret;

	ret = pcp_recvresponse(NULL, &resp);
	EXPECT_EQ(ret, PCP_ERR_INVAL);
}

TEST_F(RecvResponseTest, NoResponse) {
	int ret;

	ret = pcp_recvresponse(&pcp, NULL);
	EXPECT_EQ(ret, PCP_ERR_INVAL);
}

TEST_F(RecvResponseTest, NotInitialized) {
	pcp_response_t resp;
	int ret;

	ret = pcp_recvresponse(&pcp, &resp);
	EXPECT_EQ(ret, PCP_ERR_INVAL);
}

TEST_F(RecvResponseTest, Closed) {
	struct sockaddr_storage server;
	struct sockaddr *srv;
	struct sockaddr_in *srv4;
	pcp_response_t resp;
	int ret;

	srv = (struct sockaddr *) &server;
	srv4 = (struct sockaddr_in *) &server;

	memset(&server, 0, sizeof(server));
	srv4->sin_family = AF_INET;
	ret = inet_pton(AF_INET, SERVER4, &srv4->sin_addr);
	ASSERT_EQ(ret, 1);
	ret = pcp_init(srv, NULL, &pcp);
	ASSERT_EQ(ret, PCP_OK);
	ret = pcp_close(&pcp);
	EXPECT_EQ(ret, PCP_OK);
	ret = pcp_recvresponse(&pcp, &resp);
	EXPECT_EQ(ret, PCP_ERR_INVAL);
}

TEST_F(RecvResponseTest, NoService) {
	struct sockaddr_storage server;
	struct sockaddr *srv;
	struct sockaddr_in *srv4;
	pcp_response_t resp;
	int ret;

	srv = (struct sockaddr *) &server;
	srv4 = (struct sockaddr_in *) &server;

	memset(&server, 0, sizeof(server));
	srv4->sin_family = AF_INET;
	ret = inet_pton(AF_INET, SERVER4, &srv4->sin_addr);
	ASSERT_EQ(ret, 1);
	ret = pcp_init(srv, NULL, &pcp);
	ASSERT_EQ(ret, PCP_OK);
	// in the unlikely case we get something: throw it and retry
	do {
		ret = pcp_recvresponse(&pcp, &resp);
	} while (ret == PCP_OK);
	EXPECT_EQ(ret, PCP_TRY_AGAIN);
}

TEST_F(RecvResponseTest, ClosedSocket) {
	struct sockaddr_storage server;
	struct sockaddr *srv;
	struct sockaddr_in *srv4;
	pcp_response_t resp;
	int ret, fd;

	srv = (struct sockaddr *) &server;
	srv4 = (struct sockaddr_in *) &server;

	memset(&server, 0, sizeof(server));
	srv4->sin_family = AF_INET;
	ret = inet_pton(AF_INET, SERVER4, &srv4->sin_addr);
	ASSERT_EQ(ret, 1);
	ret = pcp_init(srv, NULL, &pcp);
	ASSERT_EQ(ret, PCP_OK);
#ifdef WIN32
	fd = INVALID_SOCKET;
#else
	fd = -1;
#endif
	ret = pcp_getsocket(&pcp, &fd);
	ASSERT_EQ(ret, PCP_OK);
#ifdef WIN32
	ASSERT_NE(fd, INVALID_SOCKET);
	(void) closesocket(fd);
#else
	ASSERT_GE(fd, 0);
	(void) close(fd);
#endif
	ret = pcp_recvresponse(&pcp, &resp);
	EXPECT_EQ(ret, PCP_ERR_RECV);
}

}

namespace {

class DecodeResponseTest : public ::testing::Test {
    protected:
	void SetUp() {
		memset(&pcp, 0, sizeof(pcp));
	}

	void TearDown() {
		(void) pcp_close(&pcp);
	}

	pcp_t pcp;
};

TEST_F(DecodeResponseTest, TooShort4) {
	pcp_response_t resp;
	int ret;
	uint8_t in[3] = { 0x01, 0x81, 0x00 };

	ret = _pcp_setresponse(&pcp, in, sizeof(in));
	ASSERT_EQ(ret, PCP_OK);
	ret = _pcp_submitresponse(&pcp, &resp);
	EXPECT_EQ(ret, PCP_ERR_RECVBAD);
	EXPECT_EQ(pcp_debug_offset, sizeof(in));
#ifdef CHECK_DECODE_LINE
	EXPECT_EQ(pcp_debug_line, 749);
#endif
}

TEST_F(DecodeResponseTest, UnsupportedVersion) {
	pcp_response_t resp;
	int ret;
	uint8_t in[4] = { 0x07, 0x81, 0x00, 0x00 };

	ret = _pcp_setresponse(&pcp, in, sizeof(in));
	ASSERT_EQ(ret, PCP_OK);
	ret = _pcp_submitresponse(&pcp, &resp);
	EXPECT_EQ(ret, PCP_ERR_RECVBAD);
	EXPECT_EQ(pcp_debug_offset, PCP_VERSION);
#ifdef CHECK_DECODE_LINE
	EXPECT_EQ(pcp_debug_line, 751);
#endif
}

TEST_F(DecodeResponseTest, NotResponse) {
	pcp_response_t resp;
	int ret;
	uint8_t in[4] = { 0x01, 0x01, 0x00, 0x00 };

	ret = _pcp_setresponse(&pcp, in, sizeof(in));
	ASSERT_EQ(ret, PCP_OK);
	ret = _pcp_submitresponse(&pcp, &resp);
	EXPECT_EQ(ret, PCP_ERR_RECVBAD);
	EXPECT_EQ(pcp_debug_offset, PCP_OPCODE);
#ifdef CHECK_DECODE_LINE
	EXPECT_EQ(pcp_debug_line, 753);
#endif
}

TEST_F(DecodeResponseTest, TooLarge) {
	pcp_response_t resp;
	int ret;
	uint8_t in[1025];

	memset(&in, 0, sizeof(in));
	in[PCP_VERSION] = 1;
	in[PCP_OPCODE] = PCP_OPCODE_MAP | PCP_OPCODE_RESPONSE;
	ret = _pcp_setresponse(&pcp, in, sizeof(in));
	ASSERT_EQ(ret, PCP_OK);
	ret = _pcp_submitresponse(&pcp, &resp);
	EXPECT_EQ(ret, PCP_ERR_RECVBAD);
	EXPECT_EQ(pcp_debug_offset, sizeof(in));
#ifdef CHECK_DECODE_LINE
	EXPECT_EQ(pcp_debug_line, 755);
#endif
}

TEST_F(DecodeResponseTest, BadAlign) {
	pcp_response_t resp;
	int ret;
	uint8_t in[7] = { 0x01, 0x81, 0x00, 0x00, 0x00, 0x00, 0x00 };

	ret = _pcp_setresponse(&pcp, in, sizeof(in));
	ASSERT_EQ(ret, PCP_OK);
	ret = _pcp_submitresponse(&pcp, &resp);
	EXPECT_EQ(ret, PCP_ERR_RECVBAD);
	EXPECT_EQ(pcp_debug_offset, sizeof(in));
#ifdef CHECK_DECODE_LINE
	EXPECT_EQ(pcp_debug_line, 757);
#endif
}

TEST_F(DecodeResponseTest, TooShort24) {
	pcp_response_t resp;
	int ret;
	uint8_t in[20];

	memset(&in, 0, sizeof(in));
	in[PCP_VERSION] = 1;
	in[PCP_OPCODE] = PCP_OPCODE_MAP | PCP_OPCODE_RESPONSE;
	ret = _pcp_setresponse(&pcp, in, sizeof(in));
	ASSERT_EQ(ret, PCP_OK);
	ret = _pcp_submitresponse(&pcp, &resp);
	EXPECT_EQ(ret, PCP_ERR_RECVBAD);
	EXPECT_EQ(pcp_debug_offset, sizeof(in));
#ifdef CHECK_DECODE_LINE
	EXPECT_EQ(pcp_debug_line, 759);
#endif
}

TEST_F(DecodeResponseTest, RapidRecovery) {
	pcp_response_t resp;
	uint32_t n32;
	int ret;
	uint8_t in[24] = {
		0x01, 0x80, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x01, 0xe2, 0x40, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };

	ret = _pcp_setresponse(&pcp, in, sizeof(in));
	ASSERT_EQ(ret, PCP_OK);
	ret = _pcp_submitresponse(&pcp, &resp);
	EXPECT_EQ(ret, PCP_TRY_AGAIN);
	EXPECT_EQ(resp.result, in[PCP_RESULT_CODE]);
	memcpy(&n32, in + PCP_EPOCH, 4);
	EXPECT_EQ(resp.epoch, ntohl(n32));
	EXPECT_EQ(resp.assigned.opcode, in[PCP_OPCODE]);
	EXPECT_EQ(resp.assigned.lifetime, 0);
}

TEST_F(DecodeResponseTest, NoRequest) {
	pcp_response_t resp;
	int ret;
	uint8_t in[PCP_OPCODE_HDR];

	memset(&in, 0, sizeof(in));
	in[PCP_VERSION] = 1;
	in[PCP_OPCODE] = PCP_OPCODE_MAP | PCP_OPCODE_RESPONSE;
	ret = _pcp_setresponse(&pcp, in, sizeof(in));
	ASSERT_EQ(ret, PCP_OK);
	ret = _pcp_submitresponse(&pcp, &resp);
	EXPECT_EQ(ret, PCP_ERR_RECVBAD);
	EXPECT_EQ(pcp_debug_offset, PCP_OPCODE);
#ifdef CHECK_DECODE_LINE
	EXPECT_EQ(pcp_debug_line, 769);
#endif
}

TEST_F(DecodeResponseTest, NotMatchingOpCode) {
	pcp_response_t resp;
	int ret;
	uint8_t req[4] = { 0x01, 0x09 };
	uint8_t in[PCP_OPCODE_HDR];

	memset(&in, 0, sizeof(in));
	in[PCP_VERSION] = 1;
	in[PCP_OPCODE] = PCP_OPCODE_MAP | PCP_OPCODE_RESPONSE;
	ret = _pcp_setresponse(&pcp, in, sizeof(in));
	ASSERT_EQ(ret, PCP_OK);
	ret = _pcp_setrequest(&pcp, req, sizeof(req));
	EXPECT_EQ(ret, PCP_OK);
	ret = _pcp_submitresponse(&pcp, &resp);
	EXPECT_EQ(ret, PCP_ERR_RECVBAD);
	EXPECT_EQ(pcp_debug_offset, PCP_OPCODE);
#ifdef CHECK_DECODE_LINE
	EXPECT_EQ(pcp_debug_line, 772);
#endif
}

TEST_F(DecodeResponseTest, UnsupportedOpCode) {
	pcp_response_t resp;
	int ret;
	uint8_t req[4] = { 0x01, 0x09 };
	uint8_t in[PCP_OPCODE_HDR];

	memset(&in, 0, sizeof(in));
	in[PCP_VERSION] = 1;
	in[PCP_OPCODE] = req[PCP_OPCODE] | PCP_OPCODE_RESPONSE;
	ret = _pcp_setresponse(&pcp, in, sizeof(in));
	ASSERT_EQ(ret, PCP_OK);
	ret = _pcp_setrequest(&pcp, req, sizeof(req));
	EXPECT_EQ(ret, PCP_OK);
	ret = _pcp_submitresponse(&pcp, &resp);
	EXPECT_EQ(ret, PCP_ERR_RECVBAD);
	EXPECT_EQ(pcp_debug_offset, PCP_OPCODE);
#ifdef CHECK_DECODE_LINE
	EXPECT_EQ(pcp_debug_line, 778);
#endif
}

TEST_F(DecodeResponseTest, OpCode) {
	pcp_response_t resp;
	int ret;
	uint8_t req[4];
	uint8_t in[PCP_OPCODE_HDR];

	memset(&req, 0, sizeof(req));
	req[PCP_VERSION] = 1;
	req[PCP_OPCODE] = PCP_OPCODE_MAP;
	memset(&in, 0, sizeof(in));
	in[PCP_VERSION] = 1;
	in[PCP_OPCODE] = PCP_OPCODE_MAP | PCP_OPCODE_RESPONSE;
	ret = _pcp_setresponse(&pcp, in, sizeof(in));
	ASSERT_EQ(ret, PCP_OK);
	ret = _pcp_setrequest(&pcp, req, sizeof(req));
	EXPECT_EQ(ret, PCP_OK);
	ret = _pcp_submitresponse(&pcp, &resp);
	EXPECT_NE(ret, PCP_OK);
	EXPECT_EQ(resp.assigned.opcode, PCP_OPCODE_MAP | PCP_OPCODE_RESPONSE);
	memset(&req, 0, sizeof(req));
	req[PCP_VERSION] = 1;
	req[PCP_OPCODE] = PCP_OPCODE_PEER;
	memset(&in, 0, sizeof(in));
	in[PCP_VERSION] = 1;
	in[PCP_OPCODE] = PCP_OPCODE_PEER | PCP_OPCODE_RESPONSE;
	ret = _pcp_setresponse(&pcp, in, sizeof(in));
	ASSERT_EQ(ret, PCP_OK);
	ret = _pcp_setrequest(&pcp, req, sizeof(req));
	EXPECT_EQ(ret, PCP_OK);
	ret = _pcp_submitresponse(&pcp, &resp);
	EXPECT_NE(ret, PCP_OK);
	EXPECT_EQ(resp.assigned.opcode, PCP_OPCODE_PEER | PCP_OPCODE_RESPONSE);
}

TEST_F(DecodeResponseTest, Result) {
	pcp_response_t resp;
	int ret;
	uint8_t req[4];
	uint8_t in[PCP_OPCODE_HDR];

	memset(&req, 0, sizeof(req));
	req[PCP_VERSION] = 1;
	req[PCP_OPCODE] = PCP_OPCODE_MAP;
	memset(&in, 0, sizeof(in));
	in[PCP_VERSION] = 1;
	in[PCP_OPCODE] = PCP_OPCODE_MAP | PCP_OPCODE_RESPONSE;
	in[PCP_RESULT_CODE] = 217;
	ret = _pcp_setresponse(&pcp, in, sizeof(in));
	ASSERT_EQ(ret, PCP_OK);
	ret = _pcp_setrequest(&pcp, req, sizeof(req));
	EXPECT_EQ(ret, PCP_OK);
	ret = _pcp_submitresponse(&pcp, &resp);
	EXPECT_NE(ret, PCP_OK);
	EXPECT_EQ(resp.result, in[PCP_RESULT_CODE]);
}

TEST_F(DecodeResponseTest, Lifetime) {
	pcp_response_t resp;
	uint32_t n32;
	int ret;
	uint8_t req[4];
	uint8_t in[PCP_OPCODE_HDR];

	memset(&req, 0, sizeof(req));
	req[PCP_VERSION] = 1;
	req[PCP_OPCODE] = PCP_OPCODE_MAP;
	memset(&in, 0, sizeof(in));
	in[PCP_VERSION] = 1;
	in[PCP_OPCODE] = PCP_OPCODE_MAP | PCP_OPCODE_RESPONSE;
	n32 = htonl(123456789);
	memcpy(in + PCP_LIFETIME, &n32, 4);
	ret = _pcp_setresponse(&pcp, in, sizeof(in));
	ASSERT_EQ(ret, PCP_OK);
	ret = _pcp_setrequest(&pcp, req, sizeof(req));
	EXPECT_EQ(ret, PCP_OK);
	ret = _pcp_submitresponse(&pcp, &resp);
	EXPECT_NE(ret, PCP_OK);
	EXPECT_EQ(resp.assigned.lifetime, 123456789);
}

TEST_F(DecodeResponseTest, Epoch) {
	pcp_response_t resp;
	uint32_t n32;
	int ret;
	uint8_t req[4];
	uint8_t in[PCP_OPCODE_HDR];

	memset(&req, 0, sizeof(req));
	req[PCP_VERSION] = 1;
	req[PCP_OPCODE] = PCP_OPCODE_MAP;
	memset(&in, 0, sizeof(in));
	in[PCP_VERSION] = 1;
	in[PCP_OPCODE] = PCP_OPCODE_MAP | PCP_OPCODE_RESPONSE;
	n32 = htonl(876543210);
	memcpy(in + PCP_EPOCH, &n32, 4);
	ret = _pcp_setresponse(&pcp, in, sizeof(in));
	ASSERT_EQ(ret, PCP_OK);
	ret = _pcp_setrequest(&pcp, req, sizeof(req));
	EXPECT_EQ(ret, PCP_OK);
	ret = _pcp_submitresponse(&pcp, &resp);
	EXPECT_NE(ret, PCP_OK);
	EXPECT_EQ(resp.epoch, 876543210);
}

TEST_F(DecodeResponseTest, TooShortMAP) {
	pcp_request_t req;
	pcp_response_t resp;
	int ret;
	uint8_t *in, msg[PCP_LEN_MAP - 4];
	uint16_t len;

	memset(&req, 0, sizeof(req));
	req.opcode = PCP_OPCODE_MAP;
	ret = pcp_makerequest(&pcp, &req, NULL);
	ASSERT_EQ(ret, PCP_OK);
	in = NULL;
	len = 0;
	ret = _pcp_getrequest(&pcp, &in, &len);
	ASSERT_EQ(ret, PCP_OK);
	ASSERT_TRUE(in != NULL);
	EXPECT_EQ(len, PCP_LEN_MAP);
	memcpy(msg, in, sizeof(msg));
	msg[PCP_OPCODE] |= PCP_OPCODE_RESPONSE;
	ret = _pcp_setresponse(&pcp, msg, sizeof(msg));
	ASSERT_EQ(ret, PCP_OK);
	ret = _pcp_submitresponse(&pcp, &resp);
	EXPECT_EQ(ret, PCP_ERR_RECVBAD);
	EXPECT_EQ(pcp_debug_offset, sizeof(msg));
#ifdef CHECK_DECODE_LINE
	EXPECT_EQ(pcp_debug_line, 791);
#endif
}

TEST_F(DecodeResponseTest, TooShortPEER) {
	pcp_request_t req;
	pcp_response_t resp;
	int ret;
	uint8_t *in, msg[PCP_LEN_PEER - 4];
	uint16_t len;

	memset(&req, 0, sizeof(req));
	req.opcode = PCP_OPCODE_PEER;
	ret = pcp_makerequest(&pcp, &req, NULL);
	ASSERT_EQ(ret, PCP_OK);
	in = NULL;
	len = 0;
	ret = _pcp_getrequest(&pcp, &in, &len);
	ASSERT_EQ(ret, PCP_OK);
	ASSERT_TRUE(in != NULL);
	EXPECT_EQ(len, PCP_LEN_PEER);
	memcpy(msg, in, sizeof(msg));
	msg[PCP_OPCODE] |= PCP_OPCODE_RESPONSE;
	ret = _pcp_setresponse(&pcp, msg, sizeof(msg));
	ASSERT_EQ(ret, PCP_OK);
	ret = _pcp_submitresponse(&pcp, &resp);
	EXPECT_EQ(ret, PCP_ERR_RECVBAD);
	EXPECT_EQ(pcp_debug_offset, sizeof(msg));
#ifdef CHECK_DECODE_LINE
	EXPECT_EQ(pcp_debug_line, 795);
#endif
}

TEST_F(DecodeResponseTest, ProtocolMismatchMAP) {
	pcp_request_t req;
	pcp_response_t resp;
	int ret;
	uint8_t *in, msg[PCP_LEN_MAP];
	uint16_t len;

	memset(&req, 0, sizeof(req));
	req.opcode = PCP_OPCODE_MAP;
	req.protocol = 117;
	ret = pcp_makerequest(&pcp, &req, NULL);
	ASSERT_EQ(ret, PCP_OK);
	in = NULL;
	len = 0;
	ret = _pcp_getrequest(&pcp, &in, &len);
	ASSERT_EQ(ret, PCP_OK);
	ASSERT_TRUE(in != NULL);
	EXPECT_EQ(len, PCP_LEN_MAP);
	memcpy(msg, in, sizeof(msg));
	msg[PCP_OPCODE] |= PCP_OPCODE_RESPONSE;
	msg[PCP_MAP_PROTOCOL] = 71;
	ret = _pcp_setresponse(&pcp, msg, sizeof(msg));
	ASSERT_EQ(ret, PCP_OK);
	ret = _pcp_submitresponse(&pcp, &resp);
	EXPECT_EQ(ret, PCP_ERR_RECVBAD);
	EXPECT_EQ(pcp_debug_offset, PCP_MAP_PROTOCOL);
#ifdef CHECK_DECODE_LINE
	EXPECT_EQ(pcp_debug_line, 802);
#endif
}

TEST_F(DecodeResponseTest, ProtocolMAP) {
	pcp_request_t req;
	pcp_response_t resp;
	int ret;
	uint8_t *in, msg[PCP_LEN_MAP];
	uint16_t len;

	memset(&req, 0, sizeof(req));
	req.opcode = PCP_OPCODE_MAP;
	req.protocol = 117;
	ret = pcp_makerequest(&pcp, &req, NULL);
	ASSERT_EQ(ret, PCP_OK);
	in = NULL;
	len = 0;
	ret = _pcp_getrequest(&pcp, &in, &len);
	ASSERT_EQ(ret, PCP_OK);
	ASSERT_TRUE(in != NULL);
	EXPECT_EQ(len, PCP_LEN_MAP);
	memcpy(msg, in, sizeof(msg));
	msg[PCP_OPCODE] |= PCP_OPCODE_RESPONSE;
	ret = _pcp_setresponse(&pcp, msg, sizeof(msg));
	ASSERT_EQ(ret, PCP_OK);
	ret = _pcp_submitresponse(&pcp, &resp);
	EXPECT_EQ(resp.assigned.protocol, 117);
}

TEST_F(DecodeResponseTest, InternalPortMismatchMAP) {
	pcp_request_t req;
	pcp_response_t resp;
	int ret;
	uint8_t *in, msg[PCP_LEN_MAP];
	uint16_t len;

	memset(&req, 0, sizeof(req));
	req.opcode = PCP_OPCODE_MAP;
	req.intport = 1234;
	ret = pcp_makerequest(&pcp, &req, NULL);
	ASSERT_EQ(ret, PCP_OK);
	in = NULL;
	len = 0;
	ret = _pcp_getrequest(&pcp, &in, &len);
	ASSERT_EQ(ret, PCP_OK);
	ASSERT_TRUE(in != NULL);
	EXPECT_EQ(len, PCP_LEN_MAP);
	memcpy(msg, in, sizeof(msg));
	msg[PCP_OPCODE] |= PCP_OPCODE_RESPONSE;
	msg[PCP_MAP_INTERNAL_PORT] ^= 7;
	ret = _pcp_setresponse(&pcp, msg, sizeof(msg));
	ASSERT_EQ(ret, PCP_OK);
	ret = _pcp_submitresponse(&pcp, &resp);
	EXPECT_EQ(ret, PCP_ERR_RECVBAD);
	EXPECT_EQ(pcp_debug_offset, PCP_MAP_INTERNAL_PORT);
#ifdef CHECK_DECODE_LINE
	EXPECT_EQ(pcp_debug_line, 806);
#endif
}

TEST_F(DecodeResponseTest, InternalPortMAP) {
	pcp_request_t req;
	pcp_response_t resp;
	int ret;
	uint8_t *in, msg[PCP_LEN_MAP];
	uint16_t len;

	memset(&req, 0, sizeof(req));
	req.opcode = PCP_OPCODE_MAP;
	req.intport = 1234;
	ret = pcp_makerequest(&pcp, &req, NULL);
	ASSERT_EQ(ret, PCP_OK);
	in = NULL;
	len = 0;
	ret = _pcp_getrequest(&pcp, &in, &len);
	ASSERT_EQ(ret, PCP_OK);
	ASSERT_TRUE(in != NULL);
	EXPECT_EQ(len, PCP_LEN_MAP);
	memcpy(msg, in, sizeof(msg));
	msg[PCP_OPCODE] |= PCP_OPCODE_RESPONSE;
	ret = _pcp_setresponse(&pcp, msg, sizeof(msg));
	ASSERT_EQ(ret, PCP_OK);
	ret = _pcp_submitresponse(&pcp, &resp);
	EXPECT_EQ(resp.assigned.intport, 1234);
}

TEST_F(DecodeResponseTest, ExternalPortMAP) {
	pcp_request_t req;
	pcp_response_t resp;
	int ret;
	uint8_t *in, msg[PCP_LEN_MAP];
	uint16_t len, n16;

	memset(&req, 0, sizeof(req));
	req.opcode = PCP_OPCODE_MAP;
	ret = pcp_makerequest(&pcp, &req, NULL);
	ASSERT_EQ(ret, PCP_OK);
	in = NULL;
	len = 0;
	ret = _pcp_getrequest(&pcp, &in, &len);
	ASSERT_EQ(ret, PCP_OK);
	ASSERT_TRUE(in != NULL);
	EXPECT_EQ(len, PCP_LEN_MAP);
	memcpy(msg, in, sizeof(msg));
	msg[PCP_OPCODE] |= PCP_OPCODE_RESPONSE;
	n16 = htons(1234);
	memcpy(msg + PCP_MAP_EXTERNAL_PORT, &n16, 2);
	ret = _pcp_setresponse(&pcp, msg, sizeof(msg));
	ASSERT_EQ(ret, PCP_OK);
	ret = _pcp_submitresponse(&pcp, &resp);
	EXPECT_EQ(resp.assigned.extport, 1234);
}

TEST_F(DecodeResponseTest, ExternalAddressMAP) {
	pcp_request_t req;
	pcp_response_t resp;
	int ret;
	uint8_t *in, msg[PCP_LEN_MAP], addr[16];
	uint16_t len;

	memset(&req, 0, sizeof(req));
	req.opcode = PCP_OPCODE_MAP;
	ret = pcp_makerequest(&pcp, &req, NULL);
	ASSERT_EQ(ret, PCP_OK);
	in = NULL;
	len = 0;
	ret = _pcp_getrequest(&pcp, &in, &len);
	ASSERT_EQ(ret, PCP_OK);
	ASSERT_TRUE(in != NULL);
	EXPECT_EQ(len, PCP_LEN_MAP);
	memcpy(msg, in, sizeof(msg));
	msg[PCP_OPCODE] |= PCP_OPCODE_RESPONSE;
	ret = inet_pton(AF_INET6, ADDR6, addr);
	ASSERT_EQ(ret, 1);
	memcpy(msg + PCP_MAP_EXTERNAL_ADDR, addr, 16);
	ret = _pcp_setresponse(&pcp, msg, sizeof(msg));
	ASSERT_EQ(ret, PCP_OK);
	ret = _pcp_submitresponse(&pcp, &resp);
	EXPECT_TRUE(memcmp(resp.assigned.extaddr6, addr, 16) == 0);
}

TEST_F(DecodeResponseTest, ProtocolMismatchPEER) {
	pcp_request_t req;
	pcp_response_t resp;
	int ret;
	uint8_t *in, msg[PCP_LEN_PEER];
	uint16_t len;

	memset(&req, 0, sizeof(req));
	req.opcode = PCP_OPCODE_PEER;
	req.protocol = 117;
	ret = pcp_makerequest(&pcp, &req, NULL);
	ASSERT_EQ(ret, PCP_OK);
	in = NULL;
	len = 0;
	ret = _pcp_getrequest(&pcp, &in, &len);
	ASSERT_EQ(ret, PCP_OK);
	ASSERT_TRUE(in != NULL);
	EXPECT_EQ(len, PCP_LEN_PEER);
	memcpy(msg, in, sizeof(msg));
	msg[PCP_OPCODE] |= PCP_OPCODE_RESPONSE;
	msg[PCP_PEER_PROTOCOL] = 71;
	ret = _pcp_setresponse(&pcp, msg, sizeof(msg));
	ASSERT_EQ(ret, PCP_OK);
	ret = _pcp_submitresponse(&pcp, &resp);
	EXPECT_EQ(ret, PCP_ERR_RECVBAD);
	EXPECT_EQ(pcp_debug_offset, PCP_PEER_PROTOCOL);
#ifdef CHECK_DECODE_LINE
	EXPECT_EQ(pcp_debug_line, 818);
#endif
}

TEST_F(DecodeResponseTest, ProtocolPEER) {
	pcp_request_t req;
	pcp_response_t resp;
	int ret;
	uint8_t *in, msg[PCP_LEN_PEER];
	uint16_t len;

	memset(&req, 0, sizeof(req));
	req.opcode = PCP_OPCODE_PEER;
	req.protocol = 117;
	ret = pcp_makerequest(&pcp, &req, NULL);
	ASSERT_EQ(ret, PCP_OK);
	in = NULL;
	len = 0;
	ret = _pcp_getrequest(&pcp, &in, &len);
	ASSERT_EQ(ret, PCP_OK);
	ASSERT_TRUE(in != NULL);
	EXPECT_EQ(len, PCP_LEN_PEER);
	memcpy(msg, in, sizeof(msg));
	msg[PCP_OPCODE] |= PCP_OPCODE_RESPONSE;
	ret = _pcp_setresponse(&pcp, msg, sizeof(msg));
	ASSERT_EQ(ret, PCP_OK);
	ret = _pcp_submitresponse(&pcp, &resp);
	EXPECT_EQ(resp.assigned.protocol, 117);
}

TEST_F(DecodeResponseTest, InternalPortMismatchPEER) {
	pcp_request_t req;
	pcp_response_t resp;
	int ret;
	uint8_t *in, msg[PCP_LEN_PEER];
	uint16_t len;

	memset(&req, 0, sizeof(req));
	req.opcode = PCP_OPCODE_PEER;
	req.intport = 1234;
	ret = pcp_makerequest(&pcp, &req, NULL);
	ASSERT_EQ(ret, PCP_OK);
	in = NULL;
	len = 0;
	ret = _pcp_getrequest(&pcp, &in, &len);
	ASSERT_EQ(ret, PCP_OK);
	ASSERT_TRUE(in != NULL);
	EXPECT_EQ(len, PCP_LEN_PEER);
	memcpy(msg, in, sizeof(msg));
	msg[PCP_OPCODE] |= PCP_OPCODE_RESPONSE;
	msg[PCP_PEER_INTERNAL_PORT] ^= 7;
	ret = _pcp_setresponse(&pcp, msg, sizeof(msg));
	ASSERT_EQ(ret, PCP_OK);
	ret = _pcp_submitresponse(&pcp, &resp);
	EXPECT_EQ(ret, PCP_ERR_RECVBAD);
	EXPECT_EQ(pcp_debug_offset, PCP_PEER_INTERNAL_PORT);
#ifdef CHECK_DECODE_LINE
	EXPECT_EQ(pcp_debug_line, 822);
#endif
}

TEST_F(DecodeResponseTest, InternalPortPEER) {
	pcp_request_t req;
	pcp_response_t resp;
	int ret;
	uint8_t *in, msg[PCP_LEN_PEER];
	uint16_t len;

	memset(&req, 0, sizeof(req));
	req.opcode = PCP_OPCODE_PEER;
	req.intport = 1234;
	ret = pcp_makerequest(&pcp, &req, NULL);
	ASSERT_EQ(ret, PCP_OK);
	in = NULL;
	len = 0;
	ret = _pcp_getrequest(&pcp, &in, &len);
	ASSERT_EQ(ret, PCP_OK);
	ASSERT_TRUE(in != NULL);
	EXPECT_EQ(len, PCP_LEN_PEER);
	memcpy(msg, in, sizeof(msg));
	msg[PCP_OPCODE] |= PCP_OPCODE_RESPONSE;
	ret = _pcp_setresponse(&pcp, msg, sizeof(msg));
	ASSERT_EQ(ret, PCP_OK);
	ret = _pcp_submitresponse(&pcp, &resp);
	EXPECT_EQ(resp.assigned.intport, 1234);
}

TEST_F(DecodeResponseTest, ExternalPortPEER) {
	pcp_request_t req;
	pcp_response_t resp;
	int ret;
	uint8_t *in, msg[PCP_LEN_PEER];
	uint16_t len, n16;

	memset(&req, 0, sizeof(req));
	req.opcode = PCP_OPCODE_PEER;
	ret = pcp_makerequest(&pcp, &req, NULL);
	ASSERT_EQ(ret, PCP_OK);
	in = NULL;
	len = 0;
	ret = _pcp_getrequest(&pcp, &in, &len);
	ASSERT_EQ(ret, PCP_OK);
	ASSERT_TRUE(in != NULL);
	EXPECT_EQ(len, PCP_LEN_PEER);
	memcpy(msg, in, sizeof(msg));
	msg[PCP_OPCODE] |= PCP_OPCODE_RESPONSE;
	n16 = htons(1234);
	memcpy(msg + PCP_PEER_EXTERNAL_PORT, &n16, 2);
	ret = _pcp_setresponse(&pcp, msg, sizeof(msg));
	ASSERT_EQ(ret, PCP_OK);
	ret = _pcp_submitresponse(&pcp, &resp);
	EXPECT_EQ(resp.assigned.extport, 1234);
}

TEST_F(DecodeResponseTest, ExternalAddressPEER) {
	pcp_request_t req;
	pcp_response_t resp;
	int ret;
	uint8_t *in, msg[PCP_LEN_PEER], addr[16];
	uint16_t len;

	memset(&req, 0, sizeof(req));
	req.opcode = PCP_OPCODE_PEER;
	ret = pcp_makerequest(&pcp, &req, NULL);
	ASSERT_EQ(ret, PCP_OK);
	in = NULL;
	len = 0;
	ret = _pcp_getrequest(&pcp, &in, &len);
	ASSERT_EQ(ret, PCP_OK);
	ASSERT_TRUE(in != NULL);
	EXPECT_EQ(len, PCP_LEN_PEER);
	memcpy(msg, in, sizeof(msg));
	msg[PCP_OPCODE] |= PCP_OPCODE_RESPONSE;
	ret = inet_pton(AF_INET6, ADDR6, addr);
	ASSERT_EQ(ret, 1);
	memcpy(msg + PCP_PEER_EXTERNAL_ADDR, addr, 16);
	ret = _pcp_setresponse(&pcp, msg, sizeof(msg));
	ASSERT_EQ(ret, PCP_OK);
	ret = _pcp_submitresponse(&pcp, &resp);
	EXPECT_TRUE(memcmp(resp.assigned.extaddr6, addr, 16) == 0);
}

TEST_F(DecodeResponseTest, RemotePeerPortMismatch) {
	pcp_request_t req;
	pcp_response_t resp;
	int ret;
	uint8_t *in, msg[PCP_LEN_PEER];
	uint16_t len;

	memset(&req, 0, sizeof(req));
	req.opcode = PCP_OPCODE_PEER;
	req.peerport = 1234;
	ret = pcp_makerequest(&pcp, &req, NULL);
	ASSERT_EQ(ret, PCP_OK);
	in = NULL;
	len = 0;
	ret = _pcp_getrequest(&pcp, &in, &len);
	ASSERT_EQ(ret, PCP_OK);
	ASSERT_TRUE(in != NULL);
	EXPECT_EQ(len, PCP_LEN_PEER);
	memcpy(msg, in, sizeof(msg));
	msg[PCP_OPCODE] |= PCP_OPCODE_RESPONSE;
	msg[PCP_PEER_REMOTE_PORT] ^= 7;
	ret = _pcp_setresponse(&pcp, msg, sizeof(msg));
	ASSERT_EQ(ret, PCP_OK);
	ret = _pcp_submitresponse(&pcp, &resp);
	EXPECT_EQ(ret, PCP_ERR_RECVBAD);
	EXPECT_EQ(pcp_debug_offset, PCP_PEER_REMOTE_PORT);
#ifdef CHECK_DECODE_LINE
	EXPECT_EQ(pcp_debug_line, 831);
#endif
}

TEST_F(DecodeResponseTest, RemotePeerPort) {
	pcp_request_t req;
	pcp_response_t resp;
	int ret;
	uint8_t *in, msg[PCP_LEN_PEER];
	uint16_t len;

	memset(&req, 0, sizeof(req));
	req.opcode = PCP_OPCODE_PEER;
	req.peerport = 1234;
	ret = pcp_makerequest(&pcp, &req, NULL);
	ASSERT_EQ(ret, PCP_OK);
	in = NULL;
	len = 0;
	ret = _pcp_getrequest(&pcp, &in, &len);
	ASSERT_EQ(ret, PCP_OK);
	ASSERT_TRUE(in != NULL);
	EXPECT_EQ(len, PCP_LEN_PEER);
	memcpy(msg, in, sizeof(msg));
	msg[PCP_OPCODE] |= PCP_OPCODE_RESPONSE;
	ret = _pcp_setresponse(&pcp, msg, sizeof(msg));
	ASSERT_EQ(ret, PCP_OK);
	ret = _pcp_submitresponse(&pcp, &resp);
	EXPECT_EQ(resp.assigned.peerport, 1234);
}

TEST_F(DecodeResponseTest, RemotePeerAddressMismatch) {
	pcp_request_t req;
	pcp_response_t resp;
	int ret;
	uint8_t *in, msg[PCP_LEN_PEER];
	uint16_t len;

	memset(&req, 0, sizeof(req));
	req.opcode = PCP_OPCODE_PEER;
	ret = inet_pton(AF_INET6, ADDR6, req.remotepeer6);
	ASSERT_EQ(ret, 1);
	ret = pcp_makerequest(&pcp, &req, NULL);
	ASSERT_EQ(ret, PCP_OK);
	in = NULL;
	len = 0;
	ret = _pcp_getrequest(&pcp, &in, &len);
	ASSERT_EQ(ret, PCP_OK);
	ASSERT_TRUE(in != NULL);
	EXPECT_EQ(len, PCP_LEN_PEER);
	memcpy(msg, in, sizeof(msg));
	msg[PCP_OPCODE] |= PCP_OPCODE_RESPONSE;
	msg[PCP_PEER_REMOTE_ADDR + 5] ^= 9;
	ret = _pcp_setresponse(&pcp, msg, sizeof(msg));
	ASSERT_EQ(ret, PCP_OK);
	ret = _pcp_submitresponse(&pcp, &resp);
	EXPECT_EQ(ret, PCP_ERR_RECVBAD);
	EXPECT_EQ(pcp_debug_offset, PCP_PEER_REMOTE_ADDR);
#ifdef CHECK_DECODE_LINE
	EXPECT_EQ(pcp_debug_line, 836);
#endif
}

TEST_F(DecodeResponseTest, RemotePeerAddress) {
	pcp_request_t req;
	pcp_response_t resp;
	int ret;
	uint8_t *in, msg[PCP_LEN_PEER], addr[16];
	uint16_t len;

	memset(&req, 0, sizeof(req));
	req.opcode = PCP_OPCODE_PEER;
	ret = inet_pton(AF_INET6, ADDR6, addr);
	ASSERT_EQ(ret, 1);
	memcpy(req.remotepeer6, addr, 16);
	ret = pcp_makerequest(&pcp, &req, NULL);
	ASSERT_EQ(ret, PCP_OK);
	in = NULL;
	len = 0;
	ret = _pcp_getrequest(&pcp, &in, &len);
	ASSERT_EQ(ret, PCP_OK);
	ASSERT_TRUE(in != NULL);
	EXPECT_EQ(len, PCP_LEN_PEER);
	memcpy(msg, in, sizeof(msg));
	msg[PCP_OPCODE] |= PCP_OPCODE_RESPONSE;
	ret = _pcp_setresponse(&pcp, msg, sizeof(msg));
	ASSERT_EQ(ret, PCP_OK);
	ret = _pcp_submitresponse(&pcp, &resp);
	EXPECT_TRUE(memcmp(resp.assigned.remotepeer6, addr, 16) == 0);
}

TEST_F(DecodeResponseTest, TooManyOptionsMAP) {
	uint8_t in[PCP_LEN_MAP + 24];
	uint8_t req[PCP_LEN_MAP];
	uint32_t n32;
	int ret;
	pcp_response_t resp;

	memset(&in, 0, sizeof(in));
	in[PCP_VERSION] = 1;
	in[PCP_OPCODE] = PCP_OPCODE_MAP | PCP_OPCODE_RESPONSE;
	n32 = htonl(123456);
	memcpy(in + PCP_EPOCH, &n32, 4);
	in[PCP_LEN_MAP] = 77;
	in[PCP_LEN_MAP + 4] = 17;
	in[PCP_LEN_MAP + 8] = 20;
	in[PCP_LEN_MAP + 12] = 11;
	in[PCP_LEN_MAP + 16] = 8;
	in[PCP_LEN_MAP + 20] = 177;
	ret = _pcp_setresponse(&pcp, in, sizeof(in));
	ASSERT_EQ(ret, PCP_OK);
	memset(&req, 0, sizeof(req));
	req[PCP_VERSION] = 1;
	req[PCP_OPCODE] = PCP_OPCODE_MAP;
	ret = _pcp_setrequest(&pcp, req, sizeof(req));
	ASSERT_EQ(ret, PCP_OK);
	memset(&resp, 0, sizeof(resp));
	ret = _pcp_submitresponse(&pcp, &resp);
	EXPECT_EQ(ret, PCP_ERR_TOOMANYOPTS);
}

TEST_F(DecodeResponseTest, TooManyOptionsPEER) {
	uint8_t in[PCP_LEN_PEER + 24];
	uint8_t req[PCP_LEN_PEER];
	uint32_t n32;
	int ret;
	pcp_response_t resp;

	memset(&in, 0, sizeof(in));
	in[PCP_VERSION] = 1;
	in[PCP_OPCODE] = PCP_OPCODE_PEER | PCP_OPCODE_RESPONSE;
	n32 = htonl(123456);
	memcpy(in + PCP_EPOCH, &n32, 4);
	in[PCP_LEN_PEER] = 77;
	in[PCP_LEN_PEER + 4] = 17;
	in[PCP_LEN_PEER + 8] = 20;
	in[PCP_LEN_PEER + 12] = 11;
	in[PCP_LEN_PEER + 16] = 8;
	in[PCP_LEN_PEER + 20] = 177;
	ret = _pcp_setresponse(&pcp, in, sizeof(in));
	ASSERT_EQ(ret, PCP_OK);
	memset(&req, 0, sizeof(req));
	req[PCP_VERSION] = 1;
	req[PCP_OPCODE] = PCP_OPCODE_PEER;
	ret = _pcp_setrequest(&pcp, req, sizeof(req));
	ASSERT_EQ(ret, PCP_OK);
	memset(&resp, 0, sizeof(resp));
	ret = _pcp_submitresponse(&pcp, &resp);
	EXPECT_EQ(ret, PCP_ERR_TOOMANYOPTS);
}

TEST_F(DecodeResponseTest, ThirdPartyOptionMAP) {
	pcp_request_t req;
	pcp_response_t resp;
	int ret;
	uint8_t *in, msg[PCP_LEN_MAP + 28], addr[16];
	uint16_t len, n16;

	memset(&req, 0, sizeof(req));
	req.opcode = PCP_OPCODE_MAP;
	ret = pcp_makerequest(&pcp, &req, NULL);
	ASSERT_EQ(ret, PCP_OK);
	in = NULL;
	len = 0;
	ret = _pcp_getrequest(&pcp, &in, &len);
	ASSERT_EQ(ret, PCP_OK);
	ASSERT_TRUE(in != NULL);
	EXPECT_EQ(len, PCP_LEN_MAP);
	memcpy(msg, in, PCP_LEN_MAP);
	memset(msg + len, 0, sizeof(msg) - PCP_LEN_MAP);
	msg[PCP_OPCODE] |= PCP_OPCODE_RESPONSE;
	ret = inet_pton(AF_INET6, ADDR6, addr);
	ASSERT_EQ(ret, 1);
	msg[PCP_LEN_MAP] = PCP_OPTION_THIRD_PARTY;
	n16 = htons(PCP_LEN_OPT3PTY);
	memcpy(msg + PCP_LEN_MAP + 2, &n16, 2);
	memcpy(msg + PCP_LEN_MAP + 4, addr, 16);
	msg[PCP_LEN_MAP + 20] = PCP_OPTION_THIRD_PARTY;
	n16 = htons(4);
	memcpy(msg + PCP_LEN_MAP + 22, &n16, 2);
	memcpy(msg + PCP_LEN_MAP + 24, addr, 4);
	ret = _pcp_setresponse(&pcp, msg, sizeof(msg));
	ASSERT_EQ(ret, PCP_OK);
	in = NULL;
	len = 0;
	ret = _pcp_getresponse(&pcp, &in, &len);
	EXPECT_EQ(ret, PCP_OK);
	EXPECT_TRUE(in != NULL);
	EXPECT_EQ(len, sizeof(msg));
	EXPECT_TRUE(memcmp(in, msg, sizeof(msg)) == 0);
	ret = _pcp_submitresponse(&pcp, &resp);
	ASSERT_EQ(ret, PCP_OK);
	ASSERT_EQ(resp.optcnt, 2);
	EXPECT_EQ(resp.optoffs[0], PCP_LEN_MAP);
	EXPECT_EQ(resp.options[0].code, PCP_OPTION_THIRD_PARTY);
	EXPECT_EQ(resp.options[0].length, PCP_LEN_OPT3PTY);
	EXPECT_EQ(resp.options[0].data, in + PCP_LEN_MAP + 4);
	EXPECT_TRUE(memcmp(resp.options[0].data, addr, 16) == 0);
	EXPECT_TRUE(memcmp(resp.options[0].optaddr6, addr, 16) == 0);
	EXPECT_EQ(resp.optoffs[1], PCP_LEN_MAP + 20);
	EXPECT_EQ(resp.options[1].code, PCP_OPTION_THIRD_PARTY);
	EXPECT_EQ(resp.options[1].length, 4);
	EXPECT_EQ(resp.options[1].data, in + PCP_LEN_MAP + 24);
	EXPECT_TRUE(memcmp(resp.options[1].data, addr, 4) == 0);
	EXPECT_TRUE(memcmp(resp.options[1].optaddr6, zero, 16) == 0);
}

TEST_F(DecodeResponseTest, ThirdPartyOptionPEER) {
	pcp_request_t req;
	pcp_response_t resp;
	int ret;
	uint8_t *in, msg[PCP_LEN_PEER + 28], addr[16];
	uint16_t len, n16;

	memset(&req, 0, sizeof(req));
	req.opcode = PCP_OPCODE_PEER;
	ret = pcp_makerequest(&pcp, &req, NULL);
	ASSERT_EQ(ret, PCP_OK);
	in = NULL;
	len = 0;
	ret = _pcp_getrequest(&pcp, &in, &len);
	ASSERT_EQ(ret, PCP_OK);
	ASSERT_TRUE(in != NULL);
	EXPECT_EQ(len, PCP_LEN_PEER);
	memcpy(msg, in, PCP_LEN_PEER);
	memset(msg + len, 0, sizeof(msg) - PCP_LEN_PEER);
	msg[PCP_OPCODE] |= PCP_OPCODE_RESPONSE;
	ret = inet_pton(AF_INET6, ADDR6, addr);
	ASSERT_EQ(ret, 1);
	msg[PCP_LEN_PEER] = PCP_OPTION_THIRD_PARTY;
	n16 = htons(PCP_LEN_OPT3PTY);
	memcpy(msg + PCP_LEN_PEER + 2, &n16, 2);
	memcpy(msg + PCP_LEN_PEER + 4, addr, 16);
	msg[PCP_LEN_PEER + 20] = PCP_OPTION_THIRD_PARTY;
	n16 = htons(4);
	memcpy(msg + PCP_LEN_PEER + 22, &n16, 2);
	memcpy(msg + PCP_LEN_PEER + 24, addr, 4);
	ret = _pcp_setresponse(&pcp, msg, sizeof(msg));
	ASSERT_EQ(ret, PCP_OK);
	in = NULL;
	len = 0;
	ret = _pcp_getresponse(&pcp, &in, &len);
	EXPECT_EQ(ret, PCP_OK);
	EXPECT_TRUE(in != NULL);
	EXPECT_EQ(len, sizeof(msg));
	EXPECT_TRUE(memcmp(in, msg, sizeof(msg)) == 0);
	ret = _pcp_submitresponse(&pcp, &resp);
	ASSERT_EQ(ret, PCP_OK);
	ASSERT_EQ(resp.optcnt, 2);
	EXPECT_EQ(resp.optoffs[0], PCP_LEN_PEER);
	EXPECT_EQ(resp.options[0].code, PCP_OPTION_THIRD_PARTY);
	EXPECT_EQ(resp.options[0].length, PCP_LEN_OPT3PTY);
	EXPECT_EQ(resp.options[0].data, in + PCP_LEN_PEER + 4);
	EXPECT_TRUE(memcmp(resp.options[0].data, addr, 16) == 0);
	EXPECT_TRUE(memcmp(resp.options[0].optaddr6, addr, 16) == 0);
	EXPECT_EQ(resp.optoffs[1], PCP_LEN_PEER + 20);
	EXPECT_EQ(resp.options[1].code, PCP_OPTION_THIRD_PARTY);
	EXPECT_EQ(resp.options[1].length, 4);
	EXPECT_EQ(resp.options[1].data, in + PCP_LEN_PEER + 24);
	EXPECT_TRUE(memcmp(resp.options[1].data, addr, 4) == 0);
	EXPECT_TRUE(memcmp(resp.options[1].optaddr6, zero, 16) == 0);
}

TEST_F(DecodeResponseTest, FilterOptionMAP) {
	pcp_request_t req;
	pcp_response_t resp;
	int ret;
	uint8_t *in, msg[PCP_LEN_MAP + 40], addr[16];
	uint16_t len, n16;

	memset(&req, 0, sizeof(req));
	req.opcode = PCP_OPCODE_MAP;
	ret = pcp_makerequest(&pcp, &req, NULL);
	ASSERT_EQ(ret, PCP_OK);
	in = NULL;
	len = 0;
	ret = _pcp_getrequest(&pcp, &in, &len);
	ASSERT_EQ(ret, PCP_OK);
	ASSERT_TRUE(in != NULL);
	EXPECT_EQ(len, PCP_LEN_MAP);
	memcpy(msg, in, PCP_LEN_MAP);
	memset(msg + len, 0, sizeof(msg) - PCP_LEN_MAP);
	msg[PCP_OPCODE] |= PCP_OPCODE_RESPONSE;
	ret = inet_pton(AF_INET6, ADDR6, addr);
	ASSERT_EQ(ret, 1);
	msg[PCP_LEN_MAP] = PCP_OPTION_FILTER;
	n16 = htons(PCP_LEN_OPTFLT);
	memcpy(msg + PCP_LEN_MAP + 2, &n16, 2);
	msg[PCP_LEN_MAP + 5] = 64;
	n16 = htons(1234);
	memcpy(msg + PCP_LEN_MAP + 6, &n16, 2);
	memcpy(msg + PCP_LEN_MAP + 8, addr, 16);
	msg[PCP_LEN_MAP + 24] = PCP_OPTION_FILTER;
	n16 = htons(12);
	memcpy(msg + PCP_LEN_MAP + 26, &n16, 2);
	msg[PCP_LEN_MAP + 29] = 64;
	n16 = htons(5678);
	memcpy(msg + PCP_LEN_MAP + 30, &n16, 2);
	memcpy(msg + PCP_LEN_MAP + 32, addr, 8);
	ret = _pcp_setresponse(&pcp, msg, sizeof(msg));
	ASSERT_EQ(ret, PCP_OK);
	in = NULL;
	len = 0;
	ret = _pcp_getresponse(&pcp, &in, &len);
	EXPECT_EQ(ret, PCP_OK);
	EXPECT_TRUE(in != NULL);
	EXPECT_EQ(len, sizeof(msg));
	EXPECT_TRUE(memcmp(in, msg, sizeof(msg)) == 0);
	ret = _pcp_submitresponse(&pcp, &resp);
	ASSERT_EQ(ret, PCP_OK);
	ASSERT_EQ(resp.optcnt, 2);
	EXPECT_EQ(resp.optoffs[0], PCP_LEN_MAP);
	EXPECT_EQ(resp.options[0].code, PCP_OPTION_FILTER);
	EXPECT_EQ(resp.options[0].length, PCP_LEN_OPTFLT);
	EXPECT_EQ(resp.options[0].data, in + PCP_LEN_MAP + 4);
	EXPECT_TRUE(memcmp(resp.options[0].data + 4, addr, 16) == 0);
	EXPECT_TRUE(memcmp(resp.options[0].optaddr6, addr, 16) == 0);
	EXPECT_EQ(resp.optoffs[1], PCP_LEN_MAP + 24);
	EXPECT_EQ(resp.options[1].code, PCP_OPTION_FILTER);
	EXPECT_EQ(resp.options[1].length, 12);
	EXPECT_EQ(resp.options[1].data, in + PCP_LEN_MAP + 28);
	EXPECT_TRUE(memcmp(resp.options[1].data + 4, addr, 8) == 0);
	EXPECT_TRUE(memcmp(resp.options[1].optaddr6, zero, 16) == 0);
}

TEST_F(DecodeResponseTest, FilterOptionPEER) {
	pcp_request_t req;
	pcp_response_t resp;
	int ret;
	uint8_t *in, msg[PCP_LEN_PEER + 40], addr[16];
	uint16_t len, n16;

	memset(&req, 0, sizeof(req));
	req.opcode = PCP_OPCODE_PEER;
	ret = pcp_makerequest(&pcp, &req, NULL);
	ASSERT_EQ(ret, PCP_OK);
	in = NULL;
	len = 0;
	ret = _pcp_getrequest(&pcp, &in, &len);
	ASSERT_EQ(ret, PCP_OK);
	ASSERT_TRUE(in != NULL);
	EXPECT_EQ(len, PCP_LEN_PEER);
	memcpy(msg, in, PCP_LEN_PEER);
	memset(msg + len, 0, sizeof(msg) - PCP_LEN_PEER);
	msg[PCP_OPCODE] |= PCP_OPCODE_RESPONSE;
	ret = inet_pton(AF_INET6, ADDR6, addr);
	ASSERT_EQ(ret, 1);
	msg[PCP_LEN_PEER] = PCP_OPTION_FILTER;
	n16 = htons(PCP_LEN_OPTFLT);
	memcpy(msg + PCP_LEN_PEER + 2, &n16, 2);
	msg[PCP_LEN_PEER + 5] = 64;
	n16 = htons(1234);
	memcpy(msg + PCP_LEN_PEER + 6, &n16, 2);
	memcpy(msg + PCP_LEN_PEER + 8, addr, 16);
	msg[PCP_LEN_PEER + 24] = PCP_OPTION_FILTER;
	n16 = htons(12);
	memcpy(msg + PCP_LEN_PEER + 26, &n16, 2);
	msg[PCP_LEN_PEER + 29] = 64;
	n16 = htons(5678);
	memcpy(msg + PCP_LEN_PEER + 30, &n16, 2);
	memcpy(msg + PCP_LEN_PEER + 32, addr, 8);
	ret = _pcp_setresponse(&pcp, msg, sizeof(msg));
	ASSERT_EQ(ret, PCP_OK);
	in = NULL;
	len = 0;
	ret = _pcp_getresponse(&pcp, &in, &len);
	EXPECT_EQ(ret, PCP_OK);
	EXPECT_TRUE(in != NULL);
	EXPECT_EQ(len, sizeof(msg));
	EXPECT_TRUE(memcmp(in, msg, sizeof(msg)) == 0);
	ret = _pcp_submitresponse(&pcp, &resp);
	ASSERT_EQ(ret, PCP_OK);
	ASSERT_EQ(resp.optcnt, 2);
	EXPECT_EQ(resp.optoffs[0], PCP_LEN_PEER);
	EXPECT_EQ(resp.options[0].code, PCP_OPTION_FILTER);
	EXPECT_EQ(resp.options[0].length, PCP_LEN_OPTFLT);
	EXPECT_EQ(resp.options[0].data, in + PCP_LEN_PEER + 4);
	EXPECT_TRUE(memcmp(resp.options[0].data + 4, addr, 16) == 0);
	EXPECT_TRUE(memcmp(resp.options[0].optaddr6, addr, 16) == 0);
	EXPECT_EQ(resp.optoffs[1], PCP_LEN_PEER + 24);
	EXPECT_EQ(resp.options[1].code, PCP_OPTION_FILTER);
	EXPECT_EQ(resp.options[1].length, 12);
	EXPECT_EQ(resp.options[1].data, in + PCP_LEN_PEER + 28);
	EXPECT_TRUE(memcmp(resp.options[1].data + 4, addr, 8) == 0);
	EXPECT_TRUE(memcmp(resp.options[1].optaddr6, zero, 16) == 0);
}

TEST_F(DecodeResponseTest, OptionsMAP) {
	pcp_request_t req;
	pcp_response_t resp;
	int ret;
	uint8_t *in, msg[PCP_LEN_MAP + 32];
	uint16_t len;
	uint8_t data1[2] = { 'a', 'b' };
	uint8_t data2[5] = { 'f', 'e', 'd', 'c', 'b' };
	uint8_t data3[8] = { '0', '1', '2', '3', '4', '5', '6', '7' };
	const uint8_t boptions[32] = {
		0x11, 0x00, 0x00, 0x02,
		0x61, 0x62, 0x00, 0x00,
		0xd9, 0x00, 0x00, 0x05,
		0x66, 0x65, 0x64, 0x63, 0x62, 0x00, 0x00, 0x00,
		0x75, 0x00, 0x00, 0x08,
		0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37 };

	memset(&req, 0, sizeof(req));
	req.opcode = PCP_OPCODE_MAP;
	ret = pcp_makerequest(&pcp, &req, NULL);
	ASSERT_EQ(ret, PCP_OK);
	in = NULL;
	len = 0;
	ret = _pcp_getrequest(&pcp, &in, &len);
	ASSERT_EQ(ret, PCP_OK);
	ASSERT_TRUE(in != NULL);
	EXPECT_EQ(len, PCP_LEN_MAP);
	memcpy(msg, in, PCP_LEN_MAP);
	memset(msg + len, 0, sizeof(msg) - PCP_LEN_MAP);
	msg[PCP_OPCODE] |= PCP_OPCODE_RESPONSE;
	memcpy(msg + PCP_LEN_MAP, boptions, sizeof(boptions));
	ret = _pcp_setresponse(&pcp, msg, sizeof(msg));
	ASSERT_EQ(ret, PCP_OK);
	in = NULL;
	len = 0;
	ret = _pcp_getresponse(&pcp, &in, &len);
	EXPECT_EQ(ret, PCP_OK);
	EXPECT_TRUE(in != NULL);
	EXPECT_EQ(len, sizeof(msg));
	EXPECT_TRUE(memcmp(in, msg, sizeof(msg)) == 0);
	ret = _pcp_submitresponse(&pcp, &resp);
	ASSERT_EQ(ret, PCP_OK);
	ASSERT_EQ(resp.optcnt, 3);
	EXPECT_EQ(resp.optoffs[0], PCP_LEN_MAP);
	EXPECT_EQ(resp.options[0].code, 0x11);
	EXPECT_EQ(resp.options[0].length, 2);
	EXPECT_EQ(resp.options[0].data, in + PCP_LEN_MAP + 4);
	EXPECT_TRUE(memcmp(resp.options[0].data, data1, 2) == 0);
	EXPECT_EQ(resp.optoffs[1], PCP_LEN_MAP + 8);
	EXPECT_EQ(resp.options[1].code, 0xd9);
	EXPECT_EQ(resp.options[1].length, 5);
	EXPECT_EQ(resp.options[1].data, in + PCP_LEN_MAP + 12);
	EXPECT_TRUE(memcmp(resp.options[1].data, data2, 5) == 0);
	EXPECT_EQ(resp.optoffs[2], PCP_LEN_MAP + 20);
	EXPECT_EQ(resp.options[2].code, 0x75);
	EXPECT_EQ(resp.options[2].length, 8);
	EXPECT_EQ(resp.options[2].data, in + PCP_LEN_MAP + 24);
	EXPECT_TRUE(memcmp(resp.options[2].data, data3, 8) == 0);	
}

TEST_F(DecodeResponseTest, OptionsPEER) {
	pcp_request_t req;
	pcp_response_t resp;
	int ret;
	uint8_t *in, msg[PCP_LEN_PEER + 32];
	uint16_t len;
	uint8_t data1[2] = { 'a', 'b' };
	uint8_t data2[5] = { 'f', 'e', 'd', 'c', 'b' };
	uint8_t data3[8] = { '0', '1', '2', '3', '4', '5', '6', '7' };
	const uint8_t boptions[32] = {
		0x11, 0x00, 0x00, 0x02,
		0x61, 0x62, 0x00, 0x00,
		0xd9, 0x00, 0x00, 0x05,
		0x66, 0x65, 0x64, 0x63, 0x62, 0x00, 0x00, 0x00,
		0x75, 0x00, 0x00, 0x08,
		0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37 };

	memset(&req, 0, sizeof(req));
	req.opcode = PCP_OPCODE_PEER;
	ret = pcp_makerequest(&pcp, &req, NULL);
	ASSERT_EQ(ret, PCP_OK);
	in = NULL;
	len = 0;
	ret = _pcp_getrequest(&pcp, &in, &len);
	ASSERT_EQ(ret, PCP_OK);
	ASSERT_TRUE(in != NULL);
	EXPECT_EQ(len, PCP_LEN_PEER);
	memcpy(msg, in, PCP_LEN_PEER);
	memset(msg + len, 0, sizeof(msg) - PCP_LEN_PEER);
	msg[PCP_OPCODE] |= PCP_OPCODE_RESPONSE;
	memcpy(msg + PCP_LEN_PEER, boptions, sizeof(boptions));
	ret = _pcp_setresponse(&pcp, msg, sizeof(msg));
	ASSERT_EQ(ret, PCP_OK);
	in = NULL;
	len = 0;
	ret = _pcp_getresponse(&pcp, &in, &len);
	EXPECT_EQ(ret, PCP_OK);
	EXPECT_TRUE(in != NULL);
	EXPECT_EQ(len, sizeof(msg));
	EXPECT_TRUE(memcmp(in, msg, sizeof(msg)) == 0);
	ret = _pcp_submitresponse(&pcp, &resp);
	ASSERT_EQ(ret, PCP_OK);
	ASSERT_EQ(resp.optcnt, 3);
	EXPECT_EQ(resp.optoffs[0], PCP_LEN_PEER);
	EXPECT_EQ(resp.options[0].code, 0x11);
	EXPECT_EQ(resp.options[0].length, 2);
	EXPECT_EQ(resp.options[0].data, in + PCP_LEN_PEER + 4);
	EXPECT_TRUE(memcmp(resp.options[0].data, data1, 2) == 0);
	EXPECT_EQ(resp.optoffs[1], PCP_LEN_PEER + 8);
	EXPECT_EQ(resp.options[1].code, 0xd9);
	EXPECT_EQ(resp.options[1].length, 5);
	EXPECT_EQ(resp.options[1].data, in + PCP_LEN_PEER + 12);
	EXPECT_TRUE(memcmp(resp.options[1].data, data2, 5) == 0);
	EXPECT_EQ(resp.optoffs[2], PCP_LEN_PEER + 20);
	EXPECT_EQ(resp.options[2].code, 0x75);
	EXPECT_EQ(resp.options[2].length, 8);
	EXPECT_EQ(resp.options[2].data, in + PCP_LEN_PEER + 24);
	EXPECT_TRUE(memcmp(resp.options[2].data, data3, 8) == 0);	
}

TEST_F(DecodeResponseTest, ReturnOK) {
	pcp_request_t req;
	pcp_response_t resp;
	struct timeval tm;
	int ret;
	uint8_t *in, msg[PCP_LEN_MAP];
	uint16_t len;

	memset(&req, 0, sizeof(req));
	req.opcode = PCP_OPCODE_MAP;
	req.lifetime = 100;
	ret = pcp_makerequest(&pcp, &req, NULL);
	ASSERT_EQ(ret, PCP_OK);
	in = NULL;
	len = 0;
	ret = _pcp_getrequest(&pcp, &in, &len);
	ASSERT_EQ(ret, PCP_OK);
	ASSERT_TRUE(in != NULL);
	EXPECT_EQ(len, PCP_LEN_MAP);
	memcpy(msg, in, sizeof(msg));
	msg[PCP_OPCODE] |= PCP_OPCODE_RESPONSE;
	ret = _pcp_setresponse(&pcp, msg, sizeof(msg));
	ASSERT_EQ(ret, PCP_OK);
	ret = _pcp_submitresponse(&pcp, &resp);
	EXPECT_EQ(ret, PCP_OK);
	ret = pcp_gettimeout(&pcp, &tm, 0);
	EXPECT_EQ(ret, PCP_OK);
	EXPECT_EQ(tm.tv_sec, 0);
	EXPECT_EQ(tm.tv_usec, 0);
}

TEST_F(DecodeResponseTest, ReturnError) {
	pcp_request_t req;
	pcp_response_t resp;
	struct timeval tm;
	uint32_t n32;
	int ret, tries;
	uint8_t *in, msg[PCP_LEN_MAP];
	uint16_t len;

	memset(&req, 0, sizeof(req));
	req.opcode = PCP_OPCODE_MAP;
	req.lifetime = 10;
	ret = pcp_makerequest(&pcp, &req, NULL);
	ASSERT_EQ(ret, PCP_OK);
	in = NULL;
	len = 0;
	ret = _pcp_getrequest(&pcp, &in, &len);
	ASSERT_EQ(ret, PCP_OK);
	ASSERT_TRUE(in != NULL);
	EXPECT_EQ(len, PCP_LEN_MAP);
	memcpy(msg, in, sizeof(msg));
	msg[PCP_OPCODE] |= PCP_OPCODE_RESPONSE;
	msg[PCP_RESULT_CODE] = 100;
	ret = _pcp_setresponse(&pcp, msg, sizeof(msg));
	EXPECT_EQ(ret, PCP_OK);
	ret = _pcp_submitresponse(&pcp, &resp);
	EXPECT_EQ(ret, PCP_ERR_PROTOBASE - 100);
	EXPECT_EQ(resp.result, 100);
	ret = pcp_gettimeout(&pcp, &tm, 0);
	EXPECT_EQ(ret, PCP_OK);
	EXPECT_EQ(tm.tv_sec, 0);
	EXPECT_EQ(tm.tv_usec, 0);
	ret = pcp_setreset(&pcp, 1, NULL);
	EXPECT_EQ(ret, PCP_OK);
	ret = _pcp_submitresponse(&pcp, &resp);
	EXPECT_EQ(ret, PCP_TRY_AGAIN);
	EXPECT_EQ(resp.result, 100);
	ret = pcp_gettimeout(&pcp, &tm, 1);
	EXPECT_EQ(ret, PCP_OK);
	if (tm.tv_sec == 10) {
		EXPECT_EQ(tm.tv_usec, 0);
	} else {
		EXPECT_EQ(tm.tv_sec, 9);
		EXPECT_GE(tm.tv_usec, 998500);
	}
	tries = -1;
	ret = pcp_gettries(&pcp, &tries);
	EXPECT_EQ(ret, PCP_OK);
	EXPECT_EQ(tries, 0);
	ret = pcp_makerequest(&pcp, &req, NULL);
	EXPECT_EQ(ret, PCP_OK);
	n32 = htonl(PCP_RESET_THRESHOLD + 1);
	memcpy(msg + PCP_LIFETIME, &n32, 4);
	ret = _pcp_setresponse(&pcp, msg, sizeof(msg));
	EXPECT_EQ(ret, PCP_OK);
	ret = _pcp_submitresponse(&pcp, &resp);
	EXPECT_EQ(ret, PCP_ERR_PROTOBASE - 100);
	EXPECT_EQ(resp.result, 100);
	ret = pcp_gettimeout(&pcp, &tm, 0);
	EXPECT_EQ(ret, PCP_OK);
	EXPECT_EQ(tm.tv_sec, 0);
	EXPECT_EQ(tm.tv_usec, 0);
}

}

namespace {

class FdInitTest : public ::testing::Test {
    protected:
	void SetUp() {
		memset(&pcp, 0, sizeof(pcp));
	}

	void TearDown() {
		(void) pcp_close(&pcp);
	}

	pcp_t pcp;
};

TEST_F(FdInitTest, NoPCP) {
	int ret;

	ret = _pcp_fdinit(NULL, 123);
	EXPECT_EQ(ret, PCP_ERR_INVAL);
}

TEST_F(FdInitTest, Any) {
	int ret, fd, fd2;

	fd = socket(AF_INET, SOCK_DGRAM, 0);
#ifdef WIN32
	ASSERT_NE(fd, INVALID_SOCKET);
	(void) closesocket(fd);
#else
	ASSERT_GE(fd, 0);
	(void) close(fd);
#endif
	ret = _pcp_fdinit(&pcp, fd);
	ASSERT_EQ(ret, PCP_OK);
	ret = pcp_getsocket(&pcp, &fd2);
	EXPECT_EQ(ret, PCP_OK);
	EXPECT_EQ(fd, fd2);
	ret = pcp_close(&pcp);
	EXPECT_EQ(ret, PCP_ERR_SYSCALL);
}

TEST_F(FdInitTest, Dup) {
	struct sockaddr_storage server;
	struct sockaddr *srv;
	struct sockaddr_in *srv4;
	pcp_request_t req;
	int ret, fd, fd2;
#ifdef WIN32
	WSAPROTOCOL_INFO pi;
	int pid;
#endif

	srv = (struct sockaddr *) &server;
	srv4 = (struct sockaddr_in *) &server;

	memset(&server, 0, sizeof(server));
	srv4->sin_family = AF_INET;
	ret = inet_pton(AF_INET, SERVER4, &srv4->sin_addr);
	ASSERT_EQ(ret, 1);
	ret = pcp_init(srv, NULL, &pcp);
	ASSERT_EQ(ret, PCP_OK);
#ifdef WIN32
	fd = INVALID_SOCKET;
#else
	fd = -1;
#endif
	ret = pcp_getsocket(&pcp, &fd);
	ASSERT_EQ(ret, PCP_OK);
#ifdef WIN32
	ASSERT_NE(fd, INVALID_SOCKET);
	pid = _getpid();
	ret = WSADuplicateSocket(fd, pid, &pi);
	ASSERT_EQ(ret, 0);
	fd2 = WSASocket(pi.iAddressFamily, pi.iSocketType,
			pi.iProtocol, &pi, 0, 0);
	ASSERT_NE(fd2, INVALID_SOCKET);
	(void) closesocket(fd);
#else
	ASSERT_GE(fd, 0);
	fd2 = dup(fd);
	ASSERT_GE(fd2, 0);
	(void) close(fd);
#endif
	ret = _pcp_fdinit(&pcp, fd2);
	ASSERT_EQ(ret, PCP_OK);
	memset(&req, 0, sizeof(req));
	req.opcode = PCP_OPCODE_PEER;
	ret = pcp_makerequest(&pcp, &req, NULL);
	ASSERT_EQ(ret, PCP_OK);
	EXPECT_TRUE(memcmp(req.intaddr6, zero, 16) != 0);
	ret = pcp_sendrequest(&pcp);
	EXPECT_EQ(ret, PCP_OK);
}

}

namespace {

class FillSrcAddrTest : public ::testing::Test {
    protected:
	void SetUp() {
		memset(&pcp, 0, sizeof(pcp));
	}

	void TearDown() {
		(void) pcp_close(&pcp);
	}

	pcp_t pcp;
};

TEST_F(FillSrcAddrTest, NoPCP) {
	pcp_request_t req;
	int ret;

	ret = _pcp_fillsrcaddr(NULL, &req);
	EXPECT_EQ(ret, PCP_ERR_INVAL);
}

TEST_F(FillSrcAddrTest, NoRequest) {
	int ret;

	ret = _pcp_fillsrcaddr(&pcp, NULL);
	EXPECT_EQ(ret, PCP_ERR_INVAL);
}

TEST_F(FillSrcAddrTest, NotInitialized) {
	pcp_request_t req;
	int ret;

	ret = _pcp_fillsrcaddr(&pcp, &req);
	EXPECT_EQ(ret, PCP_ERR_INVAL);
}

TEST_F(FillSrcAddrTest, Closed) {
	struct sockaddr_storage server;
	struct sockaddr *srv;
	struct sockaddr_in *srv4;
	pcp_request_t req;
	int ret;

	srv = (struct sockaddr *) &server;
	srv4 = (struct sockaddr_in *) &server;

	memset(&server, 0, sizeof(server));
	srv4->sin_family = AF_INET;
	ret = inet_pton(AF_INET, SERVER4, &srv4->sin_addr);
	ASSERT_EQ(ret, 1);
	ret = pcp_init(srv, NULL, &pcp);
	ASSERT_EQ(ret, PCP_OK);
	ret = pcp_close(&pcp);
	EXPECT_EQ(ret, PCP_OK);
	ret = _pcp_fillsrcaddr(&pcp, &req);
	EXPECT_EQ(ret, PCP_ERR_INVAL);
}

TEST_F(FillSrcAddrTest, ClosedSocket) {
	struct sockaddr_storage server;
	struct sockaddr *srv;
	struct sockaddr_in *srv4;
	pcp_request_t req;
	int ret, fd;

	srv = (struct sockaddr *) &server;
	srv4 = (struct sockaddr_in *) &server;

	memset(&server, 0, sizeof(server));
	srv4->sin_family = AF_INET;
	ret = inet_pton(AF_INET, SERVER4, &srv4->sin_addr);
	ASSERT_EQ(ret, 1);
	ret = pcp_init(srv, NULL, &pcp);
	ASSERT_EQ(ret, PCP_OK);
#ifdef WIN32
	fd = INVALID_SOCKET;
#else
	fd = -1;
#endif
	ret = pcp_getsocket(&pcp, &fd);
	ASSERT_EQ(ret, PCP_OK);
#ifdef WIN32
	ASSERT_NE(fd, INVALID_SOCKET);
	(void) closesocket(fd);
#else
	ASSERT_GE(fd, 0);
	(void) close(fd);
#endif
	ret = _pcp_fillsrcaddr(&pcp, &req);
	EXPECT_EQ(ret, PCP_ERR_SYSCALL);
}

TEST_F(FillSrcAddrTest, IPv4Source) {
	struct sockaddr_storage server, source;
	struct sockaddr *srv, *src;
	struct sockaddr_in *srv4, *src4;
	pcp_request_t req;
	socklen_t len;
	int ret, fd;

	srv = (struct sockaddr *) &server;
	src = (struct sockaddr *) &source;
	srv4 = (struct sockaddr_in *) &server;
	src4 = (struct sockaddr_in *) &source;

	memset(&server, 0, sizeof(server));
	srv4->sin_family = AF_INET;
	ret = inet_pton(AF_INET, SERVER4, &srv4->sin_addr);
	ASSERT_EQ(ret, 1);
	ret = pcp_init(srv, NULL, &pcp);
	EXPECT_EQ(ret, PCP_OK);
#ifdef WIN32
	fd = INVALID_SOCKET;
#else
	fd = -1;
#endif
	ret = pcp_getsocket(&pcp, &fd);
	ASSERT_EQ(ret, PCP_OK);
#ifdef WIN32
	ASSERT_NE(fd, INVALID_SOCKET);
#else
	ASSERT_GE(fd, 0);
#endif
	memset(&source, 0, sizeof(source));
	len = sizeof(source);
	ret = getsockname(fd, src, &len);
	EXPECT_EQ(ret, 0);
	EXPECT_EQ(src4->sin_family, AF_INET);
	EXPECT_EQ(len, sizeof(*src4));
	memset(&req, 0, sizeof(req));
	ret = _pcp_fillsrcaddr(&pcp, &req);
	EXPECT_EQ(ret, PCP_OK);
	EXPECT_TRUE(memcmp(req.intaddr6, pcp_mapped, 12) == 0);
	EXPECT_TRUE(memcmp(req.intaddr6 + 12, &src4->sin_addr, 4) == 0);
}

TEST_F(FillSrcAddrTest, IPv6AutoFill) {
	struct sockaddr_storage server, source;
	struct sockaddr *srv, *src;
	struct sockaddr_in6 *srv6, *src6;
	pcp_request_t req;
	socklen_t len;
	int ret, fd;

	srv = (struct sockaddr *) &server;
	src = (struct sockaddr *) &source;
	srv6 = (struct sockaddr_in6 *) &server;
	src6 = (struct sockaddr_in6 *) &source;

	memset(&server, 0, sizeof(server));
	srv6->sin6_family = AF_INET6;
#ifdef BSD
	srv6->sin6_len = sizeof(*srv6);
#endif
	ret = inet_pton(AF_INET6, SERVER6, &srv6->sin6_addr);
	ASSERT_EQ(ret, 1);
	ret = pcp_init(srv, NULL, &pcp);
	EXPECT_EQ(ret, PCP_OK);
#ifdef WIN32
	fd = INVALID_SOCKET;
#else
	fd = -1;
#endif
	ret = pcp_getsocket(&pcp, &fd);
	ASSERT_EQ(ret, PCP_OK);
#ifdef WIN32
	ASSERT_NE(fd, INVALID_SOCKET);
#else
	ASSERT_GE(fd, 0);
#endif
	memset(&source, 0, sizeof(source));
	len = sizeof(source);
	ret = getsockname(fd, src, &len);
	EXPECT_EQ(ret, 0);
	EXPECT_EQ(src6->sin6_family, AF_INET6);
	EXPECT_EQ(len, sizeof(*src6));
	memset(&req, 0, sizeof(req));
	memset(&req, 0, sizeof(req));
	ret = _pcp_fillsrcaddr(&pcp, &req);
	EXPECT_EQ(ret, PCP_OK);
	EXPECT_TRUE(memcmp(req.intaddr6, &src6->sin6_addr, 16) == 0);
}

}

namespace {

class SetRequestTest : public ::testing::Test {
    protected:
	void SetUp() {
		memset(&pcp, 0, sizeof(pcp));
	}

	void TearDown() {
		(void) pcp_close(&pcp);
	}

	pcp_t pcp;
};

TEST_F(SetRequestTest, NoPCP) {
	uint8_t req[PCP_LEN_MAP];
	int ret;

	ret = _pcp_setrequest(NULL, NULL, 0);
	EXPECT_EQ(ret, PCP_ERR_INVAL);
	ret = _pcp_setrequest(NULL, req, sizeof(req));
	EXPECT_EQ(ret, PCP_ERR_INVAL);
}

TEST_F(SetRequestTest, NoRequest) {
	int ret;

	ret = _pcp_setrequest(&pcp, NULL, PCP_LEN_MAP);
	EXPECT_EQ(ret, PCP_ERR_INVAL);
}

TEST_F(SetRequestTest, Set) {
	uint8_t req[PCP_LEN_MAP], *in;
	uint16_t len;
	int ret;

	ret = _pcp_setrequest(&pcp, NULL, 0);
	EXPECT_EQ(ret, PCP_OK);
	req[17] = 17;
	ret = _pcp_setrequest(&pcp, req, sizeof(req));
	EXPECT_EQ(ret, PCP_OK);
	in = NULL;
	len = 0;
	ret = _pcp_getrequest(&pcp, &in, &len);
	EXPECT_EQ(ret, PCP_OK);
	EXPECT_TRUE(in != NULL);
	EXPECT_EQ(len, sizeof(req));
	EXPECT_TRUE(memcmp(in, req, sizeof(req)) == 0);
	ret = _pcp_setrequest(&pcp, NULL, 0);
	EXPECT_EQ(ret, PCP_OK);
	ret = _pcp_getrequest(&pcp, &in, &len);
	EXPECT_EQ(ret, PCP_OK);
	EXPECT_TRUE(in == NULL);
	EXPECT_EQ(len, 0);
}

}

namespace {

class GetRequestTest : public ::testing::Test {
    protected:
	void SetUp() {
		memset(&pcp, 0, sizeof(pcp));
	}

	void TearDown() {
		(void) pcp_close(&pcp);
	}

	pcp_t pcp;
};

TEST_F(GetRequestTest, NoPCP) {
	uint8_t *req;
	uint16_t len;
	int ret;

	ret = _pcp_getrequest(NULL, &req, &len);
	EXPECT_EQ(ret, PCP_ERR_INVAL);
}

TEST_F(GetRequestTest, NoRequest) {
	uint16_t len;
	int ret;

	ret = _pcp_getrequest(&pcp, NULL, &len);
	EXPECT_EQ(ret, PCP_ERR_INVAL);
}

TEST_F(GetRequestTest, NoRequestLength) {
	uint8_t *req;
	int ret;

	ret = _pcp_getrequest(&pcp, &req, NULL);
	EXPECT_EQ(ret, PCP_ERR_INVAL);
}

TEST_F(GetRequestTest, Get) {
	pcp_request_t req;
	uint8_t *in;
	uint16_t len;
	int ret;

	ret = pcp_setautofill(&pcp, 0, NULL);
	EXPECT_EQ(ret, PCP_OK);
	memset(&req, 0, sizeof(req));
	req.opcode = PCP_OPCODE_MAP;
	ret = pcp_makerequest(&pcp, &req, NULL);
	EXPECT_EQ(ret, PCP_OK);
	in = NULL;
	len = 0;
	ret = _pcp_getrequest(&pcp, &in, &len);
	EXPECT_EQ(ret, PCP_OK);
	EXPECT_TRUE(in != NULL);
	EXPECT_EQ(len, PCP_LEN_MAP);
	ret = pcp_makerequest(&pcp, NULL, NULL);
	EXPECT_EQ(ret, PCP_OK);
	ret = _pcp_getrequest(&pcp, &in, &len);
	EXPECT_EQ(ret, PCP_OK);
	EXPECT_TRUE(in == NULL);
	EXPECT_EQ(len, 0);
}

}

namespace {

class SetResponseTest : public ::testing::Test {
    protected:
	void SetUp() {
		memset(&pcp, 0, sizeof(pcp));
	}

	void TearDown() {
		(void) pcp_close(&pcp);
	}

	pcp_t pcp;
};

TEST_F(SetResponseTest, NoPCP) {
	uint8_t resp[PCP_LEN_MAP];
	int ret;

	ret = _pcp_setresponse(NULL, NULL, 0);
	EXPECT_EQ(ret, PCP_ERR_INVAL);
	ret = _pcp_setresponse(NULL, resp, sizeof(resp));
	EXPECT_EQ(ret, PCP_ERR_INVAL);
}

TEST_F(SetResponseTest, NoResponse) {
	int ret;

	ret = _pcp_setresponse(&pcp, NULL, PCP_LEN_MAP);
	EXPECT_EQ(ret, PCP_ERR_INVAL);
}

TEST_F(SetResponseTest, TooLarge) {
	int ret;

	ret = _pcp_setresponse(&pcp, buf, PCP_RESPLENGTH);
	EXPECT_EQ(ret, PCP_OK);
	ret = _pcp_setresponse(&pcp, buf, PCP_RESPLENGTH + 1);
	EXPECT_EQ(ret, PCP_ERR_INVAL);
}

TEST_F(SetResponseTest, Set) {
	uint8_t resp[PCP_LEN_MAP], *in;
	uint16_t len;
	int ret;

	ret = _pcp_setresponse(&pcp, NULL, 0);
	EXPECT_EQ(ret, PCP_OK);
	resp[17] = 17;
	ret = _pcp_setresponse(&pcp, resp, sizeof(resp));
	EXPECT_EQ(ret, PCP_OK);
	in = NULL;
	len = 0;
	ret = _pcp_getresponse(&pcp, &in, &len);
	EXPECT_EQ(ret, PCP_OK);
	EXPECT_TRUE(in != NULL);
	EXPECT_EQ(len, sizeof(resp));
	EXPECT_TRUE(memcmp(in, resp, sizeof(resp)) == 0);
	ret = _pcp_setresponse(&pcp, NULL, 0);
	EXPECT_EQ(ret, PCP_OK);
	ret = _pcp_getresponse(&pcp, &in, &len);
	EXPECT_EQ(ret, PCP_OK);
	EXPECT_EQ(len, 0);
}

}

namespace {

class GetResponseTest : public ::testing::Test {
    protected:
	void SetUp() {
		memset(&pcp, 0, sizeof(pcp));
	}

	void TearDown() {
		(void) pcp_close(&pcp);
	}

	pcp_t pcp;
};

TEST_F(GetResponseTest, NoPCP) {
	uint8_t *resp;
	uint16_t len;
	int ret;

	ret = _pcp_getresponse(NULL, &resp, &len);
	EXPECT_EQ(ret, PCP_ERR_INVAL);
}

TEST_F(GetResponseTest, NoResponse) {
	uint16_t len;
	int ret;

	ret = _pcp_getresponse(&pcp, NULL, &len);
	EXPECT_EQ(ret, PCP_ERR_INVAL);
}

TEST_F(GetResponseTest, NoResponseLength) {
	uint8_t *resp;
	int ret;

	ret = _pcp_getresponse(&pcp, &resp, NULL);
	EXPECT_EQ(ret, PCP_ERR_INVAL);
}

}

namespace {

class SubmitResponseTest : public ::testing::Test {
    protected:
	void SetUp() {
		memset(&pcp, 0, sizeof(pcp));
	}

	void TearDown() {
		(void) pcp_close(&pcp);
	}

	pcp_t pcp;
};

TEST_F(SubmitResponseTest, NoPCP) {
	pcp_response_t resp;
	int ret;

	ret = _pcp_submitresponse(NULL, &resp);
	EXPECT_EQ(ret, PCP_ERR_INVAL);
}

TEST_F(SubmitResponseTest, NoResponse) {
	int ret;

	ret = _pcp_submitresponse(&pcp, NULL);
	EXPECT_EQ(ret, PCP_ERR_INVAL);
}

}

namespace {

class SetOptionTest : public ::testing::Test {
    protected:
	void SetUp() {
		options = NULL;
	}

	void TearDown() {
		pcp_freeoptions(options);
	}

	pcp_t pcp;
	pcp_option_t **options;
};

TEST_F(SetOptionTest, NoOptions) {
	pcp_option_t opt;
	int ret;

	ret = _pcp_setoption(NULL, &opt);
	EXPECT_EQ(ret, PCP_ERR_INVAL);
}

TEST_F(SetOptionTest, NoOption) {
	int ret;

	ret = _pcp_setoption(&options, NULL);
	EXPECT_EQ(ret, PCP_ERR_INVAL);
}

TEST_F(SetOptionTest, TooManyOptions) {
	pcp_option_t opt;
	int ret;

	memset(&opt, 0, sizeof(opt));
	opt.code = 1;
	ret = _pcp_setoption(&options, &opt);
	EXPECT_EQ(ret, PCP_OK);
	opt.code = 2;
	ret = _pcp_setoption(&options, &opt);
	EXPECT_EQ(ret, PCP_OK);
	opt.code = 3;
	ret = _pcp_setoption(&options, &opt);
	EXPECT_EQ(ret, PCP_OK);
	opt.code = 4;
	ret = _pcp_setoption(&options, &opt);
	EXPECT_EQ(ret, PCP_OK);
	opt.code = 5;
	ret = _pcp_setoption(&options, &opt);
	EXPECT_EQ(ret, PCP_ERR_TOOMANYOPTS);
}

TEST_F(SetOptionTest, Set) {
	pcp_option_t opt;
	uint8_t data[16];
	int ret;

	ret = pcp_third_party4(&options, 0x1234fedc);
	EXPECT_EQ(ret, PCP_OK);
	EXPECT_TRUE(options[0] != NULL);
	EXPECT_TRUE(options[1] == NULL);
	memcpy(&opt, options[0], sizeof(opt));
	EXPECT_EQ(opt.length, sizeof(data));
	ret = _pcp_setoption(&options, &opt);
	EXPECT_EQ(ret, PCP_OK);
	EXPECT_TRUE(options[0] != NULL);
	EXPECT_TRUE(options[1] != NULL);
	EXPECT_TRUE(options[2] == NULL);
	memcpy(data, opt.data, sizeof(data));
	EXPECT_EQ(options[0]->code, options[1]->code);
	EXPECT_EQ(options[0]->length, options[1]->length);
	EXPECT_TRUE(memcmp(options[0]->data, options[1]->data,
			   options[0]->length) == 0);
	options[0]->data[15] = 0x22;
	EXPECT_TRUE(memcmp(options[0]->data, options[1]->data,
			   options[0]->length) != 0);
	memcpy(data, opt.data, sizeof(data));
	opt.data = data;
	pcp_freeoptions(options);
	options = NULL;
	ret = _pcp_setoption(&options, &opt);
	EXPECT_EQ(ret, PCP_OK);
	EXPECT_TRUE(options[0] != NULL);
	EXPECT_TRUE(options[1] == NULL);
	EXPECT_EQ(options[0]->code, opt.code);
	EXPECT_EQ(options[0]->length, opt.length);
	EXPECT_TRUE(memcmp(options[0]->data, data, sizeof(data)) == 0);
	EXPECT_NE(options[0]->data, data);
}

}

int
main(int argc, char *argv[])
{
#ifdef WIN32
	WSADATA wsaData;
#endif
	int ret;

	::testing::InitGoogleTest(&argc, argv);

#ifdef WIN32
	WSAStartup(MAKEWORD(2,2), &wsaData);
#endif
	ret = RUN_ALL_TESTS();
#ifdef WIN32
	WSACleanup();
#endif
	return ret;
}
