#!/usr/bin/env python

# Copyright (C) 2011  Internet Systems Consortium, Inc. ("ISC")
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND ISC DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
# AND FITNESS.  IN NO EVENT SHALL ISC BE LIABLE FOR ANY SPECIAL, DIRECT,
# INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
# LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
# OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
# PERFORMANCE OF THIS SOFTWARE.
#

"""
Python PCP client library unit tests

Francis_Dupont@isc.org, November 2011

Usage examples:
	python libpcp_test.py
        python -m unittest -v libpcp_test
"""   

__version__ = "$Id: 797303d13eefbeff376ce250a6d942285999186c $"

import libpcp
import os
import socket
import unittest

# global settables

SERVER4 = '192.0.0.1'
SERVER6 = '2001:0:0:1::1'
ADDR4 = '10.1.2.3'
ADDR6 = '2a01:e35:2ec7:83d0:cb3a:14ff:fe1f:3b06'

MAPPED = '\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xff\xff'

# numbering:
#	classes:
#		0 = functions
#		1 = basic methods
#		2 = support methods
#		3 = options
#		4 = internal functions
#		9 = encode/decode
#	tests:
#		0 = type
#		1 = requirement
#		2 = basic behavior
#		3 = advanced

class Test01Open(unittest.TestCase):
    """PCP open (C pcp_init())"""

    def setUp(self):
        """just create the object"""
        self.pcp = libpcp.PCP()

    def tearDown(self):
        """cleanup"""
        self.pcp.close()

    def test00_create(self):
        """the created object must be not None"""
        self.assertIsNotNone(self.pcp)

    def test01_arg(self):
        """unknown keyword"""
        self.assertRaises(TypeError, self.pcp.open, [], { 'foo' : 1 })

    def test02_args(self):
        """too many (>5) arguments"""
        self.assertRaises(TypeError, self.pcp.open, [1, 2, 3, 4, 'too many'])

    def test03_arg0(self):
        """requires at least one argument (server)"""
        self.assertRaises(TypeError, self.pcp.open)

    def test04_server_type(self):
        """server is a string"""
        self.assertRaises(TypeError, self.pcp.open, [], { 'server' : 1 })

    def test05_server_none(self):
        """server is a string"""
        self.assertRaises(TypeError, self.pcp.open, [], { 'server' : None })

    def test11_server_v6(self):
        """server default family is IPv4"""
        self.assertEqual(self.pcp.open(server='::1'), libpcp.ERR_INVAL)

    def test21_server(self):
        """192.0.0.1 should work"""
        self.assertEqual(self.pcp.open(server=SERVER4), libpcp.OK)

    def test06_family_type(self):
        """family is an integer"""
        self.assertRaises(TypeError, self.pcp.open,
                          [SERVER4], { 'family' : 'foo' })

    def test07_family_bad(self):
        """only IPv4 and IPv6 are supported"""
        self.assertEqual(self.pcp.open(server=SERVER4,
                                       family=1234),
                         libpcp.ERR_INVAL)

    def test22_family_v4(self):
        """IPv4 example"""
        self.assertEqual(self.pcp.open(server=SERVER4,
                                       family=socket.AF_INET),
                         libpcp.OK)

    def test23_family_v6(self):
        """IPv6 example"""
        self.assertEqual(self.pcp.open(server=SERVER6,
                                       family=socket.AF_INET6),
                         libpcp.OK)

    def test08_source_type(self):
        """source is a string"""
        self.assertRaises(TypeError, self.pcp.open,
                          [SERVER4], { 'source' : 1 })

    def test09_source_none(self):
        """source=None is valid"""
        self.assertEqual(self.pcp.open(server=SERVER4,
                                       source=None),
                         libpcp.OK)

    def test12_source_v46(self):
        """source must be IPv4 when server is"""
        self.assertEqual(self.pcp.open(server=SERVER4,
                                       source="::"),
                         libpcp.ERR_INVAL)

    def test13_source_v64(self):
        """source must be IPv6 when server is"""
        self.assertEqual(self.pcp.open(server=SERVER6,
                                       family=socket.AF_INET6,
                                       source='0.0.0.0'),
                         libpcp.ERR_INVAL)

    def test24_source_v4(self):
        """IPv4 example"""
        self.assertEqual(self.pcp.open(server=SERVER4,
                                       source='0.0.0.0'),
                         libpcp.OK)

    def test25_source_v6(self):
        """IPv6 example"""
        self.assertEqual(self.pcp.open(server=SERVER6,
                                       family=socket.AF_INET6,
                                       source="::"),
                         libpcp.OK)

    def test26_source_bindv4(self):
        """IPv4 binding failure"""
        self.assertRaises(Exception, self.pcp.open,
                          [SERVER4], { 'source' : '1.22.77.199' })

    def test27_source_bindv6(self):
        """IPv6 binding failure"""
        self.assertRaises(Exception, self.pcp.open,
                          [SERVER6],
                          { 'family' : socket.AF_INET6, 'source' : '1::bad' })

    def test15_port_type(self):
        """port is a number"""
        self.assertRaises(TypeError, self.pcp.open,
                          [SERVER4], { 'port' : 'foo' })

    def test28_port_v4(self):
        """IPv4 port example"""
        self.assertEqual(self.pcp.open(server=SERVER4,
                                       port=5351),
                         libpcp.OK)

    def test29_port_v6(self):
        """IPv6 port example"""
        self.assertEqual(self.pcp.open(server=SERVER6,
                                       family=socket.AF_INET6,
                                       port=5351),
                         libpcp.OK)

class Test09Close(unittest.TestCase):
    """PCP close"""

    def setUp(self):
        """just create the object"""
        self.pcp = libpcp.PCP()

    def test0_arg(self):
        """close takes no argument"""
        self.assertRaises(TypeError, self.pcp.close, ['foo'])

    def test11_close0(self):
        """closing closed should work"""
        self.assertEqual(self.pcp.close(), libpcp.OK)

    def test12_close2(self):
        """multiple close should work"""
        self.pcp.close()
        self.assertEqual(self.pcp.close(), libpcp.OK)

class Test21GetSocket(unittest.TestCase):
    """PCP getsocket"""

    def setUp(self):
        """just create the object"""
        self.pcp = libpcp.PCP()

    def tearDown(self):
        """cleanup"""
        self.pcp.close()

    def test0_arg(self):
        """getsocket takes no argument"""
        self.assertRaises(TypeError, self.pcp.getsocket, ['foo'])

    def test11_notopened(self):
        """should fail on never opened"""
        self.assertRaises(Exception, self.pcp.getsocket)

    def test12_closed(self):
        """should fail on closed"""
        self.pcp.open(SERVER4)
        self.pcp.close()
        self.assertRaises(Exception, self.pcp.getsocket)

    def test21_open(self):
        """should work on opened"""
        self.pcp.open(SERVER4)
        self.assertTrue(self.pcp.getsocket() >= 0)
        
    def test13_closedsocket(self):
        """can be used to close the socket under pcp feet"""
        self.pcp.open(SERVER4)
        fd = self.pcp.getsocket()
        os.close(fd)
        self.assertRaises(Exception, self.pcp.close)

    def test22_serverv4(self):
        """check the server IPv4 address"""
        self.pcp.open(SERVER4)
        fd = self.pcp.getsocket()
        sock = socket.fromfd(fd, socket.AF_INET, socket.SOCK_DGRAM)
        self.assertEqual(sock.getpeername(), (SERVER4, libpcp.PORT))

    def test23_portv4(self):
        """check the server IPv4 port"""
        self.pcp.open(SERVER4, port=1234)
        fd = self.pcp.getsocket()
        sock = socket.fromfd(fd, socket.AF_INET, socket.SOCK_DGRAM)
        self.assertEqual(sock.getpeername(), (SERVER4, 1234))

    def test24_port0v4(self):
        """check the server IPv4 port=0"""
        self.pcp.open(SERVER4, port=0)
        fd = self.pcp.getsocket()
        sock = socket.fromfd(fd, socket.AF_INET, socket.SOCK_DGRAM)
        self.assertEqual(sock.getpeername(), (SERVER4, libpcp.PORT))

    def test25_serverv6(self):
        """check the server IPv6 address"""
        self.pcp.open(SERVER6, family=socket.AF_INET6)
        fd = self.pcp.getsocket()
        sock = socket.fromfd(fd, socket.AF_INET6, socket.SOCK_DGRAM)
        self.assertEqual(sock.getpeername(), (SERVER6, libpcp.PORT, 0, 0))

    def test26_portv6(self):
        """check the server IPv6 port"""
        self.pcp.open(SERVER6, family=socket.AF_INET6, port=1234)
        fd = self.pcp.getsocket()
        sock = socket.fromfd(fd, socket.AF_INET6, socket.SOCK_DGRAM)
        self.assertEqual(sock.getpeername(), (SERVER6, 1234, 0, 0))

    def test27_port0v6(self):
        """check the server IPv6 port=0"""
        self.pcp.open(SERVER6, family=socket.AF_INET6, port=0)
        fd = self.pcp.getsocket()
        sock = socket.fromfd(fd, socket.AF_INET6, socket.SOCK_DGRAM)
        self.assertEqual(sock.getpeername(), (SERVER6, libpcp.PORT, 0, 0))

    def test28_source4(self):
        """check the source IPv4 address"""
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        sock.connect((SERVER4, libpcp.PORT))
        addr = sock.getsockname()[0]
        sock.close()
        self.pcp.open(SERVER4, source=addr)
        fd = self.pcp.getsocket()
        sock = socket.fromfd(fd, socket.AF_INET, socket.SOCK_DGRAM)
        self.assertEqual(sock.getsockname()[0], addr)

    def test29_source6(self):
        """check the source IPv6 address"""
        sock = socket.socket(socket.AF_INET6, socket.SOCK_DGRAM)
        sock.connect((SERVER6, libpcp.PORT))
        addr = sock.getsockname()[0]
        sock.close()
        self.pcp.open(SERVER6, family=socket.AF_INET6, source=addr)
        fd = self.pcp.getsocket()
        sock = socket.fromfd(fd, socket.AF_INET6, socket.SOCK_DGRAM)
        self.assertEqual(sock.getsockname()[0], addr)

class Test22SetRCSocket(unittest.TestCase):
    """PCP setrcsocket"""

    def setUp(self):
        """just create the object"""
        self.pcp = libpcp.PCP()

    def tearDown(self):
        """cleanup"""
        self.pcp.close()

    def test01_arg0(self):
        """setrcsocket takes one argument"""
        self.assertRaises(TypeError, self.pcp.setrcsocket)

    def test02_args(self):
        """setrcsocket takes one argument"""
        self.assertRaises(TypeError, self.pcp.setrcsocket, [1, 'too many'])

    def test03_type(self):
        """argument is a number"""
        self.assertRaises(TypeError, self.pcp.setrcsocket, ['foo'])

    def test11_notopened(self):
        """should fail on never opened"""
        self.assertRaises(Exception, self.pcp.setrcsocket, [1234])

    def test12_closed(self):
        """should fail on closed"""
        self.pcp.open(SERVER4)
        self.pcp.close()
        self.assertRaises(Exception, self.pcp.setrcsocket, [1234])

    def test21_open(self):
        """should work on opened"""
        self.pcp.open(SERVER4)
        self.assertEqual(self.pcp.setrcsocket(1234), libpcp.OK)
        
    def test22_reset(self):
        """argument -1 (reset) should be accepted"""
        self.pcp.open(SERVER4)
        self.assertEqual(self.pcp.setrcsocket(-1), libpcp.OK)

class Test23GetTries(unittest.TestCase):
    """PCP gettries"""

    def setUp(self):
        """just create the object"""
        self.pcp = libpcp.PCP()

    def test0_arg(self):
        """gettries takes no argument"""
        self.assertRaises(TypeError, self.pcp.gettries, ['foo'])

    def test11_tries(self):
        """initial value is 0"""
        self.assertEqual(self.pcp.gettries(), 0)

    def test21_send(self):
        """check the value after sending two requests"""
        self.pcp.open(SERVER4)
        self.pcp.makerequest(opcode=libpcp.OPCODE_MAP)
        try:
            self.pcp.sendrequest()
            self.pcp.sendrequest()
        except Exception:
            return
        self.assertEqual(self.pcp.gettries(), 2)

    def test22_reset(self):
        """reset puts the value back to 0"""
        self.pcp.open(SERVER4)
        self.pcp.makerequest(opcode=libpcp.OPCODE_MAP)
        try:
            self.pcp.sendrequest()
        except Exception:
            return
        self.pcp.makerequest()
        self.assertEqual(self.pcp.gettries(), 0)

class Test24GetTimeout(unittest.TestCase):
    """PCP gettimeout"""

    def setUp(self):
        """just create the object"""
        self.pcp = libpcp.PCP()

    def test01_arg0(self):
        """gettimeout takes one argument"""
        self.assertRaises(TypeError, self.pcp.gettimeout)

    def test02_args(self):
        """gettimeout takes one argument"""
        self.assertRaises(TypeError, self.pcp.gettimeout, [1, 'too many'])

    def test03_type(self):
        """gettimeout takes a number"""
        self.assertRaises(TypeError, self.pcp.gettimeout, ['foo'])

    def test11_norequest(self):
        """gettimeout requires a request"""
        self.pcp.open(SERVER4)
        self.pcp.setrequest('\x00\x00')
        self.assertEqual(self.pcp.gettimeout(0), 0.0)

    def test21_initial(self):
        """initial value is 0.0"""
        self.pcp.open(SERVER4)
        self.pcp.makerequest(opcode=libpcp.OPCODE_MAP)
        self.assertEqual(self.pcp.gettimeout(0), 0.0)

    def test22_send(self):
        """value after two sends is 4.0"""
        self.pcp.open(SERVER4)
        self.pcp.makerequest(opcode=libpcp.OPCODE_MAP)
        try:
            self.pcp.sendrequest()
            self.pcp.sendrequest()
        except Exception:
            return
        tm = self.pcp.gettimeout(1)
        print tm
        self.assertTrue((tm > 3.9) and (tm < 4.1))

    def test23_reset(self):
        """reset puts again the value to 0.0"""
        self.pcp.open(SERVER4)
        self.pcp.makerequest(opcode=libpcp.OPCODE_MAP)
        try:
            self.pcp.sendrequest()
        except Exception:
            pass
        self.pcp.makerequest()
        self.assertEqual(self.pcp.gettimeout(0), 0.0)

class Test25SetReset(unittest.TestCase):
    """PCP setreset"""

    def setUp(self):
        """just create the object"""
        self.pcp = libpcp.PCP()

    def tearDown(self):
        """cleanup"""
        self.pcp.close()

    def test01_arg0(self):
        """setreset takes one argument"""
        self.assertRaises(TypeError, self.pcp.setreset)

    def test02_args(self):
        """setreset takes one argument"""
        self.assertRaises(TypeError, self.pcp.setreset, [1, 'too many'])

    def test03_type(self):
        """argument is a number"""
        self.assertRaises(TypeError, self.pcp.setreset, ['foo'])

    def test21_default(self):
        """default value is 0"""
        self.assertEqual(self.pcp.setreset(0), 0)

    def test22_set1(self):
        """set value to 1"""
        self.pcp.setreset(1)
        self.assertEqual(self.pcp.setreset(0), 1)

    def test23_set0(self):
        """set value to 1 then back to 0"""
        self.pcp.setreset(1)
        self.pcp.setreset(0)
        self.assertEqual(self.pcp.setreset(0), 0)

class Test26SetAutoFill(unittest.TestCase):
    """PCP setautofill"""

    def setUp(self):
        """just create the object"""
        self.pcp = libpcp.PCP()

    def tearDown(self):
        """cleanup"""
        self.pcp.close()

    def test01_arg0(self):
        """setautofill takes one argument"""
        self.assertRaises(TypeError, self.pcp.setautofill)

    def test02_args(self):
        """setautofill takes one argument"""
        self.assertRaises(TypeError, self.pcp.setautofill, [1, 'too many'])

    def test03_type(self):
        """argument is a number"""
        self.assertRaises(TypeError, self.pcp.setautofill, ['foo'])

    def test21_default(self):
        """default value is 1"""
        self.pcp.open(SERVER4)
        self.assertEqual(self.pcp.setautofill(1), 1)

    def test22_set0(self):
        """set value to 0"""
        self.pcp.open(SERVER4)
        self.pcp.setautofill(1)
        self.assertEqual(self.pcp.setautofill(1), 1)

    def test23_set1(self):
        """set value to 0 then back to 1"""
        self.pcp.open(SERVER4)
        self.pcp.setautofill(0)
        self.pcp.setautofill(1)
        self.assertEqual(self.pcp.setautofill(1), 1)

    def test24_autofill1(self):
        """check autofill=1 behavior"""
        self.pcp.open(SERVER4)
        self.pcp.setautofill(1)
        self.pcp.makerequest(opcode=libpcp.OPCODE_MAP)
        req = self.pcp.getrequest()
        intaddr = req[20:24]
        self.assertNotEqual(intaddr, '\x00\x00\x00\x00')

    def test25_autofill0(self):
        """check autofill=0 behavior"""
        self.pcp.open(SERVER4)
        self.pcp.setautofill(0)
        self.pcp.makerequest(opcode=libpcp.OPCODE_MAP)
        req = self.pcp.getrequest()
        intaddr = req[20:24]
        self.assertEqual(intaddr, '\x00\x00\x00\x00')

class Test31ThirdParty(unittest.TestCase):
    """internal THIRD_PARTY option builder"""

    def setUp(self):
        """just create the object"""
        self.pcp = libpcp.PCP()

    def test01_arg0(self):
        """no argument is 'empty'"""
        self.assertEqual(self.pcp.third_party(),
                         (libpcp.OPTION_THIRD_PARTY, ''))

    def test02_args(self):
        """too many (>2) arguments"""
        self.assertRaises(TypeError, self.pcp.third_party, [1, 2, 'too many'])

    def test03_family_type(self):
        """family is a number"""
        self.assertRaises(TypeError, self.pcp.third_party, ['foo', 'bar'])

    def test04_family_bad(self):
        """only 0, IPv4 and IPv6 are supported"""
        self.assertEqual(self.pcp.third_party(1234, 'foo'), libpcp.ERR_INVAL)

    def test21_family0(self):
        """family=0 is 'empty'"""
        self.assertEqual(self.pcp.third_party(0, 'foo'),
                         (libpcp.OPTION_THIRD_PARTY, ''))

    def test05_addr_type(self):
        """addr is a string"""
        self.assertRaises(TypeError, self.pcp.third_party, [0, 1])

    def test06_addr_none(self):
        """addr is a string"""
        self.assertRaises(TypeError, self.pcp.third_party, [0, None])

    def test11_v46(self):
        """family=IPv4 requires an IPv4 address"""
        self.assertEqual(self.pcp.third_party(socket.AF_INET, '::1'),
                         libpcp.ERR_INVAL)

    def test12_v64(self):
        """family=IPv6 requires an IPv6 address"""
        self.assertEqual(self.pcp.third_party(socket.AF_INET6, '127.0.0.1'),
                         libpcp.ERR_INVAL)

    def test22_v4(self):
        """IPv4 example"""
        taddr = ADDR4
        baddr = socket.inet_pton(socket.AF_INET, taddr)
        self.assertEqual(self.pcp.third_party(socket.AF_INET, taddr),
                         (libpcp.OPTION_THIRD_PARTY, MAPPED + baddr))

    def test23_v6(self):
        """IPv6 example"""
        taddr = ADDR6
        baddr = socket.inet_pton(socket.AF_INET6, taddr)
        self.assertEqual(self.pcp.third_party(socket.AF_INET6, taddr),
                         (libpcp.OPTION_THIRD_PARTY, baddr))

class Test32PreferFailure(unittest.TestCase):
    """internal PREFER_FAILURE option builder"""

    def setUp(self):
        """just create the object"""
        self.pcp = libpcp.PCP()

    def test0_arg(self):
        """prefer_failure takes no argument"""
        self.assertRaises(TypeError, self.pcp.prefer_failure, ['foo'])

    def test2_prefer_failure(self):
        """prefer_failure example"""
        self.assertEqual(self.pcp.prefer_failure(),
                         (libpcp.OPTION_PREF_FAIL, ''))

class Test33Filter(unittest.TestCase):
    """internal FILTER option builder"""

    def setUp(self):
        """just create the object"""
        self.pcp = libpcp.PCP()

    def test01_arg3(self):
        """4 arguments are required"""
        self.assertRaises(TypeError, self.pcp.filter, [1, 2, 'foo'])

    def test02_arg5(self):
        """too many (>4) arguments"""
        self.assertRaises(TypeError, self.pcp.filter,
                          [1, 2, 'foo', 3, 'too many'])

    def test03_family_type(self):
        """family is a number"""
        self.assertRaises(TypeError, self.pcp.filter, ['foo', 2, 'bar', 0])

    def test04_family_bad(self):
        """only IPv4 and IPv6 are supported"""
        self.assertEqual(self.pcp.filter(1234, 2, 'foo', 0), libpcp.ERR_INVAL)

    def test05_prefixlen_type(self):
        """prefix length is a number"""
        self.assertRaises(TypeError, self.pcp.filter, [1, 'foo', 'bar', 0])

    def test06_addr_type(self):
        """address is a string"""
        self.assertRaises(TypeError, self.pcp.filter, [1, 2, 3, 0])

    def test07_addr_none(self):
        """address is a string"""
        self.assertRaises(TypeError, self.pcp.filter, [1, 2, None, 0])

    def test08_port_type(self):
        """port is a number"""
        self.assertRaises(TypeError, self.pcp.filter, [1, 2, 'foo', 'bar'])

    def test11_v46(self):
        """family=IPv4 requires an IPv4 prefix"""
        self.assertEqual(self.pcp.filter(socket.AF_INET, 32, '::1', 0),
                         libpcp.ERR_INVAL)

    def test12_v64(self):
        """family=IPv6 requires an IPv6 prefix"""
        self.assertEqual(self.pcp.filter(socket.AF_INET6, 32, '127.0.0.1', 0),
                         libpcp.ERR_INVAL)

    def test21_v4(self):
        """IPv4 example"""
        taddr = ADDR4
        baddr = socket.inet_pton(socket.AF_INET, taddr)
        self.assertEqual(self.pcp.filter(socket.AF_INET, 32, taddr, 1234),
                         (libpcp.OPTION_FILTER,
                          '\x00\x20\x04\xd2' + MAPPED + baddr))

    def test22_v6(self):
        """IPv6 example"""
        taddr = ADDR6
        baddr = socket.inet_pton(socket.AF_INET6, taddr)
        self.assertEqual(self.pcp.filter(socket.AF_INET6, 128, taddr, 1234),
                         (libpcp.OPTION_FILTER, '\x00\x80\x04\xd2' + baddr))

class Test11MakeRequest(unittest.TestCase):
    """PCP makerequest"""

    def setUp(self):
        """just create the object"""
        self.pcp = libpcp.PCP()

    def tearDown(self):
        """cleanup"""
        self.pcp.close()

    def test001_arg(self):
        """unknown keyword"""
        self.assertRaises(TypeError, self.pcp.makerequest, [], { 'foo' : 1 })

    def test002_args(self):
        """too many (>10) arguments"""
        self.assertRaises(TypeError, self.pcp.makerequest,
                          [1, 2, 3, 'foo', 'bar', 'xyz', 7, 8, 9, [],
                           'too many'])

    def test003_arg0(self):
        """no argument just resets"""
        self.pcp.open(SERVER4)
        self.assertEqual(self.pcp.makerequest(), libpcp.OK)

    def test004_opcode_type(self):
        """opcode is a number"""
        self.assertRaises(TypeError, self.pcp.makerequest,
                          [], { 'opcode' : 'foo' })

    def test11_opcode_bad(self):
        """only MAP and PEER are supported"""
        self.pcp.open(SERVER4)
        self.assertRaises(Exception, self.pcp.makerequest,
                          [], { 'opcode' : 1234 })

    def test21_opcode_map(self):
        """MAP example"""
        self.pcp.open(SERVER4)
        self.assertEqual(self.pcp.makerequest(opcode=libpcp.OPCODE_MAP),
                         libpcp.OK)

    def test22_opcode_peer(self):
        """PEER example"""
        self.pcp.open(SERVER4)                                          
        self.assertEqual(self.pcp.makerequest(opcode=libpcp.OPCODE_PEER),
                         libpcp.OK)

    def test005_lifetime_type(self):
        """lifetime is a number"""
        self.assertRaises(TypeError, self.pcp.makerequest,
                          [], { 'lifetime' : 'foo' })

    def test006_protocol_type(self):
        """protocol is a number"""
        self.assertRaises(TypeError, self.pcp.makerequest,
                          [], { 'protocol' : 'foo' })

    def test007_intaddr_type(self):
        """internal address is a string"""
        self.assertRaises(TypeError, self.pcp.makerequest,
                          [], { 'intaddr' : 1 })

    def test008_intaddr_none(self):
        """intaddr=None is valid"""
        self.pcp.setautofill(0)
        self.assertEqual(self.pcp.makerequest(opcode=libpcp.OPCODE_MAP,
                                              intaddr=None),
                         libpcp.OK)

    def test12_intaddr_bad(self):
        """bad internal address"""
        self.pcp.setautofill(0)
        self.assertEqual(self.pcp.makerequest(opcode=libpcp.OPCODE_MAP,
                                              intaddr='foobar'),
                         libpcp.ERR_INVAL)

    def test009_extaddr_type(self):
        """external address is a string"""
        self.assertRaises(TypeError, self.pcp.makerequest,
                          [], { 'extaddr' : 1 })

    def test010_extaddr_none(self):
        """extaddr=None is valid"""
        self.pcp.setautofill(0)
        self.assertEqual(self.pcp.makerequest(opcode=libpcp.OPCODE_MAP,
                                              extaddr=None),
                         libpcp.OK)

    def test13_extaddr_bad(self):
        """bad external address"""
        self.pcp.setautofill(0)
        self.assertEqual(self.pcp.makerequest(opcode=libpcp.OPCODE_MAP,
                                              extaddr='foobar'),
                         libpcp.ERR_INVAL)

    def test011_remotepeer_type(self):
        """remote peer address is a string"""
        self.assertRaises(TypeError, self.pcp.makerequest,
                          [], { 'remotepeer' : 1 })

    def test012_remotepeer_none(self):
        """remotepeer=None is valid"""
        self.pcp.setautofill(0)
        self.assertEqual(self.pcp.makerequest(opcode=libpcp.OPCODE_MAP,
                                              remotepeer=None),
                         libpcp.OK)

    def test14_remotepeer_bad(self):
        """bad remote peer address"""
        self.pcp.setautofill(0)
        self.assertEqual(self.pcp.makerequest(opcode=libpcp.OPCODE_PEER,
                                              remotepeer='foobar'),
                         libpcp.ERR_INVAL)

    def test15_remotepeer_map(self):
        """remotepeer is ignored on MAP"""
        self.pcp.setautofill(0)
        self.assertEqual(self.pcp.makerequest(opcode=libpcp.OPCODE_MAP,
                                              remotepeer='foobar'),
                         libpcp.OK)

    def test013_intport_type(self):
        """internal port is a number"""
        self.assertRaises(TypeError, self.pcp.makerequest,
                          [], { 'intport' : 'foo' })

    def test014_extport_type(self):
        """external port is a number"""
        self.assertRaises(TypeError, self.pcp.makerequest,
                          [], { 'extport' : 'foo' })

    def test015_peerport_type(self):
        """remote peer port is a number"""
        self.assertRaises(TypeError, self.pcp.makerequest,
                          [], { 'peerport' : 'foo' })

    def test016_options_type(self):
        """options is a list"""
        self.assertRaises(TypeError, self.pcp.makerequest,
                          [], { 'options' : 'foo' })

    def test16_options_size1024(self):
        """request size limited to 1024"""
        optdata = '\x00' * 972
        self.pcp.setautofill(0)
        self.assertEqual(self.pcp.makerequest(opcode=libpcp.OPCODE_MAP,
                                              options=[ (123, optdata) ]),
                         libpcp.OK)

    def test17_options_size1025(self):
        """request size limited to 1024"""
        optdata = '\x00' * 973
        self.pcp.setautofill(0)
        self.assertRaises(Exception, self.pcp.makerequest, [],
                          { 'opcode' : libpcp.OPCODE_MAP,
                            'options': [ (123, optdata) ] })

    def test18_options_count(self):
        """no more than 4 options"""
        self.pcp.setautofill(0)
        self.assertRaises(TypeError, self.pcp.makerequest, [],
                          { 'opcode' : libpcp.OPCODE_MAP,
                            'options' : [1, 2, 3, 4, 'too many'] })

    def test017_options_tuple(self):
        """options items must be tuples"""
        self.pcp.setautofill(0)
        self.assertRaises(TypeError, self.pcp.makerequest, [],
                          { 'opcode' : libpcp.OPCODE_MAP,
                            'options' : [ 'foo'] })

    def test018_options_pair(self):
        """options items must be pairs (2-tuples)"""
        self.pcp.setautofill(0)
        self.assertRaises(TypeError, self.pcp.makerequest, [],
                          { 'opcode' : libpcp.OPCODE_MAP,
                            'options' : [ (1, 2, 3) ] })

    def test019_options_first(self):
        """in options, first member of the pair is a number"""
        self.pcp.setautofill(0)
        self.assertRaises(TypeError, self.pcp.makerequest, [],
                          { 'opcode' : libpcp.OPCODE_MAP,
                            'options' : [ ('foo', 'bar') ] })

    def test020_options_second(self):
        """in options, second member of the pair is a string"""
        self.pcp.setautofill(0)
        self.assertRaises(TypeError, self.pcp.makerequest, [],
                          { 'opcode' : libpcp.OPCODE_MAP,
                            'options' : [ (1, 2) ] })

    def test021_options_sndnone(self):
        """in options, second member of the pair is a string"""
        self.pcp.setautofill(0)
        self.assertRaises(TypeError, self.pcp.makerequest, [],
                          { 'opcode' : libpcp.OPCODE_MAP,
                            'options' : [ (1, None) ] })

    def test19_autofill(self):
        """autofill kept to 1 and no open"""
        self.pcp.setautofill(1)
        self.assertRaises(Exception, self.pcp.makerequest, [libpcp.OPCODE_MAP])

class Test91EncodeRequest(unittest.TestCase):
    """internal request encoding"""

    def setUp(self):
        """just create the object"""
        self.pcp = libpcp.PCP()

    def tearDown(self):
        """cleanup"""
        self.pcp.close()

    def test01_version(self):
        """any => req[version]=1"""
        self.pcp.makerequest(opcode=libpcp.OPCODE_MAP)
        req = self.pcp.getrequest()
        self.assertEqual(req[0], '\x01')

    def test02_opcode_map(self):
        """opcode=map => req[opcode]=1"""
        self.pcp.makerequest(opcode=libpcp.OPCODE_MAP)
        req = self.pcp.getrequest()
        self.assertEqual(req[1], '\x01')

    def test03_map_size(self):
        """opcode=map => size=48"""
        self.pcp.makerequest(opcode=libpcp.OPCODE_MAP)
        req = self.pcp.getrequest()
        self.assertEqual(len(req), 48)

    def test04_opcode_peer(self):
        """opcode=peer => req[opcode]=2"""
        self.pcp.makerequest(opcode=libpcp.OPCODE_PEER)
        req = self.pcp.getrequest()
        self.assertEqual(req[1], '\x02')

    def test05_peer_size(self):
        """opcode=peer => size=68"""
        self.pcp.makerequest(opcode=libpcp.OPCODE_PEER)
        req = self.pcp.getrequest()
        self.assertEqual(len(req), 68)

    def test06_lifetime(self):
        """lifetime=123456789 => req[lifetime]=0x075bcd15"""
        self.pcp.makerequest(opcode=libpcp.OPCODE_MAP, lifetime=123456789)
        req = self.pcp.getrequest()
        self.assertEqual(req[4:8], '\x07\x5b\xcd\x15')

    def test07_protocol_map(self):
        """opcode=MAP protocol=UDP => req[protocol]=17"""
        self.pcp.makerequest(opcode=libpcp.OPCODE_MAP,
                             protocol=socket.IPPROTO_UDP)
        req = self.pcp.getrequest()
        self.assertEqual(req[24], '\x11')

    def test08_protocol_peer(self):
        """opcode=PEER protocol=TCP => req[protocol]=6"""
        self.pcp.makerequest(opcode=libpcp.OPCODE_PEER,
                             protocol=socket.IPPROTO_TCP)
        req = self.pcp.getrequest()
        self.assertEqual(req[24], '\x06')

    def test10_intaddr_v4(self):
        """intaddr=10.1.2.3 => req[source]=::ffff:10.1.2.3"""
        self.pcp.makerequest(opcode=libpcp.OPCODE_MAP,
                             intaddr=ADDR4)
        req = self.pcp.getrequest()
        addr = socket.inet_pton(socket.AF_INET6, '::ffff:10.1.2.3')
        self.assertEqual(req[8:24], addr)

    def test11_intaddr_v6(self):
        """intaddr=2001::1234:fedc => req[source]=2001::1234:fedc"""
        self.pcp.makerequest(opcode=libpcp.OPCODE_MAP,
                             intaddr='2001::1234:fedc')
        req = self.pcp.getrequest()
        addr = socket.inet_pton(socket.AF_INET6, '2001::1234:fedc')
        self.assertEqual(req[8:24], addr)

    def test12_extaddr_map_v4(self):
        """opcode=MAP extaddr=10.1.2.3 => req[extaddr]=10.1.2.3"""
        self.pcp.makerequest(opcode=libpcp.OPCODE_MAP,
                             extaddr=ADDR4)
        req = self.pcp.getrequest()
        addr = socket.inet_pton(socket.AF_INET6, '::ffff:10.1.2.3')
        self.assertEqual(req[32:48], addr)

    def test13_extaddr_map_v6(self):
        """opcode=MAP extaddr=2001::1234:fedc =>
           req[extaddr]=2001::1234:fedc"""
        self.pcp.makerequest(opcode=libpcp.OPCODE_MAP,
                             extaddr='2001::1234:fedc')
        req = self.pcp.getrequest()
        addr = socket.inet_pton(socket.AF_INET6, '2001::1234:fedc')
        self.assertEqual(req[32:48], addr)

    def test14_extaddr_peer_v4(self):
        """opcode=PEER extaddr=10.1.2.3 => req[extaddr]=10.1.2.3"""
        self.pcp.makerequest(opcode=libpcp.OPCODE_PEER,
                             extaddr=ADDR4)
        req = self.pcp.getrequest()
        addr = socket.inet_pton(socket.AF_INET6, '::ffff:10.1.2.3')
        self.assertEqual(req[32:48], addr)

    def test15_extaddr_peer_v6(self):
        """opcode=PEER extaddr=2001::1234:fedc =>
           req[extaddr]=2001::1234:fedc"""
        self.pcp.makerequest(opcode=libpcp.OPCODE_PEER,
                             extaddr='2001::1234:fedc')
        req = self.pcp.getrequest()
        addr = socket.inet_pton(socket.AF_INET6, '2001::1234:fedc')
        self.assertEqual(req[32:48], addr)

    def test20_remotepeer_v4(self):
        """remotepeer=10.1.2.3 => req[remotepeer]=10.1.2.3"""
        self.pcp.makerequest(opcode=libpcp.OPCODE_PEER,
                             remotepeer=ADDR4)
        req = self.pcp.getrequest()
        addr = socket.inet_pton(socket.AF_INET6, '::ffff:10.1.2.3')
        self.assertEqual(req[52:68], addr)

    def test21_remotepeer_v6(self):
        """remotepeer=2001::1234:fedc => req[remotepeer]=2001::1234:fedc"""
        self.pcp.makerequest(opcode=libpcp.OPCODE_PEER,
                             remotepeer='2001::1234:fedc')
        req = self.pcp.getrequest()
        addr = socket.inet_pton(socket.AF_INET6, '2001::1234:fedc')
        self.assertEqual(req[52:68], addr)

    def test16_intport_map(self):
        """opcode=MAP intport=1234 => req[intport]=0x04d2"""
        self.pcp.makerequest(opcode=libpcp.OPCODE_MAP, intport=1234)
        req = self.pcp.getrequest()
        self.assertEqual(req[28:30], '\x04\xd2')

    def test17_intport_peer(self):
        """opcode=PEER intport=1234 => req[intport]=0x04d2"""
        self.pcp.makerequest(opcode=libpcp.OPCODE_PEER, intport=1234)
        req = self.pcp.getrequest()
        self.assertEqual(req[28:30], '\x04\xd2')

    def test18_extport_map(self):
        """opcode=MAP extport=1234 => req[extport]=0x04d2"""
        self.pcp.makerequest(opcode=libpcp.OPCODE_MAP, extport=1234)
        req = self.pcp.getrequest()
        self.assertEqual(req[30:32], '\x04\xd2')

    def test19_extport_peer(self):
        """opcode=PEER extport=1234 => req[extport]=0x04d2"""
        self.pcp.makerequest(opcode=libpcp.OPCODE_PEER, extport=1234)
        req = self.pcp.getrequest()
        self.assertEqual(req[30:32], '\x04\xd2')

    def test22_peerport(self):
        """peerport=1234 => req[peerport]=0x04d2"""
        self.pcp.makerequest(opcode=libpcp.OPCODE_PEER, peerport=1234)
        req = self.pcp.getrequest()
        self.assertEqual(req[48:50], '\x04\xd2')

    def test31_reserved_map(self):
        """opcode=MAP => all reserved fields at zero"""
        self.pcp.makerequest(opcode=libpcp.OPCODE_MAP,
                             lifetime=123456789,
                             protocol=socket.IPPROTO_UDP,
                             intaddr='2001::1234:fedc',
                             extaddr='2123::fedc:1234',
                             intport=1234, extport=5678)
        req = self.pcp.getrequest()
        rsvd = req[2:4] + req[25:28]
        self.assertEqual(rsvd, '\x00' * 5)
                             
    def test32_reserved_peer(self):
        """opcode=PEER => all reserved fields at zero"""
        self.pcp.makerequest(opcode=libpcp.OPCODE_PEER,
                             lifetime=123456789,
                             protocol=socket.IPPROTO_UDP,
                             intaddr='2001::1234:fedc',
                             extaddr='2123::fedc:1234',
                             remotepeer='2321::4321:cdef',
                             intport=1234, extport=5678, peerport=9876)
        req = self.pcp.getrequest()
        rsvd = req[2:4] + req[25:28] + req[50:52]
        self.assertEqual(rsvd, '\x00' * 7)

    def test41_options_map(self):
        """opcode=MAP options => at the end in order with padding"""
        options = [ (17, 'ab'), (217, 'fedcb'), (117, '01234567') ]
        boptions = '\x11\x00\x00\x02\x61\x62\x00\x00'
        boptions += '\xd9\x00\x00\x05\x66\x65\x64\x63\x62\x00\x00\x00'
        boptions += '\x75\x00\x00\x08\x30\x31\x32\x33\x34\x35\x36\x37'
        self.pcp.makerequest(opcode=libpcp.OPCODE_MAP, options=options)
        req = self.pcp.getrequest()
        self.assertEqual(req[48:], boptions)

    def test42_options_peer(self):
        """opcode=PEER options => at the end in order with padding"""
        options = [ (17, 'ab'), (217, 'fedcb'), (117, '01234567') ]
        boptions = '\x11\x00\x00\x02\x61\x62\x00\x00'
        boptions += '\xd9\x00\x00\x05\x66\x65\x64\x63\x62\x00\x00\x00'
        boptions += '\x75\x00\x00\x08\x30\x31\x32\x33\x34\x35\x36\x37'
        self.pcp.makerequest(opcode=libpcp.OPCODE_PEER, options=options)
        req = self.pcp.getrequest()
        self.assertEqual(req[68:], boptions)

    def test43_option_large(self):
        """option with length > 256"""
        options = [ (111, '\xa4' * 499) ]
        boptions = '\x6f\x00\x01\xf3' + ('\xa4' * 499) + '\x00'
        self.pcp.makerequest(opcode=libpcp.OPCODE_MAP, options=options)
        req = self.pcp.getrequest()
        self.assertEqual(req[48:], boptions)

class Test12SendRequest(unittest.TestCase):
    """PCP sendrequest"""

    def setUp(self):
        """just create the object"""
        self.pcp = libpcp.PCP()

    def tearDown(self):
        """cleanup"""
        try:
            self.pcp.close()
        except Exception:
            pass

    def test0_arg(self):
        """sendrequest doesn't take an argument"""
        self.assertRaises(TypeError, self.pcp.sendrequest, [1])

    def test11_notopened(self):
        """should fail on never opened"""
        self.assertRaises(Exception, self.pcp.sendrequest)

    def test12_closed(self):
        """should fail on closed"""
        self.pcp.open(SERVER4)
        self.pcp.close()
        self.assertRaises(Exception, self.pcp.sendrequest)

    def test13_norequest(self):
        """requires a request"""
        self.pcp.open(SERVER4)
        self.assertRaises(Exception, self.pcp.sendrequest)


    def test14_closedsocket(self):
        """fails if the socket was closed under pcp feet"""
        self.pcp.open(SERVER4)
        self.pcp.makerequest(opcode=libpcp.OPCODE_MAP)
        fd = self.pcp.getsocket()
        os.close(fd)
        self.assertRaises(Exception, self.pcp.sendrequest)

class Test13RecvResponse(unittest.TestCase):
    """PCP sendrequest"""

    def setUp(self):
        """just create the object"""
        self.pcp = libpcp.PCP()

    def tearDown(self):
        """cleanup"""
        self.pcp.close()

    def test0_arg(self):
        """recvresponse doesn't take an argument"""
        self.assertRaises(TypeError, self.pcp.recvresponse, [1])

    def test11_notopened(self):
        """should fail on never opened"""
        self.assertRaises(Exception, self.pcp.recvresponse)

    def test12_closed(self):
        """should fail on closed"""
        self.pcp.open(SERVER4)
        self.pcp.close()
        self.assertRaises(Exception, self.pcp.recvresponse)

    def test13_noservice(self):
        """should fail when no service is available"""
        self.pcp.open(SERVER4)
        self.assertEqual(self.pcp.recvresponse(), (libpcp.TRY_AGAIN, {}))

class Test41FdOpen(unittest.TestCase):
    """PCP fdopen (C _pcp_fdinit())"""

    def setUp(self):
        """just create the object"""
        self.pcp = libpcp.PCP()

    def test01_arg0(self):
        """fdopen takes one argument"""
        self.assertRaises(TypeError, self.pcp.fdopen)

    def test02_args(self):
        """fdopen takes one argument"""
        self.assertRaises(TypeError, self.pcp.fdopen, [1, 'too many'])

    def test03_type(self):
        """argument is a number"""
        self.assertRaises(TypeError, self.pcp.fdopen, ['foo'])

    def test21_fdopen(self):
        """should work"""
        self.assertEqual(self.pcp.fdopen(123), libpcp.OK)

    def test22_fdopen2(self):
        """check the value"""
        self.pcp.fdopen(123)
        self.assertEqual(self.pcp.getsocket(), 123)

    def test23_close(self):
        """close() should fail"""
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        fd = sock.fileno()
        sock.close()
        self.pcp.fdopen(fd)
        self.assertRaises(Exception, self.pcp.close)

    def test24_dup(self):
        """dup/close/fdopen sequence"""
        self.pcp.open(SERVER4)
        fd = os.dup(self.pcp.getsocket())
        os.close(self.pcp.getsocket())
        self.pcp.fdopen(fd)
        self.assertEqual(self.pcp.close(), libpcp.OK)

class Test42SetRequest(unittest.TestCase):
    """PCP setrequest"""

    def setUp(self):
        """just create the object"""
        self.pcp = libpcp.PCP()

    def tearDown(self):
        """cleanup"""
        self.pcp.close()

    def test01_arg0(self):
        """setrequest takes one argument"""
        self.assertRaises(TypeError, self.pcp.setrequest)

    def test02_args(self):
        """too many (>1) arguments"""
        self.assertRaises(TypeError, self.pcp.setrequest, ['foo', 'bar'])

    def test03_type(self):
        """the argument is a string"""
        self.assertRaises(TypeError, self.pcp.setrequest, [1])

    def test04_none(self):
        """the argument is a string"""
        self.assertRaises(TypeError, self.pcp.setrequest, [None])

    def test21_setrequest(self):
        """should work"""
        self.assertEqual(self.pcp.setrequest('\x01\x00'), libpcp.OK)

    def test22_getrequest(self):
        """get o set == id"""
        self.pcp.setrequest('\x01\x00')
        self.assertEqual(self.pcp.getrequest(), '\x01\x00')

class Test43Getrequest(unittest.TestCase):
    """PCP getrequest"""

    def setUp(self):
        """just create the object"""
        self.pcp = libpcp.PCP()

    def tearDown(self):
        """cleanup"""
        self.pcp.close()

    def test0_args(self):
        """getrequest uses no argument"""
        self.assertRaises(TypeError, self.pcp.getrequest, ['foo'])

    def test21_getrequest(self):
        """should work"""
        self.pcp.open(SERVER4)
        self.pcp.makerequest(libpcp.OPCODE_MAP)
        req = self.pcp.getrequest()
        self.assertEqual(req[:4], '\x01\x01\x00\x00')

    def test22_last(self):
        """get the last request"""
        self.pcp.open(SERVER4)
        self.pcp.makerequest(libpcp.OPCODE_MAP)
        self.pcp.makerequest(libpcp.OPCODE_PEER)
        req = self.pcp.getrequest()
        self.assertEqual(req[:4], '\x01\x02\x00\x00')

    def test23_setrequest(self):
        """set o get == id"""
        self.pcp.open(SERVER4)
        self.pcp.makerequest(libpcp.OPCODE_PEER)
        req = self.pcp.getrequest()
        self.pcp.makerequest(libpcp.OPCODE_MAP)
        self.pcp.setrequest(req)
        self.assertEqual(self.pcp.getrequest(), req)

class Test44SetResponse(unittest.TestCase):
    """PCP setresponse"""

    def setUp(self):
        """just create the object"""
        self.pcp = libpcp.PCP()

    def tearDown(self):
        """cleanup"""
        self.pcp.close()

    def test01_arg0(self):
        """setresponse takes one argument"""
        self.assertRaises(TypeError, self.pcp.setresponse)

    def test02_args(self):
        """too many (>1) arguments"""
        self.assertRaises(TypeError, self.pcp.setresponse, ['foo', 'bar'])

    def test03_type(self):
        """the argument is a string"""
        self.assertRaises(TypeError, self.pcp.setresponse, [1])

    def test04_none(self):
        """the argument is a string"""
        self.assertRaises(TypeError, self.pcp.setresponse, [None])

    def test21_setresponse(self):
        """should work"""
        self.assertEqual(self.pcp.setresponse('\x01\x00'), libpcp.OK)

    def test22_getresponse(self):
        """get o set == id"""
        self.pcp.setresponse('\x01\x00')
        self.assertEqual(self.pcp.getresponse(), '\x01\x00')

class Test45Getresponse(unittest.TestCase):
    """PCP getresponse"""

    def setUp(self):
        """just create the object"""
        self.pcp = libpcp.PCP()

    def tearDown(self):
        """cleanup"""
        self.pcp.close()

    def test0_args(self):
        """getresponse uses no argument"""
        self.assertRaises(TypeError, self.pcp.getresponse, ['foo'])

    def test2_setresponse(self):
        """get o set == id"""
        self.pcp.open(SERVER4)
        self.pcp.makerequest(libpcp.OPCODE_MAP)
        req = self.pcp.getrequest()
        resp = '\x01\x81\x00\x01' + req[4:]
        self.pcp.setresponse(resp)
        self.assertEqual(self.pcp.getresponse(), resp)

class Test46SubmitResponse(unittest.TestCase):
    """PCP submitresponse"""

    def setUp(self):
        """just create the object"""
        self.pcp = libpcp.PCP()

    def tearDown(self):
        """cleanup"""
        self.pcp.close()

    def test0_args(self):
        """submitresponse uses no argument"""
        self.assertRaises(TypeError, self.pcp.submitresponse, ['foo'])

    def test11_short4(self):
        """the response length must >= 4"""
        self.pcp.setresponse('\x01\x81\x00')
        self.assertRaises(Exception, self.pcp.submitresponse)

    def test12_version(self):
        """the version [0] must be 1"""
        self.pcp.setresponse('\x00\x81\x00\x00')
        self.assertRaises(Exception, self.pcp.submitresponse)

    def test13_response(self):
        """the opcode [1] must be a response (>= 128)"""
        self.pcp.setresponse('\x01\x01\x00\x00')
        self.assertRaises(Exception, self.pcp.submitresponse)

    def test14_oversize(self):
        """the response length must be <= 1024"""
        self.pcp.setresponse('\x01\x81' + '\x00' * 1023)
        self.assertRaises(Exception, self.pcp.submitresponse)

    def test15_align(self):
        """the response must be aligned on a 4 octet boundary"""
        self.pcp.setresponse('\x01\x81'+ '\x00' * 5)
        self.assertRaises(Exception, self.pcp.submitresponse)

    def test16_short24(self):
        """the response length must >= 24"""
        self.pcp.setresponse('\x01\x81'+ '\x00' * 18)
        self.assertRaises(Exception, self.pcp.submitresponse)

    def test17_norequest(self):
        """without a request should return bad opcode"""
        resp =  '\x01\x81\x00\x00' + '\x00\x00\x00\x00'
        resp += '\x00\x01\xe2\x40' + '\x00\x00\x00\x00'
        resp += '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        self.pcp.setresponse(resp)
        self.assertRaises(Exception, self.pcp.submitresponse)

    def test18_opcodemismatch(self):
        """request and response opcodes must match"""
        resp =  '\x01\x81\x00\x00' + '\x00\x00\x00\x00'
        resp += '\x00\x01\xe2\x40' + '\x00\x00\x00\x00'
        resp += '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        req =   '\x01\x99\x00\x00'
        self.pcp.setresponse(resp)
        self.pcp.setrequest(req)
        self.assertRaises(Exception, self.pcp.submitresponse)

    def test19_badopcode(self):
        """only MAP and PEER are supported"""
        resp =  '\x01\x99\x00\x00' + '\x00\x00\x00\x00'
        resp += '\x00\x01\xe2\x40' + '\x00\x00\x00\x00'
        resp += '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        req =   '\x01\x19\x00\x00'
        self.pcp.setresponse(resp)
        self.pcp.setrequest(req)
        self.assertRaises(Exception, self.pcp.submitresponse)

    def test20_short_map(self):
        """MAP undersize"""
        resp =  '\x01\x81\x00\x00' + '\x00\x00\x00\x00'
        resp += '\x00\x01\xe2\x40' + '\x00\x00\x00\x00'
        resp += '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        req =   '\x01\x01\x00\x00'
        self.pcp.setresponse(resp)
        self.pcp.setrequest(req)
        self.assertRaises(Exception, self.pcp.submitresponse)

    def test21_short_peer(self):
        """PEER undersize"""
        resp =  '\x01\x82\x00\x00' + '\x00\x00\x00\x00'
        resp += '\x00\x01\xe2\x40' + '\x00\x00\x00\x00'
        resp += '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        req =   '\x01\x02\x00\x00'
        self.pcp.setresponse(resp)
        self.pcp.setrequest(req)
        self.assertRaises(Exception, self.pcp.submitresponse)

    def test22_protomismatch_map(self):
        """request and response MAP protocols must match"""
        resp =  '\x01\x81\x00\x00' + '\x00\x00\x00\x00'
        resp += '\x00\x01\xe2\x40' + '\x00\x00\x00\x00'
        resp += '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        resp += '\x99\x00\x00\x00' + '\x00\x00\x00\x00'
        resp += '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        resp += '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        req =   '\x01\x01\x00\x00' + '\x00\x00\x00\x00'
        req +=  '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        req +=  '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        req +=  '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        req +=  '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        req +=  '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        self.pcp.setresponse(resp)
        self.pcp.setrequest(req)
        self.assertRaises(Exception, self.pcp.submitresponse)

    def test23_protomismatch_peer(self):
        """request and response PEER protocols must match"""
        resp =  '\x01\x82\x00\x00' + '\x00\x00\x00\x00'
        resp += '\x00\x01\xe2\x40' + '\x00\x00\x00\x00'
        resp += '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        resp += '\x99\x00\x00\x00' + '\x00\x00\x00\x00'
        resp += '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        resp += '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        resp += '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        resp += '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        resp += '\x00\x00\x00\x00'
        req =   '\x01\x02\x00\x00' + '\x00\x00\x00\x00'
        req +=  '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        req +=  '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        req +=  '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        req +=  '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        req +=  '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        req +=  '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        req +=  '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        req +=  '\x00\x00\x00\x00'
        self.pcp.setresponse(resp)
        self.pcp.setrequest(req)
        self.assertRaises(Exception, self.pcp.submitresponse)

    def test24_intportmismatch_map(self):
        """request and response MAP internal ports must match"""
        resp =  '\x01\x81\x00\x00' + '\x00\x00\x00\x00'
        resp += '\x00\x01\xe2\x40' + '\x00\x00\x00\x00'
        resp += '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        resp += '\x00\x00\x00\x00' + '\x66\x77\x00\x00'
        resp += '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        resp += '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        req =   '\x01\x01\x00\x00' + '\x00\x00\x00\x00'
        req +=  '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        req +=  '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        req +=  '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        req +=  '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        req +=  '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        self.pcp.setresponse(resp)
        self.pcp.setrequest(req)
        self.assertRaises(Exception, self.pcp.submitresponse)

    def test25_intportmismatch_peer(self):
        """request and response PEER internal ports must match"""
        resp =  '\x01\x82\x00\x00' + '\x00\x00\x00\x00'
        resp += '\x00\x01\xe2\x40' + '\x00\x00\x00\x00'
        resp += '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        resp += '\x00\x00\x00\x00' + '\x66\x77\x00\x00'
        resp += '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        resp += '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        resp += '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        resp += '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        resp += '\x00\x00\x00\x00'
        req =   '\x01\x02\x00\x00' + '\x00\x00\x00\x00'
        req +=  '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        req +=  '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        req +=  '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        req +=  '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        req +=  '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        req +=  '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        req +=  '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        req +=  '\x00\x00\x00\x00'
        self.pcp.setresponse(resp)
        self.pcp.setrequest(req)
        self.assertRaises(Exception, self.pcp.submitresponse)

    def test26_peerportmismatch(self):
        """request and response PEER remote peer ports must match"""
        resp =  '\x01\x82\x00\x00' + '\x00\x00\x00\x00'
        resp += '\x00\x01\xe2\x40' + '\x00\x00\x00\x00'
        resp += '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        resp += '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        resp += '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        resp += '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        resp += '\x66\x77\x00\x00' + '\x00\x00\x00\x00'
        resp += '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        resp += '\x00\x00\x00\x00'
        req =   '\x01\x02\x00\x00' + '\x00\x00\x00\x00'
        req +=  '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        req +=  '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        req +=  '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        req +=  '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        req +=  '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        req +=  '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        req +=  '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        req +=  '\x00\x00\x00\x00'
        self.pcp.setresponse(resp)
        self.pcp.setrequest(req)
        self.assertRaises(Exception, self.pcp.submitresponse)

    def test27_peeraddressmismatch(self):
        """request and response PEER remote peer addresses must match"""
        resp =  '\x01\x82\x00\x00' + '\x00\x00\x00\x00'
        resp += '\x00\x01\xe2\x40' + '\x00\x00\x00\x00'
        resp += '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        resp += '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        resp += '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        resp += '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        resp += '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        resp += '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        resp += '\x00\x11\x00\x00'
        req =   '\x01\x02\x00\x00' + '\x00\x00\x00\x00'
        req +=  '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        req +=  '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        req +=  '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        req +=  '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        req +=  '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        req +=  '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        req +=  '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        req +=  '\x00\x00\x00\x00'
        self.pcp.setresponse(resp)
        self.pcp.setrequest(req)
        self.assertRaises(Exception, self.pcp.submitresponse)

    def test28_options_map(self):
        """too many (>5) options"""
        resp =  '\x01\x81\x00\x00' + '\x00\x00\x00\x00'
        resp += '\x00\x01\xe2\x40' + '\x00\x00\x00\x00'
        resp += '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        resp += '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        resp += '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        resp += '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        resp += '\x11\x00\x00\x00' + '\x22\x00\x00\x00'
        resp += '\x33\x00\x00\x00' + '\x44\x00\x00\x00'
        resp += '\x55\x00\x00\x00' + '\x66\x00\x00\x00'
        req =   '\x01\x01\x00\x00' + '\x00\x00\x00\x00'
        req +=  '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        req +=  '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        req +=  '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        req +=  '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        req +=  '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        req +=  '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        req +=  '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        req +=  '\x00\x00\x00\x00'
        self.pcp.setresponse(resp)
        self.pcp.setrequest(req)
        self.assertRaises(Exception, self.pcp.submitresponse)

    def test29_options_peer(self):
        """too many (>5) options"""
        resp =  '\x01\x82\x00\x00' + '\x00\x00\x00\x00'
        resp += '\x00\x01\xe2\x40' + '\x00\x00\x00\x00'
        resp += '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        resp += '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        resp += '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        resp += '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        resp += '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        resp += '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        resp += '\x00\x00\x00\x00' + '\x11\x00\x00\x00'
        resp += '\x22\x00\x00\x00' + '\x33\x00\x00\x00'
        resp += '\x44\x00\x00\x00' + '\x55\x00\x00\x00'
        resp += '\x66\x00\x00\x00'
        req =   '\x01\x02\x00\x00' + '\x00\x00\x00\x00'
        req +=  '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        req +=  '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        req +=  '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        req +=  '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        req +=  '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        req +=  '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        req +=  '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        req +=  '\x00\x00\x00\x00'
        self.pcp.setresponse(resp)
        self.pcp.setrequest(req)
        self.assertRaises(Exception, self.pcp.submitresponse)

    def test30_map(self):
        """MAP example"""
        resp =  '\x01\x81\x00\x00' + '\x00\x00\x00\x00'
        resp += '\x00\x01\xe2\x40' + '\x00\x00\x00\x00'
        resp += '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        resp += '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        resp += '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        resp += '\x00\x00\xff\xff' + '\x00\x00\x00\x00'
        req =   '\x01\x01\x00\x00' + '\x00\x00\x00\x00'
        req +=  '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        req +=  '\x00\x00\xff\xff' + '\x00\x00\x00\x00'
        req +=  '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        req +=  '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        req +=  '\x00\x00\xff\xff' + '\x00\x00\x00\x00'
        self.pcp.setresponse(resp)
        self.pcp.setrequest(req)
        self.pcp.submitresponse()

    def test31_peer(self):
        """PEER example"""
        resp =  '\x01\x82\x00\x00' + '\x00\x00\x00\x00'
        resp += '\x00\x01\xe2\x40' + '\x00\x00\x00\x00'
        resp += '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        resp += '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        resp += '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        resp += '\x00\x00\xff\xff' + '\x00\x00\x00\x00'
        resp += '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        resp += '\x00\x00\x00\x00' + '\x00\x00\xff\xff'
        resp += '\x00\x00\x00\x00'
        req =   '\x01\x02\x00\x00' + '\x00\x00\x00\x00'
        req +=  '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        req +=  '\x00\x00\xff\xff' + '\x00\x00\x00\x00'
        req +=  '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        req +=  '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        req +=  '\x00\x00\xff\xff' + '\x00\x00\x00\x00'
        req +=  '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        req +=  '\x00\x00\x00\x00' + '\x00\x00\xff\xff'
        req +=  '\x00\x00\x00\x00'
        self.pcp.setresponse(resp)
        self.pcp.setrequest(req)
        self.pcp.submitresponse()

class Test39GetOptInfo(unittest.TestCase):
    """internal option helper"""

    def setUp(self):
        """just create the object"""
        self.pcp = libpcp.PCP()

    def test01_arg(self):
        """getoptinfo requires 2 arguments"""
        self.assertRaises(TypeError, self.pcp.getoptinfo)

    def test02_args(self):
        """getoptinfo requires 2 arguments"""
        self.assertRaises(TypeError, self.pcp.getoptinfo,
                          [1, 'foo', 'too many'])

    def test03_code_type(self):
        """code is a number"""
        self.assertRaises(TypeError, self.pcp.getoptinfo, ['foo', 'bar'])

    def test04_code_bad(self):
        """only THIRD_PARTY, PREFER_FAILURE and FILTER are supported"""
        self.assertIsNone(self.pcp.getoptinfo(123, 'foo'))

    def test05_data_type(self):
        """data is a string"""
        self.assertRaises(TypeError, self.pcp.getoptinfo, [0, 1])

    def test06_data_none(self):
        """data is a string"""
        self.assertRaises(TypeError, self.pcp.getoptinfo, [0, None])

    def test21_3rdparty_all(self):
        """empty THIRD_PARTY option example"""
        self.assertEqual(self.pcp.getoptinfo(libpcp.OPTION_THIRD_PARTY, ''),
                         'all')

    def test22_3rdparty_bad(self):
        """illegal length THIRD_PARTY option"""
        self.assertEqual(self.pcp.getoptinfo(libpcp.OPTION_THIRD_PARTY, 'foo'),
                         None)

    def test23_3rdparty_v4(self):
        """IPv4 THIRD_PARTY option example"""
        taddr = ADDR4
        baddr = socket.inet_pton(socket.AF_INET, taddr)
        self.assertEqual(self.pcp.getoptinfo(libpcp.OPTION_THIRD_PARTY,
                                             MAPPED + baddr),
                         taddr)

    def test24_3rdparty_v6(self):
        """IPv6 THIRD_PARTY option example"""
        taddr = ADDR6
        baddr = socket.inet_pton(socket.AF_INET6, taddr)
        self.assertEqual(self.pcp.getoptinfo(libpcp.OPTION_THIRD_PARTY, baddr),
                         taddr)

    def test25_filter_bad(self):
        """illegal length FILTER option"""
        self.assertEqual(self.pcp.getoptinfo(libpcp.OPTION_FILTER, 'foo'),
                         None)

    def test26_filter_v4(self):
        """IPv4 FILTER option example"""
        taddr = ADDR4
        baddr = socket.inet_pton(socket.AF_INET, taddr)
        bdata = '\x00\x20\x04\xd2' + MAPPED + baddr
        self.assertEqual(self.pcp.getoptinfo(libpcp.OPTION_FILTER, bdata),
                         '10.1.2.3/32 1234')

    def test27_filter_v6(self):
        """IPv6 FILTER option example"""
        taddr = ADDR6
        baddr = socket.inet_pton(socket.AF_INET6, taddr)
        bdata = '\x00\x80\x04\xd2' + baddr
        self.assertEqual(self.pcp.getoptinfo(libpcp.OPTION_FILTER, bdata),
                         taddr + '/128 1234')

class Test92BuildRequest(unittest.TestCase):
    """internal buildrequest"""

    def setUp(self):
        """just create the object"""
        self.pcp = libpcp.PCP()

    def tearDown(self):
        """cleanup"""
        self.pcp.close()

    def test01_opcode_map(self):
        """opcode=MAP"""
        self.pcp.open(SERVER4)
        self.pcp.makerequest(opcode=libpcp.OPCODE_MAP)
        req = self.pcp.getrequest()
        self.assertEqual(ord(req[1]), libpcp.OPCODE_MAP)

    def test02_opcode_peer(self):
        """opcode=PEER"""
        self.pcp.open(SERVER4)
        self.pcp.makerequest(opcode=libpcp.OPCODE_PEER)
        req = self.pcp.getrequest()
        self.assertEqual(ord(req[1]), libpcp.OPCODE_PEER)

    def test03_lifetime(self):
        """lifetime=123456"""
        self.pcp.open(SERVER4)
        self.pcp.makerequest(opcode=libpcp.OPCODE_MAP, lifetime=123456)
        req = self.pcp.getrequest()
        self.assertEqual(req[4:8], '\x00\x01\xe2\x40')

    def test04_protocol_map(self):
        """protocol=123 in MAP"""
        self.pcp.open(SERVER4)
        self.pcp.makerequest(opcode=libpcp.OPCODE_MAP, protocol=123)
        req = self.pcp.getrequest()
        self.assertEqual(ord(req[24]), 123)

    def test05_protocol_peer(self):
        """protocol=123 in PEER"""
        self.pcp.open(SERVER4)
        self.pcp.makerequest(opcode=libpcp.OPCODE_PEER, protocol=123)
        req = self.pcp.getrequest()
        self.assertEqual(ord(req[24]), 123)

    def test06_intaddr_v4(self):
        """intaddr4=10.1.2.3"""
        taddr = ADDR4
        baddr = socket.inet_pton(socket.AF_INET, taddr)
        self.pcp.setautofill(0)
        self.pcp.makerequest(opcode=libpcp.OPCODE_MAP, intaddr=taddr)
        req = self.pcp.getrequest()
        self.assertEqual(req[8:24], MAPPED + baddr)

    def test07_intaddr_v6(self):
        """intaddr6=2a01:e35:2ec7:83d0:cb3a:14ff:fe1f:3b06"""
        taddr = ADDR6
        baddr = socket.inet_pton(socket.AF_INET6, taddr)
        self.pcp.setautofill(0)
        self.pcp.makerequest(opcode=libpcp.OPCODE_MAP, intaddr=taddr)
        req = self.pcp.getrequest()
        self.assertEqual(req[8:24], baddr)

    def test08_extaddr_mapv4(self):
        """extaddr4=10.1.2.3 in MAP"""
        taddr = ADDR4
        baddr = socket.inet_pton(socket.AF_INET, taddr)
        self.pcp.open(SERVER4)
        self.pcp.makerequest(opcode=libpcp.OPCODE_MAP, extaddr=taddr)
        req = self.pcp.getrequest()
        self.assertEqual(req[32:48], MAPPED + baddr)

    def test09_extaddr_mapv6(self):
        """extaddr6=2a01:e35:2ec7:83d0:cb3a:14ff:fe1f:3b06 in MAP"""
        taddr = ADDR6
        baddr = socket.inet_pton(socket.AF_INET6, taddr)
        self.pcp.open(SERVER4)
        self.pcp.makerequest(opcode=libpcp.OPCODE_MAP, extaddr=taddr)
        req = self.pcp.getrequest()
        self.assertEqual(req[32:48], baddr)

    def test10_extaddr_peerv4(self):
        """extaddr4=10.1.2.3 in PEER"""
        taddr = ADDR4
        baddr = socket.inet_pton(socket.AF_INET, taddr)
        self.pcp.open(SERVER4)
        self.pcp.makerequest(opcode=libpcp.OPCODE_PEER, extaddr=taddr)
        req = self.pcp.getrequest()
        self.assertEqual(req[32:48], MAPPED + baddr)

    def test11_extaddr_peerv6(self):
        """extaddr6=2a01:e35:2ec7:83d0:cb3a:14ff:fe1f:3b06 in PEER"""
        taddr = ADDR6
        baddr = socket.inet_pton(socket.AF_INET6, taddr)
        self.pcp.open(SERVER4)
        self.pcp.makerequest(opcode=libpcp.OPCODE_PEER, extaddr=taddr)
        req = self.pcp.getrequest()
        self.assertEqual(req[32:48], baddr)

    def test12_remotepeer_v4(self):
        """remotepeer4=10.1.2.3"""
        taddr = ADDR4
        baddr = socket.inet_pton(socket.AF_INET, taddr)
        self.pcp.open(SERVER4)
        self.pcp.makerequest(opcode=libpcp.OPCODE_PEER, remotepeer=taddr)
        req = self.pcp.getrequest()
        self.assertEqual(req[52:68], MAPPED + baddr)

    def test13_remotepeer_v6(self):
        """remotepeer6=2a01:e35:2ec7:83d0:cb3a:14ff:fe1f:3b06"""
        taddr = ADDR6
        baddr = socket.inet_pton(socket.AF_INET6, taddr)
        self.pcp.open(SERVER4)
        self.pcp.makerequest(opcode=libpcp.OPCODE_PEER, remotepeer=taddr)
        req = self.pcp.getrequest()
        self.assertEqual(req[52:68], baddr)

    def test14_intport_map(self):
        """intport=1234 in MAP"""
        self.pcp.open(SERVER4)
        self.pcp.makerequest(opcode=libpcp.OPCODE_MAP, intport=1234)
        req = self.pcp.getrequest()
        self.assertEqual(req[28:30], '\x04\xd2')

    def test15_intport_peer(self):
        """intport=1234 in PEER"""
        self.pcp.open(SERVER4)
        self.pcp.makerequest(opcode=libpcp.OPCODE_PEER, intport=1234)
        req = self.pcp.getrequest()
        self.assertEqual(req[28:30], '\x04\xd2')

    def test16_extport_map(self):
        """extport=1234 in MAP"""
        self.pcp.open(SERVER4)
        self.pcp.makerequest(opcode=libpcp.OPCODE_MAP, extport=1234)
        req = self.pcp.getrequest()
        self.assertEqual(req[30:32], '\x04\xd2')

    def test17_extport_peer(self):
        """extport=1234 in PEER"""
        self.pcp.open(SERVER4)
        self.pcp.makerequest(opcode=libpcp.OPCODE_PEER, extport=1234)
        req = self.pcp.getrequest()
        self.assertEqual(req[30:32], '\x04\xd2')

    def test18_peerport(self):
        """peerport=1234"""
        self.pcp.open(SERVER4)
        self.pcp.makerequest(opcode=libpcp.OPCODE_PEER, peerport=1234)
        req = self.pcp.getrequest()
        self.assertEqual(req[48:50], '\x04\xd2')

    def test20_option0_map(self):
        """options=[] in MAP"""
        self.pcp.open(SERVER4)
        self.pcp.makerequest(opcode=libpcp.OPCODE_MAP, options=[])
        self.assertEqual(len(self.pcp.getrequest()), 48)

    def test21_option0_peer(self):
        """options=[] in PEER"""
        self.pcp.open(SERVER4)
        self.pcp.makerequest(opcode=libpcp.OPCODE_PEER, options=[])
        self.assertEqual(len(self.pcp.getrequest()), 68)

    def test22_option_map(self):
        """options=[ (123, 'abc') ] in MAP"""
        self.pcp.open(SERVER4)
        self.pcp.makerequest(opcode=libpcp.OPCODE_MAP,
                             options=[ (123, 'abc') ])
        req = self.pcp.getrequest()
        self.assertTrue((len(req) == 56) and
                        (req[48:56] == '\x7b\x00\x00\x03abc\x00'))

    def test23_option_peer(self):
        """options=[ (123, 'abc') ] in PEER"""
        self.pcp.open(SERVER4)
        self.pcp.makerequest(opcode=libpcp.OPCODE_PEER,
                             options=[ (123, 'abc') ])
        req = self.pcp.getrequest()
        self.assertTrue((len(req) == 76) and
                        (req[68:76] == '\x7b\x00\x00\x03abc\x00'))

    def test24_options_map(self):
        """options=[ (123, 'abc'), (55, 'qwerty') ]"""
        self.pcp.open(SERVER4)
        self.pcp.makerequest(opcode=libpcp.OPCODE_MAP,
                             options=[ (123, 'abc'), (55, 'qwerty') ])
        req = self.pcp.getrequest()
        self.assertTrue((len(req) == 68) and
                        (req[48:56] == '\x7b\x00\x00\x03abc\x00') and
                        (req[56:68] == '\x37\x00\x00\x06qwerty\x00\x00'))

    def test25_options_peer(self):
        """options=[ (123, 'abc'), (55, 'qwerty') ]"""
        self.pcp.open(SERVER4)
        self.pcp.makerequest(opcode=libpcp.OPCODE_PEER,
                             options=[ (123, 'abc'), (55, 'qwerty') ])
        req = self.pcp.getrequest()
        self.assertTrue((len(req) == 88) and
                        (req[68:76] == '\x7b\x00\x00\x03abc\x00') and
                        (req[76:88] == '\x37\x00\x00\x06qwerty\x00\x00'))

class Test99DecodeResponse(unittest.TestCase):
    """internal decoderesponse"""

    def setUp(self):
        """just create the object"""
        self.pcp = libpcp.PCP()

    def tearDown(self):
        """cleanup"""
        self.pcp.close()

    def test00_rapid_recovery(self):
        """rapid recovery example"""
        resp =  '\x01\x80\x00\x00' + '\x00\x00\x00\x00'
        resp += '\x00\x01\xe2\x40' + '\x00\x00\x00\x00'
        resp += '\x00\x00\x00\x00' + '\x00\x00\x00\x00'
        self.pcp.setresponse(resp)
        self.assertEqual(self.pcp.submitresponse(), (libpcp.TRY_AGAIN, {}))

    def test01_opcode_map(self):
        """opcode=MAP?"""
        self.pcp.open(SERVER4)
        self.pcp.makerequest(opcode=libpcp.OPCODE_MAP)
        req = self.pcp.getrequest()
        resp = req[0] + chr(ord(req[1]) | libpcp.OPCODE_RESPONSE) + req[2:]
        self.pcp.setresponse(resp)
        (r, d) = self.pcp.submitresponse()
        self.assertEqual(d['opcode'],
                         libpcp.OPCODE_MAP | libpcp.OPCODE_RESPONSE)

    def test02_opcode_peer(self):
        """opcode=PEER?"""
        self.pcp.open(SERVER4)
        self.pcp.makerequest(opcode=libpcp.OPCODE_PEER)
        req = self.pcp.getrequest()
        resp = req[0] + chr(ord(req[1]) | libpcp.OPCODE_RESPONSE) + req[2:]
        self.pcp.setresponse(resp)
        (r, d) = self.pcp.submitresponse()
        self.assertEqual(d['opcode'],
                         libpcp.OPCODE_PEER | libpcp.OPCODE_RESPONSE)

    def test10_result(self):
        """result=0?"""
        self.pcp.open(SERVER4)
        self.pcp.makerequest(opcode=libpcp.OPCODE_MAP)
        req = self.pcp.getrequest()
        resp = req[0] + chr(ord(req[1]) | libpcp.OPCODE_RESPONSE) + req[2:]
        self.pcp.setresponse(resp)
        (r, d) = self.pcp.submitresponse()
        self.assertEqual(d['result'], libpcp.OK)

    def test11_error(self):
        """result=error"""
        self.pcp.open(SERVER4)
        self.pcp.makerequest(opcode=libpcp.OPCODE_MAP)
        req = self.pcp.getrequest()
        resp = req[0] + chr(ord(req[1]) | libpcp.OPCODE_RESPONSE)
        resp += '\x00\x01' + req[4:]
        self.pcp.setresponse(resp)
        self.assertRaises(Exception, self.pcp.submitresponse)

    def test20_lifetime(self):
        """lifetime"""
        self.pcp.open(SERVER4)
        self.pcp.makerequest(opcode=libpcp.OPCODE_MAP,
                             lifetime=123456789)
        req = self.pcp.getrequest()
        resp = req[0] + chr(ord(req[1]) | libpcp.OPCODE_RESPONSE) + req[2:]
        self.pcp.setresponse(resp)
        (r, d) = self.pcp.submitresponse()
        self.assertEqual(d['lifetime'], 123456789)

    def test21_epoch(self):
        """epoch"""
        self.pcp.open(SERVER4)
        self.pcp.makerequest(opcode=libpcp.OPCODE_MAP)
        req = self.pcp.getrequest()
        resp = req[0] + chr(ord(req[1]) | libpcp.OPCODE_RESPONSE) + req[2:8]
        resp += '\x07\x5b\xcd\x15' + req[12:]
        self.pcp.setresponse(resp)
        (r, d) = self.pcp.submitresponse()
        self.assertEqual(d['epoch'], 123456789)

    def test30_protocol_map(self):
        """protocol (MAP)"""
        self.pcp.open(SERVER4)
        self.pcp.makerequest(opcode=libpcp.OPCODE_MAP,
                             protocol=socket.IPPROTO_TCP)
        req = self.pcp.getrequest()
        resp = req[0] + chr(ord(req[1]) | libpcp.OPCODE_RESPONSE) + req[2:]
        self.pcp.setresponse(resp)
        (r, d) = self.pcp.submitresponse()
        self.assertEqual(d['protocol'], socket.IPPROTO_TCP)

    def test31_protocol_peer(self):
        """protocol (PEER)"""
        self.pcp.open(SERVER4)
        self.pcp.makerequest(opcode=libpcp.OPCODE_PEER,
                             protocol=socket.IPPROTO_UDP)
        req = self.pcp.getrequest()
        resp = req[0] + chr(ord(req[1]) | libpcp.OPCODE_RESPONSE) + req[2:]
        self.pcp.setresponse(resp)
        (r, d) = self.pcp.submitresponse()
        self.assertEqual(d['protocol'], socket.IPPROTO_UDP)

    def test40_intport_map(self):
        """internal port (MAP)"""
        self.pcp.open(SERVER4)
        self.pcp.makerequest(opcode=libpcp.OPCODE_MAP, intport=12345)
        req = self.pcp.getrequest()
        resp = req[0] + chr(ord(req[1]) | libpcp.OPCODE_RESPONSE) + req[2:]
        self.pcp.setresponse(resp)
        (r, d) = self.pcp.submitresponse()
        self.assertEqual(d['intport'], 12345)

    def test41_intport_peer(self):
        """internal port (PEER)"""
        self.pcp.open(SERVER4)
        self.pcp.makerequest(opcode=libpcp.OPCODE_PEER, intport=12345)
        req = self.pcp.getrequest()
        resp = req[0] + chr(ord(req[1]) | libpcp.OPCODE_RESPONSE) + req[2:]
        self.pcp.setresponse(resp)
        (r, d) = self.pcp.submitresponse()
        self.assertEqual(d['intport'], 12345)

    def test50_extport_map(self):
        """external port (MAP)"""
        self.pcp.open(SERVER4)
        self.pcp.makerequest(opcode=libpcp.OPCODE_MAP, extport=12345)
        req = self.pcp.getrequest()
        resp = req[0] + chr(ord(req[1]) | libpcp.OPCODE_RESPONSE) + req[2:]
        self.pcp.setresponse(resp)
        (r, d) = self.pcp.submitresponse()
        self.assertEqual(d['extport'], 12345)

    def test51_extport_peer(self):
        """external port (PEER)"""
        self.pcp.open(SERVER4)
        self.pcp.makerequest(opcode=libpcp.OPCODE_PEER, extport=12345)
        req = self.pcp.getrequest()
        resp = req[0] + chr(ord(req[1]) | libpcp.OPCODE_RESPONSE) + req[2:]
        self.pcp.setresponse(resp)
        (r, d) = self.pcp.submitresponse()
        self.assertEqual(d['extport'], 12345)

    def test60_extaddr_map_v4(self):
        """external port (MAP)"""
        self.pcp.open(SERVER4)
        self.pcp.makerequest(opcode=libpcp.OPCODE_MAP, extaddr=ADDR4)
        req = self.pcp.getrequest()
        resp = req[0] + chr(ord(req[1]) | libpcp.OPCODE_RESPONSE) + req[2:]
        self.pcp.setresponse(resp)
        (r, d) = self.pcp.submitresponse()
        self.assertEqual(d['extaddr'], ADDR4)

    def test61_extaddr_map_v6(self):
        """external port (MAP)"""
        self.pcp.open(SERVER6)
        self.pcp.makerequest(opcode=libpcp.OPCODE_MAP, extaddr=ADDR6)
        req = self.pcp.getrequest()
        resp = req[0] + chr(ord(req[1]) | libpcp.OPCODE_RESPONSE) + req[2:]
        self.pcp.setresponse(resp)
        (r, d) = self.pcp.submitresponse()
        self.assertEqual(d['extaddr'], ADDR6)

    def test62_extaddr_peer_v4(self):
        """external port (PEER)"""
        self.pcp.open(SERVER4)
        self.pcp.makerequest(opcode=libpcp.OPCODE_PEER, extaddr=ADDR4)
        req = self.pcp.getrequest()
        resp = req[0] + chr(ord(req[1]) | libpcp.OPCODE_RESPONSE) + req[2:]
        self.pcp.setresponse(resp)
        (r, d) = self.pcp.submitresponse()
        self.assertEqual(d['extaddr'], ADDR4)

    def test63_extaddr_peer_v6(self):
        """external port (PEER)"""
        self.pcp.open(SERVER6)
        self.pcp.makerequest(opcode=libpcp.OPCODE_PEER, extaddr=ADDR6)
        req = self.pcp.getrequest()
        resp = req[0] + chr(ord(req[1]) | libpcp.OPCODE_RESPONSE) + req[2:]
        self.pcp.setresponse(resp)
        (r, d) = self.pcp.submitresponse()
        self.assertEqual(d['extaddr'], ADDR6)

    def test70_peerport(self):
        """remote peer port"""
        self.pcp.open(SERVER4)
        self.pcp.makerequest(opcode=libpcp.OPCODE_PEER, peerport=12345)
        req = self.pcp.getrequest()
        resp = req[0] + chr(ord(req[1]) | libpcp.OPCODE_RESPONSE) + req[2:]
        self.pcp.setresponse(resp)
        (r, d) = self.pcp.submitresponse()
        self.assertEqual(d['peerport'], 12345)

    def test71_peeraddr_v4(self):
        """remote peer address"""
        self.pcp.open(SERVER4)
        self.pcp.makerequest(opcode=libpcp.OPCODE_PEER, remotepeer=ADDR4)
        req = self.pcp.getrequest()
        resp = req[0] + chr(ord(req[1]) | libpcp.OPCODE_RESPONSE) + req[2:]
        self.pcp.setresponse(resp)
        (r, d) = self.pcp.submitresponse()
        self.assertEqual(d['remotepeer'], ADDR4)

    def test72_peeraddr_v6(self):
        """remote peer address"""
        self.pcp.open(SERVER6)
        self.pcp.makerequest(opcode=libpcp.OPCODE_PEER, remotepeer=ADDR6)
        req = self.pcp.getrequest()
        resp = req[0] + chr(ord(req[1]) | libpcp.OPCODE_RESPONSE) + req[2:]
        self.pcp.setresponse(resp)
        (r, d) = self.pcp.submitresponse()
        self.assertEqual(d['remotepeer'], ADDR6)

    def test80_options_map(self):
        """options=[ (123, 'abc'), (55, 'qwerty') ]"""
        self.pcp.open(SERVER4)
        options=[ (123, 'abc'), (55, 'qwerty') ]
        self.pcp.makerequest(opcode=libpcp.OPCODE_MAP, options=options)
        req = self.pcp.getrequest()
        resp = req[0] + chr(ord(req[1]) | libpcp.OPCODE_RESPONSE) + req[2:]
        self.pcp.setresponse(resp)
        (r, d) = self.pcp.submitresponse()
        self.assertEqual(d['options'], options)

    def test81_options_peer(self):
        """options=[ (123, 'abc'), (55, 'qwerty') ]"""
        self.pcp.open(SERVER4)
        options=[ (123, 'abc'), (55, 'qwerty') ]
        self.pcp.makerequest(opcode=libpcp.OPCODE_PEER, options=options)
        req = self.pcp.getrequest()
        resp = req[0] + chr(ord(req[1]) | libpcp.OPCODE_RESPONSE) + req[2:]
        self.pcp.setresponse(resp)
        (r, d) = self.pcp.submitresponse()
        self.assertEqual(d['options'], options)

if __name__ == '__main__':
    unittest.main()
