/*
 * Copyright (C) 2011-2012  Internet Systems Consortium, Inc. ("ISC")
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND ISC DISCLAIMS ALL WARRANTIES WITH
 * REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS.  IN NO EVENT SHALL ISC BE LIABLE FOR ANY SPECIAL, DIRECT,
 * INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
 * OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */

/*
 * PCP ATF unit tests for plain-mode AFTR and NAT444
 *
 * Requirements/Default defines:
 *  64010 for tcp and udp is in the explicit dynamic mapping port range
 *  SERVERADDR == 2001:0:0:1::1
 *  INTERNALADDR == 10.0.1.2
 *  supposed to be not authorized as an internal address:
 *  PEERADDR == 192.0.2.2
 *  (the unknown client uses the link-local address on the shared link
 *   with the server launched with the SERVERADDR as argument:
 *   highly testbed dependent so protected by LLNAME)
 *  LLADDR == fe80::20c:29ff:fefd:de17
 *  LLSCOPEID == 3
 *  LLNAME == vrelaynat.fdupont.fr
 *  STATICPORT == 10
 */

/* $Id: e1574328971c5f1ab6a49887e3f2f624ac15a9b6 $ */

#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <time.h>

#include <atf-c.h>

#define UNUSED(x)	(x) = (x)

#define PCP_PORT		5351

#define OPCODE_RESPONSE		128	/* response bit */
#define OPCODE_ANNOUNCE		0
#define OPCODE_ANNOUNCE_RESPONSE OPCODE_ANNOUNCE | OPCODE_RESPONSE
#define OPCODE_MAP		1
#define OPCODE_MAP_RESPONSE	OPCODE_MAP | OPCODE_RESPONSE
#define OPCODE_PEER		2
#define OPCODE_PEER_RESPONSE	OPCODE_PEER | OPCODE_RESPONSE

#define OFF_VERSION		0
#define OFF_OPCODE		1
#define OFF_RESERVED		2
#define OFF_RESULT_CODE		3
#define OFF_LIFETIME		4
#define OFF_EPOCH		8
#define OFF_CLIENT_ADDR		8
#define OFF_OPCODE_HDR		24

#define OFF_MAP_PROTOCOL	(OFF_OPCODE_HDR + 0)
#define OFF_MAP_INTERNAL_PORT	(OFF_OPCODE_HDR + 4)
#define OFF_MAP_EXTERNAL_PORT	(OFF_OPCODE_HDR + 6)
#define OFF_MAP_EXTERNAL_ADDR	(OFF_OPCODE_HDR + 8)

#define OFF_PEER_PROTOCOL	(OFF_OPCODE_HDR + 0)
#define OFF_PEER_INTERNAL_PORT	(OFF_OPCODE_HDR + 4)
#define OFF_PEER_EXTERNAL_PORT	(OFF_OPCODE_HDR + 6)
#define OFF_PEER_EXTERNAL_ADDR	(OFF_OPCODE_HDR + 8)
#define OFF_PEER_REMOTE_PORT	(OFF_OPCODE_HDR + 24)
#define OFF_PEER_REMOTE_ADDR	(OFF_OPCODE_HDR + 28)

#define LEN_ANNOUNCE		24
#define LEN_MAP			48
#define LEN_PEER		68
#define LEN_MAX			1024

#define OPTION_UNPROCESSED	0	/* options not processed by server */
#define OPTION_THIRD_PARTY	1	/* third-party request */
#define OPTION_PREF_FAIL	2	/* prefer failure */
#define OPTION_FILTER		3	/* remote peer filter */
#define OPTION_OPTIONAL		128	/* flag bit */

#define ERR_SUCCESS		0	/* success */
#define ERR_UNSUPVERSION	1	/* unsupported version */
#define ERR_NOTAUTH		2	/* not authorized */
#define ERR_BADREQUEST		3	/* malformed request */
#define ERR_UNSUPOPCODE		4	/* unsupported opcode */
#define ERR_UNSUPOPTION		5	/* unsupported option */
#define ERR_BADOPTION		6	/* malformed option */
#define ERR_NETFAILURE		7	/* network failure */
#define ERR_NORESOURCES		8	/* out of resources */
#define ERR_UNSUPPROTO		9	/* unsupported protocol */
#define ERR_EXQUOTA		10	/* user exceeded quota */
#define ERR_CANTPROVIDE		11	/* cannot provide external */
#define ERR_ADDRMISMATCH	12	/* address mismatch */
#define ERR_TOOMANYPEER		13	/* excessive number of remote peers */
#define ERR_PROCESSING		192	/* processing error */

#define TEST_ORIGIN	"test"

/* CGN types */
#define NAT		0		/* simple (not CGN) NAT44 */
#define DSLITE		1		/* DS-Lite AFTR */
#define NAT444		2		/* NAT444 CGN */
#define NAT64		3		/* NAT64 CGN */
#define NAT46		4		/* NAT46 CGN */
#define NPTv6		5		/* NPTv6 (aka NAT66) CGN */
#define FW4		6		/* IPv4 firewall */
#define FW6		7		/* IPv4 firewall */
#define L2NAT		8		/* layer2 (IPv4) CGN */

#if CGN_TYPE == DSLITE
#define USE_THIRD_PARTY
#define TRANSPORT	AF_INET6
#ifndef SERVERADDR
#define SERVERADDR	"2001:0:0:1::1"
#endif
#else /* CGN_TYPE == NAT444 or undefined */
#define TRANSPORT	AF_INET
#ifndef SERVERADDR
#define SERVERADDR	"192.0.0.1"
#endif
#endif

#ifndef INTERNALADDR
#define INTERNALADDR	"10.0.1.2"
#endif
#ifndef PEERADDR
#define PEERADDR	"192.0.2.2"
#endif
#ifndef LLADDR
#define LLADDR		"fe80::20c:29ff:fefd:de17"
#endif
#ifndef LLSCOPEID
#define LLSCOPEID	3
#endif
#ifndef LLNAME
#define LLNAME		"vrelaynat.fdupont.fr"
#endif
#ifndef STATICPORT
#define STATICPORT	10
#endif

uint8_t mapped[16] =
	{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	  0x00, 0x00, 0xff, 0xff, 0x00, 0x00, 0x00, 0x00 };

uint8_t intaddr[16];
uint8_t peeraddr[16];
uint8_t extaddr[16];
int fd;
#if TRANSPORT == AF_INET6
struct sockaddr_in6 server;
#else
struct sockaddr_in server;
#endif
uint8_t srcaddr[16];
uint8_t request[1080];
ssize_t reqlen, optoff;
uint8_t response[1080];
uint32_t epoch;
struct timeval last, now;

static void
setup_tp()
{
	int ret;
	socklen_t l;
#if TRANSPORT == AF_INET6
	struct sockaddr_in6 source;
#else
	struct sockaddr_in source;
#endif

	memcpy(intaddr, mapped, 12);
	ret = inet_pton(AF_INET, INTERNALADDR, &intaddr[12]);
	ATF_REQUIRE_EQ_MSG(ret, 1, "inet_pton(%s) failed", INTERNALADDR);
	memcpy(peeraddr, mapped, 12);
	ret = inet_pton(AF_INET, PEERADDR, &peeraddr[12]);
	ATF_REQUIRE_EQ_MSG(ret, 1, "inet_pton(%s) failed", PEERADDR);
#if TRANSPORT == AF_INET6
	fd = socket(PF_INET6, SOCK_DGRAM, IPPROTO_UDP);
	ATF_REQUIRE_MSG(fd >= 0, "socket() failed");
	memset(&server, 0, sizeof(server));
	server.sin6_family = AF_INET6;
#ifdef BSD
	server.sin6_len = sizeof(server);
#endif
	ret = inet_pton(AF_INET6, SERVERADDR, &server.sin6_addr);
	ATF_REQUIRE_EQ_MSG(ret, 1, "inet_pton(%s) failed", SERVERADDR);
	server.sin6_port = htons(PCP_PORT);
#else	/* INET */
	fd = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);
	ATF_REQUIRE_MSG(fd >= 0, "socket() failed");
	memset(&server, 0, sizeof(server));
	server.sin_family = AF_INET;
#ifdef BSD
	server.sin_len = sizeof(server);
#endif
	ret = inet_pton(AF_INET, SERVERADDR, &server.sin_addr);
	ATF_REQUIRE_EQ_MSG(ret, 1, "inet_pton(%s) failed", SERVERADDR);
	server.sin_port = htons(PCP_PORT);
#endif
	ret = connect(fd, (struct sockaddr *) &server, sizeof(server));
	ATF_REQUIRE_EQ_MSG(ret, 0, "connect() failed");
	memset(&source, 0, sizeof(source));
	l = sizeof(source);
	ret = getsockname(fd, (struct sockaddr *) &source, &l);
	ATF_REQUIRE_EQ_MSG(ret, 0, "getsockname() failed");
#if TRANSPORT == AF_INET6
	ATF_REQUIRE_EQ_MSG(source.sin6_family, AF_INET6,
			   "bad source address family");
	memcpy(srcaddr, &source.sin6_addr, 16);
#else
	ATF_REQUIRE_EQ_MSG(source.sin_family, AF_INET,
			   "bad source address family");
	memcpy(srcaddr, mapped, 12);
	memcpy(&srcaddr[12], &source.sin_addr, 4);
#endif
	srand((unsigned int)time(NULL));
}

static void
setup_common(uint8_t opcode, uint32_t lifetime)
{
	uint32_t n32;

	memset(request, 0, sizeof(request));
	request[OFF_VERSION] = 1;
	request[OFF_OPCODE] = opcode;
	n32 = htonl(lifetime);
	memcpy(request + OFF_LIFETIME, &n32, 4);
	memcpy(request + OFF_CLIENT_ADDR, srcaddr, 16);
	reqlen = LEN_ANNOUNCE;
}

#ifdef notyet
static void
setup_announce()
{
	setup_common(OPCODE_ANNOUNCE, 0);
	reqlen = LEN_ANNOUNCE;
}
#endif

static void
setup_map(uint32_t lifetime, uint8_t proto, uint16_t intport,
	  uint16_t extport, uint8_t *extaddr)
{
	uint16_t n16;

	setup_common(OPCODE_MAP, lifetime);
	request[OFF_MAP_PROTOCOL] = proto;
	n16 = htons(intport);
	memcpy(request + OFF_MAP_INTERNAL_PORT, &n16, 2);
	n16 = htons(extport);
	memcpy(request + OFF_MAP_EXTERNAL_PORT, &n16, 2);
	memcpy(request + OFF_MAP_EXTERNAL_ADDR, extaddr, 16);
	reqlen = LEN_MAP;
}

static void
setup_peer(uint32_t lifetime, uint8_t proto, uint16_t intport,
	   uint16_t extport, uint8_t *extaddr,
	   uint16_t peerport, uint8_t *peeraddr)
{
	uint16_t n16;

	setup_common(OPCODE_PEER, lifetime);
	request[OFF_PEER_PROTOCOL] = proto;
	n16 = htons(intport);
	memcpy(request + OFF_PEER_INTERNAL_PORT, &n16, 2);
	n16 = htons(extport);
	memcpy(request + OFF_PEER_EXTERNAL_PORT, &n16, 2);
	memcpy(request + OFF_PEER_EXTERNAL_ADDR, extaddr, 16);
	n16 = htons(peerport);
	memcpy(request + OFF_PEER_REMOTE_PORT, &n16, 2);
	memcpy(request + OFF_PEER_REMOTE_ADDR, peeraddr, 16);
	reqlen = LEN_PEER;
}

static void
append_option(uint8_t optcode, uint16_t len)
{
	uint16_t n16;

	request[reqlen] = optcode;
	n16 = htons(len);
	memcpy(request + reqlen + 2, &n16, 2);
	optoff = reqlen + 4;
	reqlen += 4 + ((len + 3) & ~3);
}

static void
append_third_party(uint8_t *addr)
{
	append_option(OPTION_THIRD_PARTY, 16);
	if (addr != NULL)
		memcpy(request + optoff, addr, 16);
}

static void
append_prefer_failure()
{
	append_option(OPTION_PREF_FAIL, 0);
}

static void
append_filter(uint8_t pfxlen, uint16_t port, uint8_t *addr)
{
	uint16_t n16;

	append_option(OPTION_FILTER, 20);
	request[optoff + 1] = pfxlen;
	n16 = htons(port);
	memcpy(request + optoff + 2, &n16, 2);
	memcpy(request + optoff + 4, addr, 16);
}

static void
cleanup()
{
	if (fd >= 0)
		(void) close(fd);
	fd = -1;
}

static void
checkepoch()
{
	uint32_t n32, newepoch, cdelta, sdelta;
	int ret;
	
	ret = gettimeofday(&now, NULL);
	ATF_REQUIRE_EQ_MSG(ret, 0, "gettimeofday() failed");
	memcpy(&n32, response + OFF_EPOCH, 4);
	if (epoch != 0) {
		ATF_CHECK_MSG((now.tv_sec > last.tv_sec) ||
			      ((now.tv_sec == last.tv_sec) &&
			       (now.tv_usec >= last.tv_usec)),
			      "clock is going backward");
		newepoch = ntohl(n32);
		ATF_REQUIRE_MSG(newepoch >= epoch,
				"epoch is going backward");
		sdelta = newepoch - epoch;
		cdelta = now.tv_sec - last.tv_sec;
		if ((now.tv_usec - last.tv_usec) >= 500000)
			cdelta++;
		else if ((now.tv_usec - last.tv_usec) < -500000)
			cdelta--;
		ATF_REQUIRE_MSG(
			(cdelta + 2 >= (((sdelta << 4) - sdelta) >> 4)) &&
			(sdelta + 2 >= (((cdelta << 4) - cdelta) >> 4)) &&
			(sdelta + 3600 >= cdelta) &&
			(cdelta + 3600 >= sdelta),
			"epoch is invalid");
	}
	last = now;
	epoch = ntohl(n32);
	return;
}

static void
checkextaddr(int ispeer)
{
	static in_addr_t extaddr4;
	in_addr_t addr;

	if (ispeer) {
		if (memcmp(response + OFF_PEER_EXTERNAL_ADDR, mapped, 12) != 0)
			return;
		memcpy(&addr, response + OFF_PEER_EXTERNAL_ADDR + 12, 4);
		if (addr == 0)
			return;
		if (extaddr4 == 0) {
			extaddr4 = addr;
			return;
		}
		ATF_CHECK_EQ_MSG(extaddr4, addr, "external address mismatch");
	} else {
		if (memcmp(response + OFF_MAP_EXTERNAL_ADDR, mapped, 12) != 0)
			return;
		memcpy(&addr, response + OFF_MAP_EXTERNAL_ADDR + 12, 4);
		if (addr == 0)
			return;
		if (extaddr4 == 0) {
			extaddr4 = addr;
			return;
		}
		ATF_CHECK_EQ_MSG(extaddr4, addr, "external address mismatch");
	}
}

#define test_send(len) \
	ret = send(fd, request, len, 0); \
	ATF_REQUIRE_EQ_MSG(ret, len, "send() failed");

#define test_recv(len) \
	memset(response, 0, sizeof(response)); \
	ret = recv(fd, response, sizeof(response), 0); \
	ATF_REQUIRE_EQ_MSG(ret, len, "recv() failed");

#define test_result(result) \
	ATF_REQUIRE_EQ_MSG(response[OFF_RESULT_CODE], result, "not expected response");

static void
delall()
{
	ssize_t ret;

	setup_map(0, 0, 0, 0, mapped);
	append_third_party(mapped);
	test_send(reqlen);
	test_recv(reqlen);
	checkepoch();
	test_result(ERR_SUCCESS);
}

static void
delclient()
{
	ssize_t ret;

	setup_map(0, 0, 0, 0, mapped);
	append_third_party(intaddr);
	test_send(reqlen);
	test_recv(reqlen);
	checkepoch();
	test_result(ERR_SUCCESS);
}

static uint16_t
randport()
{
	/* return a random port number in the range 10000..59999 */
	return (uint16_t)rand() % 50000 + 10000;
}

ATF_TC_WITH_CLEANUP(00connect);
ATF_TC_HEAD(00connect, tc) {
	atf_tc_set_md_var(tc, "descr", "connect to PCP server");
}
ATF_TC_BODY(00connect, tc) {
	UNUSED(tc);
	/* send ANNOUNCE, or MAP request for discard port, 5 sec lifetime */
}	
ATF_TC_CLEANUP(00connect, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(01tooshort);
ATF_TC_HEAD(01tooshort, tc) {
	atf_tc_set_md_var(tc, "descr", "too short");
	atf_tc_set_md_var(tc, "timeout", "2");
}
ATF_TC_BODY(01tooshort, tc) {
	ssize_t ret;

	UNUSED(tc);
	setup_common(OPCODE_MAP, 0);
	test_send(3);
	memset(response, 0, sizeof(response));
	atf_tc_expect_timeout("too short requests are discarded");
	ret = recv(fd, response, 128, 0);
}
ATF_TC_CLEANUP(01tooshort, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(02response);
ATF_TC_HEAD(02response, tc) {
	atf_tc_set_md_var(tc, "descr", "response");
	atf_tc_set_md_var(tc, "timeout", "2");
}
ATF_TC_BODY(02response, tc) {
	ssize_t ret;

	UNUSED(tc);
	setup_common(OPCODE_MAP_RESPONSE, 0);
	test_send(LEN_MAP);
	memset(response, 0, sizeof(response));
	atf_tc_expect_timeout("responses are discarded");
	ret = recv(fd, response, 128, 0);
}
ATF_TC_CLEANUP(02response, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(03natpmp);
ATF_TC_HEAD(03natpmp, tc) {
	atf_tc_set_md_var(tc, "descr", "legacy NATPMP");
	atf_tc_set_md_var(tc, "timeout", "2");
}
ATF_TC_BODY(03natpmp, tc) {
	ssize_t ret;
	uint8_t expected[4];

	UNUSED(tc);
	memset(request, 0, sizeof(request));
	/* NAT-PMP external addr request: version = 0, opcode = 0 */
	test_send(2);
	test_recv(4);
	expected[OFF_VERSION] = 1;
	expected[OFF_OPCODE] = OPCODE_RESPONSE;
	expected[OFF_RESERVED] = 0;
	expected[OFF_RESULT_CODE] = ERR_UNSUPVERSION;
	ret = memcmp(response, expected, 4);
	ATF_REQUIRE_EQ_MSG(ret, 0, "not expected response");
}
ATF_TC_CLEANUP(03natpmp, tc) {
	UNUSED(tc);
	cleanup();
}

#if TRANSPORT != AF_INET6 || !defined(LLADDR) || !defined(LLSCOPEID) || !defined(LLNAME)
ATF_TC(04unknownclient);
ATF_TC_HEAD(04unknownclient, tc) {
	atf_tc_set_md_var(tc, "descr", "unknown subscriber");
}
ATF_TC_BODY(04unknownclient, tc) {
	UNUSED(tc);
	atf_tc_skip("not yet implemented");
}
#else
ATF_TC_WITH_CLEANUP(04unknownclient);
ATF_TC_HEAD(04unknownclient, tc) {
	atf_tc_set_md_var(tc, "descr", "unknown subscriber");
	atf_tc_set_md_var(tc, "timeout", "2");
}
ATF_TC_BODY(04unknownclient, tc) {
	char hostname[256];
	int ret;
	socklen_t l;
	struct sockaddr_in6 source;

	UNUSED(tc);
	memset(hostname, 0, sizeof(hostname));
	ret = gethostname(hostname, sizeof(hostname));
	ATF_REQUIRE_EQ_MSG(ret, 0, "gethostname() failed");
	if (strcasecmp(hostname, LLNAME) != 0) {
		atf_tc_skip("not yet implemented");
		return;
	}
	fd = socket(PF_INET6, SOCK_DGRAM, IPPROTO_UDP);
	ATF_REQUIRE_MSG(fd >= 0, "socket() failed");
	memset(&source, 0, sizeof(source));
	source.sin6_family = AF_INET6;
#ifdef BSD
	source.sin6_len = sizeof(source);
#endif
	ret = inet_pton(AF_INET6, LLADDR, srcaddr);
	ATF_REQUIRE_EQ_MSG(ret, 1, "inet_pton(%s) failed", LLADDR);
	source.sin6_scope_id = LLSCOPEID;
	memset(&server, 0, sizeof(server));
	ret = bind(fd, (struct sockaddr *) &source, sizeof(source));
	ATF_REQUIRE_EQ_MSG(ret, 0, "bind() failed");
	server.sin6_family = AF_INET6;
#ifdef BSD
	server.sin6_len = sizeof(server);
#endif
	ret = inet_pton(AF_INET6, SERVERADDR, &server.sin6_addr);
	ATF_REQUIRE_EQ_MSG(ret, 1, "inet_pton(%s) failed", SERVERADDR);
	server.sin6_port = htons(PCP_PORT);
	ret = connect(fd, (struct sockaddr *) &server, sizeof(server));
	ATF_REQUIRE_EQ_MSG(ret, 0, "connect() failed");
	l = sizeof(source);
	ret = getsockname(fd, (struct sockaddr *) &source, &l);
	ATF_REQUIRE_EQ_MSG(ret, 0, "getsockname() failed");

	setup_map(0, 0, 0, 0, mapped);
	test_send(LEN_MAP);
	test_recv(LEN_MAP);
	test_result(ERR_NOTAUTH);
}
ATF_TC_CLEANUP(04unknownclient, tc) {
	UNUSED(tc);
	cleanup();
}
#endif

ATF_TC_WITH_CLEANUP(05toolong);
ATF_TC_HEAD(05toolong, tc) {
	atf_tc_set_md_var(tc, "descr", "too long");
	atf_tc_set_md_var(tc, "timeout", "2");
}
ATF_TC_BODY(05toolong, tc) {
	ssize_t ret;

	UNUSED(tc);
	setup_map(0, 0, 0, 0, mapped);
	test_send(1028);
	test_recv(1024);
	test_result(ERR_BADREQUEST);
}
ATF_TC_CLEANUP(05toolong, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(06badalign);
ATF_TC_HEAD(06badalign, tc) {
	atf_tc_set_md_var(tc, "descr", "bad aligned");
	atf_tc_set_md_var(tc, "timeout", "2");
}
ATF_TC_BODY(06badalign, tc) {
	ssize_t ret;

	UNUSED(tc);
	setup_map(0, 0, 0, 0, mapped);
	test_send(reqlen + 1);
	test_recv(reqlen + 4);
	test_result(ERR_BADREQUEST);
}
ATF_TC_CLEANUP(06badalign, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(07version);
ATF_TC_HEAD(07version, tc) {
	atf_tc_set_md_var(tc, "descr", "bad version");
	atf_tc_set_md_var(tc, "timeout", "2");
}
ATF_TC_BODY(07version, tc) {
	ssize_t ret;
	uint8_t expected[4];

	UNUSED(tc);
	memset(request, 0, sizeof(request));
	request[OFF_VERSION] = 7;
	request[OFF_OPCODE] = OPCODE_MAP;
	memcpy(request + OFF_MAP_EXTERNAL_ADDR, mapped, 12);
	test_send(LEN_MAP);
	ret = recv(fd, response, 128, 0);
	ATF_REQUIRE_MSG(ret >= 4, "recv() failed");
	expected[OFF_VERSION] = 1;
	expected[OFF_OPCODE] = OPCODE_MAP_RESPONSE;
	expected[OFF_RESERVED] = 0;
	expected[OFF_RESULT_CODE] = ERR_UNSUPVERSION;
	ret = memcmp(response, expected, 4);
	ATF_REQUIRE_EQ_MSG(ret, 0, "not expected response");
}
ATF_TC_CLEANUP(07version, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(080shorthdr);
ATF_TC_HEAD(080shorthdr, tc) {
	atf_tc_set_md_var(tc, "descr", "underrun (shorter than header)");
	atf_tc_set_md_var(tc, "timeout", "2");
}
ATF_TC_BODY(080shorthdr, tc) {
	ssize_t ret;

	UNUSED(tc);
	setup_map(0, 0, 0, 0, mapped);
	test_send(20);
	memset(response, 0, sizeof(response));
	atf_tc_expect_timeout("too short requests are discarded");
	ret = recv(fd, response, 128, 0);
}
ATF_TC_CLEANUP(080shorthdr, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(081underrunmap);
ATF_TC_HEAD(081underrunmap, tc) {
	atf_tc_set_md_var(tc, "descr",
		"underrun MAP (longer than header but shorter than MAP)");
	atf_tc_set_md_var(tc, "timeout", "2");
}
ATF_TC_BODY(081underrunmap, tc) {
	ssize_t ret;

	UNUSED(tc);
	setup_map(100, IPPROTO_TCP, 12346, 0, mapped);
	test_send(LEN_MAP - 4);
	test_recv(LEN_MAP - 4);
	test_result(ERR_BADREQUEST);
}
ATF_TC_CLEANUP(081underrunmap, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(082underrunpeer);
ATF_TC_HEAD(082underrunpeer, tc) {
	atf_tc_set_md_var(tc, "descr",
		"underrun PEER (longer than header but shorter than PEER)");
	atf_tc_set_md_var(tc, "timeout", "2");
}
ATF_TC_BODY(082underrunpeer, tc) {
	ssize_t ret;

	UNUSED(tc);
	setup_peer(100, IPPROTO_UDP, randport(), 0, mapped, 56789, peeraddr);
	test_send(LEN_PEER - 4);
	test_recv(LEN_PEER - 4);
	test_result(ERR_BADREQUEST);
}
ATF_TC_CLEANUP(082underrunpeer, tc) {
	UNUSED(tc);
	cleanup();
}

#if TRANSPORT == AF_INET6
ATF_TC_WITH_CLEANUP(0Amappedaddress);
ATF_TC_HEAD(0Amappedaddress, tc) {
	atf_tc_set_md_var(tc, "descr", "IPv4 mapped source address");
	atf_tc_set_md_var(tc, "timeout", "2");
}
ATF_TC_BODY(0Amappedaddress, tc) {
	ssize_t ret;

	UNUSED(tc);
	setup_map(0, 0, 0, 0, mapped);
	memcpy(request + OFF_CLIENT_ADDR, mapped, 12);
	test_send(reqlen);
	test_recv(reqlen);
	test_result(ERR_ADDRMISMATCH);
}
ATF_TC_CLEANUP(0Amappedaddress, tc) {
	UNUSED(tc);
	cleanup();
}
#else
ATF_TC_WITH_CLEANUP(0Aipv6address);
ATF_TC_HEAD(0Aipv6address, tc) {
	atf_tc_set_md_var(tc, "descr", "IPv6 source address");
	atf_tc_set_md_var(tc, "timeout", "2");
}
ATF_TC_BODY(0Aipv6address, tc) {
	ssize_t ret;

	UNUSED(tc);
	setup_map(0, 0, 0, 0, mapped);
	/* 2000::ff:SOURCE */
	request[OFF_CLIENT_ADDR] = 0x20;
	test_send(reqlen);
	test_recv(reqlen);
	test_result(ERR_ADDRMISMATCH);
}
ATF_TC_CLEANUP(0Aipv6address, tc) {
	UNUSED(tc);
	cleanup();
}
#endif

ATF_TC_WITH_CLEANUP(0Bbadaddress);
ATF_TC_HEAD(0Bbadaddress, tc) {
	atf_tc_set_md_var(tc, "descr",
			  "bad source address (changed en route)");
	atf_tc_set_md_var(tc, "timeout", "2");
}
ATF_TC_BODY(0Bbadaddress, tc) {
	ssize_t ret;

	UNUSED(tc);
	setup_map(0, 0, 0, 0, mapped);
	request[OFF_CLIENT_ADDR + 14] ^= 0x12;
	test_send(reqlen);
	test_recv(reqlen);
	test_result(ERR_ADDRMISMATCH);
}
ATF_TC_CLEANUP(0Bbadaddress, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(0Cbadopcode);
ATF_TC_HEAD(0Cbadopcode, tc) {
	atf_tc_set_md_var(tc, "descr", "bad opcode");
	atf_tc_set_md_var(tc, "timeout", "2");
}
ATF_TC_BODY(0Cbadopcode, tc) {
	ssize_t ret;

	UNUSED(tc);
	setup_common(107, 0);
	memcpy(request + OFF_MAP_EXTERNAL_ADDR, mapped, 12);
	test_send(LEN_MAP);
	test_recv(LEN_MAP);
	test_result(ERR_UNSUPOPCODE);
}
ATF_TC_CLEANUP(0Cbadopcode, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(10badproto);
ATF_TC_HEAD(10badproto, tc) {
	atf_tc_set_md_var(tc, "descr", "bad protocol (map)");
	atf_tc_set_md_var(tc, "timeout", "2");
}
ATF_TC_BODY(10badproto, tc) {
	ssize_t ret;

	UNUSED(tc);
	setup_map(0, 107, 0, 0, mapped);
	test_send(reqlen);
	test_recv(reqlen);
	test_result(ERR_UNSUPPROTO);
}
ATF_TC_CLEANUP(10badproto, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(110badextaddr);
ATF_TC_HEAD(110badextaddr, tc) {
	atf_tc_set_md_var(tc, "descr",
			  "not IPv4 mapped external address (map)");
	atf_tc_set_md_var(tc, "timeout", "2");
}
ATF_TC_BODY(110badextaddr, tc) {
	ssize_t ret;

	UNUSED(tc);
	setup_map(100, IPPROTO_TCP, 12345, 0, srcaddr);
	request[OFF_MAP_EXTERNAL_ADDR] ^= 1;
#ifdef USE_THIRD_PARTY
	append_third_party(intaddr);
#endif
	test_send(reqlen);
	test_recv(reqlen);
	test_result(ERR_BADREQUEST);
}
ATF_TC_CLEANUP(110badextaddr, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(111badproto0);
ATF_TC_HEAD(111badproto0, tc) {
	atf_tc_set_md_var(tc, "descr",
			  "bad protocol zero (map with lifetime > 0)");
	atf_tc_set_md_var(tc, "timeout", "2");
}
ATF_TC_BODY(111badproto0, tc) {
	ssize_t ret;

	UNUSED(tc);
	setup_map(100, 0, 12345, 0, mapped);
	test_send(reqlen);
	test_recv(reqlen);
	test_result(ERR_UNSUPPROTO);
}
ATF_TC_CLEANUP(111badproto0, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(120optionoverflow);
ATF_TC_HEAD(120optionoverflow, tc) {
	atf_tc_set_md_var(tc, "descr", "option overflow (map)");
	atf_tc_set_md_var(tc, "timeout", "2");
}
ATF_TC_BODY(120optionoverflow, tc) {
	ssize_t ret;

	UNUSED(tc);
	setup_map(0, 0, 0, 0, mapped);
	append_third_party(mapped);
	test_send(reqlen - 12);
	test_recv(reqlen - 12);
	test_result(ERR_BADOPTION);
}
ATF_TC_CLEANUP(120optionoverflow, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(1300mult3rdparty);
ATF_TC_HEAD(1300mult3rdparty, tc) {
	atf_tc_set_md_var(tc, "descr", "multiple 3rd party options (map)");
	atf_tc_set_md_var(tc, "timeout", "2");
}
ATF_TC_BODY(1300mult3rdparty, tc) {
	ssize_t ret;

	UNUSED(tc);
	setup_map(0, 0, 0, 0, mapped);
	append_third_party(intaddr);
	append_third_party(intaddr);
	test_send(reqlen);
	test_recv(reqlen);
	test_result(ERR_BADOPTION);
}
ATF_TC_CLEANUP(1300mult3rdparty, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(1301mult3rdparty2);
ATF_TC_HEAD(1301mult3rdparty2, tc) {
	atf_tc_set_md_var(tc, "descr",
			  "multiple 3rd party options (variant) (map)");
	atf_tc_set_md_var(tc, "timeout", "2");
}
ATF_TC_BODY(1301mult3rdparty2, tc) {
	ssize_t ret;

	UNUSED(tc);
	setup_map(0, 0, 0, 0, mapped);
	append_third_party(intaddr);
	append_filter(0, 0, mapped);
	append_third_party(intaddr);
	test_send(reqlen);
	test_recv(reqlen);
	test_result(ERR_BADOPTION);
}
ATF_TC_CLEANUP(1301mult3rdparty2, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(1313rdparty0);
ATF_TC_HEAD(1313rdparty0, tc) {
	atf_tc_set_md_var(tc, "descr",
			  "bad empty 3rd party option (map and lifetime > 0)");
	atf_tc_set_md_var(tc, "timeout", "2");
}
ATF_TC_BODY(1313rdparty0, tc) {
	ssize_t ret;

	UNUSED(tc);
	setup_map(100, IPPROTO_TCP, 12345, 0, mapped);
	append_third_party(mapped);
	test_send(reqlen);
	test_recv(reqlen);
	test_result(ERR_BADREQUEST);
}
ATF_TC_CLEANUP(1313rdparty0, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(132bad3rdpartylen);
ATF_TC_HEAD(132bad3rdpartylen, tc) {
	atf_tc_set_md_var(tc, "descr", "bad 3rd party option length (map)");
	atf_tc_set_md_var(tc, "timeout", "2");
}
ATF_TC_BODY(132bad3rdpartylen, tc) {
	ssize_t ret;

	UNUSED(tc);
	setup_map(100, IPPROTO_TCP, 12345, 0, mapped);
	append_option(OPTION_THIRD_PARTY, 12);
	memcpy(request + LEN_MAP + 4, mapped, 12);
	test_send(reqlen);
	test_recv(reqlen);
	test_result(ERR_BADOPTION);
}
ATF_TC_CLEANUP(132bad3rdpartylen, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(133sameaddr3rdparty);
ATF_TC_HEAD(133sameaddr3rdparty, tc) {
	atf_tc_set_md_var(tc, "descr",
			  "same source and 3rd party option addresses (map)");
	atf_tc_set_md_var(tc, "timeout", "2");
}
ATF_TC_BODY(133sameaddr3rdparty, tc) {
	ssize_t ret;

	UNUSED(tc);
	setup_map(0, 0, 0, 0, mapped);
	append_third_party(srcaddr);
	test_send(reqlen);
	test_recv(reqlen);
	/* the error depends on the parsing order */
	ATF_REQUIRE_MSG(response[OFF_RESULT_CODE] != ERR_SUCCESS,
			"an error must be returned");
}
ATF_TC_CLEANUP(133sameaddr3rdparty, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(134v63rdparty);
ATF_TC_HEAD(134v63rdparty, tc) {
	atf_tc_set_md_var(tc, "descr",
			  "not IPv4 mapped address in 3rd party option (map)");
	atf_tc_set_md_var(tc, "timeout", "2");
}
ATF_TC_BODY(134v63rdparty, tc) {
	ssize_t ret;

	UNUSED(tc);
	setup_map(100, IPPROTO_TCP, 12345, 0, mapped);
	append_third_party(srcaddr);
	request[optoff + 11] ^= 0x12;
	test_send(reqlen);
	test_recv(reqlen);
	test_result(ERR_BADOPTION);
}
ATF_TC_CLEANUP(134v63rdparty, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(135badaddr3rdparty);
ATF_TC_HEAD(135badaddr3rdparty, tc) {
	atf_tc_set_md_var(tc, "descr",
			  "bad 3rd party option internal address (map)");
	atf_tc_set_md_var(tc, "timeout", "2");
}
ATF_TC_BODY(135badaddr3rdparty, tc) {
	ssize_t ret;

	UNUSED(tc);
	setup_map(100, IPPROTO_TCP, 12345, 0, mapped);
	append_third_party(peeraddr);
	test_send(reqlen);
	test_recv(reqlen);
	test_result(ERR_PROCESSING);
}
ATF_TC_CLEANUP(135badaddr3rdparty, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(140multprefail);
ATF_TC_HEAD(140multprefail, tc) {
	atf_tc_set_md_var(tc, "descr", "multiple prefer failure options");
	atf_tc_set_md_var(tc, "timeout", "2");
}
ATF_TC_BODY(140multprefail, tc) {
	ssize_t ret;

	UNUSED(tc);
	setup_map(100, IPPROTO_TCP, 12345, 0, mapped);
	append_prefer_failure();
	append_prefer_failure();
	test_send(reqlen);
	test_recv(reqlen);
	test_result(ERR_BADOPTION);
}
ATF_TC_CLEANUP(140multprefail, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(141badprefaillen);
ATF_TC_HEAD(141badprefaillen, tc) {
	atf_tc_set_md_var(tc, "descr", "bad prefer failure option length");
	atf_tc_set_md_var(tc, "timeout", "2");
}
ATF_TC_BODY(141badprefaillen, tc) {
	ssize_t ret;

	UNUSED(tc);
	setup_map(100, IPPROTO_TCP, 12345, 0, mapped);
	append_option(OPTION_PREF_FAIL, 1);
	test_send(reqlen);
	test_recv(reqlen);
	test_result(ERR_BADOPTION);
}
ATF_TC_CLEANUP(141badprefaillen, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(142prefaillife);
ATF_TC_HEAD(142prefaillife, tc) {
	atf_tc_set_md_var(tc, "descr",
			  "prefer failure option with lifetime = 0");
	atf_tc_set_md_var(tc, "timeout", "2");
}
ATF_TC_BODY(142prefaillife, tc) {
	ssize_t ret;

	UNUSED(tc);
	setup_map(0, IPPROTO_TCP, 12345, 0, mapped);
	append_prefer_failure();
	test_send(reqlen);
	test_recv(reqlen);
	test_result(ERR_BADOPTION);
}
ATF_TC_CLEANUP(142prefaillife, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(143prefailnone);
ATF_TC_HEAD(143prefailnone, tc) {
	atf_tc_set_md_var(tc, "descr",
			  "prefer failure option with external "
			  "address and port = 0");
	atf_tc_set_md_var(tc, "timeout", "2");
}
ATF_TC_BODY(143prefailnone, tc) {
	ssize_t ret;

	UNUSED(tc);
	setup_map(0, IPPROTO_TCP, 12345, 0, mapped);
#ifdef USE_THIRD_PARTY
	append_third_party(intaddr);
#endif
	append_prefer_failure();
	test_send(reqlen);
	test_recv(reqlen);
	test_result(ERR_BADOPTION);
}
ATF_TC_CLEANUP(143prefailnone, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(144prefailaddr);
ATF_TC_HEAD(144prefailaddr, tc) {
	atf_tc_set_md_var(tc, "descr",
			  "prefer failure option with address = 0");
	atf_tc_set_md_var(tc, "timeout", "2");
}
ATF_TC_BODY(144prefailaddr, tc) {
	ssize_t ret;

	UNUSED(tc);
	setup_map(100, IPPROTO_TCP, 10004, 10004, mapped);
#ifdef USE_THIRD_PARTY
	append_third_party(intaddr);
#endif
	append_prefer_failure();
	test_send(reqlen);
	test_recv(reqlen);
	test_result(ERR_CANTPROVIDE);
}
ATF_TC_CLEANUP(144prefailaddr, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(145prefailport);
ATF_TC_HEAD(145prefailport, tc) {
	atf_tc_set_md_var(tc, "descr", "prefer failure option with port = 0");
	atf_tc_set_md_var(tc, "timeout", "2");
}
ATF_TC_BODY(145prefailport, tc) {
	ssize_t ret;
	uint8_t extaddr[16];

	UNUSED(tc);
	setup_map(100, IPPROTO_TCP, 10001, 0, mapped);
#ifdef USE_THIRD_PARTY
	append_third_party(intaddr);
#endif
	test_send(reqlen);
	test_recv(reqlen);
	checkepoch();
	memcpy(extaddr, response + OFF_MAP_EXTERNAL_ADDR, 16);
	test_result(ERR_SUCCESS);

	setup_map(100, IPPROTO_UDP, 10005, 0, extaddr);
#ifdef USE_THIRD_PARTY
	append_third_party(intaddr);
#endif
	append_prefer_failure();
	test_send(reqlen);
	test_recv(reqlen);
	checkepoch();
	test_result(ERR_SUCCESS);
}
ATF_TC_CLEANUP(145prefailport, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(146prefailbadaddr);
ATF_TC_HEAD(146prefailbadaddr, tc) {
	atf_tc_set_md_var(tc, "descr",
			  "prefer failure option with bad external address");
	atf_tc_set_md_var(tc, "timeout", "2");
}
ATF_TC_BODY(146prefailbadaddr, tc) {
	ssize_t ret;
	uint8_t extaddr[16];

	UNUSED(tc);
	setup_map(100, IPPROTO_TCP, 10002, 0, mapped);
#ifdef USE_THIRD_PARTY
	append_third_party(intaddr);
#endif
	test_send(reqlen);
	test_recv(reqlen);
	checkepoch();
	memcpy(extaddr, response + OFF_MAP_EXTERNAL_ADDR, 16);
	test_result(ERR_SUCCESS);

	setup_map(100, IPPROTO_UDP, 10007, 0, extaddr);
	if (request[OFF_MAP_EXTERNAL_ADDR + 15] == 10)
		request[OFF_MAP_EXTERNAL_ADDR + 15] = 100;
	else
		request[OFF_MAP_EXTERNAL_ADDR + 15] = 10;
#ifdef USE_THIRD_PARTY
	append_third_party(intaddr);
#endif
	append_prefer_failure();
	test_send(reqlen);
	test_recv(reqlen);
	checkepoch();
	test_result(ERR_CANTPROVIDE);
}
ATF_TC_CLEANUP(146prefailbadaddr, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(150filter0);
ATF_TC_HEAD(150filter0, tc) {
	atf_tc_set_md_var(tc, "descr", "filter option with lifetime = 0");
	atf_tc_set_md_var(tc, "timeout", "2");
}
ATF_TC_BODY(150filter0, tc) {
	ssize_t ret;

	UNUSED(tc);
	setup_map(0, IPPROTO_TCP, 12345, 0, mapped);
	append_filter(0, 0, mapped);
	test_send(reqlen);
	test_recv(reqlen);
	test_result(ERR_BADOPTION);
}
ATF_TC_CLEANUP(150filter0, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(151badfilterlen);
ATF_TC_HEAD(151badfilterlen, tc) {
	atf_tc_set_md_var(tc, "descr", "bad filter option length");
	atf_tc_set_md_var(tc, "timeout", "2");
}
ATF_TC_BODY(151badfilterlen, tc) {
	ssize_t ret;

	UNUSED(tc);
	setup_map(100, IPPROTO_TCP, 12345, 0, mapped);
	append_option(OPTION_FILTER, 24);
	test_send(reqlen);
	test_recv(reqlen);
	test_result(ERR_BADOPTION);
}
ATF_TC_CLEANUP(151badfilterlen, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(152toomanyfilter);
ATF_TC_HEAD(152toomanyfilter, tc) {
	atf_tc_set_md_var(tc, "descr", "too many filter options");
	atf_tc_set_md_var(tc, "timeout", "30");
}
ATF_TC_BODY(152toomanyfilter, tc) {
	uint16_t n16, eport = 0;
	unsigned int i = 0;
	ssize_t ret;
	uint8_t extaddr[16];

	UNUSED(tc);
	delall();

    again:
	ATF_REQUIRE_MSG(i < 10000, "unbound number of filters");
	setup_map(100, IPPROTO_TCP, 1234, 0, mapped);
	if (eport != 0) {
		/* external port from previous messages */
		memcpy(request + OFF_MAP_EXTERNAL_PORT, &eport, 2);
		/* external address from previous messages */
		memcpy(request + OFF_MAP_EXTERNAL_ADDR, extaddr, 16);
	}
#ifdef USE_THIRD_PARTY
	append_third_party(intaddr);
#endif
	if (eport != 0) {
		append_prefer_failure();
	}
	while (reqlen + 24 <= 1024) {
		append_filter(128, 100 + i++, peeraddr);
	}
	test_send(reqlen);
	test_recv(reqlen);
	checkepoch();
	checkextaddr(0);
	if (response[OFF_RESULT_CODE] == ERR_SUCCESS) {
		if (eport == 0) {
			memcpy(&eport, response + OFF_MAP_EXTERNAL_PORT, 2);
			memcpy(extaddr, response + OFF_MAP_EXTERNAL_ADDR, 16);
			ATF_REQUIRE_MSG(eport != 0, "bad external port (0)");
		} else {
			memcpy(&n16, response + OFF_MAP_EXTERNAL_PORT, 2);
			ATF_REQUIRE_EQ_MSG(n16, eport,
					   "bad external port (<>)");
		}
		usleep(10000);
		goto again;
	}
	test_result(ERR_TOOMANYPEER);
	if (eport != 0) {
		memcpy(&n16, response + OFF_MAP_EXTERNAL_PORT, 2);
		ATF_REQUIRE_EQ_MSG(n16, eport, "bad external port (<>)");
	}
}
ATF_TC_CLEANUP(152toomanyfilter, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(153badaddrfilter);
ATF_TC_HEAD(153badaddrfilter, tc) {
	atf_tc_set_md_var(tc, "descr", "filter option with an IPv6 address");
	atf_tc_set_md_var(tc, "timeout", "2");
}
ATF_TC_BODY(153badaddrfilter, tc) {
	ssize_t ret;

	UNUSED(tc);
	setup_map(100, IPPROTO_TCP, 12345, 0, mapped);
	append_filter(128, 0, srcaddr);
	/* 2000::/80 */
	request[optoff + 4] = 0x20;
	request[optoff + 18] ^= 0x11;
	test_send(reqlen);
 	test_recv(reqlen);
	test_result(ERR_BADOPTION);
}
ATF_TC_CLEANUP(153badaddrfilter, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(154badfilterplen);
ATF_TC_HEAD(154badfilterplen, tc) {
	atf_tc_set_md_var(tc, "descr",
			  "filter option with a bad prefix length");
	atf_tc_set_md_var(tc, "timeout", "2");
}
ATF_TC_BODY(154badfilterplen, tc) {
	ssize_t ret;

	UNUSED(tc);
	setup_map(100, IPPROTO_TCP, 12345, 0, mapped);
	append_filter(129, 0, peeraddr);
	test_send(reqlen);
	test_recv(reqlen);
	test_result(ERR_BADOPTION);
}
ATF_TC_CLEANUP(154badfilterplen, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(155badfilterpref);
ATF_TC_HEAD(155badfilterpref, tc) {
	atf_tc_set_md_var(tc, "descr", "filter option with a bad prefix");
	atf_tc_set_md_var(tc, "timeout", "2");
}
ATF_TC_BODY(155badfilterpref, tc) {
	ssize_t ret;

	UNUSED(tc);
	setup_map(100, IPPROTO_TCP, 12345, 0, mapped);
	append_filter(106, 0, peeraddr);
	test_send(reqlen);
	test_recv(reqlen);
	test_result(ERR_BADOPTION);
}
ATF_TC_CLEANUP(155badfilterpref, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(156badfilterorder);
ATF_TC_HEAD(156badfilterorder, tc) {
	atf_tc_set_md_var(tc, "descr", "resetting filter option not at head");
	atf_tc_set_md_var(tc, "timeout", "2");
}
ATF_TC_BODY(156badfilterorder, tc) {
	ssize_t ret;

	UNUSED(tc);
	setup_map(100, IPPROTO_UDP, 12345, 0, mapped);
	append_filter(128, 0, peeraddr);
	append_filter(0, 0, mapped);
	test_send(reqlen);
	test_recv(reqlen);
	test_result(ERR_BADOPTION);
}
ATF_TC_CLEANUP(156badfilterorder, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(160badoption);
ATF_TC_HEAD(160badoption, tc) {
	atf_tc_set_md_var(tc, "descr", "unknown mandatory option (map)");
	atf_tc_set_md_var(tc, "timeout", "2");
}
ATF_TC_BODY(160badoption, tc) {
	ssize_t ret;

	UNUSED(tc);
	setup_map(0, 0, 0, 0, mapped);
	append_option(107, 1);
	test_send(reqlen);
	test_recv(reqlen);
	test_result(ERR_UNSUPOPTION);
}
ATF_TC_CLEANUP(160badoption, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(161ignoption);
ATF_TC_HEAD(161ignoption, tc) {
	atf_tc_set_md_var(tc, "descr", "unknown optional option (map)");
	atf_tc_set_md_var(tc, "timeout", "2");
}
ATF_TC_BODY(161ignoption, tc) {
	ssize_t ret;

	UNUSED(tc);
	delall();
	setup_map(0, 0, 0, 0, mapped);
	append_third_party(mapped);
	append_option(107 | OPTION_OPTIONAL, 1);
	test_send(reqlen);
	test_recv(reqlen - 8);
	checkepoch();
	test_result(ERR_SUCCESS);
}
ATF_TC_CLEANUP(161ignoption, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(162bad0option);
ATF_TC_HEAD(162bad0option, tc) {
	atf_tc_set_md_var(tc, "descr", "unprocessed option (map)");
	atf_tc_set_md_var(tc, "timeout", "2");
}
ATF_TC_BODY(162bad0option, tc) {
	ssize_t ret;

	UNUSED(tc);
	setup_map(0, 0, 0, 0, mapped);
	append_option(0, 1);
	request[LEN_MAP + 4] = 107;
	test_send(reqlen);
	test_recv(reqlen);
	test_result(ERR_UNSUPOPTION);
}
ATF_TC_CLEANUP(162bad0option, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(163options);
ATF_TC_HEAD(163options, tc) {
	atf_tc_set_md_var(tc, "descr", "unknown options (map)");
	atf_tc_set_md_var(tc, "timeout", "2");
}
ATF_TC_BODY(163options, tc) {
	ssize_t ret;

	UNUSED(tc);
	setup_map(0, 0, 0, 0, mapped);
	append_option(107, 1);
	append_option(107, 2);
	append_option(107, 3);
	append_option(207, 0);
	append_option(107, 4);
	append_option(208, 6);
	test_send(reqlen);
	test_recv(reqlen);
	test_result(ERR_UNSUPOPTION);
}
ATF_TC_CLEANUP(163options, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(17intaddr);
ATF_TC_HEAD(17intaddr, tc) {
	atf_tc_set_md_var(tc, "descr", "no internal address (map)");
	atf_tc_set_md_var(tc, "timeout", "2");
}
ATF_TC_BODY(17intaddr, tc) {
#ifdef USE_THIRD_PARTY
	ssize_t ret;

	UNUSED(tc);
	setup_map(0, 0, 0, 0, mapped);
	test_send(reqlen);
	test_recv(reqlen);
	test_result(ERR_BADREQUEST);
#else
	UNUSED(tc);
	atf_tc_skip("NAT444 does not require THIRD_PARTY");
#endif
}
ATF_TC_CLEANUP(17intaddr, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(180intaddrdel);
ATF_TC_HEAD(180intaddrdel, tc) {
	atf_tc_set_md_var(tc, "descr", "delete internal port but no address");
	atf_tc_set_md_var(tc, "timeout", "2");
}
ATF_TC_BODY(180intaddrdel, tc) {
#ifdef USE_THIRD_PARTY
	ssize_t ret;

	UNUSED(tc);
	setup_map(0, IPPROTO_TCP, 12345, 0, mapped);
	append_third_party(mapped);
	test_send(reqlen);
	test_recv(reqlen);
	test_result(ERR_BADREQUEST);
#else
	UNUSED(tc);
	atf_tc_skip("NAT444 does not require THIRD_PARTY");
#endif
}
ATF_TC_CLEANUP(180intaddrdel, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(181intaddradd);
ATF_TC_HEAD(181intaddradd, tc) {
	atf_tc_set_md_var(tc, "descr", "no internal address (map)");
	atf_tc_set_md_var(tc, "timeout", "2");
}
ATF_TC_BODY(181intaddradd, tc) {
	ssize_t ret;

	UNUSED(tc);
	setup_map(100, IPPROTO_TCP, 12345, 0, mapped);
	append_third_party(mapped);
	test_send(reqlen);
	test_recv(reqlen);
	test_result(ERR_BADREQUEST);
}
ATF_TC_CLEANUP(181intaddradd, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(19intport);
ATF_TC_HEAD(19intport, tc) {
	atf_tc_set_md_var(tc, "descr", "no internal port (map)");
	atf_tc_set_md_var(tc, "timeout", "2");
}
ATF_TC_BODY(19intport, tc) {
	ssize_t ret;

	UNUSED(tc);
	setup_map(100, IPPROTO_TCP, 0, 0, mapped);
#ifdef USE_THIRD_PARTY
	append_third_party(intaddr);
#endif
	test_send(reqlen);
	test_recv(reqlen);
	test_result(ERR_BADREQUEST);
}
ATF_TC_CLEANUP(19intport, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(200deleteall);
ATF_TC_HEAD(200deleteall, tc) {
	atf_tc_set_md_var(tc, "descr", "delete all subscriber mappings");
	atf_tc_set_md_var(tc, "timeout", "2");
}
ATF_TC_BODY(200deleteall, tc) {
	UNUSED(tc);
	delall();
}
ATF_TC_CLEANUP(200deleteall, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(201deleteallbadproto);
ATF_TC_HEAD(201deleteallbadproto, tc) {
	atf_tc_set_md_var(tc, "descr",
			  "delete all subscriber mappings with proto != 0");
	atf_tc_set_md_var(tc, "timeout", "2");
}
ATF_TC_BODY(201deleteallbadproto, tc) {
	ssize_t ret;

	UNUSED(tc);
	setup_map(0, IPPROTO_TCP, 0, 0, mapped);
	append_third_party(mapped);
	test_send(reqlen);
	test_recv(reqlen);
	checkepoch();
	test_result(ERR_BADREQUEST);
}
ATF_TC_CLEANUP(201deleteallbadproto, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(202deleteallbadport);
ATF_TC_HEAD(202deleteallbadport, tc) {
	atf_tc_set_md_var(tc, "descr",
			  "delete all subscriber mappings with port != 0");
	atf_tc_set_md_var(tc, "timeout", "2");
}
ATF_TC_BODY(202deleteallbadport, tc) {
	ssize_t ret;

	UNUSED(tc);
	setup_map(0, IPPROTO_TCP, 12345, 0, mapped);
	/* all-zeroes third party option */
	append_third_party(mapped);
	test_send(reqlen);
	test_recv(reqlen);
	checkepoch();
	test_result(ERR_BADREQUEST);
}
ATF_TC_CLEANUP(202deleteallbadport, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(21deleteclient);
ATF_TC_HEAD(21deleteclient, tc) {
	atf_tc_set_md_var(tc, "descr", "delete all client mappings");
	atf_tc_set_md_var(tc, "timeout", "2");
}
ATF_TC_BODY(21deleteclient, tc) {
	UNUSED(tc);
	delclient();
}
ATF_TC_CLEANUP(21deleteclient, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(220deleteproto);
ATF_TC_HEAD(220deleteproto, tc) {
	atf_tc_set_md_var(tc, "descr", "delete all protocol mappings");
	atf_tc_set_md_var(tc, "timeout", "2");
}
ATF_TC_BODY(220deleteproto, tc) {
	ssize_t ret;

	UNUSED(tc);
	setup_map(0, IPPROTO_TCP, 0, 0, mapped);
#ifdef USE_THIRD_PARTY
	append_third_party(intaddr);
#endif
	test_send(reqlen);
	test_recv(reqlen);
	checkepoch();
	test_result(ERR_SUCCESS);
}
ATF_TC_CLEANUP(220deleteproto, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(221deleteport);
ATF_TC_HEAD(221deleteport, tc) {
	atf_tc_set_md_var(tc, "descr", "delete all port mappings");
	atf_tc_set_md_var(tc, "timeout", "2");
}
ATF_TC_BODY(221deleteport, tc) {
	ssize_t ret;

	UNUSED(tc);
	setup_map(0, 0, 12345, 0, mapped);
#ifdef USE_THIRD_PARTY
	append_third_party(intaddr);
#endif
	test_send(reqlen);
	test_recv(reqlen);
	checkepoch();
	test_result(ERR_BADREQUEST);
}
ATF_TC_CLEANUP(221deleteport, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(23deleterandom);
ATF_TC_HEAD(23deleterandom, tc) {
	atf_tc_set_md_var(tc, "descr", "delete a random mapping");
	atf_tc_set_md_var(tc, "timeout", "2");
}
ATF_TC_BODY(23deleterandom, tc) {
	ssize_t ret;

	UNUSED(tc);
	setup_map(0, IPPROTO_UDP, 12345, 0, mapped);
#ifdef USE_THIRD_PARTY
	append_third_party(intaddr);
#endif
	test_send(reqlen);
	test_recv(reqlen);
	checkepoch();
	test_result(ERR_SUCCESS);
}
ATF_TC_CLEANUP(23deleterandom, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(24addmap);
ATF_TC_HEAD(24addmap, tc) {
	atf_tc_set_md_var(tc, "descr", "add an explicit dynamic mapping");
	atf_tc_set_md_var(tc, "timeout", "2");
}
ATF_TC_BODY(24addmap, tc) {
	ssize_t ret;

	UNUSED(tc);
	delall();
	setup_map(100, IPPROTO_TCP, 12346, 0, mapped);
#ifdef USE_THIRD_PARTY
	append_third_party(intaddr);
#endif
	test_send(reqlen);
	test_recv(reqlen);
	checkepoch();
	test_result(ERR_SUCCESS);
}
ATF_TC_CLEANUP(24addmap, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(25badaddmappref);
ATF_TC_HEAD(25badaddmappref, tc) {
	atf_tc_set_md_var(tc, "descr", "prefer failure and failure");
	atf_tc_set_md_var(tc, "timeout", "2");
}
ATF_TC_BODY(25badaddmappref, tc) {
	ssize_t ret;

	UNUSED(tc);
	setup_map(100, IPPROTO_UDP, 12345, 10, mapped);
#ifdef USE_THIRD_PARTY
	append_third_party(intaddr);
#endif
	append_prefer_failure();
	test_send(reqlen);
	test_recv(reqlen);
	checkepoch();
	test_result(ERR_CANTPROVIDE);
}
ATF_TC_CLEANUP(25badaddmappref, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(26addmappref);
ATF_TC_HEAD(26addmappref, tc) {
	atf_tc_set_md_var(tc, "descr", "prefer failure and success");
	atf_tc_set_md_var(tc, "timeout", "2");
}
ATF_TC_BODY(26addmappref, tc) {
	ssize_t ret;

	UNUSED(tc);
	delall();
	setup_map(100, IPPROTO_TCP, 3333, 64010, mapped);
#ifdef USE_THIRD_PARTY
	append_third_party(intaddr);
#endif
	append_prefer_failure();
	test_send(reqlen);
	test_recv(reqlen);
	checkepoch();
	test_result(ERR_SUCCESS);
}
ATF_TC_CLEANUP(26addmappref, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(27addmapstatic);
ATF_TC_HEAD(27addmapstatic, tc) {
	atf_tc_set_md_var(tc, "descr",
			  "explicit dynamic/static mapping collision");
	atf_tc_set_md_var(tc, "timeout", "2");
}
ATF_TC_BODY(27addmapstatic, tc) {
	uint32_t n32;
	ssize_t ret;

	UNUSED(tc);
	delall();
	setup_map(100, IPPROTO_UDP, STATICPORT, 0, mapped);
#ifdef USE_THIRD_PARTY
	append_third_party(intaddr);
#endif
	test_send(reqlen);
	test_recv(reqlen);
	checkepoch();
	memcpy(&n32, response + OFF_LIFETIME, 4);
	ATF_CHECK_EQ_MSG(ntohl(n32), UINT32_MAX, "finite lifetime");
	test_result(ERR_SUCCESS);
}
ATF_TC_CLEANUP(27addmapstatic, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(280addfilter);
ATF_TC_HEAD(280addfilter, tc) {
	atf_tc_set_md_var(tc, "descr", "with add filter");
	atf_tc_set_md_var(tc, "timeout", "2");
}
ATF_TC_BODY(280addfilter, tc) {
	ssize_t ret;

	UNUSED(tc);
	delall();
	setup_map(100, IPPROTO_TCP, 12347, 0, mapped);
#ifdef USE_THIRD_PARTY
	append_third_party(intaddr);
#endif
	append_filter(128, 56789, peeraddr);
	test_send(reqlen);
	test_recv(reqlen);
	checkepoch();
	test_result(ERR_SUCCESS);
}
ATF_TC_CLEANUP(280addfilter, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(281delfilter);
ATF_TC_HEAD(281delfilter, tc) {
	atf_tc_set_md_var(tc, "descr", "with del filter");
	atf_tc_set_md_var(tc, "timeout", "2");
}
ATF_TC_BODY(281delfilter, tc) {
	ssize_t ret;

	UNUSED(tc);
	delall();
	setup_map(100, IPPROTO_UDP, 12348, 0, mapped);
#ifdef USE_THIRD_PARTY
	append_third_party(intaddr);
#endif
	append_filter(0, 0, mapped);
	test_send(reqlen);
	test_recv(reqlen);
	checkepoch();
	test_result(ERR_SUCCESS);
}
ATF_TC_CLEANUP(281delfilter, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(282replayfilter);
ATF_TC_HEAD(282replayfilter, tc) {
	atf_tc_set_md_var(tc, "descr", "with replayed add filter");
	atf_tc_set_md_var(tc, "timeout", "5");
}
ATF_TC_BODY(282replayfilter, tc) {
	uint16_t eport;
	ssize_t ret;
	uint8_t extaddr[16];

	UNUSED(tc);
	delall();
	setup_map(100, IPPROTO_TCP, 12349, 0, mapped);
#ifdef USE_THIRD_PARTY
	append_third_party(intaddr);
#endif
	append_filter(128, 56789, peeraddr);
	test_send(reqlen);
	test_recv(reqlen);
	memcpy(&eport, response + OFF_MAP_EXTERNAL_PORT, 2);
	eport = ntohs(eport);
	memcpy(extaddr, response + OFF_MAP_EXTERNAL_ADDR, 16);
	checkepoch();
	checkextaddr(0);
	test_result(ERR_SUCCESS);

	setup_map(100, IPPROTO_TCP, 12349, eport, extaddr);
	append_prefer_failure();
#ifdef USE_THIRD_PARTY
	append_third_party(intaddr);
#endif
	append_filter(128, 56789, peeraddr);
	test_send(reqlen);
	test_recv(reqlen);
	checkepoch();
	checkextaddr(0);
	test_result(ERR_SUCCESS);
}
ATF_TC_CLEANUP(282replayfilter, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(283replacefilter);
ATF_TC_HEAD(283replacefilter, tc) {
	atf_tc_set_md_var(tc, "descr", "with delete first filter");
	atf_tc_set_md_var(tc, "timeout", "2");
}
ATF_TC_BODY(283replacefilter, tc) {
	ssize_t ret;

	UNUSED(tc);
	delall();
	setup_map(100, IPPROTO_UDP, 12340, 0, mapped);
#ifdef USE_THIRD_PARTY
	append_third_party(intaddr);
#endif
	append_filter(0, 0, mapped);
	append_filter(128, 56789, peeraddr);
	test_send(reqlen);
	test_recv(reqlen);
	checkepoch();
	test_result(ERR_SUCCESS);
}
ATF_TC_CLEANUP(283replacefilter, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(284replaceport);
ATF_TC_HEAD(284replaceport, tc) {
	atf_tc_set_md_var(tc, "descr", "with replaced port in filters");
	atf_tc_set_md_var(tc, "timeout", "5");
}
ATF_TC_BODY(284replaceport, tc) {
	uint16_t eport;
	ssize_t ret;
	uint8_t extaddr[16];

	UNUSED(tc);
	delall();
	setup_map(100, IPPROTO_TCP, 12341, 0, mapped);
#ifdef USE_THIRD_PARTY
	append_third_party(intaddr);
#endif
	append_filter(128, 56789, peeraddr);
	test_send(reqlen);
	test_recv(reqlen);
	memcpy(&eport, response + OFF_MAP_EXTERNAL_PORT, 2);
	eport = ntohs(eport);
	memcpy(extaddr, response + OFF_MAP_EXTERNAL_ADDR, 16);
	checkepoch();
	checkextaddr(0);
	test_result(ERR_SUCCESS);

	setup_map(100, IPPROTO_TCP, 12341, eport, extaddr);
#ifdef USE_THIRD_PARTY
	append_third_party(intaddr);
#endif
	append_prefer_failure();
	append_filter(0, 0, mapped);
	append_filter(128, 2111, peeraddr);
	test_send(reqlen);
	test_recv(reqlen);
	checkepoch();
	checkextaddr(0);
	test_result(ERR_SUCCESS);
}
ATF_TC_CLEANUP(284replaceport, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(285replacepref);
ATF_TC_HEAD(285replacepref, tc) {
	atf_tc_set_md_var(tc, "descr", "with replaced prefix in filters");
	atf_tc_set_md_var(tc, "timeout", "5");
}
ATF_TC_BODY(285replacepref, tc) {
	uint16_t eport;
	ssize_t ret;
	uint8_t extaddr[16];

	UNUSED(tc);
	delall();
	setup_map(100, IPPROTO_UDP, 12342, 0, mapped);
#ifdef USE_THIRD_PARTY
	append_third_party(intaddr);
#endif
	append_filter(128, 56789, peeraddr);
	test_send(reqlen);
	test_recv(reqlen);
	memcpy(&eport, response + OFF_MAP_EXTERNAL_PORT, 2);
	eport = ntohs(eport);
	memcpy(extaddr, response + OFF_MAP_EXTERNAL_ADDR, 16);
	checkepoch();
	checkextaddr(0);
	test_result(ERR_SUCCESS);

	setup_map(100, IPPROTO_UDP, 12342, eport, extaddr);
#ifdef USE_THIRD_PARTY
	append_third_party(intaddr);
#endif
	append_prefer_failure();
	append_filter(0, 0, mapped);
	append_filter(128, 56789, peeraddr);
	request[LEN_MAP + 70] ^= 0x12;
	test_send(reqlen);
	test_recv(reqlen);
	checkepoch();
	checkextaddr(0);
	test_result(ERR_SUCCESS);
}
ATF_TC_CLEANUP(285replacepref, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(286replaceplen);
ATF_TC_HEAD(286replaceplen, tc) {
	atf_tc_set_md_var(tc, "descr",
			  "with replaced prefix length in filters");
	atf_tc_set_md_var(tc, "timeout", "5");
}
ATF_TC_BODY(286replaceplen, tc) {
	uint16_t eport;
	ssize_t ret;
	uint8_t extaddr[16];

	UNUSED(tc);
	delall();
	setup_map(100, IPPROTO_TCP, 12343, 0, mapped);
#ifdef USE_THIRD_PARTY
	append_third_party(intaddr);
#endif
	append_filter(128, 56789, peeraddr);
	test_send(reqlen);
	test_recv(reqlen);
	memcpy(&eport, response + OFF_MAP_EXTERNAL_PORT, 2);
	memcpy(extaddr, response + OFF_MAP_EXTERNAL_ADDR, 16);
	checkepoch();
	checkextaddr(0);
	test_result(ERR_SUCCESS);

	setup_map(100, IPPROTO_TCP, 12343, ntohs(eport), extaddr);
#ifdef USE_THIRD_PARTY
	append_third_party(intaddr);
#endif
	append_prefer_failure();
	append_filter(0, 0, mapped);
	append_filter(120, 56789, peeraddr);
	request[optoff + 19] = 0;
	test_send(reqlen);
	test_recv(reqlen);
	checkepoch();
	checkextaddr(0);
	test_result(ERR_SUCCESS);
}
ATF_TC_CLEANUP(286replaceplen, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(287deladdfilter);
ATF_TC_HEAD(287deladdfilter, tc) {
	atf_tc_set_md_var(tc, "descr", "with add, del, add filters");
	atf_tc_set_md_var(tc, "timeout", "5");
}
ATF_TC_BODY(287deladdfilter, tc) {
	uint16_t eport;
	ssize_t ret;
	uint8_t extaddr[16];

	UNUSED(tc);
	delall();

	setup_map(100, IPPROTO_UDP, 12344, 0, mapped);
#ifdef USE_THIRD_PARTY
	append_third_party(intaddr);
#endif
	append_filter(128, 56789, peeraddr);
	test_send(reqlen);
	test_recv(reqlen);
	memcpy(&eport, response + OFF_MAP_EXTERNAL_PORT, 2);
	eport = ntohs(eport);
	memcpy(extaddr, response + OFF_MAP_EXTERNAL_ADDR, 16);
	checkepoch();
	checkextaddr(0);
	test_result(ERR_SUCCESS);

	setup_map(100, IPPROTO_UDP, 12344, eport, extaddr);
	append_prefer_failure();
#ifdef USE_THIRD_PARTY
	append_third_party(intaddr);
#endif
	append_filter(0, 0, mapped);
	test_send(reqlen);
	test_recv(reqlen);
	checkepoch();
	checkextaddr(0);
	test_result(ERR_SUCCESS);

	setup_map(100, IPPROTO_UDP, 12344, eport, extaddr);
	append_prefer_failure();
#ifdef USE_THIRD_PARTY
	append_third_party(intaddr);
#endif
	append_filter(128, 2211, peeraddr);
	test_send(reqlen);
	test_recv(reqlen);
	checkepoch();
	test_result(ERR_SUCCESS);
}
ATF_TC_CLEANUP(287deladdfilter, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(288badaddfilter);
ATF_TC_HEAD(288badaddfilter, tc) {
	atf_tc_set_md_var(tc, "descr", "with add too many filters");
	atf_tc_set_md_var(tc, "timeout", "100");
}
ATF_TC_BODY(288badaddfilter, tc) {
	uint16_t n16, eport;
	int i;
	ssize_t ret;
	uint8_t extaddr[16];

	UNUSED(tc);
	delall();
	i = 1;

	setup_map(100, IPPROTO_TCP, 11234, 0, mapped);
#ifdef USE_THIRD_PARTY
	append_third_party(intaddr);
#endif
	append_filter(128, 6789, peeraddr);
	test_send(reqlen);
	test_recv(reqlen);
	memcpy(&eport, response + OFF_MAP_EXTERNAL_PORT, 2);
	memcpy(extaddr, response + OFF_MAP_EXTERNAL_ADDR, 16);
	checkepoch();
	checkextaddr(0);
	test_result(ERR_SUCCESS);

 again:
	ATF_REQUIRE_MSG(i < 10000, "unbound number of filters");
	setup_map(100, IPPROTO_TCP, 11234, ntohs(eport), extaddr);
	append_prefer_failure();
#ifdef USE_THIRD_PARTY
	append_third_party(intaddr);
#endif
	append_filter(128, 6789 + i++, peeraddr);
	test_send(reqlen);
	test_recv(reqlen);
	checkepoch();
	checkextaddr(0);
	memcpy(&n16, response + OFF_MAP_EXTERNAL_PORT, 2);
	ATF_REQUIRE_EQ_MSG(n16, eport, "bad external port");
	if (response[OFF_RESULT_CODE] == ERR_SUCCESS) {
		usleep(10000);
		goto again;
	}
	test_result(ERR_TOOMANYPEER);
}
ATF_TC_CLEANUP(288badaddfilter, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(290renewmap);
ATF_TC_HEAD(290renewmap, tc) {
	atf_tc_set_md_var(tc, "descr", "standard renewal (map)");
	atf_tc_set_md_var(tc, "timeout", "5");
}
ATF_TC_BODY(290renewmap, tc) {
	uint16_t eport;
	ssize_t ret;
	uint8_t extaddr[16];

	UNUSED(tc);
	delall();
	setup_map(100, IPPROTO_UDP, 11111, 0, mapped);
#ifdef USE_THIRD_PARTY
	append_third_party(intaddr);
#endif
	test_send(reqlen);
	test_recv(reqlen);
	memcpy(&eport, response + OFF_MAP_EXTERNAL_PORT, 2);
	eport = ntohs(eport);
	memcpy(extaddr, response + OFF_MAP_EXTERNAL_ADDR, 16);
	checkepoch();
	checkextaddr(0);
	test_result(ERR_SUCCESS);

	setup_map(200, IPPROTO_UDP, 11111, eport, extaddr);
	append_prefer_failure();
#ifdef USE_THIRD_PARTY
	append_third_party(intaddr);
#endif
	test_send(reqlen);
	test_recv(reqlen);
	checkepoch();
	checkextaddr(0);
	test_result(ERR_SUCCESS);
}
ATF_TC_CLEANUP(290renewmap, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(291deladdmap);
ATF_TC_HEAD(291deladdmap, tc) {
	atf_tc_set_md_var(tc, "descr", "add, del, add map");
	atf_tc_set_md_var(tc, "timeout", "5");
}
ATF_TC_BODY(291deladdmap, tc) {
	uint16_t eport;
	ssize_t ret;
	uint8_t extaddr[16];

	UNUSED(tc);
	delall();

	setup_map(100, IPPROTO_TCP, 12222, 0, mapped);
#ifdef USE_THIRD_PARTY
	append_third_party(intaddr);
#endif
	test_send(reqlen);
	test_recv(reqlen);
	memcpy(&eport, response + OFF_MAP_EXTERNAL_PORT, 2);
	eport = ntohs(eport);
	memcpy(extaddr, response + OFF_MAP_EXTERNAL_ADDR, 16);
	checkepoch();
	checkextaddr(0);
	test_result(ERR_SUCCESS);

	setup_map(0, IPPROTO_TCP, 12222, 0, mapped);
#ifdef USE_THIRD_PARTY
	append_third_party(intaddr);
#endif
	test_send(reqlen);
	test_recv(reqlen);
	checkepoch();
	checkextaddr(0);
	test_result(ERR_SUCCESS);

	setup_map(100, IPPROTO_TCP, 12222, eport, extaddr);
	append_prefer_failure();
#ifdef USE_THIRD_PARTY
	append_third_party(intaddr);
#endif
	test_send(reqlen);
	test_recv(reqlen);
	checkepoch();
	checkextaddr(0);
	test_result(ERR_SUCCESS);
}
ATF_TC_CLEANUP(291deladdmap, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(292badrenewmap);
ATF_TC_HEAD(292badrenewmap, tc) {
	atf_tc_set_md_var(tc, "descr", "bad renewal (map)");
	atf_tc_set_md_var(tc, "timeout", "5");
}
ATF_TC_BODY(292badrenewmap, tc) {
	uint16_t eport;
	ssize_t ret;
	uint8_t extaddr[16];

	UNUSED(tc);
	delall();

	setup_map(100, IPPROTO_UDP, 11911, 0, mapped);
#ifdef USE_THIRD_PARTY
	append_third_party(intaddr);
#endif
	test_send(reqlen);
	test_recv(reqlen);
	memcpy(&eport, response + OFF_MAP_EXTERNAL_PORT, 2);
	eport = ntohs(eport);
	memcpy(extaddr, response + OFF_MAP_EXTERNAL_ADDR, 16);
	checkepoch();
	checkextaddr(0);
	test_result(ERR_SUCCESS);

	setup_map(200, IPPROTO_UDP, 11911, eport ^ 0x101, extaddr);
	append_prefer_failure();
#ifdef USE_THIRD_PARTY
	append_third_party(intaddr);
#endif
	test_send(reqlen);
	test_recv(reqlen);
	checkepoch();
	checkextaddr(0);
	test_result(ERR_CANTPROVIDE);
}
ATF_TC_CLEANUP(292badrenewmap, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(293replaymap);
ATF_TC_HEAD(293replaymap, tc) {
	atf_tc_set_md_var(tc, "descr", "replay (map)");
	atf_tc_set_md_var(tc, "timeout", "5");
}
ATF_TC_BODY(293replaymap, tc) {
	uint16_t eport;
	ssize_t ret;
	uint8_t extaddr[16];

	UNUSED(tc);
	delall();

	setup_map(100, IPPROTO_TCP, 12321, 0, mapped);
#ifdef USE_THIRD_PARTY
	append_third_party(intaddr);
#endif
	test_send(reqlen);
	test_recv(reqlen);
	memcpy(&eport, response + OFF_MAP_EXTERNAL_PORT, 2);
	eport = ntohs(eport);
	memcpy(extaddr, response + OFF_MAP_EXTERNAL_ADDR, 16);
	checkepoch();
	checkextaddr(0);
	test_result(ERR_SUCCESS);

	test_send(reqlen);
	test_recv(reqlen);
	checkepoch();
	checkextaddr(0);
	test_result(ERR_SUCCESS);
}
ATF_TC_CLEANUP(293replaymap, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(300badproto0);
ATF_TC_HEAD(300badproto0, tc) {
	atf_tc_set_md_var(tc, "descr", "protocol == 0 (peer)");
	atf_tc_set_md_var(tc, "timeout", "2");
}
ATF_TC_BODY(300badproto0, tc) {
	uint16_t n16;
	ssize_t ret;

	UNUSED(tc);
	n16 = randport();
	setup_peer(0, 0, n16, 0, mapped, n16, peeraddr);
	request[OFF_PEER_REMOTE_ADDR + 15] ^= 1;
	test_send(reqlen);
	test_recv(reqlen);
	test_result(ERR_UNSUPPROTO);
}
ATF_TC_CLEANUP(300badproto0, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(301badproto);
ATF_TC_HEAD(301badproto, tc) {
	atf_tc_set_md_var(tc, "descr", "bad protocol (peer)");
	atf_tc_set_md_var(tc, "timeout", "2");
}
ATF_TC_BODY(301badproto, tc) {
	uint16_t n16;
	ssize_t ret;

	UNUSED(tc);
	n16 = randport();
	setup_peer(0, 107, n16, 0, mapped, n16, peeraddr);
	request[OFF_PEER_REMOTE_ADDR + 15] ^= 1;
	test_send(reqlen);
	test_recv(reqlen);
	test_result(ERR_UNSUPPROTO);
}
ATF_TC_CLEANUP(301badproto, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(310badextaddr);
ATF_TC_HEAD(310badextaddr, tc) {
	atf_tc_set_md_var(tc, "descr",
			  "not IPv4 mapped external address (peer)");
	atf_tc_set_md_var(tc, "timeout", "2");
}
ATF_TC_BODY(310badextaddr, tc) {
	uint16_t n16;
	ssize_t ret;

	UNUSED(tc);
	n16 = randport();
	setup_peer(100, IPPROTO_TCP, n16, 0, srcaddr, n16, peeraddr);
	request[OFF_PEER_EXTERNAL_ADDR] ^= 1;
#ifdef USE_THIRD_PARTY
	append_third_party(intaddr);
#endif
	test_send(reqlen);
	test_recv(reqlen);
	test_result(ERR_BADREQUEST);
}
ATF_TC_CLEANUP(310badextaddr, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(311badproto0);
ATF_TC_HEAD(311badproto0, tc) {
	atf_tc_set_md_var(tc, "descr", "bad protocol zero (peer)");
	atf_tc_set_md_var(tc, "timeout", "2");
}
ATF_TC_BODY(311badproto0, tc) {
	uint16_t n16;
	ssize_t ret;

	UNUSED(tc);
	n16 = randport();
	setup_peer(100, 0, n16, 0, mapped, n16, peeraddr);
	request[OFF_PEER_REMOTE_ADDR + 15] ^= 1;
	test_send(reqlen);
	test_recv(reqlen);
	test_result(ERR_UNSUPPROTO);
}
ATF_TC_CLEANUP(311badproto0, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(320optionoverflow);
ATF_TC_HEAD(320optionoverflow, tc) {
	atf_tc_set_md_var(tc, "descr", "option overflow (peer)");
	atf_tc_set_md_var(tc, "timeout", "2");
}
ATF_TC_BODY(320optionoverflow, tc) {
	uint16_t n16;
	ssize_t ret;

	UNUSED(tc);
	n16 = randport();
	setup_peer(100, IPPROTO_TCP, n16, 0, mapped, n16, peeraddr);
	request[OFF_PEER_REMOTE_ADDR + 15] ^= 1;
	append_third_party(mapped);
	test_send(reqlen - 12);
 	test_recv(reqlen - 12);
	test_result(ERR_BADOPTION);
}
ATF_TC_CLEANUP(320optionoverflow, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(330mult3rdparty);
ATF_TC_HEAD(330mult3rdparty, tc) {
	atf_tc_set_md_var(tc, "descr", "multiple 3rd party options (peer)");
	atf_tc_set_md_var(tc, "timeout", "2");
}
ATF_TC_BODY(330mult3rdparty, tc) {
	uint16_t n16;
	ssize_t ret;

	UNUSED(tc);
	n16 = randport();
	setup_peer(100, IPPROTO_TCP, n16, 0, mapped, n16, peeraddr);
	request[OFF_PEER_REMOTE_ADDR + 15] ^= 1;
	append_third_party(intaddr);
	append_third_party(intaddr);
	test_send(reqlen);
	test_recv(reqlen);
	test_result(ERR_BADOPTION);
}
ATF_TC_CLEANUP(330mult3rdparty, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(3313rdparty0);
ATF_TC_HEAD(3313rdparty0, tc) {
	atf_tc_set_md_var(tc, "descr", "bad empty 3rd party option (peer)");
	atf_tc_set_md_var(tc, "timeout", "2");
}
ATF_TC_BODY(3313rdparty0, tc) {
	uint16_t n16;
	ssize_t ret;

	UNUSED(tc);
	n16 = randport();
	setup_peer(100, IPPROTO_TCP, n16, 0, mapped, n16, peeraddr);
	request[OFF_PEER_REMOTE_ADDR + 15] ^= 1;
	append_third_party(mapped);
	test_send(reqlen);
	test_recv(reqlen);
	test_result(ERR_BADREQUEST);
}
ATF_TC_CLEANUP(3313rdparty0, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(332bad3rdpartylen);
ATF_TC_HEAD(332bad3rdpartylen, tc) {
	atf_tc_set_md_var(tc, "descr", "bad 3rd party option length (peer)");
	atf_tc_set_md_var(tc, "timeout", "2");
}
ATF_TC_BODY(332bad3rdpartylen, tc) {
	uint16_t n16;
	ssize_t ret;

	UNUSED(tc);
	n16 = randport();
	setup_peer(100, IPPROTO_TCP, n16, 0, mapped, n16, peeraddr);
	request[OFF_PEER_REMOTE_ADDR + 15] ^= 1;
	append_option(OPTION_THIRD_PARTY, 12);
	memcpy(request + LEN_PEER + 4, mapped, 12);
	test_send(reqlen);
	test_recv(reqlen);
	test_result(ERR_BADOPTION);
}
ATF_TC_CLEANUP(332bad3rdpartylen, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(333sameaddr3rdparty);
ATF_TC_HEAD(333sameaddr3rdparty, tc) {
	atf_tc_set_md_var(tc, "descr",
			  "same source and 3rd party option addresses (peer)");
	atf_tc_set_md_var(tc, "timeout", "2");
}
ATF_TC_BODY(333sameaddr3rdparty, tc) {
	uint16_t n16;
	ssize_t ret;

	UNUSED(tc);
	n16 = randport();
	setup_peer(100, IPPROTO_TCP, n16, 0, mapped, n16, peeraddr);
	request[OFF_PEER_REMOTE_ADDR + 15] ^= 1;
	append_third_party(srcaddr);
	test_send(reqlen);
 	test_recv(reqlen);
	/* the error depends on the parsing order */
	ATF_REQUIRE_MSG(response[OFF_RESULT_CODE] != ERR_SUCCESS,
			"an error must be returned");
}
ATF_TC_CLEANUP(333sameaddr3rdparty, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(334v63rdparty);
ATF_TC_HEAD(334v63rdparty, tc) {
	atf_tc_set_md_var(tc, "descr",
			  "not IPv4 mapped address in 3rd party option (peer)");
	atf_tc_set_md_var(tc, "timeout", "2");
}
ATF_TC_BODY(334v63rdparty, tc) {
	uint16_t n16;
	ssize_t ret;

	UNUSED(tc);
	n16 = randport();
	setup_peer(100, IPPROTO_TCP, n16, 0, mapped, n16, peeraddr);
	append_third_party(srcaddr);
	request[optoff + 11] ^= 0x12;
	test_send(reqlen);
 	test_recv(reqlen);
	test_result(ERR_BADOPTION);
}
ATF_TC_CLEANUP(334v63rdparty, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(335badaddr3rdparty);
ATF_TC_HEAD(335badaddr3rdparty, tc) {
	atf_tc_set_md_var(tc, "descr",
			  "bad 3rd party option internal address (peer)");
	atf_tc_set_md_var(tc, "timeout", "2");
}
ATF_TC_BODY(335badaddr3rdparty, tc) {
	uint16_t n16;
	ssize_t ret;

	UNUSED(tc);
	n16 = randport();
	setup_peer(100, IPPROTO_TCP, n16, 0, mapped, n16, peeraddr);
	append_third_party(peeraddr);
	test_send(reqlen);
 	test_recv(reqlen);
	test_result(ERR_PROCESSING);
}
ATF_TC_CLEANUP(335badaddr3rdparty, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(34prefail);
ATF_TC_HEAD(34prefail, tc) {
	atf_tc_set_md_var(tc, "descr", "prefer failure option (peer)");
	atf_tc_set_md_var(tc, "timeout", "2");
}
ATF_TC_BODY(34prefail, tc) {
	uint16_t n16;
	ssize_t ret;

	UNUSED(tc);
	n16 = randport();
	setup_peer(100, IPPROTO_TCP, n16, 0, mapped, n16, peeraddr);
	append_prefer_failure();
	test_send(reqlen);
 	test_recv(reqlen);
	test_result(ERR_BADOPTION);
}
ATF_TC_CLEANUP(34prefail, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(35filter);
ATF_TC_HEAD(35filter, tc) {
	atf_tc_set_md_var(tc, "descr", "filter option (peer)");
	atf_tc_set_md_var(tc, "timeout", "2");
}
ATF_TC_BODY(35filter, tc) {
	uint16_t n16;
	ssize_t ret;

	UNUSED(tc);
	n16 = randport();
	setup_peer(100, IPPROTO_TCP, n16, 0, mapped, n16, peeraddr);
	append_filter(0, 0, mapped);
	test_send(reqlen);
 	test_recv(reqlen);
	test_result(ERR_BADOPTION);
}
ATF_TC_CLEANUP(35filter, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(360badoption);
ATF_TC_HEAD(360badoption, tc) {
	atf_tc_set_md_var(tc, "descr", "unknown mandatory option (peer)");
	atf_tc_set_md_var(tc, "timeout", "2");
}
ATF_TC_BODY(360badoption, tc) {
	uint16_t n16;
	ssize_t ret;

	UNUSED(tc);
	n16 = randport();
	setup_peer(100, IPPROTO_TCP, n16, 0, mapped, n16, peeraddr);
	append_option(107, 1);
	test_send(reqlen);
	test_recv(reqlen);
	test_result(ERR_UNSUPOPTION);
}
ATF_TC_CLEANUP(360badoption, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(361ignoption);
ATF_TC_HEAD(361ignoption, tc) {
	atf_tc_set_md_var(tc, "descr", "unknown optional option (peer)");
	atf_tc_set_md_var(tc, "timeout", "2");
}
ATF_TC_BODY(361ignoption, tc) {
	uint16_t n16;
	ssize_t ret;

	UNUSED(tc);
	n16 = randport();
	setup_peer(100, IPPROTO_TCP, n16, 0, mapped, n16, peeraddr);
#ifdef USE_THIRD_PARTY
	append_third_party(intaddr);
#endif
	append_option(107 | OPTION_OPTIONAL, 1);
	test_send(reqlen);
	test_recv(reqlen - 8);
	checkepoch();
	test_result(ERR_SUCCESS);
}
ATF_TC_CLEANUP(361ignoption, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(362bad0option);
ATF_TC_HEAD(362bad0option, tc) {
	atf_tc_set_md_var(tc, "descr", "unprocessed option (peer)");
	atf_tc_set_md_var(tc, "timeout", "2");
}
ATF_TC_BODY(362bad0option, tc) {
	uint16_t n16;
	ssize_t ret;

	UNUSED(tc);
	n16 = randport();
	setup_peer(100, IPPROTO_TCP, n16, 0, mapped, n16, peeraddr);
	/*append_third_party(intaddr);*/
	append_option(0, 1);
	request[optoff] = 107;
	test_send(reqlen);
 	test_recv(reqlen);
	test_result(ERR_UNSUPOPTION);
}
ATF_TC_CLEANUP(362bad0option, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(363options);
ATF_TC_HEAD(363options, tc) {
	atf_tc_set_md_var(tc, "descr", "unknown options (peer)");
	atf_tc_set_md_var(tc, "timeout", "2");
}
ATF_TC_BODY(363options, tc) {
	uint16_t n16;
	ssize_t ret;

	UNUSED(tc);
	n16 = randport();
	setup_peer(100, IPPROTO_TCP, n16, 0, mapped, n16, peeraddr);
	append_option(107, 1);
	append_option(107, 2);
	append_option(108, 3);
	append_option(207, 0);
	append_option(107, 4);
	append_option(208, 6);
	test_send(reqlen);
	test_recv(reqlen);
	test_result(ERR_UNSUPOPTION);
}
ATF_TC_CLEANUP(363options, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(37intaddr);
ATF_TC_HEAD(37intaddr, tc) {
	atf_tc_set_md_var(tc, "descr", "no internal address (peer)");
	atf_tc_set_md_var(tc, "timeout", "2");
}
ATF_TC_BODY(37intaddr, tc) {
#ifdef USE_THIRD_PARTY
	uint16_t n16;
	ssize_t ret;

	UNUSED(tc);
	n16 = randport();
	setup_peer(100, IPPROTO_TCP, n16, 0, mapped, n16, peeraddr);
	test_send(reqlen);
	test_recv(reqlen);
	test_result(ERR_BADREQUEST);
#else
	UNUSED(tc);
	atf_tc_skip("NAT444 does not require THIRD_PARTY");
#endif
}
ATF_TC_CLEANUP(37intaddr, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(38peeraddr);
ATF_TC_HEAD(38peeraddr, tc) {
	atf_tc_set_md_var(tc, "descr", "bad remote peer address (peer)");
	atf_tc_set_md_var(tc, "timeout", "2");
}
ATF_TC_BODY(38peeraddr, tc) {
	uint16_t n16;
	ssize_t ret;

	UNUSED(tc);
	n16 = randport();
	setup_peer(100, IPPROTO_TCP, n16, 0, mapped, n16, srcaddr);
	request[OFF_PEER_REMOTE_ADDR] = 0x20;
#ifdef USE_THIRD_PARTY
	append_third_party(intaddr);
#endif
	test_send(reqlen);
	test_recv(reqlen);
	checkepoch();
	test_result(ERR_BADREQUEST);
}
ATF_TC_CLEANUP(38peeraddr, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(39intport);
ATF_TC_HEAD(39intport, tc) {
	atf_tc_set_md_var(tc, "descr", "no internal port (peer)");
	atf_tc_set_md_var(tc, "timeout", "2");
}
ATF_TC_BODY(39intport, tc) {
	uint16_t n16;
	ssize_t ret;

	UNUSED(tc);
	n16 = randport();
	setup_peer(100, IPPROTO_TCP, 0, 0, mapped, n16, peeraddr);
#ifdef USE_THIRD_PARTY
	append_third_party(intaddr);
#endif
	test_send(reqlen);
 	test_recv(reqlen);
	test_result(ERR_BADREQUEST);
}
ATF_TC_CLEANUP(39intport, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(40addpeer);
ATF_TC_HEAD(40addpeer, tc) {
	atf_tc_set_md_var(tc, "descr", "create an implicit dynamic mapping");
	atf_tc_set_md_var(tc, "timeout", "5");
}
ATF_TC_BODY(40addpeer, tc) {
	uint16_t n16;
	ssize_t ret;

	/* TCP: intaddr4:54321 extaddr:0 peeraddr4:56789 life=100 */
	UNUSED(tc);
	n16 = randport();
	setup_peer(100, IPPROTO_TCP, n16, 0, mapped, n16, peeraddr);
#ifdef USE_THIRD_PARTY
	append_third_party(intaddr);
#endif
	test_send(reqlen);
	test_recv(reqlen);
	checkepoch();
	test_result(ERR_SUCCESS);
}
ATF_TC_CLEANUP(40addpeer, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(41badaddpeerpref);
ATF_TC_HEAD(41badaddpeerpref, tc) {
	atf_tc_set_md_var(tc, "descr", "bad external port (peer)");
	atf_tc_set_md_var(tc, "timeout", "2");
}
ATF_TC_BODY(41badaddpeerpref, tc) {
	uint16_t n16;
	ssize_t ret;

	UNUSED(tc);
	n16 = randport();
	setup_peer(100, IPPROTO_TCP, n16, 10, mapped, n16, peeraddr);
#ifdef USE_THIRD_PARTY
	append_third_party(intaddr);
#endif
	test_send(reqlen);
	test_recv(reqlen);
	checkepoch();
	test_result(ERR_CANTPROVIDE);
}
ATF_TC_CLEANUP(41badaddpeerpref, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(42badaddpeer);
ATF_TC_HEAD(42badaddpeer, tc) {
	atf_tc_set_md_var(tc, "descr", 
			  "implicit/explicit dynamic mapping collision");
	atf_tc_set_md_var(tc, "timeout", "5");
}
ATF_TC_BODY(42badaddpeer, tc) {
	ssize_t ret;

	/* explicit UDP: intaddr4:54322 life=100 */
	UNUSED(tc);
	setup_map(100, IPPROTO_UDP, 54322, 0, mapped);
#ifdef USE_THIRD_PARTY
	append_third_party(intaddr);
#endif
	test_send(reqlen);
	test_recv(reqlen);
	checkepoch();
	checkextaddr(0);
	test_result(ERR_SUCCESS);

	/* implicit UDP: intaddr4:54322 -> peeraddr4:56789 life=100 */
	setup_peer(100, IPPROTO_UDP, 54322, 0, mapped, 56789, peeraddr);
#ifdef USE_THIRD_PARTY
	append_third_party(intaddr);
#endif
	test_send(reqlen);
	test_recv(reqlen);
	checkepoch();
	checkextaddr(1);
	test_result(ERR_PROCESSING);
}
ATF_TC_CLEANUP(42badaddpeer, tc) {
	UNUSED(tc);
	cleanup();
}

/* should simply return the static mapping??? */

ATF_TC_WITH_CLEANUP(43addpeerstatic);
ATF_TC_HEAD(43addpeerstatic, tc) {
	atf_tc_set_md_var(tc, "descr", 
			  "implicit dynamic/static mapping collision");
	atf_tc_set_md_var(tc, "timeout", "5");
}
ATF_TC_BODY(43addpeerstatic, tc) {
	ssize_t ret;

	/* implicit TCP: intaddr4:STATICPORT -> peeraddr4:56789 life=100 */
	UNUSED(tc);
	setup_peer(100, IPPROTO_TCP, STATICPORT, 0, mapped, 56789, peeraddr);
#ifdef USE_THIRD_PARTY
	append_third_party(intaddr);
#endif
	test_send(reqlen);
	test_recv(reqlen);
	checkepoch();
	test_result(ERR_PROCESSING);
}
ATF_TC_CLEANUP(43addpeerstatic, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(44renewpeer);
ATF_TC_HEAD(44renewpeer, tc) {
	atf_tc_set_md_var(tc, "descr", "standard renewal (peer)");
	atf_tc_set_md_var(tc, "timeout", "5");
}
ATF_TC_BODY(44renewpeer, tc) {
	uint16_t n16, eport;
	ssize_t ret;
	uint8_t extaddr[16];

	UNUSED(tc);
	n16 = randport();
	setup_peer(10, IPPROTO_UDP, n16, 0, mapped, 1111, peeraddr);
#ifdef USE_THIRD_PARTY
	append_third_party(intaddr);
#endif
	test_send(reqlen);
	test_recv(reqlen);
	memcpy(&eport, response + OFF_PEER_EXTERNAL_PORT, 2);
	memcpy(extaddr, response + OFF_PEER_EXTERNAL_ADDR, 16);
	checkepoch();
	checkextaddr(1);
	test_result(ERR_SUCCESS);

	setup_peer(100, IPPROTO_UDP, n16, ntohs(eport), extaddr, 1111, peeraddr);
#ifdef USE_THIRD_PARTY
	append_third_party(intaddr);
#endif
	test_send(reqlen);
	test_recv(reqlen);
	checkepoch();
	checkextaddr(1);
	test_result(ERR_SUCCESS);
}
ATF_TC_CLEANUP(44renewpeer, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(45badrenewpeer);
ATF_TC_HEAD(45badrenewpeer, tc) {
	atf_tc_set_md_var(tc, "descr", "bad renewal (peer)");
	atf_tc_set_md_var(tc, "timeout", "5");
}
ATF_TC_BODY(45badrenewpeer, tc) {
	uint16_t n16, eport;
	ssize_t ret;
	uint8_t extaddr[16];

	UNUSED(tc);
	n16 = randport();
	setup_peer(10, IPPROTO_TCP, n16, 0, mapped, 1111, peeraddr);
#ifdef USE_THIRD_PARTY
	append_third_party(intaddr);
#endif
	test_send(reqlen);
	test_recv(reqlen);
	memcpy(&eport, response + OFF_PEER_EXTERNAL_PORT, 2);
	memcpy(extaddr, response + OFF_PEER_EXTERNAL_ADDR, 16);
	checkepoch();
	checkextaddr(1);
	test_result(ERR_SUCCESS);

	/* modified external port from previous call */
	setup_peer(100, IPPROTO_TCP, n16, ntohs(eport ^ 0x101), extaddr, 1111, peeraddr);
#ifdef USE_THIRD_PARTY
	append_third_party(intaddr);
#endif
	test_send(reqlen);
	test_recv(reqlen);
	checkepoch();
	checkextaddr(1);
	test_result(ERR_CANTPROVIDE);
}
ATF_TC_CLEANUP(45badrenewpeer, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(46replaypeer);
ATF_TC_HEAD(46replaypeer, tc) {
	atf_tc_set_md_var(tc, "descr", "replay (peer)");
	atf_tc_set_md_var(tc, "timeout", "5");
}
ATF_TC_BODY(46replaypeer, tc) {
	uint16_t n16;
	ssize_t ret;

	UNUSED(tc);
	n16 = randport();
	setup_peer(10, IPPROTO_TCP, n16, 0, mapped, 1111, peeraddr);
#ifdef USE_THIRD_PARTY
	append_third_party(intaddr);
#endif
	test_send(reqlen);
	test_recv(reqlen);
	checkepoch();
	checkextaddr(1);
	test_result(ERR_SUCCESS);

	test_send(reqlen);
	test_recv(reqlen);
	checkepoch();
	checkextaddr(1);
	test_result(ERR_SUCCESS);
}
ATF_TC_CLEANUP(46replaypeer, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TC_WITH_CLEANUP(47peerbadaddr);
ATF_TC_HEAD(47peerbadaddr, tc) {
	atf_tc_set_md_var(tc, "descr", "bad external address (peer)");
	atf_tc_set_md_var(tc, "timeout", "2");
}
ATF_TC_BODY(47peerbadaddr, tc) {
	uint16_t n16, eport;
	ssize_t ret;
	uint8_t extaddr[16];

	UNUSED(tc);
	n16 = randport();
	setup_map(10, IPPROTO_TCP, 10003, 0, mapped);
#ifdef USE_THIRD_PARTY
	append_third_party(intaddr);
#endif
	test_send(reqlen);
	test_recv(reqlen);
	memcpy(&eport, response + OFF_PEER_EXTERNAL_PORT, 2);
	memcpy(extaddr, response + OFF_MAP_EXTERNAL_ADDR, 16);
	checkepoch();
	test_result(ERR_SUCCESS);

	setup_peer(100, IPPROTO_TCP, 10006, ntohs(eport), extaddr, 1111, peeraddr);
	if (request[OFF_PEER_EXTERNAL_ADDR + 15] == 10)
		request[OFF_PEER_EXTERNAL_ADDR + 15] = 100;
	else
		request[OFF_PEER_EXTERNAL_ADDR + 15] = 10;
#ifdef USE_THIRD_PARTY
	append_third_party(intaddr);
#endif
	test_send(reqlen);
	test_recv(reqlen);
	checkepoch();
	test_result(ERR_CANTPROVIDE);
}
ATF_TC_CLEANUP(47peerbadaddr, tc) {
	UNUSED(tc);
	cleanup();
}

ATF_TP_ADD_TCS(tp) {
	setup_tp();

	ATF_TP_ADD_TC(tp, 00connect);
	ATF_TP_ADD_TC(tp, 01tooshort);
	ATF_TP_ADD_TC(tp, 02response);
	ATF_TP_ADD_TC(tp, 03natpmp);
	ATF_TP_ADD_TC(tp, 04unknownclient);
	ATF_TP_ADD_TC(tp, 05toolong);
	ATF_TP_ADD_TC(tp, 06badalign);
	ATF_TP_ADD_TC(tp, 07version);
	ATF_TP_ADD_TC(tp, 080shorthdr);
	ATF_TP_ADD_TC(tp, 081underrunmap);
	ATF_TP_ADD_TC(tp, 082underrunpeer);
#if TRANSPORT == AF_INET6
	ATF_TP_ADD_TC(tp, 0Amappedaddress);
#else
	ATF_TP_ADD_TC(tp, 0Aipv6address);
#endif
	ATF_TP_ADD_TC(tp, 0Bbadaddress);
	ATF_TP_ADD_TC(tp, 0Cbadopcode);
	ATF_TP_ADD_TC(tp, 10badproto);
	ATF_TP_ADD_TC(tp, 110badextaddr);
	ATF_TP_ADD_TC(tp, 111badproto0);
	ATF_TP_ADD_TC(tp, 120optionoverflow);
	ATF_TP_ADD_TC(tp, 1300mult3rdparty);
	ATF_TP_ADD_TC(tp, 1301mult3rdparty2);
	ATF_TP_ADD_TC(tp, 1313rdparty0);
	ATF_TP_ADD_TC(tp, 132bad3rdpartylen);
	ATF_TP_ADD_TC(tp, 133sameaddr3rdparty);
	ATF_TP_ADD_TC(tp, 134v63rdparty);
	ATF_TP_ADD_TC(tp, 135badaddr3rdparty);
	ATF_TP_ADD_TC(tp, 140multprefail);
	ATF_TP_ADD_TC(tp, 141badprefaillen);
	ATF_TP_ADD_TC(tp, 142prefaillife);
	ATF_TP_ADD_TC(tp, 143prefailnone);
	ATF_TP_ADD_TC(tp, 144prefailaddr);
	ATF_TP_ADD_TC(tp, 145prefailport);
	ATF_TP_ADD_TC(tp, 146prefailbadaddr);
	ATF_TP_ADD_TC(tp, 150filter0);
	ATF_TP_ADD_TC(tp, 151badfilterlen);
	ATF_TP_ADD_TC(tp, 152toomanyfilter);
	ATF_TP_ADD_TC(tp, 153badaddrfilter);
	ATF_TP_ADD_TC(tp, 154badfilterplen);
	ATF_TP_ADD_TC(tp, 155badfilterpref);
	ATF_TP_ADD_TC(tp, 156badfilterorder);
	ATF_TP_ADD_TC(tp, 160badoption);
	ATF_TP_ADD_TC(tp, 161ignoption);
	ATF_TP_ADD_TC(tp, 162bad0option);
	ATF_TP_ADD_TC(tp, 163options);
	ATF_TP_ADD_TC(tp, 17intaddr);
	ATF_TP_ADD_TC(tp, 180intaddrdel);
	ATF_TP_ADD_TC(tp, 181intaddradd);
	ATF_TP_ADD_TC(tp, 19intport);
	ATF_TP_ADD_TC(tp, 200deleteall);
	ATF_TP_ADD_TC(tp, 201deleteallbadproto);
	ATF_TP_ADD_TC(tp, 202deleteallbadport);
	ATF_TP_ADD_TC(tp, 21deleteclient);
	ATF_TP_ADD_TC(tp, 220deleteproto);
	ATF_TP_ADD_TC(tp, 221deleteport);
	ATF_TP_ADD_TC(tp, 23deleterandom);
	ATF_TP_ADD_TC(tp, 24addmap);
	ATF_TP_ADD_TC(tp, 25badaddmappref);
	ATF_TP_ADD_TC(tp, 26addmappref);
	ATF_TP_ADD_TC(tp, 27addmapstatic);
	ATF_TP_ADD_TC(tp, 280addfilter);
	ATF_TP_ADD_TC(tp, 281delfilter);
	ATF_TP_ADD_TC(tp, 282replayfilter);
	ATF_TP_ADD_TC(tp, 283replacefilter);
	ATF_TP_ADD_TC(tp, 284replaceport);
	ATF_TP_ADD_TC(tp, 285replacepref);
	ATF_TP_ADD_TC(tp, 286replaceplen);
	ATF_TP_ADD_TC(tp, 287deladdfilter);
	ATF_TP_ADD_TC(tp, 288badaddfilter);
	ATF_TP_ADD_TC(tp, 290renewmap);
	ATF_TP_ADD_TC(tp, 291deladdmap);
	ATF_TP_ADD_TC(tp, 292badrenewmap);
	ATF_TP_ADD_TC(tp, 293replaymap);
	ATF_TP_ADD_TC(tp, 300badproto0);
	ATF_TP_ADD_TC(tp, 301badproto);
	ATF_TP_ADD_TC(tp, 310badextaddr);
	ATF_TP_ADD_TC(tp, 311badproto0);
	ATF_TP_ADD_TC(tp, 320optionoverflow);
	ATF_TP_ADD_TC(tp, 330mult3rdparty);
	ATF_TP_ADD_TC(tp, 3313rdparty0);
	ATF_TP_ADD_TC(tp, 332bad3rdpartylen);
	ATF_TP_ADD_TC(tp, 333sameaddr3rdparty);
	ATF_TP_ADD_TC(tp, 334v63rdparty);
	ATF_TP_ADD_TC(tp, 335badaddr3rdparty);
	ATF_TP_ADD_TC(tp, 34prefail);
	ATF_TP_ADD_TC(tp, 35filter);
	ATF_TP_ADD_TC(tp, 360badoption);
	ATF_TP_ADD_TC(tp, 361ignoption);
	ATF_TP_ADD_TC(tp, 362bad0option);
	ATF_TP_ADD_TC(tp, 363options);
	ATF_TP_ADD_TC(tp, 37intaddr);
	ATF_TP_ADD_TC(tp, 38peeraddr);
	ATF_TP_ADD_TC(tp, 39intport);
	ATF_TP_ADD_TC(tp, 40addpeer);
	ATF_TP_ADD_TC(tp, 41badaddpeerpref);
	ATF_TP_ADD_TC(tp, 42badaddpeer);
	ATF_TP_ADD_TC(tp, 43addpeerstatic);
	ATF_TP_ADD_TC(tp, 44renewpeer);
	ATF_TP_ADD_TC(tp, 45badrenewpeer);
	ATF_TP_ADD_TC(tp, 46replaypeer);
	ATF_TP_ADD_TC(tp, 47peerbadaddr);

	return atf_no_error();
}
